/*
 * Copyright (c) Facebook, Inc. and its affiliates.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include "velox/dwio/common/encryption/Encryption.h"

namespace facebook {
namespace velox {
namespace dwio {
namespace common {
namespace encryption {

bool operator==(const EncryptionProperties& a, const EncryptionProperties& b) {
  return std::addressof(a) == std::addressof(b) ||
      (typeid(a) == typeid(b) && a.equals(b));
}

} // namespace encryption
} // namespace common
} // namespace dwio
} // namespace velox
} // namespace facebook
