/*
 * Copyright (c) Facebook, Inc. and its affiliates.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include "velox/common/file/tests/TestUtils.h"

namespace facebook::velox::tests::utils {

std::vector<std::string> iobufsToStrings(
    const std::vector<folly::IOBuf>& iobufs) {
  std::vector<std::string> result;
  result.reserve(iobufs.size());
  for (auto& iobuf : iobufs) {
    auto coalesced = iobuf.cloneCoalescedAsValueWithHeadroomTailroom(0, 0);
    result.emplace_back(
        reinterpret_cast<const char*>(coalesced.data()), coalesced.length());
  }
  return result;
}

} // namespace facebook::velox::tests::utils
