/*
 * Copyright (c) Facebook, Inc. and its affiliates.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#pragma once

#include <memory>

#include "velox/expression/VectorFunction.h"

namespace facebook::velox::functions::sparksql {

/// Supported Types:
/// TINYINT
/// SMALLINT
/// INTEGER
/// BIGINT
/// HUGEINT
/// REAL
/// DOUBLE
/// BOOLEAN
/// VARCHAR
/// TIMESTAMP

/// Special cases:
/// NaN in Spark is handled differently from standard floating point semantics.
/// It is considered larger than any other numeric values. So Least/Greatest
/// functions implement this order:
/// -Inf < Inf < NaN

std::shared_ptr<exec::VectorFunction> makeLeast(
    const std::string& name,
    const std::vector<exec::VectorFunctionArg>& inputArgs,
    const core::QueryConfig& config);

std::vector<std::shared_ptr<exec::FunctionSignature>> leastSignatures();

std::shared_ptr<exec::VectorFunction> makeGreatest(
    const std::string& name,
    const std::vector<exec::VectorFunctionArg>& inputArgs,
    const core::QueryConfig& config);

std::vector<std::shared_ptr<exec::FunctionSignature>> greatestSignatures();

} // namespace facebook::velox::functions::sparksql
