/*
 * Copyright (c) Facebook, Inc. and its affiliates.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

/*
 * Copyright (c) 2024 by Rivos Inc.
 * Licensed under the Apache License, Version 2.0, see LICENSE for details.
 * SPDX-License-Identifier: Apache-2.0
 */

#pragma once

// some platforms require a system header to be included
#if defined(PLATFORM_CUDA)
#include <cuda.h>
#elif defined(PLATFORM_HIP)
#include <hip/hip_runtime.h>
#elif defined(PLATFORM_METAL)
#include <metal_stdlib>
#endif

// function attributes used for device code
#if defined(PLATFORM_CUDA) || defined(PLATFORM_HIP)
#define ATTR __device__ __forceinline__
#elif defined(PLATFORM_SYCL) || defined(PLATFORM_METAL) || \
    defined(PLATFORM_OPENCL) || defined(PLATFORM_OPENMP)
#define ATTR inline
#else
#error platform must be set
#endif
