/*
 * Copyright (c) Facebook, Inc. and its affiliates.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

// This file is generated. Do not modify it manually. To re-generate it, run:
//
//  javac velox/type/tz/GenTimeZoneNames.java
//  java velox/type/tz/GenTimeZoneNames > velox/type/tz/TimeZoneNames.cpp
//
// Use the version of Java with which you're running Presto to generate
// compatible names. This version was generated using openjdk version
// "1.8.0_362"
// @generated

#include "velox/type/tz/TimeZoneNames.h"

#include <folly/Portability.h>

// This file just contains a single function that returns an, admittedly
// large, hard coded map.
FOLLY_CLANG_DISABLE_WARNING("-Wignored-optimization-argument");

namespace facebook::velox::tz {

const std::unordered_map<std::string, TimeZoneNames>& getTimeZoneNames() {
  static auto* timeZoneNames = new std::unordered_map<
      std::string,
      TimeZoneNames>([] {
    // Work around clang compiler bug causing multi-hour compilation
    // with -fsanitize=fuzzer
    // https://github.com/llvm/llvm-project/issues/75666
    return std::unordered_map<std::string, TimeZoneNames>{
        {"America/Los_Angeles",
         {"Pacific Standard Time", "PST", "Pacific Daylight Time", "PDT"}},
        {"PST",
         {"Pacific Standard Time", "PST", "Pacific Daylight Time", "PDT"}},
        {"America/Denver",
         {"Mountain Standard Time", "MST", "Mountain Daylight Time", "MDT"}},
        {"MST",
         {"Mountain Standard Time", "MST", "Mountain Daylight Time", "MDT"}},
        {"America/Phoenix",
         {"Mountain Standard Time", "MST", "Mountain Daylight Time", "MDT"}},
        {"PNT",
         {"Mountain Standard Time", "MST", "Mountain Daylight Time", "MDT"}},
        {"America/Chicago",
         {"Central Standard Time", "CST", "Central Daylight Time", "CDT"}},
        {"CST",
         {"Central Standard Time", "CST", "Central Daylight Time", "CDT"}},
        {"America/New_York",
         {"Eastern Standard Time", "EST", "Eastern Daylight Time", "EDT"}},
        {"EST",
         {"Eastern Standard Time", "EST", "Eastern Daylight Time", "EDT"}},
        {"America/Indianapolis",
         {"Eastern Standard Time", "EST", "Eastern Daylight Time", "EDT"}},
        {"IET",
         {"Eastern Standard Time", "EST", "Eastern Daylight Time", "EDT"}},
        {"Pacific/Honolulu",
         {"Hawaii Standard Time", "HST", "Hawaii Daylight Time", "HDT"}},
        {"HST", {"Hawaii Standard Time", "HST", "Hawaii Daylight Time", "HDT"}},
        {"America/Anchorage",
         {"Alaska Standard Time", "AKST", "Alaska Daylight Time", "AKDT"}},
        {"AST",
         {"Alaska Standard Time", "AKST", "Alaska Daylight Time", "AKDT"}},
        {"America/Halifax",
         {"Atlantic Standard Time", "AST", "Atlantic Daylight Time", "ADT"}},
        {"America/Sitka",
         {"Alaska Standard Time", "AKST", "Alaska Daylight Time", "AKDT"}},
        {"America/St_Johns",
         {"Newfoundland Standard Time",
          "NST",
          "Newfoundland Daylight Time",
          "NDT"}},
        {"CNT",
         {"Newfoundland Standard Time",
          "NST",
          "Newfoundland Daylight Time",
          "NDT"}},
        {"Europe/Paris",
         {"Central European Time",
          "CET",
          "Central European Summer Time",
          "CEST"}},
        {"ECT",
         {"Central European Time",
          "CET",
          "Central European Summer Time",
          "CEST"}},
        {"GMT", {"Greenwich Mean Time", "GMT", "Greenwich Mean Time", "GMT"}},
        {"Africa/Casablanca",
         {"Western European Time",
          "WET",
          "Western European Summer Time",
          "WEST"}},
        {"Asia/Jerusalem",
         {"Israel Standard Time", "IST", "Israel Daylight Time", "IDT"}},
        {"Asia/Tokyo",
         {"Japan Standard Time", "JST", "Japan Daylight Time", "JDT"}},
        {"JST", {"Japan Standard Time", "JST", "Japan Daylight Time", "JDT"}},
        {"Europe/Bucharest",
         {"Eastern European Time",
          "EET",
          "Eastern European Summer Time",
          "EEST"}},
        {"Asia/Shanghai",
         {"China Standard Time", "CST", "China Daylight Time", "CDT"}},
        {"CTT", {"China Standard Time", "CST", "China Daylight Time", "CDT"}},
        {"UTC",
         {"Coordinated Universal Time",
          "UTC",
          "Coordinated Universal Time",
          "UTC"}},
        {"ACT",
         {"Australian Central Standard Time (Northern Territory)",
          "ACST",
          "Australian Central Daylight Time (Northern Territory)",
          "ACDT"}},
        {"AET",
         {"Australian Eastern Standard Time (New South Wales)",
          "AEST",
          "Australian Eastern Daylight Time (New South Wales)",
          "AEDT"}},
        {"AGT", {"Argentine Time", "ART", "Argentine Summer Time", "ARST"}},
        {"ART",
         {"Eastern European Time",
          "EET",
          "Eastern European Summer Time",
          "EEST"}},
        {"Africa/Abidjan",
         {"Greenwich Mean Time", "GMT", "Greenwich Mean Time", "GMT"}},
        {"Africa/Accra",
         {"Ghana Mean Time", "GMT", "Ghana Summer Time", "GHST"}},
        {"Africa/Addis_Ababa",
         {"Eastern African Time",
          "EAT",
          "Eastern African Summer Time",
          "EAST"}},
        {"Africa/Algiers",
         {"Central European Time",
          "CET",
          "Central European Summer Time",
          "CEST"}},
        {"Africa/Asmara",
         {"Eastern African Time",
          "EAT",
          "Eastern African Summer Time",
          "EAST"}},
        {"Africa/Asmera",
         {"Eastern African Time",
          "EAT",
          "Eastern African Summer Time",
          "EAST"}},
        {"Africa/Bamako",
         {"Greenwich Mean Time", "GMT", "Greenwich Mean Time", "GMT"}},
        {"Africa/Bangui",
         {"Western African Time",
          "WAT",
          "Western African Summer Time",
          "WAST"}},
        {"Africa/Banjul",
         {"Greenwich Mean Time", "GMT", "Greenwich Mean Time", "GMT"}},
        {"Africa/Bissau",
         {"Greenwich Mean Time", "GMT", "Greenwich Mean Time", "GMT"}},
        {"Africa/Blantyre",
         {"Central African Time",
          "CAT",
          "Central African Summer Time",
          "CAST"}},
        {"Africa/Brazzaville",
         {"Western African Time",
          "WAT",
          "Western African Summer Time",
          "WAST"}},
        {"Africa/Bujumbura",
         {"Central African Time",
          "CAT",
          "Central African Summer Time",
          "CAST"}},
        {"Africa/Cairo",
         {"Eastern European Time",
          "EET",
          "Eastern European Summer Time",
          "EEST"}},
        {"Africa/Ceuta",
         {"Central European Time",
          "CET",
          "Central European Summer Time",
          "CEST"}},
        {"Africa/Conakry",
         {"Greenwich Mean Time", "GMT", "Greenwich Mean Time", "GMT"}},
        {"Africa/Dakar",
         {"Greenwich Mean Time", "GMT", "Greenwich Mean Time", "GMT"}},
        {"Africa/Dar_es_Salaam",
         {"Eastern African Time",
          "EAT",
          "Eastern African Summer Time",
          "EAST"}},
        {"Africa/Djibouti",
         {"Eastern African Time",
          "EAT",
          "Eastern African Summer Time",
          "EAST"}},
        {"Africa/Douala",
         {"Western African Time",
          "WAT",
          "Western African Summer Time",
          "WAST"}},
        {"Africa/El_Aaiun",
         {"Western European Time",
          "WET",
          "Western European Summer Time",
          "WEST"}},
        {"Africa/Freetown",
         {"Greenwich Mean Time", "GMT", "Greenwich Mean Time", "GMT"}},
        {"Africa/Gaborone",
         {"Central African Time",
          "CAT",
          "Central African Summer Time",
          "CAST"}},
        {"Africa/Harare",
         {"Central African Time",
          "CAT",
          "Central African Summer Time",
          "CAST"}},
        {"Africa/Johannesburg",
         {"South Africa Standard Time",
          "SAST",
          "South Africa Summer Time",
          "SAST"}},
        {"Africa/Juba",
         {"Central African Time",
          "CAT",
          "Central African Summer Time",
          "CAST"}},
        {"Africa/Kampala",
         {"Eastern African Time",
          "EAT",
          "Eastern African Summer Time",
          "EAST"}},
        {"Africa/Khartoum",
         {"Central African Time",
          "CAT",
          "Central African Summer Time",
          "CAST"}},
        {"Africa/Kigali",
         {"Central African Time",
          "CAT",
          "Central African Summer Time",
          "CAST"}},
        {"Africa/Kinshasa",
         {"Western African Time",
          "WAT",
          "Western African Summer Time",
          "WAST"}},
        {"Africa/Lagos",
         {"Western African Time",
          "WAT",
          "Western African Summer Time",
          "WAST"}},
        {"Africa/Libreville",
         {"Western African Time",
          "WAT",
          "Western African Summer Time",
          "WAST"}},
        {"Africa/Lome",
         {"Greenwich Mean Time", "GMT", "Greenwich Mean Time", "GMT"}},
        {"Africa/Luanda",
         {"Western African Time",
          "WAT",
          "Western African Summer Time",
          "WAST"}},
        {"Africa/Lubumbashi",
         {"Central African Time",
          "CAT",
          "Central African Summer Time",
          "CAST"}},
        {"Africa/Lusaka",
         {"Central African Time",
          "CAT",
          "Central African Summer Time",
          "CAST"}},
        {"Africa/Malabo",
         {"Western African Time",
          "WAT",
          "Western African Summer Time",
          "WAST"}},
        {"Africa/Maputo",
         {"Central African Time",
          "CAT",
          "Central African Summer Time",
          "CAST"}},
        {"Africa/Maseru",
         {"South Africa Standard Time",
          "SAST",
          "South Africa Summer Time",
          "SAST"}},
        {"Africa/Mbabane",
         {"South Africa Standard Time",
          "SAST",
          "South Africa Summer Time",
          "SAST"}},
        {"Africa/Mogadishu",
         {"Eastern African Time",
          "EAT",
          "Eastern African Summer Time",
          "EAST"}},
        {"Africa/Monrovia",
         {"Greenwich Mean Time", "GMT", "Greenwich Mean Time", "GMT"}},
        {"Africa/Nairobi",
         {"Eastern African Time",
          "EAT",
          "Eastern African Summer Time",
          "EAST"}},
        {"Africa/Ndjamena",
         {"Western African Time",
          "WAT",
          "Western African Summer Time",
          "WAST"}},
        {"Africa/Niamey",
         {"Western African Time",
          "WAT",
          "Western African Summer Time",
          "WAST"}},
        {"Africa/Nouakchott",
         {"Greenwich Mean Time", "GMT", "Greenwich Mean Time", "GMT"}},
        {"Africa/Ouagadougou",
         {"Greenwich Mean Time", "GMT", "Greenwich Mean Time", "GMT"}},
        {"Africa/Porto-Novo",
         {"Western African Time",
          "WAT",
          "Western African Summer Time",
          "WAST"}},
        {"Africa/Sao_Tome",
         {"Greenwich Mean Time", "GMT", "Greenwich Mean Time", "GMT"}},
        {"Africa/Timbuktu",
         {"Greenwich Mean Time", "GMT", "Greenwich Mean Time", "GMT"}},
        {"Africa/Tripoli",
         {"Eastern European Time",
          "EET",
          "Eastern European Summer Time",
          "EEST"}},
        {"Africa/Tunis",
         {"Central European Time",
          "CET",
          "Central European Summer Time",
          "CEST"}},
        {"Africa/Windhoek",
         {"Central African Time", "CAT", "Western African Time", "WAT"}},
        {"America/Adak",
         {"Hawaii Standard Time", "HST", "Hawaii Daylight Time", "HDT"}},
        {"America/Anguilla",
         {"Atlantic Standard Time", "AST", "Atlantic Daylight Time", "ADT"}},
        {"America/Antigua",
         {"Atlantic Standard Time", "AST", "Atlantic Daylight Time", "ADT"}},
        {"America/Araguaina",
         {"Brasilia Time", "BRT", "Brasilia Summer Time", "BRST"}},
        {"America/Argentina/Buenos_Aires",
         {"Argentine Time", "ART", "Argentine Summer Time", "ARST"}},
        {"America/Argentina/Catamarca",
         {"Argentine Time", "ART", "Argentine Summer Time", "ARST"}},
        {"America/Argentina/ComodRivadavia",
         {"Argentine Time", "ART", "Argentine Summer Time", "ARST"}},
        {"America/Argentina/Cordoba",
         {"Argentine Time", "ART", "Argentine Summer Time", "ARST"}},
        {"America/Argentina/Jujuy",
         {"Argentine Time", "ART", "Argentine Summer Time", "ARST"}},
        {"America/Argentina/La_Rioja",
         {"Argentine Time", "ART", "Argentine Summer Time", "ARST"}},
        {"America/Argentina/Mendoza",
         {"Argentine Time", "ART", "Argentine Summer Time", "ARST"}},
        {"America/Argentina/Rio_Gallegos",
         {"Argentine Time", "ART", "Argentine Summer Time", "ARST"}},
        {"America/Argentina/Salta",
         {"Argentine Time", "ART", "Argentine Summer Time", "ARST"}},
        {"America/Argentina/San_Juan",
         {"Argentine Time", "ART", "Argentine Summer Time", "ARST"}},
        {"America/Argentina/San_Luis",
         {"Argentine Time", "ART", "Argentine Summer Time", "ARST"}},
        {"America/Argentina/Tucuman",
         {"Argentine Time", "ART", "Argentine Summer Time", "ARST"}},
        {"America/Argentina/Ushuaia",
         {"Argentine Time", "ART", "Argentine Summer Time", "ARST"}},
        {"America/Aruba",
         {"Atlantic Standard Time", "AST", "Atlantic Daylight Time", "ADT"}},
        {"America/Asuncion",
         {"Paraguay Time", "PYT", "Paraguay Summer Time", "PYST"}},
        {"America/Atikokan",
         {"Eastern Standard Time", "EST", "Eastern Daylight Time", "EDT"}},
        {"America/Atka",
         {"Hawaii Standard Time", "HST", "Hawaii Daylight Time", "HDT"}},
        {"America/Bahia",
         {"Brasilia Time", "BRT", "Brasilia Summer Time", "BRST"}},
        {"America/Bahia_Banderas",
         {"Central Standard Time", "CST", "Central Daylight Time", "CDT"}},
        {"America/Barbados",
         {"Atlantic Standard Time", "AST", "Atlantic Daylight Time", "ADT"}},
        {"America/Belem",
         {"Brasilia Time", "BRT", "Brasilia Summer Time", "BRST"}},
        {"America/Belize",
         {"Central Standard Time", "CST", "Central Daylight Time", "CDT"}},
        {"America/Blanc-Sablon",
         {"Atlantic Standard Time", "AST", "Atlantic Daylight Time", "ADT"}},
        {"America/Boa_Vista",
         {"Amazon Time", "AMT", "Amazon Summer Time", "AMST"}},
        {"America/Bogota",
         {"Colombia Time", "COT", "Colombia Summer Time", "COST"}},
        {"America/Boise",
         {"Mountain Standard Time", "MST", "Mountain Daylight Time", "MDT"}},
        {"America/Buenos_Aires",
         {"Argentine Time", "ART", "Argentine Summer Time", "ARST"}},
        {"America/Cambridge_Bay",
         {"Mountain Standard Time", "MST", "Mountain Daylight Time", "MDT"}},
        {"America/Campo_Grande",
         {"Amazon Time", "AMT", "Amazon Summer Time", "AMST"}},
        {"America/Cancun",
         {"Eastern Standard Time", "EST", "Eastern Daylight Time", "EDT"}},
        {"America/Caracas",
         {"Venezuela Time", "VET", "Venezuela Summer Time", "VEST"}},
        {"America/Catamarca",
         {"Argentine Time", "ART", "Argentine Summer Time", "ARST"}},
        {"America/Cayenne",
         {"French Guiana Time", "GFT", "French Guiana Summer Time", "GFST"}},
        {"America/Cayman",
         {"Eastern Standard Time", "EST", "Eastern Daylight Time", "EDT"}},
        {"America/Chihuahua",
         {"Central Standard Time", "CST", "Central Daylight Time", "CDT"}},
        {"America/Ciudad_Juarez",
         {"Mountain Standard Time", "MST", "Mountain Daylight Time", "MDT"}},
        {"America/Creston",
         {"Mountain Standard Time", "MST", "Mountain Daylight Time", "MDT"}},
        {"America/Coral_Harbour",
         {"Eastern Standard Time", "EST", "Eastern Daylight Time", "EDT"}},
        {"America/Cordoba",
         {"Argentine Time", "ART", "Argentine Summer Time", "ARST"}},
        {"America/Costa_Rica",
         {"Central Standard Time", "CST", "Central Daylight Time", "CDT"}},
        {"America/Cuiaba",
         {"Amazon Time", "AMT", "Amazon Summer Time", "AMST"}},
        {"America/Curacao",
         {"Atlantic Standard Time", "AST", "Atlantic Daylight Time", "ADT"}},
        {"America/Danmarkshavn",
         {"Greenwich Mean Time", "GMT", "Greenwich Mean Time", "GMT"}},
        {"America/Dawson",
         {"Mountain Standard Time", "MST", "Mountain Daylight Time", "MDT"}},
        {"America/Dawson_Creek",
         {"Mountain Standard Time", "MST", "Mountain Daylight Time", "MDT"}},
        {"America/Detroit",
         {"Eastern Standard Time", "EST", "Eastern Daylight Time", "EDT"}},
        {"America/Dominica",
         {"Atlantic Standard Time", "AST", "Atlantic Daylight Time", "ADT"}},
        {"America/Edmonton",
         {"Mountain Standard Time", "MST", "Mountain Daylight Time", "MDT"}},
        {"America/Eirunepe", {"Acre Time", "ACT", "Acre Summer Time", "ACST"}},
        {"America/El_Salvador",
         {"Central Standard Time", "CST", "Central Daylight Time", "CDT"}},
        {"America/Ensenada",
         {"Pacific Standard Time", "PST", "Pacific Daylight Time", "PDT"}},
        {"America/Fort_Nelson",
         {"Mountain Standard Time", "MST", "Mountain Daylight Time", "MDT"}},
        {"America/Fort_Wayne",
         {"Eastern Standard Time", "EST", "Eastern Daylight Time", "EDT"}},
        {"America/Fortaleza",
         {"Brasilia Time", "BRT", "Brasilia Summer Time", "BRST"}},
        {"America/Glace_Bay",
         {"Atlantic Standard Time", "AST", "Atlantic Daylight Time", "ADT"}},
        {"America/Godthab",
         {"Western Greenland Time",
          "WGT",
          "Western Greenland Summer Time",
          "WGST"}},
        {"America/Goose_Bay",
         {"Atlantic Standard Time", "AST", "Atlantic Daylight Time", "ADT"}},
        {"America/Grand_Turk",
         {"Eastern Standard Time", "EST", "Eastern Daylight Time", "EDT"}},
        {"America/Grenada",
         {"Atlantic Standard Time", "AST", "Atlantic Daylight Time", "ADT"}},
        {"America/Guadeloupe",
         {"Atlantic Standard Time", "AST", "Atlantic Daylight Time", "ADT"}},
        {"America/Guatemala",
         {"Central Standard Time", "CST", "Central Daylight Time", "CDT"}},
        {"America/Guayaquil",
         {"Ecuador Time", "ECT", "Ecuador Summer Time", "ECST"}},
        {"America/Guyana",
         {"Guyana Time", "GYT", "Guyana Summer Time", "GYST"}},
        {"America/Havana",
         {"Cuba Standard Time", "CST", "Cuba Daylight Time", "CDT"}},
        {"America/Hermosillo",
         {"Mountain Standard Time", "MST", "Mountain Daylight Time", "MDT"}},
        {"America/Indiana/Indianapolis",
         {"Eastern Standard Time", "EST", "Eastern Daylight Time", "EDT"}},
        {"America/Indiana/Knox",
         {"Central Standard Time", "CST", "Central Daylight Time", "CDT"}},
        {"America/Indiana/Marengo",
         {"Eastern Standard Time", "EST", "Eastern Daylight Time", "EDT"}},
        {"America/Indiana/Petersburg",
         {"Eastern Standard Time", "EST", "Eastern Daylight Time", "EDT"}},
        {"America/Indiana/Tell_City",
         {"Central Standard Time", "CST", "Central Daylight Time", "CDT"}},
        {"America/Indiana/Vevay",
         {"Eastern Standard Time", "EST", "Eastern Daylight Time", "EDT"}},
        {"America/Indiana/Vincennes",
         {"Eastern Standard Time", "EST", "Eastern Daylight Time", "EDT"}},
        {"America/Indiana/Winamac",
         {"Eastern Standard Time", "EST", "Eastern Daylight Time", "EDT"}},
        {"America/Inuvik",
         {"Mountain Standard Time", "MST", "Mountain Daylight Time", "MDT"}},
        {"America/Iqaluit",
         {"Eastern Standard Time", "EST", "Eastern Daylight Time", "EDT"}},
        {"America/Jamaica",
         {"Eastern Standard Time", "EST", "Eastern Daylight Time", "EDT"}},
        {"America/Jujuy",
         {"Argentine Time", "ART", "Argentine Summer Time", "ARST"}},
        {"America/Juneau",
         {"Alaska Standard Time", "AKST", "Alaska Daylight Time", "AKDT"}},
        {"America/Kentucky/Louisville",
         {"Eastern Standard Time", "EST", "Eastern Daylight Time", "EDT"}},
        {"America/Kentucky/Monticello",
         {"Eastern Standard Time", "EST", "Eastern Daylight Time", "EDT"}},
        {"America/Knox_IN",
         {"Central Standard Time", "CST", "Central Daylight Time", "CDT"}},
        {"America/Kralendijk",
         {"Atlantic Standard Time", "AST", "Atlantic Daylight Time", "ADT"}},
        {"America/La_Paz",
         {"Bolivia Time", "BOT", "Bolivia Summer Time", "BOST"}},
        {"America/Lima", {"Peru Time", "PET", "Peru Summer Time", "PEST"}},
        {"America/Louisville",
         {"Eastern Standard Time", "EST", "Eastern Daylight Time", "EDT"}},
        {"America/Lower_Princes",
         {"Atlantic Standard Time", "AST", "Atlantic Daylight Time", "ADT"}},
        {"America/Maceio",
         {"Brasilia Time", "BRT", "Brasilia Summer Time", "BRST"}},
        {"America/Managua",
         {"Central Standard Time", "CST", "Central Daylight Time", "CDT"}},
        {"America/Manaus",
         {"Amazon Time", "AMT", "Amazon Summer Time", "AMST"}},
        {"America/Marigot",
         {"Atlantic Standard Time", "AST", "Atlantic Daylight Time", "ADT"}},
        {"America/Martinique",
         {"Atlantic Standard Time", "AST", "Atlantic Daylight Time", "ADT"}},
        {"America/Matamoros",
         {"Central Standard Time", "CST", "Central Daylight Time", "CDT"}},
        {"America/Mazatlan",
         {"Mountain Standard Time", "MST", "Mountain Daylight Time", "MDT"}},
        {"America/Mendoza",
         {"Argentine Time", "ART", "Argentine Summer Time", "ARST"}},
        {"America/Menominee",
         {"Central Standard Time", "CST", "Central Daylight Time", "CDT"}},
        {"America/Merida",
         {"Central Standard Time", "CST", "Central Daylight Time", "CDT"}},
        {"America/Metlakatla",
         {"Alaska Standard Time", "AKST", "Alaska Daylight Time", "AKDT"}},
        {"America/Mexico_City",
         {"Central Standard Time", "CST", "Central Daylight Time", "CDT"}},
        {"America/Miquelon",
         {"Pierre & Miquelon Standard Time",
          "PMST",
          "Pierre & Miquelon Daylight Time",
          "PMDT"}},
        {"America/Moncton",
         {"Atlantic Standard Time", "AST", "Atlantic Daylight Time", "ADT"}},
        {"America/Montevideo",
         {"Uruguay Time", "UYT", "Uruguay Summer Time", "UYST"}},
        {"America/Monterrey",
         {"Central Standard Time", "CST", "Central Daylight Time", "CDT"}},
        {"America/Montreal",
         {"Eastern Standard Time", "EST", "Eastern Daylight Time", "EDT"}},
        {"America/Montserrat",
         {"Atlantic Standard Time", "AST", "Atlantic Daylight Time", "ADT"}},
        {"America/Nassau",
         {"Eastern Standard Time", "EST", "Eastern Daylight Time", "EDT"}},
        {"America/Nipigon",
         {"Eastern Standard Time", "EST", "Eastern Daylight Time", "EDT"}},
        {"America/Nome",
         {"Alaska Standard Time", "AKST", "Alaska Daylight Time", "AKDT"}},
        {"America/Noronha",
         {"Fernando de Noronha Time",
          "FNT",
          "Fernando de Noronha Summer Time",
          "FNST"}},
        {"America/North_Dakota/Beulah",
         {"Central Standard Time", "CST", "Central Daylight Time", "CDT"}},
        {"America/North_Dakota/Center",
         {"Central Standard Time", "CST", "Central Daylight Time", "CDT"}},
        {"America/North_Dakota/New_Salem",
         {"Central Standard Time", "CST", "Central Daylight Time", "CDT"}},
        {"America/Nuuk",
         {"Western Greenland Time",
          "WGT",
          "Western Greenland Summer Time",
          "WGST"}},
        {"America/Ojinaga",
         {"Central Standard Time", "CST", "Central Daylight Time", "CDT"}},
        {"America/Panama",
         {"Eastern Standard Time", "EST", "Eastern Daylight Time", "EDT"}},
        {"America/Pangnirtung",
         {"Eastern Standard Time", "EST", "Eastern Daylight Time", "EDT"}},
        {"America/Paramaribo",
         {"Suriname Time", "SRT", "Suriname Summer Time", "SRST"}},
        {"America/Port-au-Prince",
         {"Eastern Standard Time", "EST", "Eastern Daylight Time", "EDT"}},
        {"America/Port_of_Spain",
         {"Atlantic Standard Time", "AST", "Atlantic Daylight Time", "ADT"}},
        {"America/Porto_Acre",
         {"Acre Time", "ACT", "Acre Summer Time", "ACST"}},
        {"America/Porto_Velho",
         {"Amazon Time", "AMT", "Amazon Summer Time", "AMST"}},
        {"America/Puerto_Rico",
         {"Atlantic Standard Time", "AST", "Atlantic Daylight Time", "ADT"}},
        {"America/Rainy_River",
         {"Central Standard Time", "CST", "Central Daylight Time", "CDT"}},
        {"America/Rankin_Inlet",
         {"Central Standard Time", "CST", "Central Daylight Time", "CDT"}},
        {"America/Recife",
         {"Brasilia Time", "BRT", "Brasilia Summer Time", "BRST"}},
        {"America/Regina",
         {"Central Standard Time", "CST", "Central Daylight Time", "CDT"}},
        {"America/Resolute",
         {"Central Standard Time", "CST", "Central Daylight Time", "CDT"}},
        {"America/Rio_Branco",
         {"Acre Time", "ACT", "Acre Summer Time", "ACST"}},
        {"America/Rosario",
         {"Argentine Time", "ART", "Argentine Summer Time", "ARST"}},
        {"America/Santa_Isabel",
         {"Pacific Standard Time", "PST", "Pacific Daylight Time", "PDT"}},
        {"America/Santarem",
         {"Brasilia Time", "BRT", "Brasilia Summer Time", "BRST"}},
        {"America/Santiago",
         {"Chile Time", "CLT", "Chile Summer Time", "CLST"}},
        {"America/Santo_Domingo",
         {"Atlantic Standard Time", "AST", "Atlantic Daylight Time", "ADT"}},
        {"America/Sao_Paulo",
         {"Brasilia Time", "BRT", "Brasilia Summer Time", "BRST"}},
        {"America/Scoresbysund",
         {"Eastern Greenland Time",
          "EGT",
          "Eastern Greenland Summer Time",
          "EGST"}},
        {"America/Shiprock",
         {"Mountain Standard Time", "MST", "Mountain Daylight Time", "MDT"}},
        {"America/St_Barthelemy",
         {"Atlantic Standard Time", "AST", "Atlantic Daylight Time", "ADT"}},
        {"America/St_Kitts",
         {"Atlantic Standard Time", "AST", "Atlantic Daylight Time", "ADT"}},
        {"America/St_Lucia",
         {"Atlantic Standard Time", "AST", "Atlantic Daylight Time", "ADT"}},
        {"America/St_Thomas",
         {"Atlantic Standard Time", "AST", "Atlantic Daylight Time", "ADT"}},
        {"America/St_Vincent",
         {"Atlantic Standard Time", "AST", "Atlantic Daylight Time", "ADT"}},
        {"America/Swift_Current",
         {"Central Standard Time", "CST", "Central Daylight Time", "CDT"}},
        {"America/Tegucigalpa",
         {"Central Standard Time", "CST", "Central Daylight Time", "CDT"}},
        {"America/Thule",
         {"Atlantic Standard Time", "AST", "Atlantic Daylight Time", "ADT"}},
        {"America/Thunder_Bay",
         {"Eastern Standard Time", "EST", "Eastern Daylight Time", "EDT"}},
        {"America/Tijuana",
         {"Pacific Standard Time", "PST", "Pacific Daylight Time", "PDT"}},
        {"America/Toronto",
         {"Eastern Standard Time", "EST", "Eastern Daylight Time", "EDT"}},
        {"America/Tortola",
         {"Atlantic Standard Time", "AST", "Atlantic Daylight Time", "ADT"}},
        {"America/Vancouver",
         {"Pacific Standard Time", "PST", "Pacific Daylight Time", "PDT"}},
        {"America/Virgin",
         {"Atlantic Standard Time", "AST", "Atlantic Daylight Time", "ADT"}},
        {"America/Whitehorse",
         {"Mountain Standard Time", "MST", "Mountain Daylight Time", "MDT"}},
        {"America/Winnipeg",
         {"Central Standard Time", "CST", "Central Daylight Time", "CDT"}},
        {"America/Yakutat",
         {"Alaska Standard Time", "AKST", "Alaska Daylight Time", "AKDT"}},
        {"America/Yellowknife",
         {"Mountain Standard Time", "MST", "Mountain Daylight Time", "MDT"}},
        {"Antarctica/Casey",
         {"Australian Western Standard Time",
          "AWST",
          "Australian Western Daylight Time",
          "AWDT"}},
        {"Antarctica/Davis",
         {"Davis Time", "DAVT", "Davis Summer Time", "DAVST"}},
        {"Antarctica/DumontDUrville",
         {"Dumont-d'Urville Time",
          "DDUT",
          "Dumont-d'Urville Summer Time",
          "DDUST"}},
        {"Antarctica/Macquarie",
         {"Australian Eastern Standard Time (Macquarie)",
          "AEST",
          "Australian Eastern Daylight Time (Macquarie)",
          "AEDT"}},
        {"Antarctica/Mawson",
         {"Mawson Time", "MAWT", "Mawson Summer Time", "MAWST"}},
        {"Antarctica/McMurdo",
         {"New Zealand Standard Time",
          "NZST",
          "New Zealand Daylight Time",
          "NZDT"}},
        {"Antarctica/Palmer",
         {"Chile Time", "CLT", "Chile Summer Time", "CLST"}},
        {"Antarctica/Rothera",
         {"Rothera Time", "ROTT", "Rothera Summer Time", "ROTST"}},
        {"Antarctica/South_Pole",
         {"New Zealand Standard Time",
          "NZST",
          "New Zealand Daylight Time",
          "NZDT"}},
        {"Antarctica/Syowa",
         {"Syowa Time", "SYOT", "Syowa Summer Time", "SYOST"}},
        {"Antarctica/Troll",
         {"Coordinated Universal Time",
          "UTC",
          "Central European Summer Time",
          "CEST"}},
        {"Antarctica/Vostok",
         {"Vostok Time", "VOST", "Vostok Summer Time", "VOSST"}},
        {"Arctic/Longyearbyen",
         {"Central European Time",
          "CET",
          "Central European Summer Time",
          "CEST"}},
        {"Asia/Aden",
         {"Arabia Standard Time", "AST", "Arabia Daylight Time", "ADT"}},
        {"Asia/Almaty",
         {"Alma-Ata Time", "ALMT", "Alma-Ata Summer Time", "ALMST"}},
        {"Asia/Amman",
         {"Eastern European Time",
          "EET",
          "Eastern European Summer Time",
          "EEST"}},
        {"Asia/Anadyr", {"Anadyr Time", "ANAT", "Anadyr Summer Time", "ANAST"}},
        {"Asia/Aqtau", {"Aqtau Time", "AQTT", "Aqtau Summer Time", "AQTST"}},
        {"Asia/Aqtobe", {"Aqtobe Time", "AQTT", "Aqtobe Summer Time", "AQTST"}},
        {"Asia/Ashgabat",
         {"Turkmenistan Time", "TMT", "Turkmenistan Summer Time", "TMST"}},
        {"Asia/Ashkhabad",
         {"Turkmenistan Time", "TMT", "Turkmenistan Summer Time", "TMST"}},
        {"Asia/Baghdad",
         {"Arabia Standard Time", "AST", "Arabia Daylight Time", "ADT"}},
        {"Asia/Bahrain",
         {"Arabia Standard Time", "AST", "Arabia Daylight Time", "ADT"}},
        {"Asia/Baku",
         {"Azerbaijan Time", "AZT", "Azerbaijan Summer Time", "AZST"}},
        {"Asia/Bangkok",
         {"Indochina Time", "ICT", "Indochina Summer Time", "ICST"}},
        {"Asia/Beirut",
         {"Eastern European Time",
          "EET",
          "Eastern European Summer Time",
          "EEST"}},
        {"Asia/Bishkek",
         {"Kirgizstan Time", "KGT", "Kirgizstan Summer Time", "KGST"}},
        {"Asia/Brunei", {"Brunei Time", "BNT", "Brunei Summer Time", "BNST"}},
        {"Asia/Calcutta",
         {"India Standard Time", "IST", "India Daylight Time", "IDT"}},
        {"Asia/Chita",
         {"Yakutsk Time", "YAKT", "Yakutsk Summer Time", "YAKST"}},
        {"Asia/Choibalsan",
         {"Choibalsan Time", "CHOT", "Choibalsan Summer Time", "CHOST"}},
        {"Asia/Chongqing",
         {"China Standard Time", "CST", "China Daylight Time", "CDT"}},
        {"Asia/Chungking",
         {"China Standard Time", "CST", "China Daylight Time", "CDT"}},
        {"Asia/Colombo",
         {"India Standard Time", "IST", "India Daylight Time", "IDT"}},
        {"Asia/Dacca",
         {"Bangladesh Time", "BDT", "Bangladesh Summer Time", "BDST"}},
        {"Asia/Dhaka",
         {"Bangladesh Time", "BDT", "Bangladesh Summer Time", "BDST"}},
        {"Asia/Dili",
         {"Timor-Leste Time", "TLT", "Timor-Leste Summer Time", "TLST"}},
        {"Asia/Damascus",
         {"Eastern European Time",
          "EET",
          "Eastern European Summer Time",
          "EEST"}},
        {"Asia/Dubai",
         {"Gulf Standard Time", "GST", "Gulf Daylight Time", "GDT"}},
        {"Asia/Dushanbe",
         {"Tajikistan Time", "TJT", "Tajikistan Summer Time", "TJST"}},
        {"Asia/Gaza",
         {"Eastern European Time",
          "EET",
          "Eastern European Summer Time",
          "EEST"}},
        {"Asia/Harbin",
         {"China Standard Time", "CST", "China Daylight Time", "CDT"}},
        {"Asia/Hebron",
         {"Eastern European Time",
          "EET",
          "Eastern European Summer Time",
          "EEST"}},
        {"Asia/Ho_Chi_Minh",
         {"Indochina Time", "ICT", "Indochina Summer Time", "ICST"}},
        {"Asia/Hong_Kong",
         {"Hong Kong Time", "HKT", "Hong Kong Summer Time", "HKST"}},
        {"Asia/Hovd", {"Hovd Time", "HOVT", "Hovd Summer Time", "HOVST"}},
        {"Asia/Irkutsk",
         {"Irkutsk Time", "IRKT", "Irkutsk Summer Time", "IRKST"}},
        {"Asia/Istanbul", {"Turkey Time", "TRT", "Turkey Summer Time", "TRST"}},
        {"Asia/Jakarta",
         {"West Indonesia Time", "WIB", "West Indonesia Summer Time", "WIST"}},
        {"Asia/Jayapura",
         {"East Indonesia Time", "WIT", "East Indonesia Summer Time", "EIST"}},
        {"Asia/Kabul",
         {"Afghanistan Time", "AFT", "Afghanistan Summer Time", "AFST"}},
        {"Asia/Kamchatka",
         {"Petropavlovsk-Kamchatski Time",
          "PETT",
          "Petropavlovsk-Kamchatski Summer Time",
          "PETST"}},
        {"Asia/Karachi",
         {"Pakistan Time", "PKT", "Pakistan Summer Time", "PKST"}},
        {"Asia/Kashgar",
         {"Xinjiang Standard Time", "XJT", "Xinjiang Daylight Time", "XJDT"}},
        {"Asia/Kathmandu", {"Nepal Time", "NPT", "Nepal Summer Time", "NPST"}},
        {"Asia/Katmandu", {"Nepal Time", "NPT", "Nepal Summer Time", "NPST"}},
        {"Asia/Khandyga",
         {"Yakutsk Time", "YAKT", "Yakutsk Summer Time", "YAKST"}},
        {"Asia/Kolkata",
         {"India Standard Time", "IST", "India Daylight Time", "IDT"}},
        {"Asia/Krasnoyarsk",
         {"Krasnoyarsk Time", "KRAT", "Krasnoyarsk Summer Time", "KRAST"}},
        {"Asia/Kuala_Lumpur",
         {"Malaysia Time", "MYT", "Malaysia Summer Time", "MYST"}},
        {"Asia/Kuching",
         {"Malaysia Time", "MYT", "Malaysia Summer Time", "MYST"}},
        {"Asia/Kuwait",
         {"Arabia Standard Time", "AST", "Arabia Daylight Time", "ADT"}},
        {"Asia/Macao",
         {"China Standard Time", "CST", "China Daylight Time", "CDT"}},
        {"Asia/Macau",
         {"China Standard Time", "CST", "China Daylight Time", "CDT"}},
        {"Asia/Magadan",
         {"Magadan Time", "MAGT", "Magadan Summer Time", "MAGST"}},
        {"Asia/Makassar",
         {"Central Indonesia Time",
          "WITA",
          "Central Indonesia Summer Time",
          "CIST"}},
        {"Asia/Manila",
         {"Philippines Standard Time",
          "PST",
          "Philippines Daylight Time",
          "PDT"}},
        {"Asia/Muscat",
         {"Gulf Standard Time", "GST", "Gulf Daylight Time", "GDT"}},
        {"Asia/Nicosia",
         {"Eastern European Time",
          "EET",
          "Eastern European Summer Time",
          "EEST"}},
        {"Asia/Novokuznetsk",
         {"Krasnoyarsk Time", "KRAT", "Krasnoyarsk Summer Time", "KRAST"}},
        {"Asia/Novosibirsk",
         {"Novosibirsk Time", "NOVT", "Novosibirsk Summer Time", "NOVST"}},
        {"Asia/Oral", {"Oral Time", "ORAT", "Oral Summer Time", "ORAST"}},
        {"Asia/Omsk", {"Omsk Time", "OMST", "Omsk Summer Time", "OMSST"}},
        {"Asia/Phnom_Penh",
         {"Indochina Time", "ICT", "Indochina Summer Time", "ICST"}},
        {"Asia/Pontianak",
         {"West Indonesia Time", "WIB", "West Indonesia Summer Time", "WIST"}},
        {"Asia/Pyongyang",
         {"Korea Standard Time", "KST", "Korea Daylight Time", "KDT"}},
        {"Asia/Qatar",
         {"Arabia Standard Time", "AST", "Arabia Daylight Time", "ADT"}},
        {"Asia/Qostanay",
         {"Kostanay Standard Time", "QOST", "Kostanay Summer Time", "QOSST"}},
        {"Asia/Qyzylorda",
         {"Qyzylorda Time", "QYZT", "Qyzylorda Summer Time", "QYZST"}},
        {"Asia/Rangoon",
         {"Myanmar Time", "MMT", "Myanmar Summer Time", "MMST"}},
        {"Asia/Riyadh",
         {"Arabia Standard Time", "AST", "Arabia Daylight Time", "ADT"}},
        {"Asia/Saigon",
         {"Indochina Time", "ICT", "Indochina Summer Time", "ICST"}},
        {"Asia/Sakhalin",
         {"Sakhalin Time", "SAKT", "Sakhalin Summer Time", "SAKST"}},
        {"Asia/Samarkand",
         {"Uzbekistan Time", "UZT", "Uzbekistan Summer Time", "UZST"}},
        {"Asia/Seoul",
         {"Korea Standard Time", "KST", "Korea Daylight Time", "KDT"}},
        {"Asia/Singapore",
         {"Singapore Time", "SGT", "Singapore Summer Time", "SGST"}},
        {"Asia/Srednekolymsk",
         {"Srednekolymsk Time",
          "SRET",
          "Srednekolymsk Daylight Time",
          "SREDT"}},
        {"Asia/Taipei",
         {"China Standard Time", "CST", "China Daylight Time", "CDT"}},
        {"Asia/Tel_Aviv",
         {"Israel Standard Time", "IST", "Israel Daylight Time", "IDT"}},
        {"Asia/Tashkent",
         {"Uzbekistan Time", "UZT", "Uzbekistan Summer Time", "UZST"}},
        {"Asia/Tbilisi",
         {"Georgia Time", "GET", "Georgia Summer Time", "GEST"}},
        {"Asia/Tehran",
         {"Iran Standard Time", "IRST", "Iran Daylight Time", "IRDT"}},
        {"Asia/Thimbu", {"Bhutan Time", "BTT", "Bhutan Summer Time", "BTST"}},
        {"Asia/Thimphu", {"Bhutan Time", "BTT", "Bhutan Summer Time", "BTST"}},
        {"Asia/Ujung_Pandang",
         {"Central Indonesia Time",
          "WITA",
          "Central Indonesia Summer Time",
          "CIST"}},
        {"Asia/Ulaanbaatar",
         {"Ulaanbaatar Time", "ULAT", "Ulaanbaatar Summer Time", "ULAST"}},
        {"Asia/Ulan_Bator",
         {"Ulaanbaatar Time", "ULAT", "Ulaanbaatar Summer Time", "ULAST"}},
        {"Asia/Urumqi",
         {"Xinjiang Standard Time", "XJT", "Xinjiang Daylight Time", "XJDT"}},
        {"Asia/Ust-Nera",
         {"Ust-Nera Time", "VLAT", "Ust-Nera Summer Time", "VLAST"}},
        {"Asia/Vientiane",
         {"Indochina Time", "ICT", "Indochina Summer Time", "ICST"}},
        {"Asia/Vladivostok",
         {"Vladivostok Time", "VLAT", "Vladivostok Summer Time", "VLAST"}},
        {"Asia/Yakutsk",
         {"Yakutsk Time", "YAKT", "Yakutsk Summer Time", "YAKST"}},
        {"Asia/Yangon", {"Myanmar Time", "MMT", "Myanmar Summer Time", "MMST"}},
        {"Asia/Yekaterinburg",
         {"Yekaterinburg Time", "YEKT", "Yekaterinburg Summer Time", "YEKST"}},
        {"Asia/Yerevan",
         {"Armenia Time", "AMT", "Armenia Summer Time", "AMST"}},
        {"Atlantic/Azores",
         {"Azores Time", "AZOT", "Azores Summer Time", "AZOST"}},
        {"Atlantic/Bermuda",
         {"Atlantic Standard Time", "AST", "Atlantic Daylight Time", "ADT"}},
        {"Atlantic/Canary",
         {"Western European Time",
          "WET",
          "Western European Summer Time",
          "WEST"}},
        {"Atlantic/Cape_Verde",
         {"Cape Verde Time", "CVT", "Cape Verde Summer Time", "CVST"}},
        {"Atlantic/Faeroe",
         {"Western European Time",
          "WET",
          "Western European Summer Time",
          "WEST"}},
        {"Atlantic/Faroe",
         {"Western European Time",
          "WET",
          "Western European Summer Time",
          "WEST"}},
        {"Atlantic/Jan_Mayen",
         {"Central European Time",
          "CET",
          "Central European Summer Time",
          "CEST"}},
        {"Atlantic/Madeira",
         {"Western European Time",
          "WET",
          "Western European Summer Time",
          "WEST"}},
        {"Atlantic/Reykjavik",
         {"Greenwich Mean Time", "GMT", "Greenwich Mean Time", "GMT"}},
        {"Atlantic/South_Georgia",
         {"South Georgia Standard Time",
          "GST",
          "South Georgia Daylight Time",
          "GDT"}},
        {"Atlantic/St_Helena",
         {"Greenwich Mean Time", "GMT", "Greenwich Mean Time", "GMT"}},
        {"Atlantic/Stanley",
         {"Falkland Is. Time", "FKT", "Falkland Is. Summer Time", "FKST"}},
        {"Australia/ACT",
         {"Australian Eastern Standard Time (New South Wales)",
          "AEST",
          "Australian Eastern Daylight Time (New South Wales)",
          "AEDT"}},
        {"Australia/Adelaide",
         {"Australian Central Standard Time (South Australia)",
          "ACST",
          "Australian Central Daylight Time (South Australia)",
          "ACDT"}},
        {"Australia/Brisbane",
         {"Australian Eastern Standard Time (Queensland)",
          "AEST",
          "Australian Eastern Daylight Time (Queensland)",
          "AEDT"}},
        {"Australia/Broken_Hill",
         {"Australian Central Standard Time (South Australia/New South Wales)",
          "ACST",
          "Australian Central Daylight Time (South Australia/New South Wales)",
          "ACDT"}},
        {"Australia/Canberra",
         {"Australian Eastern Standard Time (New South Wales)",
          "AEST",
          "Australian Eastern Daylight Time (New South Wales)",
          "AEDT"}},
        {"Australia/Currie",
         {"Australian Eastern Standard Time (New South Wales)",
          "AEST",
          "Australian Eastern Daylight Time (New South Wales)",
          "AEDT"}},
        {"Australia/Darwin",
         {"Australian Central Standard Time (Northern Territory)",
          "ACST",
          "Australian Central Daylight Time (Northern Territory)",
          "ACDT"}},
        {"Australia/Eucla",
         {"Australian Central Western Standard Time",
          "ACWST",
          "Australian Central Western Daylight Time",
          "ACWDT"}},
        {"Australia/Hobart",
         {"Australian Eastern Standard Time (Tasmania)",
          "AEST",
          "Australian Eastern Daylight Time (Tasmania)",
          "AEDT"}},
        {"Australia/LHI",
         {"Lord Howe Standard Time",
          "LHST",
          "Lord Howe Daylight Time",
          "LHDT"}},
        {"Australia/Lindeman",
         {"Australian Eastern Standard Time (Queensland)",
          "AEST",
          "Australian Eastern Daylight Time (Queensland)",
          "AEDT"}},
        {"Australia/Lord_Howe",
         {"Lord Howe Standard Time",
          "LHST",
          "Lord Howe Daylight Time",
          "LHDT"}},
        {"Australia/Melbourne",
         {"Australian Eastern Standard Time (Victoria)",
          "AEST",
          "Australian Eastern Daylight Time (Victoria)",
          "AEDT"}},
        {"Australia/North",
         {"Australian Central Standard Time (Northern Territory)",
          "ACST",
          "Australian Central Daylight Time (Northern Territory)",
          "ACDT"}},
        {"Australia/NSW",
         {"Australian Eastern Standard Time (New South Wales)",
          "AEST",
          "Australian Eastern Daylight Time (New South Wales)",
          "AEDT"}},
        {"Australia/Perth",
         {"Australian Western Standard Time",
          "AWST",
          "Australian Western Daylight Time",
          "AWDT"}},
        {"Australia/Queensland",
         {"Australian Eastern Standard Time (Queensland)",
          "AEST",
          "Australian Eastern Daylight Time (Queensland)",
          "AEDT"}},
        {"Australia/South",
         {"Australian Central Standard Time (South Australia)",
          "ACST",
          "Australian Central Daylight Time (South Australia)",
          "ACDT"}},
        {"Australia/Sydney",
         {"Australian Eastern Standard Time (New South Wales)",
          "AEST",
          "Australian Eastern Daylight Time (New South Wales)",
          "AEDT"}},
        {"Australia/Tasmania",
         {"Australian Eastern Standard Time (Tasmania)",
          "AEST",
          "Australian Eastern Daylight Time (Tasmania)",
          "AEDT"}},
        {"Australia/Victoria",
         {"Australian Eastern Standard Time (Victoria)",
          "AEST",
          "Australian Eastern Daylight Time (Victoria)",
          "AEDT"}},
        {"Australia/West",
         {"Australian Western Standard Time",
          "AWST",
          "Australian Western Daylight Time",
          "AWDT"}},
        {"Australia/Yancowinna",
         {"Australian Central Standard Time (South Australia/New South Wales)",
          "ACST",
          "Australian Central Daylight Time (South Australia/New South Wales)",
          "ACDT"}},
        {"BET", {"Brasilia Time", "BRT", "Brasilia Summer Time", "BRST"}},
        {"BST", {"Bangladesh Time", "BDT", "Bangladesh Summer Time", "BDST"}},
        {"Brazil/Acre", {"Acre Time", "ACT", "Acre Summer Time", "ACST"}},
        {"Brazil/DeNoronha",
         {"Fernando de Noronha Time",
          "FNT",
          "Fernando de Noronha Summer Time",
          "FNST"}},
        {"Brazil/East",
         {"Brasilia Time", "BRT", "Brasilia Summer Time", "BRST"}},
        {"Brazil/West", {"Amazon Time", "AMT", "Amazon Summer Time", "AMST"}},
        {"Canada/Atlantic",
         {"Atlantic Standard Time", "AST", "Atlantic Daylight Time", "ADT"}},
        {"Canada/Central",
         {"Central Standard Time", "CST", "Central Daylight Time", "CDT"}},
        {"Canada/Eastern",
         {"Eastern Standard Time", "EST", "Eastern Daylight Time", "EDT"}},
        {"Canada/Mountain",
         {"Mountain Standard Time", "MST", "Mountain Daylight Time", "MDT"}},
        {"Canada/Newfoundland",
         {"Newfoundland Standard Time",
          "NST",
          "Newfoundland Daylight Time",
          "NDT"}},
        {"Canada/Pacific",
         {"Pacific Standard Time", "PST", "Pacific Daylight Time", "PDT"}},
        {"Canada/Yukon",
         {"Mountain Standard Time", "MST", "Mountain Daylight Time", "MDT"}},
        {"Canada/Saskatchewan",
         {"Central Standard Time", "CST", "Central Daylight Time", "CDT"}},
        {"CAT",
         {"Central African Time",
          "CAT",
          "Central African Summer Time",
          "CAST"}},
        {"CET",
         {"Central European Time",
          "CET",
          "Central European Summer Time",
          "CEST"}},
        {"Chile/Continental",
         {"Chile Time", "CLT", "Chile Summer Time", "CLST"}},
        {"Chile/EasterIsland",
         {"Easter Is. Time", "EAST", "Easter Is. Summer Time", "EASST"}},
        {"CST6CDT",
         {"Central Standard Time", "CST", "Central Daylight Time", "CDT"}},
        {"Cuba", {"Cuba Standard Time", "CST", "Cuba Daylight Time", "CDT"}},
        {"EAT",
         {"Eastern African Time",
          "EAT",
          "Eastern African Summer Time",
          "EAST"}},
        {"EET",
         {"Eastern European Time",
          "EET",
          "Eastern European Summer Time",
          "EEST"}},
        {"Egypt",
         {"Eastern European Time",
          "EET",
          "Eastern European Summer Time",
          "EEST"}},
        {"Eire", {"Greenwich Mean Time", "GMT", "Irish Summer Time", "IST"}},
        {"EST5EDT",
         {"Eastern Standard Time", "EST", "Eastern Daylight Time", "EDT"}},
        {"Etc/Greenwich",
         {"Greenwich Mean Time", "GMT", "Greenwich Mean Time", "GMT"}},
        {"Etc/UCT",
         {"Coordinated Universal Time",
          "UTC",
          "Coordinated Universal Time",
          "UTC"}},
        {"Etc/Universal",
         {"Coordinated Universal Time",
          "UTC",
          "Coordinated Universal Time",
          "UTC"}},
        {"Etc/UTC",
         {"Coordinated Universal Time",
          "UTC",
          "Coordinated Universal Time",
          "UTC"}},
        {"Etc/Zulu",
         {"Coordinated Universal Time",
          "UTC",
          "Coordinated Universal Time",
          "UTC"}},
        {"Europe/Amsterdam",
         {"Central European Time",
          "CET",
          "Central European Summer Time",
          "CEST"}},
        {"Europe/Andorra",
         {"Central European Time",
          "CET",
          "Central European Summer Time",
          "CEST"}},
        {"Europe/Athens",
         {"Eastern European Time",
          "EET",
          "Eastern European Summer Time",
          "EEST"}},
        {"Europe/Belfast",
         {"Greenwich Mean Time", "GMT", "British Summer Time", "BST"}},
        {"Europe/Belgrade",
         {"Central European Time",
          "CET",
          "Central European Summer Time",
          "CEST"}},
        {"Europe/Berlin",
         {"Central European Time",
          "CET",
          "Central European Summer Time",
          "CEST"}},
        {"Europe/Bratislava",
         {"Central European Time",
          "CET",
          "Central European Summer Time",
          "CEST"}},
        {"Europe/Brussels",
         {"Central European Time",
          "CET",
          "Central European Summer Time",
          "CEST"}},
        {"Europe/Budapest",
         {"Central European Time",
          "CET",
          "Central European Summer Time",
          "CEST"}},
        {"Europe/Busingen",
         {"Central European Time",
          "CET",
          "Central European Summer Time",
          "CEST"}},
        {"Europe/Chisinau",
         {"Eastern European Time",
          "EET",
          "Eastern European Summer Time",
          "EEST"}},
        {"Europe/Copenhagen",
         {"Central European Time",
          "CET",
          "Central European Summer Time",
          "CEST"}},
        {"Europe/Dublin",
         {"Greenwich Mean Time", "GMT", "Irish Summer Time", "IST"}},
        {"Europe/Gibraltar",
         {"Central European Time",
          "CET",
          "Central European Summer Time",
          "CEST"}},
        {"Europe/Guernsey",
         {"Greenwich Mean Time", "GMT", "British Summer Time", "BST"}},
        {"Europe/Helsinki",
         {"Eastern European Time",
          "EET",
          "Eastern European Summer Time",
          "EEST"}},
        {"Europe/Isle_of_Man",
         {"Greenwich Mean Time", "GMT", "British Summer Time", "BST"}},
        {"Europe/Istanbul",
         {"Turkey Time", "TRT", "Turkey Summer Time", "TRST"}},
        {"Europe/Jersey",
         {"Greenwich Mean Time", "GMT", "British Summer Time", "BST"}},
        {"Europe/Kaliningrad",
         {"Eastern European Time",
          "EET",
          "Eastern European Summer Time",
          "EEST"}},
        {"Europe/Kiev",
         {"Eastern European Time",
          "EET",
          "Eastern European Summer Time",
          "EEST"}},
        {"Europe/Lisbon",
         {"Western European Time",
          "WET",
          "Western European Summer Time",
          "WEST"}},
        {"Europe/Ljubljana",
         {"Central European Time",
          "CET",
          "Central European Summer Time",
          "CEST"}},
        {"Europe/London",
         {"Greenwich Mean Time", "GMT", "British Summer Time", "BST"}},
        {"Europe/Luxembourg",
         {"Central European Time",
          "CET",
          "Central European Summer Time",
          "CEST"}},
        {"Europe/Madrid",
         {"Central European Time",
          "CET",
          "Central European Summer Time",
          "CEST"}},
        {"Europe/Malta",
         {"Central European Time",
          "CET",
          "Central European Summer Time",
          "CEST"}},
        {"Europe/Mariehamn",
         {"Eastern European Time",
          "EET",
          "Eastern European Summer Time",
          "EEST"}},
        {"Europe/Minsk",
         {"Moscow Standard Time", "MSK", "Moscow Daylight Time", "MSD"}},
        {"Europe/Monaco",
         {"Central European Time",
          "CET",
          "Central European Summer Time",
          "CEST"}},
        {"Europe/Moscow",
         {"Moscow Standard Time", "MSK", "Moscow Daylight Time", "MSD"}},
        {"Europe/Nicosia",
         {"Eastern European Time",
          "EET",
          "Eastern European Summer Time",
          "EEST"}},
        {"Europe/Oslo",
         {"Central European Time",
          "CET",
          "Central European Summer Time",
          "CEST"}},
        {"Europe/Podgorica",
         {"Central European Time",
          "CET",
          "Central European Summer Time",
          "CEST"}},
        {"Europe/Prague",
         {"Central European Time",
          "CET",
          "Central European Summer Time",
          "CEST"}},
        {"Europe/Riga",
         {"Eastern European Time",
          "EET",
          "Eastern European Summer Time",
          "EEST"}},
        {"Europe/Rome",
         {"Central European Time",
          "CET",
          "Central European Summer Time",
          "CEST"}},
        {"Europe/Samara",
         {"Samara Time", "SAMT", "Samara Summer Time", "SAMST"}},
        {"Europe/San_Marino",
         {"Central European Time",
          "CET",
          "Central European Summer Time",
          "CEST"}},
        {"Europe/Sarajevo",
         {"Central European Time",
          "CET",
          "Central European Summer Time",
          "CEST"}},
        {"Europe/Simferopol",
         {"Moscow Standard Time", "MSK", "Moscow Daylight Time", "MSD"}},
        {"Europe/Skopje",
         {"Central European Time",
          "CET",
          "Central European Summer Time",
          "CEST"}},
        {"Europe/Sofia",
         {"Eastern European Time",
          "EET",
          "Eastern European Summer Time",
          "EEST"}},
        {"Europe/Stockholm",
         {"Central European Time",
          "CET",
          "Central European Summer Time",
          "CEST"}},
        {"Europe/Tallinn",
         {"Eastern European Time",
          "EET",
          "Eastern European Summer Time",
          "EEST"}},
        {"Europe/Tirane",
         {"Central European Time",
          "CET",
          "Central European Summer Time",
          "CEST"}},
        {"Europe/Tiraspol",
         {"Eastern European Time",
          "EET",
          "Eastern European Summer Time",
          "EEST"}},
        {"Europe/Uzhgorod",
         {"Eastern European Time",
          "EET",
          "Eastern European Summer Time",
          "EEST"}},
        {"Europe/Vaduz",
         {"Central European Time",
          "CET",
          "Central European Summer Time",
          "CEST"}},
        {"Europe/Vatican",
         {"Central European Time",
          "CET",
          "Central European Summer Time",
          "CEST"}},
        {"Europe/Vienna",
         {"Central European Time",
          "CET",
          "Central European Summer Time",
          "CEST"}},
        {"Europe/Vilnius",
         {"Eastern European Time",
          "EET",
          "Eastern European Summer Time",
          "EEST"}},
        {"Europe/Volgograd",
         {"Moscow Standard Time", "MSK", "Moscow Daylight Time", "MSD"}},
        {"Europe/Warsaw",
         {"Central European Time",
          "CET",
          "Central European Summer Time",
          "CEST"}},
        {"Europe/Zagreb",
         {"Central European Time",
          "CET",
          "Central European Summer Time",
          "CEST"}},
        {"Europe/Zaporozhye",
         {"Eastern European Time",
          "EET",
          "Eastern European Summer Time",
          "EEST"}},
        {"Europe/Zurich",
         {"Central European Time",
          "CET",
          "Central European Summer Time",
          "CEST"}},
        {"GB", {"Greenwich Mean Time", "GMT", "British Summer Time", "BST"}},
        {"GB-Eire",
         {"Greenwich Mean Time", "GMT", "British Summer Time", "BST"}},
        {"Greenwich",
         {"Greenwich Mean Time", "GMT", "Greenwich Mean Time", "GMT"}},
        {"Hongkong",
         {"Hong Kong Time", "HKT", "Hong Kong Summer Time", "HKST"}},
        {"Iceland",
         {"Greenwich Mean Time", "GMT", "Greenwich Mean Time", "GMT"}},
        {"Iran", {"Iran Standard Time", "IRST", "Iran Daylight Time", "IRDT"}},
        {"IST", {"India Standard Time", "IST", "India Daylight Time", "IDT"}},
        {"Indian/Antananarivo",
         {"Eastern African Time",
          "EAT",
          "Eastern African Summer Time",
          "EAST"}},
        {"Indian/Chagos",
         {"Indian Ocean Territory Time",
          "IOT",
          "Indian Ocean Territory Summer Time",
          "IOST"}},
        {"Indian/Christmas",
         {"Christmas Island Time",
          "CXT",
          "Christmas Island Summer Time",
          "CXST"}},
        {"Indian/Cocos",
         {"Cocos Islands Time", "CCT", "Cocos Islands Summer Time", "CCST"}},
        {"Indian/Comoro",
         {"Eastern African Time",
          "EAT",
          "Eastern African Summer Time",
          "EAST"}},
        {"Indian/Kerguelen",
         {"French Southern & Antarctic Lands Time",
          "TFT",
          "French Southern & Antarctic Lands Summer Time",
          "TFST"}},
        {"Indian/Mahe",
         {"Seychelles Time", "SCT", "Seychelles Summer Time", "SCST"}},
        {"Indian/Maldives",
         {"Maldives Time", "MVT", "Maldives Summer Time", "MVST"}},
        {"Indian/Mauritius",
         {"Mauritius Time", "MUT", "Mauritius Summer Time", "MUST"}},
        {"Indian/Mayotte",
         {"Eastern African Time",
          "EAT",
          "Eastern African Summer Time",
          "EAST"}},
        {"Indian/Reunion",
         {"Reunion Time", "RET", "Reunion Summer Time", "REST"}},
        {"Israel",
         {"Israel Standard Time", "IST", "Israel Daylight Time", "IDT"}},
        {"Jamaica",
         {"Eastern Standard Time", "EST", "Eastern Daylight Time", "EDT"}},
        {"Japan", {"Japan Standard Time", "JST", "Japan Daylight Time", "JDT"}},
        {"Kwajalein",
         {"Marshall Islands Time",
          "MHT",
          "Marshall Islands Summer Time",
          "MHST"}},
        {"Libya",
         {"Eastern European Time",
          "EET",
          "Eastern European Summer Time",
          "EEST"}},
        {"MET",
         {"Middle Europe Time", "MET", "Middle Europe Summer Time", "MEST"}},
        {"Mexico/BajaNorte",
         {"Pacific Standard Time", "PST", "Pacific Daylight Time", "PDT"}},
        {"Mexico/BajaSur",
         {"Mountain Standard Time", "MST", "Mountain Daylight Time", "MDT"}},
        {"Mexico/General",
         {"Central Standard Time", "CST", "Central Daylight Time", "CDT"}},
        {"MIT",
         {"West Samoa Standard Time",
          "WSST",
          "West Samoa Daylight Time",
          "WSDT"}},
        {"MST7MDT",
         {"Mountain Standard Time", "MST", "Mountain Daylight Time", "MDT"}},
        {"Navajo",
         {"Mountain Standard Time", "MST", "Mountain Daylight Time", "MDT"}},
        {"NET", {"Armenia Time", "AMT", "Armenia Summer Time", "AMST"}},
        {"NST",
         {"New Zealand Standard Time",
          "NZST",
          "New Zealand Daylight Time",
          "NZDT"}},
        {"NZ",
         {"New Zealand Standard Time",
          "NZST",
          "New Zealand Daylight Time",
          "NZDT"}},
        {"NZ-CHAT",
         {"Chatham Standard Time", "CHAST", "Chatham Daylight Time", "CHADT"}},
        {"PLT", {"Pakistan Time", "PKT", "Pakistan Summer Time", "PKST"}},
        {"Portugal",
         {"Western European Time",
          "WET",
          "Western European Summer Time",
          "WEST"}},
        {"PRT",
         {"Atlantic Standard Time", "AST", "Atlantic Daylight Time", "ADT"}},
        {"Pacific/Apia",
         {"West Samoa Standard Time",
          "WSST",
          "West Samoa Daylight Time",
          "WSDT"}},
        {"Pacific/Auckland",
         {"New Zealand Standard Time",
          "NZST",
          "New Zealand Daylight Time",
          "NZDT"}},
        {"Pacific/Bougainville",
         {"Bougainville Standard Time",
          "BST",
          "Bougainville Daylight Time",
          "BST"}},
        {"Pacific/Chatham",
         {"Chatham Standard Time", "CHAST", "Chatham Daylight Time", "CHADT"}},
        {"Pacific/Chuuk", {"Chuuk Time", "CHUT", "Chuuk Summer Time", "CHUST"}},
        {"Pacific/Easter",
         {"Easter Is. Time", "EAST", "Easter Is. Summer Time", "EASST"}},
        {"Pacific/Efate",
         {"Vanuatu Time", "VUT", "Vanuatu Summer Time", "VUST"}},
        {"Pacific/Enderbury",
         {"Phoenix Is. Time", "PHOT", "Phoenix Is. Summer Time", "PHOST"}},
        {"Pacific/Fakaofo",
         {"Tokelau Time", "TKT", "Tokelau Summer Time", "TKST"}},
        {"Pacific/Fiji", {"Fiji Time", "FJT", "Fiji Summer Time", "FJST"}},
        {"Pacific/Funafuti",
         {"Tuvalu Time", "TVT", "Tuvalu Summer Time", "TVST"}},
        {"Pacific/Galapagos",
         {"Galapagos Time", "GALT", "Galapagos Summer Time", "GALST"}},
        {"Pacific/Gambier",
         {"Gambier Time", "GAMT", "Gambier Summer Time", "GAMST"}},
        {"Pacific/Guadalcanal",
         {"Solomon Is. Time", "SBT", "Solomon Is. Summer Time", "SBST"}},
        {"Pacific/Guam",
         {"Chamorro Standard Time", "ChST", "Chamorro Daylight Time", "ChDT"}},
        {"Pacific/Johnston",
         {"Hawaii Standard Time", "HST", "Hawaii Daylight Time", "HDT"}},
        {"Pacific/Kiritimati",
         {"Line Is. Time", "LINT", "Line Is. Summer Time", "LINST"}},
        {"Pacific/Kosrae",
         {"Kosrae Time", "KOST", "Kosrae Summer Time", "KOSST"}},
        {"Pacific/Kwajalein",
         {"Marshall Islands Time",
          "MHT",
          "Marshall Islands Summer Time",
          "MHST"}},
        {"Pacific/Majuro",
         {"Marshall Islands Time",
          "MHT",
          "Marshall Islands Summer Time",
          "MHST"}},
        {"Pacific/Marquesas",
         {"Marquesas Time", "MART", "Marquesas Summer Time", "MARST"}},
        {"Pacific/Midway",
         {"Samoa Standard Time", "SST", "Samoa Daylight Time", "SDT"}},
        {"Pacific/Nauru", {"Nauru Time", "NRT", "Nauru Summer Time", "NRST"}},
        {"Pacific/Niue", {"Niue Time", "NUT", "Niue Summer Time", "NUST"}},
        {"Pacific/Norfolk",
         {"Norfolk Time", "NFT", "Norfolk Summer Time", "NFST"}},
        {"Pacific/Noumea",
         {"New Caledonia Time", "NCT", "New Caledonia Summer Time", "NCST"}},
        {"Pacific/Pago_Pago",
         {"Samoa Standard Time", "SST", "Samoa Daylight Time", "SDT"}},
        {"Pacific/Palau", {"Palau Time", "PWT", "Palau Summer Time", "PWST"}},
        {"Pacific/Pitcairn",
         {"Pitcairn Standard Time", "PST", "Pitcairn Daylight Time", "PDT"}},
        {"Pacific/Pohnpei",
         {"Pohnpei Time", "PONT", "Pohnpei Summer Time", "PONST"}},
        {"Pacific/Ponape",
         {"Pohnpei Time", "PONT", "Pohnpei Summer Time", "PONST"}},
        {"Pacific/Port_Moresby",
         {"Papua New Guinea Time",
          "PGT",
          "Papua New Guinea Summer Time",
          "PGST"}},
        {"Pacific/Rarotonga",
         {"Cook Is. Time", "CKT", "Cook Is. Summer Time", "CKHST"}},
        {"Pacific/Saipan",
         {"Chamorro Standard Time", "ChST", "Chamorro Daylight Time", "ChDT"}},
        {"Pacific/Samoa",
         {"Samoa Standard Time", "SST", "Samoa Daylight Time", "SDT"}},
        {"Pacific/Tahiti",
         {"Tahiti Time", "TAHT", "Tahiti Summer Time", "TAHST"}},
        {"Pacific/Tarawa",
         {"Gilbert Is. Time", "GILT", "Gilbert Is. Summer Time", "GILST"}},
        {"Pacific/Tongatapu",
         {"Tonga Time", "TOT", "Tonga Summer Time", "TOST"}},
        {"Pacific/Truk", {"Chuuk Time", "CHUT", "Chuuk Summer Time", "CHUST"}},
        {"Pacific/Wake", {"Wake Time", "WAKT", "Wake Summer Time", "WAKST"}},
        {"Pacific/Wallis",
         {"Wallis & Futuna Time",
          "WFT",
          "Wallis & Futuna Summer Time",
          "WFST"}},
        {"Pacific/Yap", {"Chuuk Time", "CHUT", "Chuuk Summer Time", "CHUST"}},
        {"Poland",
         {"Central European Time",
          "CET",
          "Central European Summer Time",
          "CEST"}},
        {"PRC", {"China Standard Time", "CST", "China Daylight Time", "CDT"}},
        {"PST8PDT",
         {"Pacific Standard Time", "PST", "Pacific Daylight Time", "PDT"}},
        {"ROK", {"Korea Standard Time", "KST", "Korea Daylight Time", "KDT"}},
        {"Singapore",
         {"Singapore Time", "SGT", "Singapore Summer Time", "SGST"}},
        {"SST", {"Solomon Is. Time", "SBT", "Solomon Is. Summer Time", "SBST"}},
        {"SystemV/AST4",
         {"Atlantic Standard Time", "AST", "Atlantic Daylight Time", "ADT"}},
        {"SystemV/AST4ADT",
         {"Atlantic Standard Time", "AST", "Atlantic Daylight Time", "ADT"}},
        {"SystemV/CST6",
         {"Central Standard Time", "CST", "Central Daylight Time", "CDT"}},
        {"SystemV/CST6CDT",
         {"Central Standard Time", "CST", "Central Daylight Time", "CDT"}},
        {"SystemV/EST5",
         {"Eastern Standard Time", "EST", "Eastern Daylight Time", "EDT"}},
        {"SystemV/EST5EDT",
         {"Eastern Standard Time", "EST", "Eastern Daylight Time", "EDT"}},
        {"SystemV/HST10",
         {"Hawaii Standard Time", "HST", "Hawaii Daylight Time", "HDT"}},
        {"SystemV/MST7",
         {"Mountain Standard Time", "MST", "Mountain Daylight Time", "MDT"}},
        {"SystemV/MST7MDT",
         {"Mountain Standard Time", "MST", "Mountain Daylight Time", "MDT"}},
        {"SystemV/PST8",
         {"Pacific Standard Time", "PST", "Pacific Daylight Time", "PDT"}},
        {"SystemV/PST8PDT",
         {"Pacific Standard Time", "PST", "Pacific Daylight Time", "PDT"}},
        {"SystemV/YST9",
         {"Alaska Standard Time", "AKST", "Alaska Daylight Time", "AKDT"}},
        {"SystemV/YST9YDT",
         {"Alaska Standard Time", "AKST", "Alaska Daylight Time", "AKDT"}},
        {"Turkey", {"Turkey Time", "TRT", "Turkey Summer Time", "TRST"}},
        {"UCT",
         {"Coordinated Universal Time",
          "UTC",
          "Coordinated Universal Time",
          "UTC"}},
        {"Universal",
         {"Coordinated Universal Time",
          "UTC",
          "Coordinated Universal Time",
          "UTC"}},
        {"US/Alaska",
         {"Alaska Standard Time", "AKST", "Alaska Daylight Time", "AKDT"}},
        {"US/Aleutian",
         {"Hawaii Standard Time", "HST", "Hawaii Daylight Time", "HDT"}},
        {"US/Arizona",
         {"Mountain Standard Time", "MST", "Mountain Daylight Time", "MDT"}},
        {"US/Central",
         {"Central Standard Time", "CST", "Central Daylight Time", "CDT"}},
        {"US/Eastern",
         {"Eastern Standard Time", "EST", "Eastern Daylight Time", "EDT"}},
        {"US/Hawaii",
         {"Hawaii Standard Time", "HST", "Hawaii Daylight Time", "HDT"}},
        {"US/Indiana-Starke",
         {"Central Standard Time", "CST", "Central Daylight Time", "CDT"}},
        {"US/East-Indiana",
         {"Eastern Standard Time", "EST", "Eastern Daylight Time", "EDT"}},
        {"US/Michigan",
         {"Eastern Standard Time", "EST", "Eastern Daylight Time", "EDT"}},
        {"US/Mountain",
         {"Mountain Standard Time", "MST", "Mountain Daylight Time", "MDT"}},
        {"US/Pacific",
         {"Pacific Standard Time", "PST", "Pacific Daylight Time", "PDT"}},
        {"US/Samoa",
         {"Samoa Standard Time", "SST", "Samoa Daylight Time", "SDT"}},
        {"VST", {"Indochina Time", "ICT", "Indochina Summer Time", "ICST"}},
        {"W-SU",
         {"Moscow Standard Time", "MSK", "Moscow Daylight Time", "MSD"}},
        {"WET",
         {"Western European Time",
          "WET",
          "Western European Summer Time",
          "WEST"}},
        {"Zulu",
         {"Coordinated Universal Time",
          "UTC",
          "Coordinated Universal Time",
          "UTC"}},
    };
  }());
  return *timeZoneNames;
}

} // namespace facebook::velox::tz
