/*
 * Copyright (c) Facebook, Inc. and its affiliates.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#pragma once

#include "velox/dwio/common/ReaderFactory.h"
#include "velox/dwio/dwrf/reader/DwrfReader.h"

namespace facebook::velox::orc {

class OrcReaderFactory : public dwio::common::ReaderFactory {
 public:
  OrcReaderFactory() : ReaderFactory(dwio::common::FileFormat::ORC) {}

  std::unique_ptr<dwio::common::Reader> createReader(
      std::unique_ptr<dwio::common::BufferedInput> input,
      const dwio::common::ReaderOptions& options) override {
    return velox::dwrf::DwrfReader::create(std::move(input), options);
  }
};

void registerOrcReaderFactory();

void unregisterOrcReaderFactory();

} // namespace facebook::velox::orc
