/*
 * Copyright (c) Facebook, Inc. and its affiliates.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#pragma once

#include <glog/logging.h>
#include <string>
#include "folly/GLog.h"

#define VELOX_WAVE_LOG_PREFIX "WAVE"
#define VELOX_WAVE_LOG(severity) LOG(severity) << VELOX_WAVE_LOG_PREFIX
#define VELOX_WAVE_LOG_EVERY_MS(severity, ms) \
  FB_LOG_EVERY_MS(severity, ms) << VELOX_WAVE_LOG_PREFIX

namespace facebook::velox::wave {
/// Throws a VeloxException with 'message'.
void waveError(const std::string& message);
} // namespace facebook::velox::wave
