/*
 * Copyright (c) Facebook, Inc. and its affiliates.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

// This header must be included in any dynamically loaded user-defined
// function library. It enforces the correct function declaration with
// the proper signature.
#pragma once

#include "velox/functions/Udf.h"

extern "C" {
// The "registry()" declaration and "check()" function ensure the correct
// registry signature function is defined by the user.
void registry();
void check() {
  registry();
}
}
