/*
 * Copyright (c) Facebook, Inc. and its affiliates.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#pragma once

// Macro to declare execution domain if the header can be included in a C++ file
// that is not compiled with nvcc
#if defined(__CUDACC_RTC__)
#define WAVE_DEVICE_HOST __device__
#elif defined(__CUDACC__)
#define WAVE_DEVICE_HOST __device__ __host__
#else
#define WAVE_DEVICE_HOST
#endif
