/*
 * Copyright (c) Facebook, Inc. and its affiliates.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include "velox/dwio/dwrf/common/IntEncoder.h"

namespace facebook::velox::dwrf {

template <bool isSigned>
uint64_t IntEncoder<isSigned>::flush() {
  output_->BackUp(bufferLength_ - bufferPosition_);
  auto dataSize = output_->flush();
  bufferLength_ = bufferPosition_ = 0;
  return dataSize;
}

template uint64_t IntEncoder<true>::flush();
template uint64_t IntEncoder<false>::flush();

} // namespace facebook::velox::dwrf
