
/*
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

/*
 * Copyright (c) 2014-2019 Steven G. Johnson, Jiahao Chen, Peter Colberg, Tony
 * Kelman, Scott P. Jones, and other contributors. Copyright (c) 2009 Public
 * Software Group e. V., Berlin, Germany
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
 * DEALINGS IN THE SOFTWARE.
 */

/**
 * @mainpage
 *
 * utf8proc is a free/open-source (MIT/expat licensed) C library
 * providing Unicode normalization, case-folding, and other operations
 * for strings in the UTF-8 encoding, supporting up-to-date Unicode versions.
 * See the utf8proc home page (http://julialang.org/utf8proc/)
 * for downloads and other information, or the source code on github
 * (https://github.com/JuliaLang/utf8proc).
 *
 * For the utf8proc API documentation, see: @ref utf8proc.h
 *
 * The features of utf8proc include:
 *
 * - Transformation of strings (@ref utf8proc_map) to:
 *    - decompose (@ref UTF8PROC_DECOMPOSE) or compose (@ref UTF8PROC_COMPOSE)
 * Unicode combining characters
 * (http://en.wikipedia.org/wiki/Combining_character)
 *    - canonicalize Unicode compatibility characters (@ref UTF8PROC_COMPAT)
 *    - strip "ignorable" (@ref UTF8PROC_IGNORE) characters, control characters
 * (@ref UTF8PROC_STRIPCC), or combining characters such as accents (@ref
 * UTF8PROC_STRIPMARK)
 *    - case-folding (@ref UTF8PROC_CASEFOLD)
 * - Unicode normalization: @ref utf8proc_NFD, @ref utf8proc_NFC, @ref
 * utf8proc_NFKD, @ref utf8proc_NFKC
 * - Detecting grapheme boundaries (@ref utf8proc_grapheme_break and @ref
 * UTF8PROC_CHARBOUND)
 * - Character-width computation: @ref utf8proc_charwidth
 * - Classification of characters by Unicode category: @ref utf8proc_category
 * and @ref utf8proc_category_string
 * - Encode (@ref utf8proc_encode_char) and decode (@ref utf8proc_iterate)
 * Unicode codepoints to/from UTF-8.
 */

/** @file */

static const utf8proc_uint16_t utf8proc_sequences[] = {
    97,    98,    99,    100,   101,   102,   103,   104,   105,   106,   107,
    108,   109,   110,   111,   112,   113,   114,   115,   116,   117,   118,
    119,   120,   121,   122,   32,    32,    776,   32,    772,   50,    51,
    32,    769,   956,   32,    807,   49,    49,    8260,  52,    49,    8260,
    50,    51,    8260,  52,    65,    768,   224,   65,    769,   225,   65,
    770,   226,   65,    771,   227,   65,    776,   228,   65,    778,   229,
    230,   67,    807,   231,   69,    768,   232,   69,    769,   233,   69,
    770,   234,   69,    776,   235,   73,    768,   236,   73,    769,   237,
    73,    770,   238,   73,    776,   239,   240,   78,    771,   241,   79,
    768,   242,   79,    769,   243,   79,    770,   244,   79,    771,   245,
    79,    776,   246,   248,   85,    768,   249,   85,    769,   250,   85,
    770,   251,   85,    776,   252,   89,    769,   253,   254,   115,   115,
    97,    768,   97,    769,   97,    770,   97,    771,   97,    776,   97,
    778,   99,    807,   101,   768,   101,   769,   101,   770,   101,   776,
    105,   768,   105,   769,   105,   770,   105,   776,   110,   771,   111,
    768,   111,   769,   111,   770,   111,   771,   111,   776,   117,   768,
    117,   769,   117,   770,   117,   776,   121,   769,   121,   776,   65,
    772,   257,   97,    772,   65,    774,   259,   97,    774,   65,    808,
    261,   97,    808,   67,    769,   263,   99,    769,   67,    770,   265,
    99,    770,   67,    775,   267,   99,    775,   67,    780,   269,   99,
    780,   68,    780,   271,   100,   780,   273,   69,    772,   275,   101,
    772,   69,    774,   277,   101,   774,   69,    775,   279,   101,   775,
    69,    808,   281,   101,   808,   69,    780,   283,   101,   780,   71,
    770,   285,   103,   770,   71,    774,   287,   103,   774,   71,    775,
    289,   103,   775,   71,    807,   291,   103,   807,   72,    770,   293,
    104,   770,   295,   73,    771,   297,   105,   771,   73,    772,   299,
    105,   772,   73,    774,   301,   105,   774,   73,    808,   303,   105,
    808,   73,    775,   105,   775,   73,    74,    307,   105,   106,   74,
    770,   309,   106,   770,   75,    807,   311,   107,   807,   76,    769,
    314,   108,   769,   76,    807,   316,   108,   807,   76,    780,   318,
    108,   780,   76,    183,   320,   108,   183,   322,   78,    769,   324,
    110,   769,   78,    807,   326,   110,   807,   78,    780,   328,   110,
    780,   700,   110,   331,   79,    772,   333,   111,   772,   79,    774,
    335,   111,   774,   79,    779,   337,   111,   779,   339,   82,    769,
    341,   114,   769,   82,    807,   343,   114,   807,   82,    780,   345,
    114,   780,   83,    769,   347,   115,   769,   83,    770,   349,   115,
    770,   83,    807,   351,   115,   807,   83,    780,   353,   115,   780,
    84,    807,   355,   116,   807,   84,    780,   357,   116,   780,   359,
    85,    771,   361,   117,   771,   85,    772,   363,   117,   772,   85,
    774,   365,   117,   774,   85,    778,   367,   117,   778,   85,    779,
    369,   117,   779,   85,    808,   371,   117,   808,   87,    770,   373,
    119,   770,   89,    770,   375,   121,   770,   89,    776,   255,   90,
    769,   378,   122,   769,   90,    775,   380,   122,   775,   90,    780,
    382,   122,   780,   595,   387,   389,   596,   392,   598,   599,   396,
    477,   601,   603,   402,   608,   611,   617,   616,   409,   623,   626,
    629,   79,    795,   417,   111,   795,   419,   421,   640,   424,   643,
    429,   648,   85,    795,   432,   117,   795,   650,   651,   436,   438,
    658,   441,   445,   68,    381,   454,   68,    382,   100,   382,   76,
    74,    457,   76,    106,   108,   106,   78,    74,    460,   78,    106,
    110,   106,   65,    780,   462,   97,    780,   73,    780,   464,   105,
    780,   79,    780,   466,   111,   780,   85,    780,   468,   117,   780,
    220,   772,   470,   252,   772,   220,   769,   472,   252,   769,   220,
    780,   474,   252,   780,   220,   768,   476,   252,   768,   196,   772,
    479,   228,   772,   550,   772,   481,   551,   772,   198,   772,   483,
    230,   772,   485,   71,    780,   487,   103,   780,   75,    780,   489,
    107,   780,   79,    808,   491,   111,   808,   490,   772,   493,   491,
    772,   439,   780,   495,   658,   780,   106,   780,   68,    90,    499,
    68,    122,   100,   122,   71,    769,   501,   103,   769,   405,   447,
    78,    768,   505,   110,   768,   197,   769,   507,   229,   769,   198,
    769,   509,   230,   769,   216,   769,   511,   248,   769,   65,    783,
    513,   97,    783,   65,    785,   515,   97,    785,   69,    783,   517,
    101,   783,   69,    785,   519,   101,   785,   73,    783,   521,   105,
    783,   73,    785,   523,   105,   785,   79,    783,   525,   111,   783,
    79,    785,   527,   111,   785,   82,    783,   529,   114,   783,   82,
    785,   531,   114,   785,   85,    783,   533,   117,   783,   85,    785,
    535,   117,   785,   83,    806,   537,   115,   806,   84,    806,   539,
    116,   806,   541,   72,    780,   543,   104,   780,   414,   547,   549,
    65,    775,   551,   97,    775,   69,    807,   553,   101,   807,   214,
    772,   555,   246,   772,   213,   772,   557,   245,   772,   79,    775,
    559,   111,   775,   558,   772,   561,   559,   772,   89,    772,   563,
    121,   772,   11365, 572,   410,   11366, 578,   384,   649,   652,   583,
    585,   587,   589,   591,   614,   633,   635,   641,   32,    774,   32,
    775,   32,    778,   32,    808,   32,    771,   32,    779,   661,   768,
    769,   787,   776,   769,   953,   881,   883,   697,   887,   32,    837,
    59,    1011,  168,   769,   913,   769,   940,   183,   917,   769,   941,
    919,   769,   942,   921,   769,   943,   927,   769,   972,   933,   769,
    973,   937,   769,   974,   970,   769,   953,   776,   769,   945,   946,
    947,   948,   949,   950,   951,   952,   954,   955,   957,   958,   959,
    960,   961,   963,   964,   965,   966,   967,   968,   969,   921,   776,
    970,   933,   776,   971,   945,   769,   949,   769,   951,   769,   953,
    769,   971,   769,   965,   776,   769,   953,   776,   965,   776,   959,
    769,   965,   769,   969,   769,   983,   933,   978,   769,   978,   776,
    985,   987,   989,   991,   993,   995,   997,   999,   1001,  1003,  1005,
    1007,  962,   920,   1016,  931,   1010,  1019,  891,   892,   893,   1045,
    768,   1104,  1045,  776,   1105,  1106,  1043,  769,   1107,  1108,  1109,
    1110,  1030,  776,   1111,  1112,  1113,  1114,  1115,  1050,  769,   1116,
    1048,  768,   1117,  1059,  774,   1118,  1119,  1072,  1073,  1074,  1075,
    1076,  1077,  1078,  1079,  1080,  1048,  774,   1081,  1082,  1083,  1084,
    1085,  1086,  1087,  1088,  1089,  1090,  1091,  1092,  1093,  1094,  1095,
    1096,  1097,  1098,  1099,  1100,  1101,  1102,  1103,  1080,  774,   1077,
    768,   1077,  776,   1075,  769,   1110,  776,   1082,  769,   1080,  768,
    1091,  774,   1121,  1123,  1125,  1127,  1129,  1131,  1133,  1135,  1137,
    1139,  1141,  1140,  783,   1143,  1141,  783,   1145,  1147,  1149,  1151,
    1153,  1163,  1165,  1167,  1169,  1171,  1173,  1175,  1177,  1179,  1181,
    1183,  1185,  1187,  1189,  1191,  1193,  1195,  1197,  1199,  1201,  1203,
    1205,  1207,  1209,  1211,  1213,  1215,  1231,  1046,  774,   1218,  1078,
    774,   1220,  1222,  1224,  1226,  1228,  1230,  1040,  774,   1233,  1072,
    774,   1040,  776,   1235,  1072,  776,   1237,  1045,  774,   1239,  1077,
    774,   1241,  1240,  776,   1243,  1241,  776,   1046,  776,   1245,  1078,
    776,   1047,  776,   1247,  1079,  776,   1249,  1048,  772,   1251,  1080,
    772,   1048,  776,   1253,  1080,  776,   1054,  776,   1255,  1086,  776,
    1257,  1256,  776,   1259,  1257,  776,   1069,  776,   1261,  1101,  776,
    1059,  772,   1263,  1091,  772,   1059,  776,   1265,  1091,  776,   1059,
    779,   1267,  1091,  779,   1063,  776,   1269,  1095,  776,   1271,  1067,
    776,   1273,  1099,  776,   1275,  1277,  1279,  1281,  1283,  1285,  1287,
    1289,  1291,  1293,  1295,  1297,  1299,  1301,  1303,  1305,  1307,  1309,
    1311,  1313,  1315,  1317,  1319,  1321,  1323,  1325,  1327,  1377,  1378,
    1379,  1380,  1381,  1382,  1383,  1384,  1385,  1386,  1387,  1388,  1389,
    1390,  1391,  1392,  1393,  1394,  1395,  1396,  1397,  1398,  1399,  1400,
    1401,  1402,  1403,  1404,  1405,  1406,  1407,  1408,  1409,  1410,  1411,
    1412,  1413,  1414,  1381,  1410,  1575,  1619,  1575,  1620,  1608,  1620,
    1575,  1621,  1610,  1620,  1575,  1652,  1608,  1652,  1735,  1652,  1610,
    1652,  1749,  1620,  1729,  1620,  1746,  1620,  2344,  2364,  2352,  2364,
    2355,  2364,  2325,  2364,  2326,  2364,  2327,  2364,  2332,  2364,  2337,
    2364,  2338,  2364,  2347,  2364,  2351,  2364,  2503,  2494,  2503,  2519,
    2465,  2492,  2466,  2492,  2479,  2492,  2610,  2620,  2616,  2620,  2582,
    2620,  2583,  2620,  2588,  2620,  2603,  2620,  2887,  2902,  2887,  2878,
    2887,  2903,  2849,  2876,  2850,  2876,  2962,  3031,  3014,  3006,  3015,
    3006,  3014,  3031,  3142,  3158,  3263,  3285,  3270,  3285,  3270,  3286,
    3270,  3266,  3274,  3285,  3398,  3390,  3399,  3390,  3398,  3415,  3545,
    3530,  3545,  3535,  3548,  3530,  3545,  3551,  3661,  3634,  3789,  3762,
    3755,  3737,  3755,  3745,  3851,  3906,  4023,  3916,  4023,  3921,  4023,
    3926,  4023,  3931,  4023,  3904,  4021,  3953,  3954,  3953,  3956,  4018,
    3968,  4018,  3969,  4019,  3968,  4019,  3969,  3953,  3968,  3986,  4023,
    3996,  4023,  4001,  4023,  4006,  4023,  4011,  4023,  3984,  4021,  4133,
    4142,  11520, 11521, 11522, 11523, 11524, 11525, 11526, 11527, 11528, 11529,
    11530, 11531, 11532, 11533, 11534, 11535, 11536, 11537, 11538, 11539, 11540,
    11541, 11542, 11543, 11544, 11545, 11546, 11547, 11548, 11549, 11550, 11551,
    11552, 11553, 11554, 11555, 11556, 11557, 11559, 11565, 4316,  5104,  5105,
    5106,  5107,  5108,  5109,  6917,  6965,  6919,  6965,  6921,  6965,  6923,
    6965,  6925,  6965,  6929,  6965,  6970,  6965,  6972,  6965,  6974,  6965,
    6975,  6965,  6978,  6965,  42571, 4304,  4305,  4306,  4307,  4308,  4309,
    4310,  4311,  4312,  4313,  4314,  4315,  4317,  4318,  4319,  4320,  4321,
    4322,  4323,  4324,  4325,  4326,  4327,  4328,  4329,  4330,  4331,  4332,
    4333,  4334,  4335,  4336,  4337,  4338,  4339,  4340,  4341,  4342,  4343,
    4344,  4345,  4346,  4349,  4350,  4351,  65,    198,   66,    68,    69,
    398,   71,    72,    73,    74,    75,    76,    77,    78,    79,    546,
    80,    82,    84,    85,    87,    592,   593,   7426,  604,   7446,  7447,
    7453,  7461,  594,   597,   607,   609,   613,   618,   7547,  669,   621,
    7557,  671,   625,   624,   627,   628,   632,   642,   427,   7452,  656,
    657,   65,    805,   7681,  97,    805,   66,    775,   7683,  98,    775,
    66,    803,   7685,  98,    803,   66,    817,   7687,  98,    817,   199,
    769,   7689,  231,   769,   68,    775,   7691,  100,   775,   68,    803,
    7693,  100,   803,   68,    817,   7695,  100,   817,   68,    807,   7697,
    100,   807,   68,    813,   7699,  100,   813,   274,   768,   7701,  275,
    768,   274,   769,   7703,  275,   769,   69,    813,   7705,  101,   813,
    69,    816,   7707,  101,   816,   552,   774,   7709,  553,   774,   70,
    775,   7711,  102,   775,   71,    772,   7713,  103,   772,   72,    775,
    7715,  104,   775,   72,    803,   7717,  104,   803,   72,    776,   7719,
    104,   776,   72,    807,   7721,  104,   807,   72,    814,   7723,  104,
    814,   73,    816,   7725,  105,   816,   207,   769,   7727,  239,   769,
    75,    769,   7729,  107,   769,   75,    803,   7731,  107,   803,   75,
    817,   7733,  107,   817,   76,    803,   7735,  108,   803,   7734,  772,
    7737,  7735,  772,   76,    817,   7739,  108,   817,   76,    813,   7741,
    108,   813,   77,    769,   7743,  109,   769,   77,    775,   7745,  109,
    775,   77,    803,   7747,  109,   803,   78,    775,   7749,  110,   775,
    78,    803,   7751,  110,   803,   78,    817,   7753,  110,   817,   78,
    813,   7755,  110,   813,   213,   769,   7757,  245,   769,   213,   776,
    7759,  245,   776,   332,   768,   7761,  333,   768,   332,   769,   7763,
    333,   769,   80,    769,   7765,  112,   769,   80,    775,   7767,  112,
    775,   82,    775,   7769,  114,   775,   82,    803,   7771,  114,   803,
    7770,  772,   7773,  7771,  772,   82,    817,   7775,  114,   817,   83,
    775,   7777,  115,   775,   83,    803,   7779,  115,   803,   346,   775,
    7781,  347,   775,   352,   775,   7783,  353,   775,   7778,  775,   7785,
    7779,  775,   84,    775,   7787,  116,   775,   84,    803,   7789,  116,
    803,   84,    817,   7791,  116,   817,   84,    813,   7793,  116,   813,
    85,    804,   7795,  117,   804,   85,    816,   7797,  117,   816,   85,
    813,   7799,  117,   813,   360,   769,   7801,  361,   769,   362,   776,
    7803,  363,   776,   86,    771,   7805,  118,   771,   86,    803,   7807,
    118,   803,   87,    768,   7809,  119,   768,   87,    769,   7811,  119,
    769,   87,    776,   7813,  119,   776,   87,    775,   7815,  119,   775,
    87,    803,   7817,  119,   803,   88,    775,   7819,  120,   775,   88,
    776,   7821,  120,   776,   89,    775,   7823,  121,   775,   90,    770,
    7825,  122,   770,   90,    803,   7827,  122,   803,   90,    817,   7829,
    122,   817,   104,   817,   116,   776,   119,   778,   121,   778,   97,
    702,   383,   775,   65,    803,   7841,  97,    803,   65,    777,   7843,
    97,    777,   194,   769,   7845,  226,   769,   194,   768,   7847,  226,
    768,   194,   777,   7849,  226,   777,   194,   771,   7851,  226,   771,
    7840,  770,   7853,  7841,  770,   258,   769,   7855,  259,   769,   258,
    768,   7857,  259,   768,   258,   777,   7859,  259,   777,   258,   771,
    7861,  259,   771,   7840,  774,   7863,  7841,  774,   69,    803,   7865,
    101,   803,   69,    777,   7867,  101,   777,   69,    771,   7869,  101,
    771,   202,   769,   7871,  234,   769,   202,   768,   7873,  234,   768,
    202,   777,   7875,  234,   777,   202,   771,   7877,  234,   771,   7864,
    770,   7879,  7865,  770,   73,    777,   7881,  105,   777,   73,    803,
    7883,  105,   803,   79,    803,   7885,  111,   803,   79,    777,   7887,
    111,   777,   212,   769,   7889,  244,   769,   212,   768,   7891,  244,
    768,   212,   777,   7893,  244,   777,   212,   771,   7895,  244,   771,
    7884,  770,   7897,  7885,  770,   416,   769,   7899,  417,   769,   416,
    768,   7901,  417,   768,   416,   777,   7903,  417,   777,   416,   771,
    7905,  417,   771,   416,   803,   7907,  417,   803,   85,    803,   7909,
    117,   803,   85,    777,   7911,  117,   777,   431,   769,   7913,  432,
    769,   431,   768,   7915,  432,   768,   431,   777,   7917,  432,   777,
    431,   771,   7919,  432,   771,   431,   803,   7921,  432,   803,   89,
    768,   7923,  121,   768,   89,    803,   7925,  121,   803,   89,    777,
    7927,  121,   777,   89,    771,   7929,  121,   771,   7931,  7933,  7935,
    945,   787,   945,   788,   7936,  768,   7937,  768,   7936,  769,   7937,
    769,   7936,  834,   7937,  834,   913,   787,   7936,  913,   788,   7937,
    7944,  768,   7938,  7945,  768,   7939,  7944,  769,   7940,  7945,  769,
    7941,  7944,  834,   7942,  7945,  834,   7943,  949,   787,   949,   788,
    7952,  768,   7953,  768,   7952,  769,   7953,  769,   917,   787,   7952,
    917,   788,   7953,  7960,  768,   7954,  7961,  768,   7955,  7960,  769,
    7956,  7961,  769,   7957,  951,   787,   951,   788,   7968,  768,   7969,
    768,   7968,  769,   7969,  769,   7968,  834,   7969,  834,   919,   787,
    7968,  919,   788,   7969,  7976,  768,   7970,  7977,  768,   7971,  7976,
    769,   7972,  7977,  769,   7973,  7976,  834,   7974,  7977,  834,   7975,
    953,   787,   953,   788,   7984,  768,   7985,  768,   7984,  769,   7985,
    769,   7984,  834,   7985,  834,   921,   787,   7984,  921,   788,   7985,
    7992,  768,   7986,  7993,  768,   7987,  7992,  769,   7988,  7993,  769,
    7989,  7992,  834,   7990,  7993,  834,   7991,  959,   787,   959,   788,
    8000,  768,   8001,  768,   8000,  769,   8001,  769,   927,   787,   8000,
    927,   788,   8001,  8008,  768,   8002,  8009,  768,   8003,  8008,  769,
    8004,  8009,  769,   8005,  965,   787,   965,   788,   8016,  768,   965,
    787,   768,   8017,  768,   8016,  769,   965,   787,   769,   8017,  769,
    8016,  834,   965,   787,   834,   8017,  834,   933,   788,   8017,  8025,
    768,   8019,  8025,  769,   8021,  8025,  834,   8023,  969,   787,   969,
    788,   8032,  768,   8033,  768,   8032,  769,   8033,  769,   8032,  834,
    8033,  834,   937,   787,   8032,  937,   788,   8033,  8040,  768,   8034,
    8041,  768,   8035,  8040,  769,   8036,  8041,  769,   8037,  8040,  834,
    8038,  8041,  834,   8039,  945,   768,   949,   768,   951,   768,   953,
    768,   959,   768,   965,   768,   969,   768,   7936,  837,   7936,  953,
    7937,  837,   7937,  953,   7938,  837,   7938,  953,   7939,  837,   7939,
    953,   7940,  837,   7940,  953,   7941,  837,   7941,  953,   7942,  837,
    7942,  953,   7943,  837,   7943,  953,   7944,  837,   7945,  837,   7946,
    837,   7947,  837,   7948,  837,   7949,  837,   7950,  837,   7951,  837,
    7968,  837,   7968,  953,   7969,  837,   7969,  953,   7970,  837,   7970,
    953,   7971,  837,   7971,  953,   7972,  837,   7972,  953,   7973,  837,
    7973,  953,   7974,  837,   7974,  953,   7975,  837,   7975,  953,   7976,
    837,   7977,  837,   7978,  837,   7979,  837,   7980,  837,   7981,  837,
    7982,  837,   7983,  837,   8032,  837,   8032,  953,   8033,  837,   8033,
    953,   8034,  837,   8034,  953,   8035,  837,   8035,  953,   8036,  837,
    8036,  953,   8037,  837,   8037,  953,   8038,  837,   8038,  953,   8039,
    837,   8039,  953,   8040,  837,   8041,  837,   8042,  837,   8043,  837,
    8044,  837,   8045,  837,   8046,  837,   8047,  837,   945,   774,   945,
    772,   8048,  837,   8048,  953,   945,   837,   945,   953,   940,   837,
    940,   953,   945,   834,   8118,  837,   945,   834,   953,   913,   774,
    8112,  913,   772,   8113,  913,   768,   8048,  902,   8049,  913,   837,
    32,    787,   32,    834,   168,   834,   8052,  837,   8052,  953,   951,
    837,   951,   953,   942,   837,   942,   953,   951,   834,   8134,  837,
    951,   834,   953,   917,   768,   8050,  904,   8051,  919,   768,   8052,
    905,   8053,  919,   837,   8127,  768,   8127,  769,   8127,  834,   953,
    774,   953,   772,   970,   768,   953,   776,   768,   912,   953,   834,
    970,   834,   953,   776,   834,   921,   774,   8144,  921,   772,   8145,
    921,   768,   8054,  906,   8055,  8190,  768,   8190,  769,   8190,  834,
    965,   774,   965,   772,   971,   768,   965,   776,   768,   944,   961,
    787,   961,   788,   965,   834,   971,   834,   965,   776,   834,   933,
    774,   8160,  933,   772,   8161,  933,   768,   8058,  910,   8059,  929,
    788,   8165,  168,   768,   901,   96,    8060,  837,   8060,  953,   969,
    837,   969,   953,   974,   837,   974,   953,   969,   834,   8182,  837,
    969,   834,   953,   927,   768,   8056,  908,   8057,  937,   768,   8060,
    911,   8061,  937,   837,   180,   32,    788,   8194,  8195,  8208,  32,
    819,   46,    46,    46,    46,    46,    46,    8242,  8242,  8242,  8242,
    8242,  8245,  8245,  8245,  8245,  8245,  33,    33,    32,    773,   63,
    63,    63,    33,    33,    63,    8242,  8242,  8242,  8242,  48,    52,
    53,    54,    55,    56,    57,    43,    8722,  61,    40,    41,    82,
    115,   97,    47,    99,    97,    47,    115,   67,    176,   67,    99,
    47,    111,   99,    47,    117,   400,   176,   70,    78,    111,   81,
    83,    77,    84,    69,    76,    84,    77,    90,    937,   197,   70,
    8526,  1488,  1489,  1490,  1491,  70,    65,    88,    915,   928,   8721,
    49,    8260,  55,    49,    8260,  57,    49,    8260,  49,    48,    49,
    8260,  51,    50,    8260,  51,    49,    8260,  53,    50,    8260,  53,
    51,    8260,  53,    52,    8260,  53,    49,    8260,  54,    53,    8260,
    54,    49,    8260,  56,    51,    8260,  56,    53,    8260,  56,    55,
    8260,  56,    49,    8260,  8560,  73,    73,    8561,  73,    73,    73,
    8562,  73,    86,    8563,  86,    8564,  86,    73,    8565,  86,    73,
    73,    8566,  86,    73,    73,    73,    8567,  73,    88,    8568,  88,
    8569,  88,    73,    8570,  88,    73,    73,    8571,  8572,  8573,  8574,
    8575,  105,   105,   105,   105,   105,   105,   118,   118,   105,   118,
    105,   105,   118,   105,   105,   105,   105,   120,   120,   105,   120,
    105,   105,   8580,  48,    8260,  51,    8592,  824,   8594,  824,   8596,
    824,   8656,  824,   8660,  824,   8658,  824,   8707,  824,   8712,  824,
    8715,  824,   8739,  824,   8741,  824,   8747,  8747,  8747,  8747,  8747,
    8750,  8750,  8750,  8750,  8750,  8764,  824,   8771,  824,   8773,  824,
    8776,  824,   61,    824,   8801,  824,   8781,  824,   60,    824,   62,
    824,   8804,  824,   8805,  824,   8818,  824,   8819,  824,   8822,  824,
    8823,  824,   8826,  824,   8827,  824,   8834,  824,   8835,  824,   8838,
    824,   8839,  824,   8866,  824,   8872,  824,   8873,  824,   8875,  824,
    8828,  824,   8829,  824,   8849,  824,   8850,  824,   8882,  824,   8883,
    824,   8884,  824,   8885,  824,   12296, 12297, 49,    48,    49,    49,
    49,    50,    49,    51,    49,    52,    49,    53,    49,    54,    49,
    55,    49,    56,    49,    57,    50,    48,    40,    49,    41,    40,
    50,    41,    40,    51,    41,    40,    52,    41,    40,    53,    41,
    40,    54,    41,    40,    55,    41,    40,    56,    41,    40,    57,
    41,    40,    49,    48,    41,    40,    49,    49,    41,    40,    49,
    50,    41,    40,    49,    51,    41,    40,    49,    52,    41,    40,
    49,    53,    41,    40,    49,    54,    41,    40,    49,    55,    41,
    40,    49,    56,    41,    40,    49,    57,    41,    40,    50,    48,
    41,    49,    46,    50,    46,    51,    46,    52,    46,    53,    46,
    54,    46,    55,    46,    56,    46,    57,    46,    49,    48,    46,
    49,    49,    46,    49,    50,    46,    49,    51,    46,    49,    52,
    46,    49,    53,    46,    49,    54,    46,    49,    55,    46,    49,
    56,    46,    49,    57,    46,    50,    48,    46,    40,    97,    41,
    40,    98,    41,    40,    99,    41,    40,    100,   41,    40,    101,
    41,    40,    102,   41,    40,    103,   41,    40,    104,   41,    40,
    105,   41,    40,    106,   41,    40,    107,   41,    40,    108,   41,
    40,    109,   41,    40,    110,   41,    40,    111,   41,    40,    112,
    41,    40,    113,   41,    40,    114,   41,    40,    115,   41,    40,
    116,   41,    40,    117,   41,    40,    118,   41,    40,    119,   41,
    40,    120,   41,    40,    121,   41,    40,    122,   41,    9424,  9425,
    9426,  9427,  9428,  9429,  9430,  9431,  9432,  9433,  9434,  9435,  9436,
    9437,  9438,  9439,  9440,  9441,  83,    9442,  9443,  9444,  9445,  9446,
    9447,  89,    9448,  9449,  8747,  8747,  8747,  8747,  58,    58,    61,
    61,    61,    61,    61,    61,    10973, 824,   11312, 11313, 11314, 11315,
    11316, 11317, 11318, 11319, 11320, 11321, 11322, 11323, 11324, 11325, 11326,
    11327, 11328, 11329, 11330, 11331, 11332, 11333, 11334, 11335, 11336, 11337,
    11338, 11339, 11340, 11341, 11342, 11343, 11344, 11345, 11346, 11347, 11348,
    11349, 11350, 11351, 11352, 11353, 11354, 11355, 11356, 11357, 11358, 11361,
    619,   7549,  637,   11368, 11370, 11372, 11379, 11382, 575,   576,   11393,
    11395, 11397, 11399, 11401, 11403, 11405, 11407, 11409, 11411, 11413, 11415,
    11417, 11419, 11421, 11423, 11425, 11427, 11429, 11431, 11433, 11435, 11437,
    11439, 11441, 11443, 11445, 11447, 11449, 11451, 11453, 11455, 11457, 11459,
    11461, 11463, 11465, 11467, 11469, 11471, 11473, 11475, 11477, 11479, 11481,
    11483, 11485, 11487, 11489, 11491, 11500, 11502, 11507, 11617, 27597, 40863,
    19968, 20008, 20022, 20031, 20057, 20101, 20108, 20128, 20154, 20799, 20837,
    20843, 20866, 20886, 20907, 20960, 20981, 20992, 21147, 21241, 21269, 21274,
    21304, 21313, 21340, 21353, 21378, 21430, 21448, 21475, 22231, 22303, 22763,
    22786, 22794, 22805, 22823, 22899, 23376, 23424, 23544, 23567, 23586, 23608,
    23662, 23665, 24027, 24037, 24049, 24062, 24178, 24186, 24191, 24308, 24318,
    24331, 24339, 24400, 24417, 24435, 24515, 25096, 25142, 25163, 25903, 25908,
    25991, 26007, 26020, 26041, 26080, 26085, 26352, 26376, 26408, 27424, 27490,
    27513, 27571, 27595, 27604, 27611, 27663, 27668, 27700, 28779, 29226, 29238,
    29243, 29247, 29255, 29273, 29275, 29356, 29572, 29577, 29916, 29926, 29976,
    29983, 29992, 30000, 30091, 30098, 30326, 30333, 30382, 30399, 30446, 30683,
    30690, 30707, 31034, 31160, 31166, 31348, 31435, 31481, 31859, 31992, 32566,
    32593, 32650, 32701, 32769, 32780, 32786, 32819, 32895, 32905, 33251, 33258,
    33267, 33276, 33292, 33307, 33311, 33390, 33394, 33400, 34381, 34411, 34880,
    34892, 34915, 35198, 35211, 35282, 35328, 35895, 35910, 35925, 35960, 35997,
    36196, 36208, 36275, 36523, 36554, 36763, 36784, 36789, 37009, 37193, 37318,
    37324, 37329, 38263, 38272, 38428, 38582, 38585, 38632, 38737, 38750, 38754,
    38761, 38859, 38893, 38899, 38913, 39080, 39131, 39135, 39318, 39321, 39340,
    39592, 39640, 39647, 39717, 39727, 39730, 39740, 39770, 40165, 40565, 40575,
    40613, 40635, 40643, 40653, 40657, 40697, 40701, 40718, 40723, 40736, 40763,
    40778, 40786, 40845, 40860, 40864, 12306, 21316, 21317, 12363, 12441, 12365,
    12441, 12367, 12441, 12369, 12441, 12371, 12441, 12373, 12441, 12375, 12441,
    12377, 12441, 12379, 12441, 12381, 12441, 12383, 12441, 12385, 12441, 12388,
    12441, 12390, 12441, 12392, 12441, 12399, 12441, 12399, 12442, 12402, 12441,
    12402, 12442, 12405, 12441, 12405, 12442, 12408, 12441, 12408, 12442, 12411,
    12441, 12411, 12442, 12358, 12441, 32,    12441, 32,    12442, 12445, 12441,
    12424, 12426, 12459, 12441, 12461, 12441, 12463, 12441, 12465, 12441, 12467,
    12441, 12469, 12441, 12471, 12441, 12473, 12441, 12475, 12441, 12477, 12441,
    12479, 12441, 12481, 12441, 12484, 12441, 12486, 12441, 12488, 12441, 12495,
    12441, 12495, 12442, 12498, 12441, 12498, 12442, 12501, 12441, 12501, 12442,
    12504, 12441, 12504, 12442, 12507, 12441, 12507, 12442, 12454, 12441, 12527,
    12441, 12528, 12441, 12529, 12441, 12530, 12441, 12541, 12441, 12467, 12488,
    4352,  4353,  4522,  4354,  4524,  4525,  4355,  4356,  4357,  4528,  4529,
    4530,  4531,  4532,  4533,  4378,  4358,  4359,  4360,  4385,  4361,  4362,
    4363,  4364,  4365,  4366,  4367,  4368,  4369,  4370,  4449,  4450,  4451,
    4452,  4453,  4454,  4455,  4456,  4457,  4458,  4459,  4460,  4461,  4462,
    4463,  4464,  4465,  4466,  4467,  4468,  4469,  4448,  4372,  4373,  4551,
    4552,  4556,  4558,  4563,  4567,  4569,  4380,  4573,  4575,  4381,  4382,
    4384,  4386,  4387,  4391,  4393,  4395,  4396,  4397,  4398,  4399,  4402,
    4406,  4416,  4423,  4428,  4593,  4594,  4439,  4440,  4441,  4484,  4485,
    4488,  4497,  4498,  4500,  4510,  4513,  19977, 22235, 19978, 20013, 19979,
    30002, 19993, 19969, 22825, 22320, 40,    4352,  41,    40,    4354,  41,
    40,    4355,  41,    40,    4357,  41,    40,    4358,  41,    40,    4359,
    41,    40,    4361,  41,    40,    4363,  41,    40,    4364,  41,    40,
    4366,  41,    40,    4367,  41,    40,    4368,  41,    40,    4369,  41,
    40,    4370,  41,    40,    4352,  4449,  41,    40,    4354,  4449,  41,
    40,    4355,  4449,  41,    40,    4357,  4449,  41,    40,    4358,  4449,
    41,    40,    4359,  4449,  41,    40,    4361,  4449,  41,    40,    4363,
    4449,  41,    40,    4364,  4449,  41,    40,    4366,  4449,  41,    40,
    4367,  4449,  41,    40,    4368,  4449,  41,    40,    4369,  4449,  41,
    40,    4370,  4449,  41,    40,    4364,  4462,  41,    40,    4363,  4457,
    4364,  4453,  4523,  41,    40,    4363,  4457,  4370,  4462,  41,    40,
    19968, 41,    40,    20108, 41,    40,    19977, 41,    40,    22235, 41,
    40,    20116, 41,    40,    20845, 41,    40,    19971, 41,    40,    20843,
    41,    40,    20061, 41,    40,    21313, 41,    40,    26376, 41,    40,
    28779, 41,    40,    27700, 41,    40,    26408, 41,    40,    37329, 41,
    40,    22303, 41,    40,    26085, 41,    40,    26666, 41,    40,    26377,
    41,    40,    31038, 41,    40,    21517, 41,    40,    29305, 41,    40,
    36001, 41,    40,    31069, 41,    40,    21172, 41,    40,    20195, 41,
    40,    21628, 41,    40,    23398, 41,    40,    30435, 41,    40,    20225,
    41,    40,    36039, 41,    40,    21332, 41,    40,    31085, 41,    40,
    20241, 41,    40,    33258, 41,    40,    33267, 41,    21839, 24188, 31631,
    80,    84,    69,    50,    49,    50,    50,    50,    51,    50,    52,
    50,    53,    50,    54,    50,    55,    50,    56,    50,    57,    51,
    48,    51,    49,    51,    50,    51,    51,    51,    52,    51,    53,
    4352,  4449,  4354,  4449,  4355,  4449,  4357,  4449,  4358,  4449,  4359,
    4449,  4361,  4449,  4363,  4449,  4364,  4449,  4366,  4449,  4367,  4449,
    4368,  4449,  4369,  4449,  4370,  4449,  4366,  4449,  4535,  4352,  4457,
    4364,  4462,  4363,  4468,  4363,  4462,  20116, 20845, 19971, 20061, 26666,
    26377, 31038, 21517, 29305, 36001, 31069, 21172, 31192, 30007, 36969, 20778,
    21360, 27880, 38917, 20241, 20889, 27491, 24038, 21491, 21307, 23447, 23398,
    30435, 20225, 36039, 21332, 22812, 51,    54,    51,    55,    51,    56,
    51,    57,    52,    48,    52,    49,    52,    50,    52,    51,    52,
    52,    52,    53,    52,    54,    52,    55,    52,    56,    52,    57,
    53,    48,    49,    26376, 50,    26376, 51,    26376, 52,    26376, 53,
    26376, 54,    26376, 55,    26376, 56,    26376, 57,    26376, 49,    48,
    26376, 49,    49,    26376, 49,    50,    26376, 72,    103,   101,   114,
    103,   101,   86,    76,    84,    68,    12450, 12452, 12454, 12456, 12458,
    12459, 12461, 12463, 12465, 12467, 12469, 12471, 12473, 12475, 12477, 12479,
    12481, 12484, 12486, 12488, 12490, 12491, 12492, 12493, 12494, 12495, 12498,
    12501, 12504, 12507, 12510, 12511, 12512, 12513, 12514, 12516, 12518, 12520,
    12521, 12522, 12523, 12524, 12525, 12527, 12528, 12529, 12530, 20196, 21644,
    12450, 12497, 12540, 12488, 12450, 12523, 12501, 12449, 12450, 12531, 12506,
    12450, 12450, 12540, 12523, 12452, 12491, 12531, 12464, 12452, 12531, 12481,
    12454, 12457, 12531, 12456, 12473, 12463, 12540, 12489, 12456, 12540, 12459,
    12540, 12458, 12531, 12473, 12458, 12540, 12512, 12459, 12452, 12522, 12459,
    12521, 12483, 12488, 12459, 12525, 12522, 12540, 12460, 12525, 12531, 12460,
    12531, 12510, 12462, 12460, 12462, 12491, 12540, 12461, 12517, 12522, 12540,
    12462, 12523, 12480, 12540, 12461, 12525, 12461, 12525, 12464, 12521, 12512,
    12461, 12525, 12513, 12540, 12488, 12523, 12461, 12525, 12527, 12483, 12488,
    12464, 12521, 12512, 12464, 12521, 12512, 12488, 12531, 12463, 12523, 12476,
    12452, 12525, 12463, 12525, 12540, 12493, 12465, 12540, 12473, 12467, 12523,
    12490, 12467, 12540, 12509, 12469, 12452, 12463, 12523, 12469, 12531, 12481,
    12540, 12512, 12471, 12522, 12531, 12464, 12475, 12531, 12481, 12475, 12531,
    12488, 12480, 12540, 12473, 12487, 12471, 12489, 12523, 12488, 12531, 12490,
    12494, 12494, 12483, 12488, 12495, 12452, 12484, 12497, 12540, 12475, 12531,
    12488, 12497, 12540, 12484, 12496, 12540, 12524, 12523, 12500, 12450, 12473,
    12488, 12523, 12500, 12463, 12523, 12500, 12467, 12499, 12523, 12501, 12449,
    12521, 12483, 12489, 12501, 12451, 12540, 12488, 12502, 12483, 12471, 12455,
    12523, 12501, 12521, 12531, 12504, 12463, 12479, 12540, 12523, 12506, 12477,
    12506, 12491, 12498, 12504, 12523, 12484, 12506, 12531, 12473, 12506, 12540,
    12472, 12505, 12540, 12479, 12509, 12452, 12531, 12488, 12508, 12523, 12488,
    12507, 12531, 12509, 12531, 12489, 12507, 12540, 12523, 12507, 12540, 12531,
    12510, 12452, 12463, 12525, 12510, 12452, 12523, 12510, 12483, 12495, 12510,
    12523, 12463, 12510, 12531, 12471, 12519, 12531, 12511, 12463, 12525, 12531,
    12511, 12522, 12511, 12522, 12496, 12540, 12523, 12513, 12460, 12513, 12460,
    12488, 12531, 12513, 12540, 12488, 12523, 12516, 12540, 12489, 12516, 12540,
    12523, 12518, 12450, 12531, 12522, 12483, 12488, 12523, 12522, 12521, 12523,
    12500, 12540, 12523, 12540, 12502, 12523, 12524, 12512, 12524, 12531, 12488,
    12466, 12531, 12527, 12483, 12488, 48,    28857, 49,    28857, 50,    28857,
    51,    28857, 52,    28857, 53,    28857, 54,    28857, 55,    28857, 56,
    28857, 57,    28857, 49,    48,    28857, 49,    49,    28857, 49,    50,
    28857, 49,    51,    28857, 49,    52,    28857, 49,    53,    28857, 49,
    54,    28857, 49,    55,    28857, 49,    56,    28857, 49,    57,    28857,
    50,    48,    28857, 50,    49,    28857, 50,    50,    28857, 50,    51,
    28857, 50,    52,    28857, 104,   80,    97,    100,   97,    65,    85,
    98,    97,    114,   111,   86,    112,   99,    100,   109,   100,   109,
    178,   100,   109,   179,   73,    85,    24179, 25104, 26157, 21644, 22823,
    27491, 26126, 27835, 26666, 24335, 20250, 31038, 112,   65,    110,   65,
    956,   65,    109,   65,    107,   65,    75,    66,    77,    66,    71,
    66,    99,    97,    108,   107,   99,    97,    108,   112,   70,    110,
    70,    956,   70,    956,   103,   109,   103,   107,   103,   72,    122,
    107,   72,    122,   77,    72,    122,   71,    72,    122,   84,    72,
    122,   956,   8467,  109,   8467,  100,   8467,  107,   8467,  102,   109,
    110,   109,   956,   109,   109,   109,   99,    109,   107,   109,   109,
    109,   178,   99,    109,   178,   109,   178,   107,   109,   178,   109,
    109,   179,   99,    109,   179,   109,   179,   107,   109,   179,   109,
    8725,  115,   109,   8725,  115,   178,   80,    97,    107,   80,    97,
    77,    80,    97,    71,    80,    97,    114,   97,    100,   114,   97,
    100,   8725,  115,   114,   97,    100,   8725,  115,   178,   112,   115,
    110,   115,   956,   115,   109,   115,   112,   86,    110,   86,    956,
    86,    109,   86,    107,   86,    77,    86,    112,   87,    110,   87,
    956,   87,    109,   87,    107,   87,    77,    87,    107,   937,   77,
    937,   97,    46,    109,   46,    66,    113,   99,    99,    99,    100,
    67,    8725,  107,   103,   67,    111,   46,    100,   66,    71,    121,
    104,   97,    72,    80,    105,   110,   75,    75,    75,    77,    107,
    116,   108,   109,   108,   110,   108,   111,   103,   108,   120,   109,
    98,    109,   105,   108,   109,   111,   108,   80,    72,    112,   46,
    109,   46,    80,    80,    77,    80,    82,    115,   114,   83,    118,
    87,    98,    86,    8725,  109,   65,    8725,  109,   49,    26085, 50,
    26085, 51,    26085, 52,    26085, 53,    26085, 54,    26085, 55,    26085,
    56,    26085, 57,    26085, 49,    48,    26085, 49,    49,    26085, 49,
    50,    26085, 49,    51,    26085, 49,    52,    26085, 49,    53,    26085,
    49,    54,    26085, 49,    55,    26085, 49,    56,    26085, 49,    57,
    26085, 50,    48,    26085, 50,    49,    26085, 50,    50,    26085, 50,
    51,    26085, 50,    52,    26085, 50,    53,    26085, 50,    54,    26085,
    50,    55,    26085, 50,    56,    26085, 50,    57,    26085, 51,    48,
    26085, 51,    49,    26085, 103,   97,    108,   42561, 42563, 42565, 42567,
    42569, 42573, 42575, 42577, 42579, 42581, 42583, 42585, 42587, 42589, 42591,
    42593, 42595, 42597, 42599, 42601, 42603, 42605, 42625, 42627, 42629, 42631,
    42633, 42635, 42637, 42639, 42641, 42643, 42645, 42647, 42649, 42651, 42787,
    42789, 42791, 42793, 42795, 42797, 42799, 42803, 42805, 42807, 42809, 42811,
    42813, 42815, 42817, 42819, 42821, 42823, 42825, 42827, 42829, 42831, 42833,
    42835, 42837, 42839, 42841, 42843, 42845, 42847, 42849, 42851, 42853, 42855,
    42857, 42859, 42861, 42863, 42874, 42876, 7545,  42879, 42881, 42883, 42885,
    42887, 42892, 42897, 42899, 42903, 42905, 42907, 42909, 42911, 42913, 42915,
    42917, 42919, 42921, 620,   670,   647,   43859, 42933, 42935, 42937, 42939,
    42941, 42943, 42947, 42900, 7566,  42952, 42954, 42998, 294,   43831, 43858,
    653,   5024,  5025,  5026,  5027,  5028,  5029,  5030,  5031,  5032,  5033,
    5034,  5035,  5036,  5037,  5038,  5039,  5040,  5041,  5042,  5043,  5044,
    5045,  5046,  5047,  5048,  5049,  5050,  5051,  5052,  5053,  5054,  5055,
    5056,  5057,  5058,  5059,  5060,  5061,  5062,  5063,  5064,  5065,  5066,
    5067,  5068,  5069,  5070,  5071,  5072,  5073,  5074,  5075,  5076,  5077,
    5078,  5079,  5080,  5081,  5082,  5083,  5084,  5085,  5086,  5087,  5088,
    5089,  5090,  5091,  5092,  5093,  5094,  5095,  5096,  5097,  5098,  5099,
    5100,  5101,  5102,  5103,  35912, 26356, 36040, 28369, 20018, 21477, 22865,
    21895, 22856, 25078, 30313, 32645, 34367, 34746, 35064, 37007, 27138, 27931,
    28889, 29662, 33853, 37226, 39409, 20098, 21365, 27396, 29211, 34349, 40478,
    23888, 28651, 34253, 35172, 25289, 33240, 34847, 24266, 26391, 28010, 29436,
    37070, 20358, 20919, 21214, 25796, 27347, 29200, 30439, 34310, 34396, 36335,
    38706, 39791, 40442, 30860, 31103, 32160, 33737, 37636, 35542, 22751, 24324,
    31840, 32894, 29282, 30922, 36034, 38647, 22744, 23650, 27155, 28122, 28431,
    32047, 32311, 38475, 21202, 32907, 20956, 20940, 31260, 32190, 33777, 38517,
    35712, 25295, 35582, 20025, 23527, 24594, 29575, 30064, 21271, 30971, 20415,
    24489, 19981, 27852, 25976, 32034, 21443, 22622, 30465, 33865, 35498, 27578,
    27784, 25342, 33509, 25504, 30053, 20142, 20841, 20937, 26753, 31975, 33391,
    35538, 37327, 21237, 21570, 24300, 26053, 28670, 31018, 38317, 39530, 40599,
    40654, 26310, 27511, 36706, 24180, 24976, 25088, 25754, 28451, 29001, 29833,
    31178, 32244, 32879, 36646, 34030, 36899, 37706, 21015, 21155, 21693, 28872,
    35010, 24265, 24565, 25467, 27566, 31806, 29557, 20196, 22265, 23994, 24604,
    29618, 29801, 32666, 32838, 37428, 38646, 38728, 38936, 20363, 31150, 37300,
    38584, 24801, 20102, 20698, 23534, 23615, 26009, 29134, 30274, 34044, 36988,
    26248, 38446, 21129, 26491, 26611, 27969, 28316, 29705, 30041, 30827, 32016,
    39006, 25134, 38520, 20523, 23833, 28138, 36650, 24459, 24900, 26647, 38534,
    21033, 21519, 23653, 26131, 26446, 26792, 27877, 29702, 30178, 32633, 35023,
    35041, 38626, 21311, 28346, 21533, 29136, 29848, 34298, 38563, 40023, 40607,
    26519, 28107, 33256, 31520, 31890, 29376, 28825, 35672, 20160, 33590, 21050,
    20999, 24230, 25299, 31958, 23429, 27934, 26292, 36667, 38477, 24275, 20800,
    21952, 22618, 26228, 20958, 29482, 30410, 31036, 31070, 31077, 31119, 38742,
    31934, 34322, 35576, 36920, 37117, 39151, 39164, 39208, 40372, 37086, 38583,
    20398, 20711, 20813, 21193, 21220, 21329, 21917, 22022, 22120, 22592, 22696,
    23652, 24724, 24936, 24974, 25074, 25935, 26082, 26257, 26757, 28023, 28186,
    28450, 29038, 29227, 29730, 30865, 31049, 31048, 31056, 31062, 31117, 31118,
    31296, 31361, 31680, 32265, 32321, 32626, 32773, 33261, 33401, 33879, 35088,
    35222, 35585, 35641, 36051, 36104, 36790, 38627, 38911, 38971, 24693, 55376,
    57070, 33304, 20006, 20917, 20840, 20352, 20805, 20864, 21191, 21242, 21845,
    21913, 21986, 22707, 22852, 22868, 23138, 23336, 24274, 24281, 24425, 24493,
    24792, 24910, 24840, 24928, 25140, 25540, 25628, 25682, 25942, 26395, 26454,
    28379, 28363, 28702, 30631, 29237, 29359, 29809, 29958, 30011, 30237, 30239,
    30427, 30452, 30538, 30528, 30924, 31409, 31867, 32091, 32574, 33618, 33775,
    34681, 35137, 35206, 35519, 35531, 35565, 35722, 36664, 36978, 37273, 37494,
    38524, 38875, 38923, 39698, 55370, 56394, 55370, 56388, 55372, 57301, 15261,
    16408, 16441, 55380, 56905, 55383, 56528, 55391, 57043, 40771, 40846, 102,
    102,   102,   105,   102,   108,   102,   102,   105,   102,   102,   108,
    383,   116,   115,   116,   1396,  1398,  1396,  1381,  1396,  1387,  1406,
    1398,  1396,  1389,  1497,  1460,  1522,  1463,  1506,  1492,  1499,  1500,
    1501,  1512,  1514,  1513,  1473,  1513,  1474,  64329, 1473,  64329, 1474,
    1488,  1463,  1488,  1464,  1488,  1468,  1489,  1468,  1490,  1468,  1491,
    1468,  1492,  1468,  1493,  1468,  1494,  1468,  1496,  1468,  1497,  1468,
    1498,  1468,  1499,  1468,  1500,  1468,  1502,  1468,  1504,  1468,  1505,
    1468,  1507,  1468,  1508,  1468,  1510,  1468,  1511,  1468,  1512,  1468,
    1513,  1468,  1514,  1468,  1493,  1465,  1489,  1471,  1499,  1471,  1508,
    1471,  1488,  1500,  1649,  1659,  1662,  1664,  1658,  1663,  1657,  1700,
    1702,  1668,  1667,  1670,  1671,  1677,  1676,  1678,  1672,  1688,  1681,
    1705,  1711,  1715,  1713,  1722,  1723,  1728,  1729,  1726,  1746,  1747,
    1709,  1735,  1734,  1736,  1655,  1739,  1733,  1737,  1744,  1609,  1574,
    1575,  1574,  1749,  1574,  1608,  1574,  1735,  1574,  1734,  1574,  1736,
    1574,  1744,  1574,  1609,  1740,  1574,  1580,  1574,  1581,  1574,  1605,
    1574,  1610,  1576,  1580,  1576,  1581,  1576,  1582,  1576,  1605,  1576,
    1609,  1576,  1610,  1578,  1580,  1578,  1581,  1578,  1582,  1578,  1605,
    1578,  1609,  1578,  1610,  1579,  1580,  1579,  1605,  1579,  1609,  1579,
    1610,  1580,  1581,  1580,  1605,  1581,  1580,  1581,  1605,  1582,  1580,
    1582,  1581,  1582,  1605,  1587,  1580,  1587,  1581,  1587,  1582,  1587,
    1605,  1589,  1581,  1589,  1605,  1590,  1580,  1590,  1581,  1590,  1582,
    1590,  1605,  1591,  1581,  1591,  1605,  1592,  1605,  1593,  1580,  1593,
    1605,  1594,  1580,  1594,  1605,  1601,  1580,  1601,  1581,  1601,  1582,
    1601,  1605,  1601,  1609,  1601,  1610,  1602,  1581,  1602,  1605,  1602,
    1609,  1602,  1610,  1603,  1575,  1603,  1580,  1603,  1581,  1603,  1582,
    1603,  1604,  1603,  1605,  1603,  1609,  1603,  1610,  1604,  1580,  1604,
    1581,  1604,  1582,  1604,  1605,  1604,  1609,  1604,  1610,  1605,  1580,
    1605,  1581,  1605,  1582,  1605,  1605,  1605,  1609,  1605,  1610,  1606,
    1580,  1606,  1581,  1606,  1582,  1606,  1605,  1606,  1609,  1606,  1610,
    1607,  1580,  1607,  1605,  1607,  1609,  1607,  1610,  1610,  1580,  1610,
    1581,  1610,  1582,  1610,  1605,  1610,  1609,  1610,  1610,  1584,  1648,
    1585,  1648,  1609,  1648,  32,    1612,  1617,  32,    1613,  1617,  32,
    1614,  1617,  32,    1615,  1617,  32,    1616,  1617,  32,    1617,  1648,
    1574,  1585,  1574,  1586,  1574,  1606,  1576,  1585,  1576,  1586,  1576,
    1606,  1578,  1585,  1578,  1586,  1578,  1606,  1579,  1585,  1579,  1586,
    1579,  1606,  1605,  1575,  1606,  1585,  1606,  1586,  1606,  1606,  1610,
    1585,  1610,  1586,  1610,  1606,  1574,  1582,  1574,  1607,  1576,  1607,
    1578,  1607,  1589,  1582,  1604,  1607,  1606,  1607,  1607,  1648,  1610,
    1607,  1579,  1607,  1587,  1607,  1588,  1605,  1588,  1607,  1600,  1614,
    1617,  1600,  1615,  1617,  1600,  1616,  1617,  1591,  1609,  1591,  1610,
    1593,  1609,  1593,  1610,  1594,  1609,  1594,  1610,  1587,  1609,  1587,
    1610,  1588,  1609,  1588,  1610,  1581,  1609,  1581,  1610,  1580,  1609,
    1580,  1610,  1582,  1609,  1582,  1610,  1589,  1609,  1589,  1610,  1590,
    1609,  1590,  1610,  1588,  1580,  1588,  1581,  1588,  1582,  1588,  1585,
    1587,  1585,  1589,  1585,  1590,  1585,  1575,  1611,  1578,  1580,  1605,
    1578,  1581,  1580,  1578,  1581,  1605,  1578,  1582,  1605,  1578,  1605,
    1580,  1578,  1605,  1581,  1578,  1605,  1582,  1580,  1605,  1581,  1581,
    1605,  1610,  1581,  1605,  1609,  1587,  1581,  1580,  1587,  1580,  1581,
    1587,  1580,  1609,  1587,  1605,  1581,  1587,  1605,  1580,  1587,  1605,
    1605,  1589,  1581,  1581,  1589,  1605,  1605,  1588,  1581,  1605,  1588,
    1580,  1610,  1588,  1605,  1582,  1588,  1605,  1605,  1590,  1581,  1609,
    1590,  1582,  1605,  1591,  1605,  1581,  1591,  1605,  1605,  1591,  1605,
    1610,  1593,  1580,  1605,  1593,  1605,  1605,  1593,  1605,  1609,  1594,
    1605,  1605,  1594,  1605,  1610,  1594,  1605,  1609,  1601,  1582,  1605,
    1602,  1605,  1581,  1602,  1605,  1605,  1604,  1581,  1605,  1604,  1581,
    1610,  1604,  1581,  1609,  1604,  1580,  1580,  1604,  1582,  1605,  1604,
    1605,  1581,  1605,  1581,  1580,  1605,  1581,  1605,  1605,  1581,  1610,
    1605,  1580,  1581,  1605,  1580,  1605,  1605,  1582,  1580,  1605,  1582,
    1605,  1605,  1580,  1582,  1607,  1605,  1580,  1607,  1605,  1605,  1606,
    1581,  1605,  1606,  1581,  1609,  1606,  1580,  1605,  1606,  1580,  1609,
    1606,  1605,  1610,  1606,  1605,  1609,  1610,  1605,  1605,  1576,  1582,
    1610,  1578,  1580,  1610,  1578,  1580,  1609,  1578,  1582,  1610,  1578,
    1582,  1609,  1578,  1605,  1610,  1578,  1605,  1609,  1580,  1605,  1610,
    1580,  1581,  1609,  1580,  1605,  1609,  1587,  1582,  1609,  1589,  1581,
    1610,  1588,  1581,  1610,  1590,  1581,  1610,  1604,  1580,  1610,  1604,
    1605,  1610,  1610,  1581,  1610,  1610,  1580,  1610,  1610,  1605,  1610,
    1605,  1605,  1610,  1602,  1605,  1610,  1606,  1581,  1610,  1593,  1605,
    1610,  1603,  1605,  1610,  1606,  1580,  1581,  1605,  1582,  1610,  1604,
    1580,  1605,  1603,  1605,  1605,  1580,  1581,  1610,  1581,  1580,  1610,
    1605,  1580,  1610,  1601,  1605,  1610,  1576,  1581,  1610,  1587,  1582,
    1610,  1606,  1580,  1610,  1589,  1604,  1746,  1602,  1604,  1746,  1575,
    1604,  1604,  1607,  1575,  1603,  1576,  1585,  1605,  1581,  1605,  1583,
    1589,  1604,  1593,  1605,  1585,  1587,  1608,  1604,  1593,  1604,  1610,
    1607,  1608,  1587,  1604,  1605,  1589,  1604,  1609,  17,    1589,  1604,
    1609,  32,    1575,  1604,  1604,  1607,  32,    1593,  1604,  1610,  1607,
    32,    1608,  1587,  1604,  1605,  7,     1580,  1604,  32,    1580,  1604,
    1575,  1604,  1607,  1585,  1740,  1575,  1604,  44,    12289, 12290, 58,
    33,    63,    12310, 12311, 8230,  8229,  8212,  8211,  95,    123,   125,
    12308, 12309, 12304, 12305, 12298, 12299, 12300, 12301, 12302, 12303, 91,
    93,    8254,  35,    38,    42,    45,    60,    62,    92,    36,    37,
    64,    32,    1611,  1600,  1611,  32,    1612,  32,    1613,  32,    1614,
    1600,  1614,  32,    1615,  1600,  1615,  32,    1616,  1600,  1616,  32,
    1617,  1600,  1617,  32,    1618,  1600,  1618,  1569,  1570,  1571,  1572,
    1573,  1574,  1575,  1576,  1577,  1578,  1579,  1580,  1581,  1582,  1583,
    1584,  1585,  1586,  1587,  1588,  1589,  1590,  1591,  1592,  1593,  1594,
    1601,  1602,  1603,  1604,  1605,  1606,  1607,  1608,  1610,  1604,  1570,
    1604,  1571,  1604,  1573,  1604,  1575,  34,    39,    47,    65345, 65346,
    65347, 65348, 65349, 65350, 65351, 65352, 65353, 65354, 65355, 65356, 65357,
    65358, 65359, 65360, 65361, 65362, 65363, 65364, 65365, 65366, 65367, 65368,
    65369, 65370, 94,    124,   126,   10629, 10630, 12539, 12449, 12451, 12453,
    12455, 12457, 12515, 12517, 12519, 12483, 12540, 12531, 12441, 12442, 12644,
    12593, 12594, 12595, 12596, 12597, 12598, 12599, 12600, 12601, 12602, 12603,
    12604, 12605, 12606, 12607, 12608, 12609, 12610, 12611, 12612, 12613, 12614,
    12615, 12616, 12617, 12618, 12619, 12620, 12621, 12622, 12623, 12624, 12625,
    12626, 12627, 12628, 12629, 12630, 12631, 12632, 12633, 12634, 12635, 12636,
    12637, 12638, 12639, 12640, 12641, 12642, 12643, 162,   163,   172,   175,
    166,   165,   8361,  9474,  8592,  8593,  8594,  8595,  9632,  9675,  55297,
    56360, 55297, 56361, 55297, 56362, 55297, 56363, 55297, 56364, 55297, 56365,
    55297, 56366, 55297, 56367, 55297, 56368, 55297, 56369, 55297, 56370, 55297,
    56371, 55297, 56372, 55297, 56373, 55297, 56374, 55297, 56375, 55297, 56376,
    55297, 56377, 55297, 56378, 55297, 56379, 55297, 56380, 55297, 56381, 55297,
    56382, 55297, 56383, 55297, 56384, 55297, 56385, 55297, 56386, 55297, 56387,
    55297, 56388, 55297, 56389, 55297, 56390, 55297, 56391, 55297, 56392, 55297,
    56393, 55297, 56394, 55297, 56395, 55297, 56396, 55297, 56397, 55297, 56398,
    55297, 56399, 55297, 56536, 55297, 56537, 55297, 56538, 55297, 56539, 55297,
    56540, 55297, 56541, 55297, 56542, 55297, 56543, 55297, 56544, 55297, 56545,
    55297, 56546, 55297, 56547, 55297, 56548, 55297, 56549, 55297, 56550, 55297,
    56551, 55297, 56552, 55297, 56553, 55297, 56554, 55297, 56555, 55297, 56556,
    55297, 56557, 55297, 56558, 55297, 56559, 55297, 56560, 55297, 56561, 55297,
    56562, 55297, 56563, 55297, 56564, 55297, 56565, 55297, 56566, 55297, 56567,
    55297, 56568, 55297, 56569, 55297, 56570, 55297, 56571, 55299, 56512, 55299,
    56513, 55299, 56514, 55299, 56515, 55299, 56516, 55299, 56517, 55299, 56518,
    55299, 56519, 55299, 56520, 55299, 56521, 55299, 56522, 55299, 56523, 55299,
    56524, 55299, 56525, 55299, 56526, 55299, 56527, 55299, 56528, 55299, 56529,
    55299, 56530, 55299, 56531, 55299, 56532, 55299, 56533, 55299, 56534, 55299,
    56535, 55299, 56536, 55299, 56537, 55299, 56538, 55299, 56539, 55299, 56540,
    55299, 56541, 55299, 56542, 55299, 56543, 55299, 56544, 55299, 56545, 55299,
    56546, 55299, 56547, 55299, 56548, 55299, 56549, 55299, 56550, 55299, 56551,
    55299, 56552, 55299, 56553, 55299, 56554, 55299, 56555, 55299, 56556, 55299,
    56557, 55299, 56558, 55299, 56559, 55299, 56560, 55299, 56561, 55299, 56562,
    55300, 56473, 55300, 56506, 55300, 56475, 55300, 56506, 55300, 56485, 55300,
    56506, 55300, 56625, 55300, 56615, 55300, 56626, 55300, 56615, 55300, 57159,
    55300, 57150, 55300, 57159, 55300, 57175, 55301, 56505, 55301, 56506, 55301,
    56505, 55301, 56496, 55301, 56505, 55301, 56509, 55301, 56760, 55301, 56751,
    55301, 56761, 55301, 56751, 55302, 56512, 55302, 56513, 55302, 56514, 55302,
    56515, 55302, 56516, 55302, 56517, 55302, 56518, 55302, 56519, 55302, 56520,
    55302, 56521, 55302, 56522, 55302, 56523, 55302, 56524, 55302, 56525, 55302,
    56526, 55302, 56527, 55302, 56528, 55302, 56529, 55302, 56530, 55302, 56531,
    55302, 56532, 55302, 56533, 55302, 56534, 55302, 56535, 55302, 56536, 55302,
    56537, 55302, 56538, 55302, 56539, 55302, 56540, 55302, 56541, 55302, 56542,
    55302, 56543, 55302, 56629, 55302, 56624, 55323, 56928, 55323, 56929, 55323,
    56930, 55323, 56931, 55323, 56932, 55323, 56933, 55323, 56934, 55323, 56935,
    55323, 56936, 55323, 56937, 55323, 56938, 55323, 56939, 55323, 56940, 55323,
    56941, 55323, 56942, 55323, 56943, 55323, 56944, 55323, 56945, 55323, 56946,
    55323, 56947, 55323, 56948, 55323, 56949, 55323, 56950, 55323, 56951, 55323,
    56952, 55323, 56953, 55323, 56954, 55323, 56955, 55323, 56956, 55323, 56957,
    55323, 56958, 55323, 56959, 55348, 56663, 55348, 56677, 55348, 56664, 55348,
    56677, 55348, 56671, 55348, 56686, 55348, 56671, 55348, 56687, 55348, 56671,
    55348, 56688, 55348, 56671, 55348, 56689, 55348, 56671, 55348, 56690, 55348,
    56761, 55348, 56677, 55348, 56762, 55348, 56677, 55348, 56763, 55348, 56686,
    55348, 56764, 55348, 56686, 55348, 56763, 55348, 56687, 55348, 56764, 55348,
    56687, 305,   567,   913,   914,   916,   917,   918,   919,   921,   922,
    923,   924,   925,   926,   927,   929,   1012,  932,   934,   935,   936,
    8711,  8706,  1013,  977,   1008,  981,   1009,  982,   988,   55354, 56610,
    55354, 56611, 55354, 56612, 55354, 56613, 55354, 56614, 55354, 56615, 55354,
    56616, 55354, 56617, 55354, 56618, 55354, 56619, 55354, 56620, 55354, 56621,
    55354, 56622, 55354, 56623, 55354, 56624, 55354, 56625, 55354, 56626, 55354,
    56627, 55354, 56628, 55354, 56629, 55354, 56630, 55354, 56631, 55354, 56632,
    55354, 56633, 55354, 56634, 55354, 56635, 55354, 56636, 55354, 56637, 55354,
    56638, 55354, 56639, 55354, 56640, 55354, 56641, 55354, 56642, 55354, 56643,
    1646,  1697,  1647,  48,    46,    48,    44,    49,    44,    50,    44,
    51,    44,    52,    44,    53,    44,    54,    44,    55,    44,    56,
    44,    57,    44,    40,    65,    41,    40,    66,    41,    40,    67,
    41,    40,    68,    41,    40,    69,    41,    40,    70,    41,    40,
    71,    41,    40,    72,    41,    40,    73,    41,    40,    74,    41,
    40,    75,    41,    40,    76,    41,    40,    77,    41,    40,    78,
    41,    40,    79,    41,    40,    80,    41,    40,    81,    41,    40,
    82,    41,    40,    83,    41,    40,    84,    41,    40,    85,    41,
    40,    86,    41,    40,    87,    41,    40,    88,    41,    40,    89,
    41,    40,    90,    41,    12308, 83,    12309, 67,    68,    87,    90,
    72,    86,    83,    68,    83,    83,    80,    80,    86,    87,    67,
    77,    67,    77,    68,    77,    82,    68,    74,    12411, 12363, 12467,
    12467, 23383, 21452, 12487, 22810, 35299, 20132, 26144, 28961, 21069, 24460,
    20877, 26032, 21021, 32066, 36009, 22768, 21561, 28436, 25237, 25429, 36938,
    25351, 25171, 31105, 31354, 21512, 28288, 30003, 21106, 21942, 37197, 12308,
    26412, 12309, 12308, 19977, 12309, 12308, 20108, 12309, 12308, 23433, 12309,
    12308, 28857, 12309, 12308, 25171, 12309, 12308, 30423, 12309, 12308, 21213,
    12309, 12308, 25943, 12309, 24471, 21487, 20029, 20024, 20033, 55360, 56610,
    20320, 20411, 20482, 20602, 20633, 20687, 13470, 55361, 56890, 20820, 20836,
    20855, 55361, 56604, 13497, 20839, 55361, 56651, 20887, 20900, 20172, 20908,
    55396, 56799, 20995, 13535, 21051, 21062, 21111, 13589, 21253, 21254, 21321,
    21338, 21363, 21373, 21375, 55362, 56876, 28784, 21450, 21471, 55362, 57187,
    21483, 21489, 21510, 21662, 21560, 21576, 21608, 21666, 21750, 21776, 21843,
    21859, 21892, 21931, 21939, 21954, 22294, 22295, 22097, 22132, 22766, 22478,
    22516, 22541, 22411, 22578, 22577, 22700, 55365, 56548, 22770, 22775, 22790,
    22818, 22882, 55365, 57000, 55365, 57066, 23020, 23067, 23079, 23000, 23142,
    14062, 14076, 23304, 23358, 55366, 56776, 23491, 23512, 23539, 55366, 57112,
    23551, 23558, 24403, 14209, 23648, 23744, 23693, 55367, 56804, 23875, 55367,
    56806, 23918, 23915, 23932, 24033, 24034, 14383, 24061, 24104, 24125, 24169,
    14434, 55368, 56707, 14460, 24240, 24243, 24246, 55400, 57234, 55368, 57137,
    33281, 24354, 14535, 55372, 57016, 55384, 56794, 24418, 24427, 14563, 24474,
    24525, 24535, 24569, 24705, 14650, 14620, 55369, 57044, 24775, 24904, 24908,
    24954, 25010, 24996, 25007, 25054, 25104, 25115, 25181, 25265, 25300, 25424,
    55370, 57100, 25405, 25340, 25448, 25475, 25572, 55370, 57329, 25634, 25541,
    25513, 14894, 25705, 25726, 25757, 25719, 14956, 25964, 55372, 56330, 26083,
    26360, 26185, 15129, 15112, 15076, 20882, 20885, 26368, 26268, 32941, 17369,
    26401, 26462, 26451, 55372, 57283, 15177, 26618, 26501, 26706, 55373, 56429,
    26766, 26655, 26900, 26946, 27043, 27114, 27304, 55373, 56995, 27355, 15384,
    27425, 55374, 56487, 27476, 15438, 27506, 27551, 27579, 55374, 56973, 55367,
    56587, 55374, 57082, 27726, 55375, 56508, 27839, 27853, 27751, 27926, 27966,
    28009, 28024, 28037, 55375, 56606, 27956, 28207, 28270, 15667, 28359, 55375,
    57041, 28153, 28526, 55375, 57182, 55375, 57230, 28614, 28729, 28699, 15766,
    28746, 28797, 28791, 28845, 55361, 56613, 28997, 55376, 56931, 29084, 55376,
    57259, 29224, 29264, 55377, 56840, 29312, 29333, 55377, 57141, 55378, 56340,
    29562, 29579, 16044, 29605, 16056, 29767, 29788, 29829, 29898, 16155, 29988,
    55379, 56374, 30014, 55379, 56466, 55368, 56735, 30224, 55379, 57249, 55379,
    57272, 55380, 56388, 16380, 16392, 55380, 56563, 55380, 56562, 55380, 56601,
    55380, 56627, 30494, 30495, 30603, 16454, 16534, 55381, 56349, 30798, 16611,
    55381, 56870, 55381, 56986, 55381, 57029, 31211, 16687, 31306, 31311, 55382,
    56700, 55382, 56999, 31470, 16898, 55382, 57259, 31686, 31689, 16935, 55383,
    56448, 31954, 17056, 31976, 31971, 32000, 55383, 57222, 32099, 17153, 32199,
    32258, 32325, 17204, 55384, 56872, 55384, 56903, 17241, 55384, 57049, 32634,
    55384, 57150, 32661, 32762, 55385, 56538, 55385, 56611, 32864, 55385, 56744,
    32880, 55372, 57183, 17365, 32946, 33027, 17419, 33086, 23221, 55385, 57255,
    55385, 57269, 55372, 57235, 55372, 57244, 33284, 36766, 17515, 33425, 33419,
    33437, 21171, 33457, 33459, 33469, 33510, 55386, 57148, 33565, 33635, 33709,
    33571, 33725, 33767, 33619, 33738, 33740, 33756, 55387, 56374, 55387, 56683,
    55387, 56533, 17707, 34033, 34035, 34070, 55388, 57290, 34148, 55387, 57132,
    17757, 17761, 55387, 57265, 55388, 56530, 17771, 34384, 34407, 34409, 34473,
    34440, 34574, 34530, 34600, 34667, 34694, 17879, 34785, 34817, 17913, 34912,
    55389, 56935, 35031, 35038, 17973, 35066, 13499, 55390, 56494, 55390, 56678,
    18110, 18119, 35488, 55391, 56488, 36011, 36033, 36123, 36215, 55391, 57135,
    55362, 56324, 36299, 36284, 36336, 55362, 56542, 36564, 55393, 56786, 55393,
    56813, 37012, 37105, 37137, 55393, 57134, 37147, 37432, 37591, 37592, 37500,
    37881, 37909, 55394, 57338, 38283, 18837, 38327, 55395, 56695, 18918, 38595,
    23986, 38691, 55396, 56645, 55396, 56858, 19054, 19062, 38880, 55397, 56330,
    19122, 55397, 56470, 38953, 55397, 56758, 39138, 19251, 39209, 39335, 39362,
    39422, 19406, 55398, 57136, 40000, 40189, 19662, 19693, 40295, 55400, 56526,
    19704, 55400, 56581, 55400, 56846, 55400, 56977, 19798, 40702, 40709, 40719,
    40726, 55401, 56832, 7838,  192,   193,   194,   195,   196,   199,   200,
    201,   202,   203,   204,   205,   206,   207,   208,   209,   210,   211,
    212,   213,   214,   216,   217,   218,   219,   220,   221,   222,   376,
    256,   258,   260,   262,   264,   266,   268,   270,   272,   274,   276,
    278,   280,   282,   284,   286,   288,   290,   292,   296,   298,   300,
    302,   306,   308,   310,   313,   315,   317,   319,   321,   323,   325,
    327,   330,   332,   334,   336,   338,   340,   342,   344,   346,   348,
    350,   352,   354,   356,   358,   360,   362,   364,   366,   368,   370,
    372,   374,   377,   379,   381,   579,   386,   388,   391,   395,   401,
    502,   408,   573,   544,   416,   418,   420,   423,   428,   431,   435,
    437,   440,   444,   503,   453,   452,   456,   455,   459,   458,   461,
    463,   465,   467,   469,   471,   473,   475,   478,   480,   482,   484,
    486,   488,   490,   492,   494,   498,   497,   500,   504,   506,   508,
    510,   512,   514,   516,   518,   520,   522,   524,   526,   528,   530,
    532,   534,   536,   538,   540,   542,   548,   550,   552,   554,   556,
    558,   560,   562,   571,   11390, 11391, 577,   582,   584,   586,   588,
    590,   11375, 11373, 11376, 385,   390,   393,   394,   399,   42923, 403,
    42924, 404,   42893, 42922, 407,   406,   42926, 11362, 42925, 412,   11374,
    413,   415,   11364, 422,   42949, 425,   42929, 430,   580,   433,   434,
    581,   439,   42930, 42928, 880,   882,   886,   1021,  1022,  1023,  938,
    939,   975,   984,   986,   990,   992,   994,   996,   998,   1000,  1002,
    1004,  1006,  1017,  895,   1015,  1018,  1040,  1041,  1042,  1043,  1044,
    1045,  1046,  1047,  1048,  1049,  1050,  1051,  1052,  1053,  1054,  1055,
    1056,  1057,  1058,  1059,  1060,  1061,  1062,  1063,  1064,  1065,  1066,
    1067,  1068,  1069,  1070,  1071,  1024,  1025,  1026,  1027,  1028,  1029,
    1030,  1031,  1032,  1033,  1034,  1035,  1036,  1037,  1038,  1039,  1120,
    1122,  1124,  1126,  1128,  1130,  1132,  1134,  1136,  1138,  1140,  1142,
    1144,  1146,  1148,  1150,  1152,  1162,  1164,  1166,  1168,  1170,  1172,
    1174,  1176,  1178,  1180,  1182,  1184,  1186,  1188,  1190,  1192,  1194,
    1196,  1198,  1200,  1202,  1204,  1206,  1208,  1210,  1212,  1214,  1217,
    1219,  1221,  1223,  1225,  1227,  1229,  1216,  1232,  1234,  1236,  1238,
    1240,  1242,  1244,  1246,  1248,  1250,  1252,  1254,  1256,  1258,  1260,
    1262,  1264,  1266,  1268,  1270,  1272,  1274,  1276,  1278,  1280,  1282,
    1284,  1286,  1288,  1290,  1292,  1294,  1296,  1298,  1300,  1302,  1304,
    1306,  1308,  1310,  1312,  1314,  1316,  1318,  1320,  1322,  1324,  1326,
    1329,  1330,  1331,  1332,  1333,  1334,  1335,  1336,  1337,  1338,  1339,
    1340,  1341,  1342,  1343,  1344,  1345,  1346,  1347,  1348,  1349,  1350,
    1351,  1352,  1353,  1354,  1355,  1356,  1357,  1358,  1359,  1360,  1361,
    1362,  1363,  1364,  1365,  1366,  7312,  7313,  7314,  7315,  7316,  7317,
    7318,  7319,  7320,  7321,  7322,  7323,  7324,  7325,  7326,  7327,  7328,
    7329,  7330,  7331,  7332,  7333,  7334,  7335,  7336,  7337,  7338,  7339,
    7340,  7341,  7342,  7343,  7344,  7345,  7346,  7347,  7348,  7349,  7350,
    7351,  7352,  7353,  7354,  7357,  7358,  7359,  43888, 43889, 43890, 43891,
    43892, 43893, 43894, 43895, 43896, 43897, 43898, 43899, 43900, 43901, 43902,
    43903, 43904, 43905, 43906, 43907, 43908, 43909, 43910, 43911, 43912, 43913,
    43914, 43915, 43916, 43917, 43918, 43919, 43920, 43921, 43922, 43923, 43924,
    43925, 43926, 43927, 43928, 43929, 43930, 43931, 43932, 43933, 43934, 43935,
    43936, 43937, 43938, 43939, 43940, 43941, 43942, 43943, 43944, 43945, 43946,
    43947, 43948, 43949, 43950, 43951, 43952, 43953, 43954, 43955, 43956, 43957,
    43958, 43959, 43960, 43961, 43962, 43963, 43964, 43965, 43966, 43967, 5112,
    5113,  5114,  5115,  5116,  5117,  42570, 42877, 11363, 42950, 7680,  7682,
    7684,  7686,  7688,  7690,  7692,  7694,  7696,  7698,  7700,  7702,  7704,
    7706,  7708,  7710,  7712,  7714,  7716,  7718,  7720,  7722,  7724,  7726,
    7728,  7730,  7732,  7734,  7736,  7738,  7740,  7742,  7744,  7746,  7748,
    7750,  7752,  7754,  7756,  7758,  7760,  7762,  7764,  7766,  7768,  7770,
    7772,  7774,  7776,  7778,  7780,  7782,  7784,  7786,  7788,  7790,  7792,
    7794,  7796,  7798,  7800,  7802,  7804,  7806,  7808,  7810,  7812,  7814,
    7816,  7818,  7820,  7822,  7824,  7826,  7828,  223,   7840,  7842,  7844,
    7846,  7848,  7850,  7852,  7854,  7856,  7858,  7860,  7862,  7864,  7866,
    7868,  7870,  7872,  7874,  7876,  7878,  7880,  7882,  7884,  7886,  7888,
    7890,  7892,  7894,  7896,  7898,  7900,  7902,  7904,  7906,  7908,  7910,
    7912,  7914,  7916,  7918,  7920,  7922,  7924,  7926,  7928,  7930,  7932,
    7934,  7944,  7945,  7946,  7947,  7948,  7949,  7950,  7951,  7960,  7961,
    7962,  7963,  7964,  7965,  7976,  7977,  7978,  7979,  7980,  7981,  7982,
    7983,  7992,  7993,  7994,  7995,  7996,  7997,  7998,  7999,  8008,  8009,
    8010,  8011,  8012,  8013,  8025,  8027,  8029,  8031,  8040,  8041,  8042,
    8043,  8044,  8045,  8046,  8047,  8122,  8123,  8136,  8137,  8138,  8139,
    8154,  8155,  8184,  8185,  8170,  8171,  8186,  8187,  8072,  8073,  8074,
    8075,  8076,  8077,  8078,  8079,  8064,  8065,  8066,  8067,  8068,  8069,
    8070,  8071,  8088,  8089,  8090,  8091,  8092,  8093,  8094,  8095,  8080,
    8081,  8082,  8083,  8084,  8085,  8086,  8087,  8104,  8105,  8106,  8107,
    8108,  8109,  8110,  8111,  8096,  8097,  8098,  8099,  8100,  8101,  8102,
    8103,  8120,  8121,  8124,  8115,  8140,  8131,  8152,  8153,  8168,  8169,
    8172,  8188,  8179,  8498,  8544,  8545,  8546,  8547,  8548,  8549,  8550,
    8551,  8552,  8553,  8554,  8555,  8556,  8557,  8558,  8559,  8579,  9398,
    9399,  9400,  9401,  9402,  9403,  9404,  9405,  9406,  9407,  9408,  9409,
    9410,  9411,  9412,  9413,  9414,  9415,  9416,  9417,  9418,  9419,  9420,
    9421,  9422,  9423,  11264, 11265, 11266, 11267, 11268, 11269, 11270, 11271,
    11272, 11273, 11274, 11275, 11276, 11277, 11278, 11279, 11280, 11281, 11282,
    11283, 11284, 11285, 11286, 11287, 11288, 11289, 11290, 11291, 11292, 11293,
    11294, 11295, 11296, 11297, 11298, 11299, 11300, 11301, 11302, 11303, 11304,
    11305, 11306, 11307, 11308, 11309, 11310, 11360, 570,   574,   11367, 11369,
    11371, 11378, 11381, 11392, 11394, 11396, 11398, 11400, 11402, 11404, 11406,
    11408, 11410, 11412, 11414, 11416, 11418, 11420, 11422, 11424, 11426, 11428,
    11430, 11432, 11434, 11436, 11438, 11440, 11442, 11444, 11446, 11448, 11450,
    11452, 11454, 11456, 11458, 11460, 11462, 11464, 11466, 11468, 11470, 11472,
    11474, 11476, 11478, 11480, 11482, 11484, 11486, 11488, 11490, 11499, 11501,
    11506, 4256,  4257,  4258,  4259,  4260,  4261,  4262,  4263,  4264,  4265,
    4266,  4267,  4268,  4269,  4270,  4271,  4272,  4273,  4274,  4275,  4276,
    4277,  4278,  4279,  4280,  4281,  4282,  4283,  4284,  4285,  4286,  4287,
    4288,  4289,  4290,  4291,  4292,  4293,  4295,  4301,  42560, 42562, 42564,
    42566, 42568, 42572, 42574, 42576, 42578, 42580, 42582, 42584, 42586, 42588,
    42590, 42592, 42594, 42596, 42598, 42600, 42602, 42604, 42624, 42626, 42628,
    42630, 42632, 42634, 42636, 42638, 42640, 42642, 42644, 42646, 42648, 42650,
    42786, 42788, 42790, 42792, 42794, 42796, 42798, 42802, 42804, 42806, 42808,
    42810, 42812, 42814, 42816, 42818, 42820, 42822, 42824, 42826, 42828, 42830,
    42832, 42834, 42836, 42838, 42840, 42842, 42844, 42846, 42848, 42850, 42852,
    42854, 42856, 42858, 42860, 42862, 42873, 42875, 42878, 42880, 42882, 42884,
    42886, 42891, 42896, 42898, 42948, 42902, 42904, 42906, 42908, 42910, 42912,
    42914, 42916, 42918, 42920, 42932, 42934, 42936, 42938, 42940, 42942, 42946,
    42951, 42953, 42997, 42931, 65313, 65314, 65315, 65316, 65317, 65318, 65319,
    65320, 65321, 65322, 65323, 65324, 65325, 65326, 65327, 65328, 65329, 65330,
    65331, 65332, 65333, 65334, 65335, 65336, 65337, 65338, 55297, 56320, 55297,
    56321, 55297, 56322, 55297, 56323, 55297, 56324, 55297, 56325, 55297, 56326,
    55297, 56327, 55297, 56328, 55297, 56329, 55297, 56330, 55297, 56331, 55297,
    56332, 55297, 56333, 55297, 56334, 55297, 56335, 55297, 56336, 55297, 56337,
    55297, 56338, 55297, 56339, 55297, 56340, 55297, 56341, 55297, 56342, 55297,
    56343, 55297, 56344, 55297, 56345, 55297, 56346, 55297, 56347, 55297, 56348,
    55297, 56349, 55297, 56350, 55297, 56351, 55297, 56352, 55297, 56353, 55297,
    56354, 55297, 56355, 55297, 56356, 55297, 56357, 55297, 56358, 55297, 56359,
    55297, 56496, 55297, 56497, 55297, 56498, 55297, 56499, 55297, 56500, 55297,
    56501, 55297, 56502, 55297, 56503, 55297, 56504, 55297, 56505, 55297, 56506,
    55297, 56507, 55297, 56508, 55297, 56509, 55297, 56510, 55297, 56511, 55297,
    56512, 55297, 56513, 55297, 56514, 55297, 56515, 55297, 56516, 55297, 56517,
    55297, 56518, 55297, 56519, 55297, 56520, 55297, 56521, 55297, 56522, 55297,
    56523, 55297, 56524, 55297, 56525, 55297, 56526, 55297, 56527, 55297, 56528,
    55297, 56529, 55297, 56530, 55297, 56531, 55299, 56448, 55299, 56449, 55299,
    56450, 55299, 56451, 55299, 56452, 55299, 56453, 55299, 56454, 55299, 56455,
    55299, 56456, 55299, 56457, 55299, 56458, 55299, 56459, 55299, 56460, 55299,
    56461, 55299, 56462, 55299, 56463, 55299, 56464, 55299, 56465, 55299, 56466,
    55299, 56467, 55299, 56468, 55299, 56469, 55299, 56470, 55299, 56471, 55299,
    56472, 55299, 56473, 55299, 56474, 55299, 56475, 55299, 56476, 55299, 56477,
    55299, 56478, 55299, 56479, 55299, 56480, 55299, 56481, 55299, 56482, 55299,
    56483, 55299, 56484, 55299, 56485, 55299, 56486, 55299, 56487, 55299, 56488,
    55299, 56489, 55299, 56490, 55299, 56491, 55299, 56492, 55299, 56493, 55299,
    56494, 55299, 56495, 55299, 56496, 55299, 56497, 55299, 56498, 55302, 56480,
    55302, 56481, 55302, 56482, 55302, 56483, 55302, 56484, 55302, 56485, 55302,
    56486, 55302, 56487, 55302, 56488, 55302, 56489, 55302, 56490, 55302, 56491,
    55302, 56492, 55302, 56493, 55302, 56494, 55302, 56495, 55302, 56496, 55302,
    56497, 55302, 56498, 55302, 56499, 55302, 56500, 55302, 56501, 55302, 56502,
    55302, 56503, 55302, 56504, 55302, 56505, 55302, 56506, 55302, 56507, 55302,
    56508, 55302, 56509, 55302, 56510, 55302, 56511, 55323, 56896, 55323, 56897,
    55323, 56898, 55323, 56899, 55323, 56900, 55323, 56901, 55323, 56902, 55323,
    56903, 55323, 56904, 55323, 56905, 55323, 56906, 55323, 56907, 55323, 56908,
    55323, 56909, 55323, 56910, 55323, 56911, 55323, 56912, 55323, 56913, 55323,
    56914, 55323, 56915, 55323, 56916, 55323, 56917, 55323, 56918, 55323, 56919,
    55323, 56920, 55323, 56921, 55323, 56922, 55323, 56923, 55323, 56924, 55323,
    56925, 55323, 56926, 55323, 56927, 55354, 56576, 55354, 56577, 55354, 56578,
    55354, 56579, 55354, 56580, 55354, 56581, 55354, 56582, 55354, 56583, 55354,
    56584, 55354, 56585, 55354, 56586, 55354, 56587, 55354, 56588, 55354, 56589,
    55354, 56590, 55354, 56591, 55354, 56592, 55354, 56593, 55354, 56594, 55354,
    56595, 55354, 56596, 55354, 56597, 55354, 56598, 55354, 56599, 55354, 56600,
    55354, 56601, 55354, 56602, 55354, 56603, 55354, 56604, 55354, 56605, 55354,
    56606, 55354, 56607, 55354, 56608, 55354, 56609,
};

static const utf8proc_uint16_t utf8proc_stage1table[] = {
    0,     256,   512,   768,   1024,  1280,  1536,  1792,  2048,  2304,  2560,
    2816,  3072,  3328,  3584,  3840,  4096,  4352,  4608,  4864,  5120,  5376,
    5632,  5888,  6144,  6400,  6656,  6912,  7168,  7424,  7680,  7936,  8192,
    8448,  8704,  8960,  9216,  9472,  9728,  9984,  10240, 10496, 10752, 11008,
    11264, 11520, 11776, 12032, 12288, 12544, 12800, 13056, 13312, 13312, 13312,
    13312, 13312, 13312, 13312, 13312, 13312, 13312, 13312, 13312, 13312, 13312,
    13312, 13312, 13312, 13312, 13312, 13312, 13312, 13312, 13312, 13312, 13312,
    13568, 13312, 13312, 13312, 13312, 13312, 13312, 13312, 13312, 13312, 13312,
    13312, 13312, 13312, 13312, 13312, 13312, 13312, 13312, 13312, 13312, 13312,
    13312, 13312, 13312, 13312, 13312, 13312, 13312, 13312, 13312, 13312, 13312,
    13312, 13312, 13312, 13312, 13312, 13312, 13312, 13312, 13312, 13312, 13312,
    13312, 13312, 13312, 13312, 13312, 13312, 13312, 13312, 13312, 13312, 13312,
    13312, 13312, 13312, 13312, 13312, 13312, 13312, 13312, 13312, 13312, 13312,
    13312, 13312, 13312, 13312, 13312, 13312, 13312, 13312, 13312, 13312, 13312,
    13312, 13312, 13312, 13312, 13312, 13824, 14080, 13312, 13312, 13312, 14336,
    5376,  14592, 14848, 15104, 15360, 15616, 15872, 16128, 16384, 16640, 16896,
    17152, 17408, 17664, 16128, 16384, 16640, 16896, 17152, 17408, 17664, 16128,
    16384, 16640, 16896, 17152, 17408, 17664, 16128, 16384, 16640, 16896, 17152,
    17408, 17664, 16128, 16384, 16640, 16896, 17152, 17408, 17664, 16128, 16384,
    16640, 16896, 17152, 17408, 17664, 16128, 17920, 18176, 18176, 18176, 18176,
    18176, 18176, 18176, 18176, 18432, 18432, 18432, 18432, 18432, 18432, 18432,
    18432, 18432, 18432, 18432, 18432, 18432, 18432, 18432, 18432, 18432, 18432,
    18432, 18432, 18432, 18432, 18432, 18432, 18432, 18688, 18944, 19200, 19456,
    19712, 19968, 20224, 20480, 20736, 20992, 21248, 21504, 21760, 5376,  22016,
    22272, 22528, 22784, 23040, 23296, 23552, 23808, 24064, 24320, 24576, 24832,
    25088, 25344, 25600, 25856, 26112, 26368, 26624, 26880, 27136, 27392, 27648,
    27904, 28160, 5376,  5376,  5376,  28416, 28672, 28928, 27136, 27136, 27136,
    27136, 27136, 27136, 27136, 27136, 27136, 27136, 5376,  5376,  5376,  5376,
    29184, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136,
    27136, 27136, 27136, 27136, 27136, 5376,  5376,  29440, 27136, 27136, 27136,
    27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136,
    27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136,
    27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 5376,  5376,  29696,
    29952, 27136, 27136, 30208, 30464, 13312, 13312, 13312, 13312, 13312, 13312,
    13312, 13312, 13312, 13312, 13312, 13312, 13312, 13312, 13312, 13312, 13312,
    13312, 13312, 13312, 13312, 13312, 13312, 30720, 13312, 13312, 13312, 13312,
    30976, 31232, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136,
    27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136,
    27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136,
    27136, 27136, 27136, 13312, 31488, 31744, 27136, 27136, 27136, 27136, 27136,
    27136, 27136, 27136, 27136, 32000, 27136, 27136, 27136, 27136, 27136, 27136,
    27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136,
    27136, 27136, 32256, 32512, 32768, 33024, 33280, 33536, 33792, 34048, 10240,
    10240, 34304, 27136, 27136, 27136, 27136, 27136, 34560, 34816, 35072, 27136,
    27136, 27136, 27136, 27136, 35328, 35584, 27136, 27136, 35840, 36096, 36352,
    27136, 36608, 36864, 37120, 37376, 37632, 37888, 38144, 38400, 38656, 38912,
    39168, 39424, 27136, 27136, 27136, 27136, 13312, 13312, 13312, 13312, 13312,
    13312, 13312, 13312, 13312, 13312, 13312, 13312, 13312, 13312, 13312, 13312,
    13312, 13312, 13312, 13312, 13312, 13312, 13312, 13312, 13312, 13312, 13312,
    13312, 13312, 13312, 13312, 13312, 13312, 13312, 13312, 13312, 13312, 13312,
    13312, 13312, 13312, 13312, 13312, 13312, 13312, 13312, 13312, 13312, 13312,
    13312, 13312, 13312, 13312, 13312, 13312, 13312, 13312, 13312, 13312, 13312,
    13312, 13312, 13312, 13312, 13312, 13312, 13312, 13312, 13312, 13312, 13312,
    13312, 13312, 13312, 13312, 13312, 13312, 13312, 13312, 13312, 13312, 13312,
    13312, 13312, 13312, 13312, 13312, 13312, 13312, 13312, 13312, 13312, 13312,
    13312, 13312, 13312, 13312, 13312, 13312, 13312, 13312, 13312, 13312, 13312,
    13312, 13312, 13312, 13312, 13312, 13312, 13312, 13312, 13312, 13312, 13312,
    13312, 13312, 13312, 13312, 13312, 13312, 13312, 13312, 13312, 13312, 13312,
    13312, 13312, 13312, 13312, 13312, 13312, 13312, 13312, 13312, 13312, 13312,
    13312, 13312, 13312, 13312, 13312, 13312, 13312, 13312, 13312, 13312, 13312,
    13312, 13312, 13312, 13312, 13312, 13312, 13312, 13312, 13312, 13312, 13312,
    13312, 13312, 13312, 13312, 13312, 13312, 13312, 39680, 13312, 13312, 13312,
    13312, 13312, 13312, 13312, 13312, 13312, 13312, 13312, 13312, 13312, 13312,
    13312, 13312, 39936, 40192, 13312, 13312, 13312, 13312, 13312, 13312, 13312,
    13312, 13312, 13312, 13312, 13312, 13312, 13312, 13312, 13312, 13312, 13312,
    13312, 13312, 13312, 40448, 13312, 13312, 13312, 13312, 13312, 13312, 13312,
    13312, 13312, 13312, 13312, 13312, 13312, 13312, 13312, 13312, 13312, 13312,
    13312, 13312, 13312, 13312, 13312, 13312, 13312, 13312, 13312, 13312, 40704,
    27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136,
    27136, 40960, 41216, 41472, 27136, 27136, 27136, 27136, 27136, 13312, 13312,
    13312, 13312, 13312, 13312, 13312, 13312, 13312, 13312, 13312, 13312, 13312,
    13312, 13312, 13312, 13312, 13312, 13312, 41728, 27136, 27136, 27136, 27136,
    27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136,
    27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136,
    27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136,
    27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136,
    27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136,
    27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136,
    27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136,
    27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136,
    27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136,
    27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136,
    27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136,
    27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136,
    27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136,
    27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136,
    27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136,
    27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136,
    27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136,
    27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136,
    27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136,
    27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136,
    27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136,
    27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136,
    27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136,
    27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136,
    27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136,
    27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136,
    27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136,
    27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136,
    27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136,
    27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136,
    27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136,
    27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136,
    27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136,
    27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136,
    27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136,
    27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136,
    27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136,
    27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136,
    27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136,
    27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136,
    27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136,
    27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136,
    27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136,
    27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136,
    27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136,
    27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136,
    27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136,
    27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136,
    27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136,
    27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136,
    27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136,
    27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136,
    27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136,
    27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136,
    27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136,
    27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136,
    27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136,
    27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136,
    27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136,
    27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136,
    27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136,
    27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136,
    27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136,
    27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136,
    27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136,
    27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136,
    27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136,
    27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136,
    27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136,
    27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136,
    27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136,
    27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136,
    27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136,
    27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136,
    27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136,
    27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136,
    27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136,
    27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136,
    27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136,
    27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136,
    27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136,
    27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136,
    27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136,
    27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136,
    27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136,
    27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136,
    27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136,
    27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136,
    27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136,
    27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136,
    27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136,
    27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136,
    27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136,
    27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136,
    27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136,
    27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136,
    27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136,
    27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136,
    27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136,
    27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136,
    27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136,
    27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136,
    27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136,
    27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136,
    27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136,
    27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136,
    27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136,
    27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136,
    27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136,
    27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136,
    27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136,
    27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136,
    27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136,
    27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136,
    27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136,
    27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136,
    27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136,
    27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136,
    27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136,
    27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136,
    27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136,
    27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136,
    27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136,
    27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136,
    27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136,
    27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136,
    27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136,
    27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136,
    27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136,
    27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136,
    27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136,
    27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136,
    27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136,
    27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136,
    27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136,
    27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136,
    27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136,
    27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136,
    27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136,
    27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136,
    27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136,
    27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136,
    27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136,
    27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136,
    27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136,
    27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136,
    27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136,
    27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136,
    27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136,
    27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136,
    27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136,
    27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136,
    27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136,
    27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136,
    27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136,
    27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136,
    27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136,
    27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136,
    27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136,
    27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136,
    27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136,
    27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136,
    27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136,
    27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136,
    27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136,
    27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136,
    27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136,
    27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136,
    27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136,
    27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136,
    27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136,
    27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136,
    27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136,
    27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136,
    27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136,
    27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136,
    27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136,
    27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136,
    27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136,
    27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136,
    27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136,
    27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136,
    27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136,
    27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136,
    27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136,
    27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136,
    27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136,
    27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136,
    27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136,
    27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136,
    27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136,
    27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136,
    27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136,
    27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136,
    27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136,
    27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136,
    27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136,
    27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136,
    27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136,
    27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136,
    27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136,
    27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136,
    27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136,
    27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136,
    27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136,
    27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136,
    27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136,
    27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136,
    27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136,
    27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136,
    27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136,
    27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136,
    27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136,
    27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136,
    27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136,
    27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136,
    27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136,
    27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136,
    27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136,
    27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136,
    27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136,
    27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136,
    27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136,
    27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136,
    27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136,
    27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136,
    27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136,
    27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136,
    27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136,
    27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136,
    27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136,
    27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136,
    27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136,
    27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136,
    27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136,
    27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136,
    27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136,
    27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136,
    27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136,
    27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136,
    27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136,
    27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136,
    27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136,
    27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136,
    27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136,
    27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136,
    27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136,
    27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136,
    27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136,
    27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136,
    27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136,
    27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136,
    27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136,
    27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 41984, 42240,
    27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136,
    27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136,
    27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136,
    27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136,
    27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136,
    27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136,
    27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136,
    27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136,
    27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136,
    27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136,
    27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136,
    27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136,
    27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136,
    27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136,
    27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136,
    27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136,
    27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136,
    27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136,
    27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136,
    27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136,
    27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136,
    27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136,
    27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136, 27136,
    27136, 18432, 18432, 18432, 18432, 18432, 18432, 18432, 18432, 18432, 18432,
    18432, 18432, 18432, 18432, 18432, 18432, 18432, 18432, 18432, 18432, 18432,
    18432, 18432, 18432, 18432, 18432, 18432, 18432, 18432, 18432, 18432, 18432,
    18432, 18432, 18432, 18432, 18432, 18432, 18432, 18432, 18432, 18432, 18432,
    18432, 18432, 18432, 18432, 18432, 18432, 18432, 18432, 18432, 18432, 18432,
    18432, 18432, 18432, 18432, 18432, 18432, 18432, 18432, 18432, 18432, 18432,
    18432, 18432, 18432, 18432, 18432, 18432, 18432, 18432, 18432, 18432, 18432,
    18432, 18432, 18432, 18432, 18432, 18432, 18432, 18432, 18432, 18432, 18432,
    18432, 18432, 18432, 18432, 18432, 18432, 18432, 18432, 18432, 18432, 18432,
    18432, 18432, 18432, 18432, 18432, 18432, 18432, 18432, 18432, 18432, 18432,
    18432, 18432, 18432, 18432, 18432, 18432, 18432, 18432, 18432, 18432, 18432,
    18432, 18432, 18432, 18432, 18432, 18432, 18432, 18432, 18432, 18432, 18432,
    18432, 18432, 18432, 18432, 18432, 18432, 18432, 18432, 18432, 18432, 18432,
    18432, 18432, 18432, 18432, 18432, 18432, 18432, 18432, 18432, 18432, 18432,
    18432, 18432, 18432, 18432, 18432, 18432, 18432, 18432, 18432, 18432, 18432,
    18432, 18432, 18432, 18432, 18432, 18432, 18432, 18432, 18432, 18432, 18432,
    18432, 18432, 18432, 18432, 18432, 18432, 18432, 18432, 18432, 18432, 18432,
    18432, 18432, 18432, 18432, 18432, 18432, 18432, 18432, 18432, 18432, 18432,
    18432, 18432, 18432, 18432, 18432, 18432, 18432, 18432, 18432, 18432, 18432,
    18432, 18432, 18432, 18432, 18432, 18432, 18432, 18432, 18432, 18432, 18432,
    18432, 18432, 18432, 18432, 18432, 18432, 18432, 18432, 18432, 18432, 18432,
    18432, 18432, 18432, 18432, 18432, 18432, 18432, 18432, 18432, 18432, 18432,
    18432, 18432, 18432, 18432, 18432, 18432, 18432, 18432, 18432, 18432, 18432,
    18432, 18432, 18432, 42496, 18432, 18432, 18432, 18432, 18432, 18432, 18432,
    18432, 18432, 18432, 18432, 18432, 18432, 18432, 18432, 18432, 18432, 18432,
    18432, 18432, 18432, 18432, 18432, 18432, 18432, 18432, 18432, 18432, 18432,
    18432, 18432, 18432, 18432, 18432, 18432, 18432, 18432, 18432, 18432, 18432,
    18432, 18432, 18432, 18432, 18432, 18432, 18432, 18432, 18432, 18432, 18432,
    18432, 18432, 18432, 18432, 18432, 18432, 18432, 18432, 18432, 18432, 18432,
    18432, 18432, 18432, 18432, 18432, 18432, 18432, 18432, 18432, 18432, 18432,
    18432, 18432, 18432, 18432, 18432, 18432, 18432, 18432, 18432, 18432, 18432,
    18432, 18432, 18432, 18432, 18432, 18432, 18432, 18432, 18432, 18432, 18432,
    18432, 18432, 18432, 18432, 18432, 18432, 18432, 18432, 18432, 18432, 18432,
    18432, 18432, 18432, 18432, 18432, 18432, 18432, 18432, 18432, 18432, 18432,
    18432, 18432, 18432, 18432, 18432, 18432, 18432, 18432, 18432, 18432, 18432,
    18432, 18432, 18432, 18432, 18432, 18432, 18432, 18432, 18432, 18432, 18432,
    18432, 18432, 18432, 18432, 18432, 18432, 18432, 18432, 18432, 18432, 18432,
    18432, 18432, 18432, 18432, 18432, 18432, 18432, 18432, 18432, 18432, 18432,
    18432, 18432, 18432, 18432, 18432, 18432, 18432, 18432, 18432, 18432, 18432,
    18432, 18432, 18432, 18432, 18432, 18432, 18432, 18432, 18432, 18432, 18432,
    18432, 18432, 18432, 18432, 18432, 18432, 18432, 18432, 18432, 18432, 18432,
    18432, 18432, 18432, 18432, 18432, 18432, 18432, 18432, 18432, 18432, 18432,
    18432, 18432, 18432, 18432, 18432, 18432, 18432, 18432, 18432, 18432, 18432,
    18432, 18432, 18432, 18432, 18432, 18432, 18432, 18432, 18432, 18432, 18432,
    18432, 18432, 18432, 18432, 18432, 18432, 18432, 18432, 18432, 18432, 18432,
    18432, 18432, 18432, 18432, 18432, 18432, 18432, 18432, 18432, 18432, 18432,
    18432, 18432, 18432, 18432, 18432, 18432, 42496,
};

static const utf8proc_uint16_t utf8proc_stage2table[] = {
    1,    2,    2,    2,    2,    2,    2,    2,    2,    3,    4,    3,
    5,    6,    2,    2,    2,    2,    2,    2,    2,    2,    2,    2,
    2,    2,    2,    2,    7,    7,    7,    3,    8,    9,    9,    10,
    11,   10,   9,    9,    12,   13,   9,    14,   15,   16,   15,   15,
    17,   17,   17,   17,   17,   17,   17,   17,   17,   17,   15,   9,
    18,   19,   20,   9,    9,    21,   22,   23,   24,   25,   26,   27,
    28,   29,   30,   31,   32,   33,   34,   35,   36,   37,   38,   39,
    40,   41,   42,   43,   44,   45,   46,   12,   9,    13,   47,   48,
    47,   49,   50,   51,   52,   53,   54,   55,   56,   57,   58,   59,
    60,   61,   62,   63,   64,   65,   66,   67,   68,   69,   70,   71,
    72,   73,   74,   12,   75,   13,   75,   2,    2,    2,    2,    2,
    2,    7,    2,    2,    2,    2,    2,    2,    2,    2,    2,    2,
    2,    2,    2,    2,    2,    2,    2,    2,    2,    2,    2,    2,
    2,    2,    2,    2,    76,   9,    11,   11,   11,   11,   77,   9,
    78,   79,   80,   81,   75,   82,   79,   83,   84,   85,   86,   87,
    88,   89,   9,    9,    90,   91,   92,   93,   94,   95,   96,   9,
    97,   98,   99,   100,  101,  102,  103,  104,  105,  106,  107,  108,
    109,  110,  111,  112,  113,  114,  115,  116,  117,  118,  119,  75,
    120,  121,  122,  123,  124,  125,  126,  127,  128,  129,  130,  131,
    132,  133,  134,  135,  136,  137,  138,  139,  140,  141,  142,  143,
    144,  145,  146,  147,  148,  149,  150,  75,   151,  152,  153,  154,
    155,  156,  157,  158,  159,  160,  161,  162,  163,  164,  165,  166,
    167,  168,  169,  170,  171,  172,  173,  174,  175,  176,  177,  178,
    179,  180,  181,  182,  183,  184,  185,  186,  187,  188,  189,  190,
    191,  192,  193,  194,  195,  196,  197,  198,  199,  200,  201,  202,
    203,  204,  205,  206,  207,  208,  209,  210,  211,  212,  213,  214,
    215,  216,  217,  218,  219,  220,  221,  222,  223,  224,  225,  226,
    227,  228,  229,  230,  231,  232,  233,  234,  235,  236,  237,  238,
    239,  240,  241,  242,  243,  244,  245,  246,  247,  248,  249,  250,
    251,  252,  253,  254,  255,  256,  257,  258,  259,  260,  261,  262,
    263,  264,  265,  266,  267,  268,  269,  270,  271,  272,  273,  274,
    275,  276,  277,  278,  279,  280,  281,  282,  283,  284,  285,  286,
    287,  288,  289,  290,  291,  292,  293,  294,  295,  296,  297,  298,
    299,  215,  300,  301,  302,  303,  304,  305,  306,  307,  308,  309,
    310,  311,  312,  215,  313,  314,  315,  316,  317,  318,  319,  320,
    321,  322,  323,  324,  325,  326,  215,  215,  327,  328,  329,  330,
    331,  332,  333,  334,  335,  336,  337,  338,  339,  340,  215,  341,
    342,  343,  215,  344,  341,  341,  341,  341,  345,  346,  347,  348,
    349,  350,  351,  352,  353,  354,  355,  356,  357,  358,  359,  360,
    361,  362,  363,  364,  365,  366,  367,  368,  369,  370,  371,  372,
    373,  374,  375,  376,  377,  378,  379,  380,  381,  382,  383,  384,
    385,  386,  387,  388,  389,  390,  391,  392,  393,  394,  395,  396,
    397,  398,  399,  400,  401,  402,  403,  404,  405,  406,  407,  408,
    409,  410,  411,  412,  413,  414,  415,  416,  417,  418,  419,  420,
    421,  422,  423,  424,  425,  426,  427,  428,  429,  430,  431,  432,
    433,  434,  435,  436,  437,  215,  438,  439,  440,  441,  442,  443,
    444,  445,  446,  447,  448,  449,  450,  451,  452,  453,  454,  455,
    215,  215,  215,  215,  215,  215,  456,  457,  458,  459,  460,  461,
    462,  463,  464,  465,  466,  467,  468,  469,  470,  471,  472,  473,
    474,  475,  476,  477,  478,  479,  480,  481,  482,  215,  483,  484,
    215,  485,  215,  486,  487,  215,  215,  215,  488,  489,  215,  490,
    215,  491,  492,  215,  493,  494,  495,  496,  497,  215,  215,  498,
    215,  499,  500,  215,  215,  501,  215,  215,  215,  215,  215,  215,
    215,  502,  215,  215,  503,  215,  504,  505,  215,  215,  215,  506,
    507,  508,  509,  510,  511,  215,  215,  215,  215,  215,  512,  215,
    341,  215,  215,  215,  215,  215,  215,  215,  215,  513,  514,  215,
    215,  215,  215,  215,  215,  215,  215,  215,  215,  215,  215,  215,
    215,  215,  215,  215,  515,  516,  517,  518,  519,  520,  521,  522,
    523,  524,  524,  525,  525,  525,  525,  525,  525,  525,  47,   47,
    47,   47,   524,  524,  524,  524,  524,  524,  524,  524,  524,  524,
    525,  525,  47,   47,   47,   47,   47,   47,   526,  527,  528,  529,
    530,  531,  47,   47,   532,  533,  534,  535,  536,  47,   47,   47,
    47,   47,   47,   47,   524,  47,   525,  47,   47,   47,   47,   47,
    47,   47,   47,   47,   47,   47,   47,   47,   47,   47,   47,   47,
    537,  538,  539,  540,  541,  542,  543,  544,  545,  546,  547,  548,
    549,  542,  542,  550,  542,  551,  542,  552,  553,  554,  555,  555,
    555,  555,  554,  556,  555,  555,  555,  555,  555,  557,  557,  558,
    559,  560,  561,  562,  563,  555,  555,  555,  555,  564,  565,  555,
    566,  567,  555,  555,  568,  568,  568,  568,  569,  555,  555,  555,
    555,  542,  542,  542,  570,  571,  572,  573,  574,  575,  542,  555,
    555,  555,  542,  542,  542,  555,  555,  576,  542,  542,  542,  555,
    555,  555,  555,  542,  554,  555,  555,  542,  577,  578,  578,  577,
    578,  578,  577,  542,  542,  542,  542,  542,  542,  542,  542,  542,
    542,  542,  542,  542,  579,  580,  581,  582,  583,  47,   584,  585,
    0,    0,    586,  587,  588,  589,  590,  591,  0,    0,    0,    0,
    88,   592,  593,  594,  595,  596,  597,  0,    598,  0,    599,  600,
    601,  602,  603,  604,  605,  606,  607,  608,  609,  610,  611,  612,
    613,  614,  615,  616,  617,  618,  0,    619,  620,  621,  622,  623,
    624,  625,  626,  627,  628,  629,  630,  631,  632,  633,  634,  635,
    636,  637,  638,  639,  640,  641,  642,  643,  644,  645,  646,  647,
    648,  649,  650,  651,  652,  653,  654,  655,  656,  657,  658,  659,
    660,  661,  662,  663,  664,  665,  666,  667,  668,  669,  670,  671,
    672,  673,  674,  675,  676,  677,  678,  679,  680,  681,  682,  683,
    684,  685,  686,  687,  688,  689,  690,  691,  692,  693,  694,  695,
    696,  697,  698,  699,  700,  701,  75,   702,  703,  704,  705,  706,
    215,  707,  708,  709,  710,  711,  712,  713,  714,  715,  716,  717,
    718,  719,  720,  721,  722,  723,  724,  725,  726,  727,  728,  729,
    730,  731,  732,  733,  734,  735,  736,  737,  738,  739,  740,  741,
    742,  743,  744,  745,  746,  747,  748,  749,  750,  751,  752,  753,
    754,  755,  756,  757,  758,  759,  760,  761,  762,  763,  764,  765,
    766,  767,  768,  769,  770,  771,  772,  773,  774,  775,  776,  777,
    778,  779,  780,  781,  782,  783,  784,  785,  786,  787,  788,  789,
    790,  791,  792,  793,  794,  795,  796,  797,  798,  799,  800,  801,
    802,  803,  804,  805,  806,  807,  808,  809,  810,  811,  812,  813,
    814,  815,  816,  817,  818,  819,  820,  821,  822,  823,  824,  825,
    826,  827,  828,  829,  830,  831,  832,  833,  834,  835,  836,  837,
    838,  839,  840,  542,  542,  542,  542,  542,  841,  841,  842,  843,
    844,  845,  846,  847,  848,  849,  850,  851,  852,  853,  854,  855,
    856,  857,  858,  859,  860,  861,  862,  863,  864,  865,  866,  867,
    868,  869,  870,  871,  872,  873,  874,  875,  876,  877,  878,  879,
    880,  881,  882,  883,  884,  885,  886,  887,  888,  889,  890,  891,
    892,  893,  894,  895,  896,  897,  898,  899,  900,  901,  902,  903,
    904,  905,  906,  907,  908,  909,  910,  911,  912,  913,  914,  915,
    916,  917,  918,  919,  920,  921,  922,  923,  924,  925,  926,  927,
    928,  929,  930,  931,  932,  933,  934,  935,  936,  937,  938,  939,
    940,  941,  942,  943,  944,  945,  946,  947,  948,  949,  950,  951,
    952,  953,  954,  955,  956,  957,  958,  959,  960,  961,  962,  963,
    964,  965,  966,  967,  968,  969,  970,  971,  972,  973,  974,  975,
    976,  977,  978,  979,  980,  981,  982,  983,  984,  985,  986,  987,
    988,  989,  990,  991,  992,  993,  994,  995,  996,  997,  998,  999,
    1000, 1001, 1002, 1003, 1004, 1005, 1006, 1007, 0,    1008, 1009, 1010,
    1011, 1012, 1013, 1014, 1015, 1016, 1017, 1018, 1019, 1020, 1021, 1022,
    1023, 1024, 1025, 1026, 1027, 1028, 1029, 1030, 1031, 1032, 1033, 1034,
    1035, 1036, 1037, 1038, 1039, 1040, 1041, 1042, 1043, 1044, 1045, 0,
    0,    525,  1046, 1046, 1046, 1046, 1046, 1046, 215,  1047, 1048, 1049,
    1050, 1051, 1052, 1053, 1054, 1055, 1056, 1057, 1058, 1059, 1060, 1061,
    1062, 1063, 1064, 1065, 1066, 1067, 1068, 1069, 1070, 1071, 1072, 1073,
    1074, 1075, 1076, 1077, 1078, 1079, 1080, 1081, 1082, 1083, 1084, 1085,
    215,  1046, 1086, 0,    0,    77,   77,   11,   0,    555,  542,  542,
    542,  542,  555,  542,  542,  542,  1087, 555,  542,  542,  542,  542,
    542,  542,  555,  555,  555,  555,  555,  555,  542,  542,  555,  542,
    542,  1087, 1088, 542,  1089, 1090, 1091, 1092, 1093, 1094, 1095, 1096,
    1097, 1098, 1098, 1099, 1100, 1101, 1102, 1103, 1104, 1105, 1106, 1104,
    542,  555,  1104, 1097, 0,    0,    0,    0,    0,    0,    0,    0,
    1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107,
    1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107,
    1107, 1107, 1107, 0,    0,    0,    0,    1107, 1107, 1107, 1107, 1104,
    1104, 0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    1108, 1108, 1108, 1108, 1108, 1108, 75,   75,   1109, 10,   10,   1110,
    15,   1111, 77,   77,   542,  542,  542,  542,  542,  542,  542,  542,
    1112, 1113, 1114, 1111, 1115, 0,    1111, 1111, 1116, 1116, 1117, 1118,
    1119, 1120, 1121, 1122, 1116, 1116, 1116, 1116, 1116, 1116, 1116, 1116,
    1116, 1116, 1116, 1116, 1116, 1116, 1116, 1116, 1116, 1116, 1116, 1116,
    1116, 1116, 1116, 1116, 1123, 1116, 1116, 1116, 1116, 1116, 1116, 1116,
    1124, 1116, 1125, 1126, 1127, 1128, 1112, 1113, 1114, 1129, 1130, 1131,
    1132, 1133, 555,  542,  542,  542,  542,  542,  555,  542,  542,  555,
    1134, 1134, 1134, 1134, 1134, 1134, 1134, 1134, 1134, 1134, 10,   1135,
    1135, 1111, 1116, 1116, 1136, 1116, 1116, 1116, 1116, 1137, 1138, 1139,
    1140, 1116, 1116, 1116, 1116, 1116, 1116, 1116, 1116, 1116, 1116, 1116,
    1116, 1116, 1116, 1116, 1116, 1116, 1116, 1116, 1116, 1116, 1116, 1116,
    1116, 1116, 1116, 1116, 1116, 1116, 1116, 1116, 1116, 1116, 1116, 1116,
    1116, 1116, 1116, 1116, 1116, 1116, 1116, 1116, 1116, 1116, 1116, 1116,
    1116, 1116, 1116, 1116, 1116, 1116, 1116, 1116, 1116, 1116, 1116, 1116,
    1116, 1116, 1116, 1116, 1116, 1116, 1116, 1116, 1116, 1116, 1116, 1116,
    1141, 1142, 1143, 1116, 1116, 1116, 1116, 1116, 1116, 1116, 1116, 1116,
    1116, 1116, 1116, 1116, 1116, 1116, 1144, 1145, 1111, 1146, 542,  542,
    542,  542,  542,  542,  542,  1108, 77,   542,  542,  542,  542,  555,
    542,  1123, 1123, 542,  542,  77,   555,  542,  542,  555,  1116, 1116,
    17,   17,   17,   17,   17,   17,   17,   17,   17,   17,   1116, 1116,
    1116, 1147, 1147, 1116, 1111, 1111, 1111, 1111, 1111, 1111, 1111, 1111,
    1111, 1111, 1111, 1111, 1111, 1111, 0,    1148, 1116, 1149, 1116, 1116,
    1116, 1116, 1116, 1116, 1116, 1116, 1116, 1116, 1116, 1116, 1116, 1116,
    1116, 1116, 1116, 1116, 1116, 1116, 1116, 1116, 1116, 1116, 1116, 1116,
    1116, 1116, 1116, 1116, 542,  555,  542,  542,  555,  542,  542,  555,
    555,  555,  542,  555,  555,  542,  555,  542,  542,  542,  555,  542,
    555,  542,  555,  542,  555,  542,  542,  0,    0,    1116, 1116, 1116,
    1116, 1116, 1116, 1116, 1116, 1116, 1116, 1116, 1116, 1116, 1116, 1116,
    1116, 1116, 1116, 1116, 1116, 1116, 1116, 1116, 1116, 1116, 1116, 1116,
    1116, 1116, 1116, 1116, 1116, 1116, 1116, 1116, 1116, 1116, 1116, 1116,
    1116, 1116, 1116, 1116, 1116, 1116, 1116, 1116, 1116, 1116, 1116, 1116,
    1116, 1116, 1116, 1116, 1116, 1116, 1116, 1116, 1116, 1116, 1116, 1116,
    1116, 1116, 1116, 1116, 1116, 1116, 1116, 1116, 1116, 1116, 1116, 1116,
    1116, 1116, 1116, 1116, 1116, 1116, 1116, 1116, 1116, 1116, 1116, 1116,
    1116, 1116, 1150, 1150, 1150, 1150, 1150, 1150, 1150, 1150, 1150, 1150,
    1150, 1116, 0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    1151, 1151, 1151, 1151, 1151, 1151, 1151, 1151,
    1151, 1151, 1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107,
    1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107,
    1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107, 542,
    542,  542,  542,  542,  542,  542,  555,  542,  1152, 1152, 77,   9,
    9,    9,    1152, 0,    0,    555,  1153, 1153, 1107, 1107, 1107, 1107,
    1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107,
    1107, 1107, 1107, 1107, 1107, 1107, 542,  542,  542,  542,  1152, 542,
    542,  542,  542,  542,  542,  542,  542,  542,  1152, 542,  542,  542,
    1152, 542,  542,  542,  542,  542,  0,    0,    1104, 1104, 1104, 1104,
    1104, 1104, 1104, 1104, 1104, 1104, 1104, 1104, 1104, 1104, 1104, 0,
    1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107,
    1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107,
    1107, 555,  555,  555,  0,    0,    1104, 0,    1116, 1116, 1116, 1116,
    1116, 1116, 1116, 1116, 1116, 1116, 1116, 0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    1116, 1116, 1116, 1116, 1116, 1116, 1116, 1116, 1116, 1116, 1116, 1116,
    1116, 1116, 1116, 1116, 1116, 1116, 1116, 1116, 1116, 0,    1116, 1116,
    1116, 1116, 1116, 1116, 1116, 1116, 1116, 1116, 1116, 1116, 1116, 1116,
    1116, 1116, 1116, 1116, 0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    555,  542,  542,  542,  542,  542,  542,  542,  542,
    542,  542,  542,  542,  542,  542,  1108, 555,  542,  542,  555,  542,
    542,  555,  542,  542,  542,  555,  555,  555,  1126, 1127, 1128, 542,
    542,  542,  555,  542,  542,  555,  555,  542,  542,  542,  542,  542,
    1150, 1150, 1150, 1154, 341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  1155, 1156, 341,  341,  341,  341,  341,  341,
    1157, 1158, 341,  1159, 1160, 341,  341,  341,  341,  341,  1150, 1154,
    1161, 341,  1154, 1154, 1154, 1150, 1150, 1150, 1150, 1150, 1150, 1150,
    1150, 1154, 1154, 1154, 1154, 1162, 1154, 1154, 341,  542,  555,  542,
    542,  1150, 1150, 1150, 1163, 1164, 1165, 1166, 1167, 1168, 1169, 1170,
    341,  341,  1150, 1150, 1046, 1046, 1171, 1171, 1171, 1171, 1171, 1171,
    1171, 1171, 1171, 1171, 1046, 525,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  1150, 1154, 1154,
    0,    341,  341,  341,  341,  341,  341,  341,  341,  0,    0,    341,
    341,  0,    0,    341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  0,    341,  341,  341,  341,  341,  341,  341,  0,    341,  0,
    0,    0,    341,  341,  341,  341,  0,    0,    1172, 341,  1173, 1154,
    1154, 1150, 1150, 1150, 1150, 0,    0,    1174, 1154, 0,    0,    1175,
    1176, 1162, 341,  0,    0,    0,    0,    0,    0,    0,    0,    1177,
    0,    0,    0,    0,    1178, 1179, 0,    1180, 341,  341,  1150, 1150,
    0,    0,    1171, 1171, 1171, 1171, 1171, 1171, 1171, 1171, 1171, 1171,
    341,  341,  11,   11,   1181, 1181, 1181, 1181, 1181, 1181, 840,  11,
    341,  1046, 542,  0,    0,    1150, 1150, 1154, 0,    341,  341,  341,
    341,  341,  341,  0,    0,    0,    0,    341,  341,  0,    0,    341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  0,    341,  341,
    341,  341,  341,  341,  341,  0,    341,  1182, 0,    341,  1183, 0,
    341,  341,  0,    0,    1172, 0,    1154, 1154, 1154, 1150, 1150, 0,
    0,    0,    0,    1150, 1150, 0,    0,    1150, 1150, 1162, 0,    0,
    0,    1150, 0,    0,    0,    0,    0,    0,    0,    1184, 1185, 1186,
    341,  0,    1187, 0,    0,    0,    0,    0,    0,    0,    1171, 1171,
    1171, 1171, 1171, 1171, 1171, 1171, 1171, 1171, 1150, 1150, 341,  341,
    341,  1150, 1046, 0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    1150, 1150, 1154, 0,    341,  341,  341,  341,  341,  341,  341,
    341,  341,  0,    341,  341,  341,  0,    341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  0,    341,  341,  341,  341,  341,  341,
    341,  0,    341,  341,  0,    341,  341,  341,  341,  341,  0,    0,
    1172, 341,  1154, 1154, 1154, 1150, 1150, 1150, 1150, 1150, 0,    1150,
    1150, 1154, 0,    1154, 1154, 1162, 0,    0,    341,  0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    341,  341,  1150, 1150, 0,    0,    1171, 1171, 1171, 1171, 1171, 1171,
    1171, 1171, 1171, 1171, 1046, 11,   0,    0,    0,    0,    0,    0,
    0,    341,  1150, 1150, 1150, 1150, 1150, 1150, 0,    1150, 1154, 1154,
    0,    341,  341,  341,  341,  341,  341,  341,  341,  0,    0,    341,
    341,  0,    0,    341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  0,    341,  341,  341,  341,  341,  341,  341,  0,    341,  341,
    0,    341,  341,  341,  341,  341,  0,    0,    1172, 341,  1188, 1150,
    1154, 1150, 1150, 1150, 1150, 0,    0,    1189, 1190, 0,    0,    1191,
    1192, 1162, 0,    0,    0,    0,    0,    0,    0,    1150, 1193, 1194,
    0,    0,    0,    0,    1195, 1196, 0,    341,  341,  341,  1150, 1150,
    0,    0,    1171, 1171, 1171, 1171, 1171, 1171, 1171, 1171, 1171, 1171,
    840,  341,  1181, 1181, 1181, 1181, 1181, 1181, 0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    1150, 341,  0,    341,  341,  341,
    341,  341,  341,  0,    0,    0,    341,  341,  341,  0,    1197, 341,
    1198, 341,  0,    0,    0,    341,  341,  0,    341,  0,    341,  341,
    0,    0,    0,    341,  341,  0,    0,    0,    341,  341,  341,  0,
    0,    0,    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  0,    0,    0,    0,    1199, 1154, 1150, 1154, 1154, 0,
    0,    0,    1200, 1201, 1154, 0,    1202, 1203, 1204, 1162, 0,    0,
    341,  0,    0,    0,    0,    0,    0,    1205, 0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    1171, 1171,
    1171, 1171, 1171, 1171, 1171, 1171, 1171, 1171, 1181, 1181, 1181, 77,
    77,   77,   77,   77,   77,   11,   77,   0,    0,    0,    0,    0,
    1150, 1154, 1154, 1154, 1150, 341,  341,  341,  341,  341,  341,  341,
    341,  0,    341,  341,  341,  0,    341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  0,    341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  0,    0,
    0,    341,  1150, 1150, 1150, 1154, 1154, 1154, 1154, 0,    1206, 1150,
    1207, 0,    1150, 1150, 1150, 1162, 0,    0,    0,    0,    0,    0,
    0,    1208, 1209, 0,    341,  341,  341,  0,    0,    0,    0,    0,
    341,  341,  1150, 1150, 0,    0,    1171, 1171, 1171, 1171, 1171, 1171,
    1171, 1171, 1171, 1171, 0,    0,    0,    0,    0,    0,    0,    1046,
    1210, 1210, 1210, 1210, 1210, 1210, 1210, 840,  341,  1150, 1154, 1154,
    1046, 341,  341,  341,  341,  341,  341,  341,  341,  0,    341,  341,
    341,  0,    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  0,    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    0,    341,  341,  341,  341,  341,  0,    0,    1172, 341,  1154, 1211,
    1212, 1154, 1213, 1154, 1154, 0,    1214, 1215, 1216, 0,    1217, 1218,
    1150, 1162, 0,    0,    0,    0,    0,    0,    0,    1219, 1220, 0,
    0,    0,    0,    0,    0,    0,    341,  0,    341,  341,  1150, 1150,
    0,    0,    1171, 1171, 1171, 1171, 1171, 1171, 1171, 1171, 1171, 1171,
    0,    341,  341,  0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    1150, 1150, 1154, 1154, 341,  341,  341,  341,
    341,  341,  341,  341,  341,  0,    341,  341,  341,  0,    341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  1162, 1162, 341,  1221, 1154, 1154, 1150, 1150, 1150,
    1150, 0,    1222, 1223, 1154, 0,    1224, 1225, 1226, 1162, 1227, 840,
    0,    0,    0,    0,    341,  341,  341,  1228, 1181, 1181, 1181, 1181,
    1181, 1181, 1181, 341,  341,  341,  1150, 1150, 0,    0,    1171, 1171,
    1171, 1171, 1171, 1171, 1171, 1171, 1171, 1171, 1181, 1181, 1181, 1181,
    1181, 1181, 1181, 1181, 1181, 840,  341,  341,  341,  341,  341,  341,
    0,    1150, 1154, 1154, 0,    341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  0,
    0,    0,    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  0,    341,  341,  341,  341,  341,  341,  341,  341,  341,
    0,    341,  0,    0,    341,  341,  341,  341,  341,  341,  341,  0,
    0,    0,    1229, 0,    0,    0,    0,    1230, 1154, 1154, 1150, 1150,
    1150, 0,    1150, 0,    1154, 1231, 1232, 1154, 1233, 1234, 1235, 1236,
    0,    0,    0,    0,    0,    0,    1171, 1171, 1171, 1171, 1171, 1171,
    1171, 1171, 1171, 1171, 0,    0,    1154, 1154, 1046, 0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  1150, 341,  1237,
    1150, 1150, 1150, 1150, 1238, 1238, 1162, 0,    0,    0,    0,    11,
    341,  341,  341,  341,  341,  341,  525,  1150, 1239, 1239, 1239, 1239,
    1150, 1150, 1150, 1046, 1171, 1171, 1171, 1171, 1171, 1171, 1171, 1171,
    1171, 1171, 1046, 1046, 0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    341,  341,  0,    341,  0,    341,  341,
    341,  341,  341,  0,    341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  0,    341,  0,    341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  1150, 341,  1240, 1150, 1150, 1150, 1150,
    1241, 1241, 1162, 1150, 1150, 341,  0,    0,    341,  341,  341,  341,
    341,  0,    525,  0,    1242, 1242, 1242, 1242, 1150, 1150, 0,    0,
    1171, 1171, 1171, 1171, 1171, 1171, 1171, 1171, 1171, 1171, 0,    0,
    1243, 1244, 341,  341,  0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    341,  840,  840,  840,  1046, 1046, 1046, 1046, 1046, 1046, 1046, 1046,
    1245, 1046, 1046, 1046, 1046, 1046, 1046, 840,  1046, 840,  840,  840,
    555,  555,  840,  840,  840,  840,  840,  840,  1171, 1171, 1171, 1171,
    1171, 1171, 1171, 1171, 1171, 1171, 1181, 1181, 1181, 1181, 1181, 1181,
    1181, 1181, 1181, 1181, 840,  555,  840,  555,  840,  1246, 12,   13,
    12,   13,   1154, 1154, 341,  341,  341,  1247, 341,  341,  341,  341,
    0,    341,  341,  341,  341,  1248, 341,  341,  341,  341,  1249, 341,
    341,  341,  341,  1250, 341,  341,  341,  341,  1251, 341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  1252, 341,  341,
    341,  0,    0,    0,    0,    1253, 1254, 1255, 1256, 1257, 1258, 1259,
    1260, 1261, 1254, 1254, 1254, 1254, 1150, 1154, 1254, 1262, 542,  542,
    1162, 1046, 542,  542,  341,  341,  341,  341,  341,  1150, 1150, 1150,
    1150, 1150, 1150, 1263, 1150, 1150, 1150, 1150, 0,    1150, 1150, 1150,
    1150, 1264, 1150, 1150, 1150, 1150, 1265, 1150, 1150, 1150, 1150, 1266,
    1150, 1150, 1150, 1150, 1267, 1150, 1150, 1150, 1150, 1150, 1150, 1150,
    1150, 1150, 1150, 1150, 1150, 1268, 1150, 1150, 1150, 0,    840,  840,
    840,  840,  840,  840,  840,  840,  555,  840,  840,  840,  840,  840,
    840,  0,    840,  840,  1046, 1046, 1046, 1046, 1046, 840,  840,  840,
    840,  1046, 1046, 0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  1269, 1270, 341,  341,  341,  341,  1271,
    1271, 1150, 1272, 1150, 1150, 1154, 1150, 1150, 1150, 1150, 1150, 1172,
    1271, 1162, 1162, 1154, 1154, 1150, 1150, 341,  1171, 1171, 1171, 1171,
    1171, 1171, 1171, 1171, 1171, 1171, 1046, 1046, 1046, 1046, 1046, 1046,
    341,  341,  341,  341,  341,  341,  1154, 1154, 1150, 1150, 341,  341,
    341,  341,  1150, 1150, 1150, 341,  1271, 1271, 1271, 341,  341,  1271,
    1271, 1271, 1271, 1271, 1271, 1271, 341,  341,  341,  1150, 1150, 1150,
    1150, 341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  1150, 1271, 1154, 1150, 1150, 1271, 1271, 1271, 1271, 1271,
    1271, 555,  341,  1271, 1171, 1171, 1171, 1171, 1171, 1171, 1171, 1171,
    1171, 1171, 1271, 1271, 1271, 1150, 840,  840,  1273, 1274, 1275, 1276,
    1277, 1278, 1279, 1280, 1281, 1282, 1283, 1284, 1285, 1286, 1287, 1288,
    1289, 1290, 1291, 1292, 1293, 1294, 1295, 1296, 1297, 1298, 1299, 1300,
    1301, 1302, 1303, 1304, 1305, 1306, 1307, 1308, 1309, 1310, 0,    1311,
    0,    0,    0,    0,    0,    1312, 0,    0,    1313, 1314, 1315, 1316,
    1317, 1318, 1319, 1320, 1321, 1322, 1323, 1324, 1325, 1326, 1327, 1328,
    1329, 1330, 1331, 1332, 1333, 1334, 1335, 1336, 1337, 1338, 1339, 1340,
    1341, 1342, 1343, 1344, 1345, 1346, 1347, 1348, 1349, 1350, 1351, 1352,
    1353, 1354, 1355, 1046, 1356, 1357, 1358, 1359, 1360, 1360, 1360, 1360,
    1360, 1360, 1360, 1360, 1360, 1360, 1360, 1360, 1360, 1360, 1360, 1360,
    1360, 1360, 1360, 1360, 1360, 1360, 1360, 1360, 1360, 1360, 1360, 1360,
    1360, 1360, 1360, 1360, 1360, 1360, 1360, 1360, 1360, 1360, 1360, 1360,
    1360, 1360, 1360, 1360, 1360, 1360, 1360, 1360, 1360, 1360, 1360, 1360,
    1360, 1360, 1360, 1360, 1360, 1360, 1360, 1360, 1360, 1360, 1360, 1360,
    1360, 1360, 1360, 1360, 1360, 1360, 1360, 1360, 1360, 1360, 1360, 1360,
    1360, 1360, 1360, 1360, 1360, 1360, 1360, 1360, 1360, 1360, 1360, 1360,
    1360, 1360, 1360, 1360, 1360, 1360, 1360, 1361, 1362, 1363, 1363, 1363,
    1363, 1363, 1363, 1363, 1363, 1363, 1363, 1363, 1363, 1363, 1363, 1363,
    1363, 1363, 1363, 1363, 1363, 1363, 1363, 1363, 1363, 1363, 1363, 1363,
    1363, 1363, 1363, 1363, 1363, 1363, 1363, 1363, 1363, 1363, 1363, 1363,
    1363, 1363, 1363, 1363, 1363, 1363, 1363, 1363, 1363, 1363, 1363, 1363,
    1363, 1363, 1363, 1363, 1363, 1363, 1363, 1363, 1363, 1363, 1363, 1363,
    1363, 1363, 1363, 1363, 1363, 1363, 1363, 1363, 1364, 1364, 1364, 1364,
    1364, 1364, 1364, 1364, 1364, 1364, 1364, 1364, 1364, 1364, 1364, 1364,
    1364, 1364, 1364, 1364, 1364, 1364, 1364, 1364, 1364, 1364, 1364, 1364,
    1364, 1364, 1364, 1364, 1364, 1364, 1364, 1364, 1364, 1364, 1364, 1364,
    1364, 1364, 1364, 1364, 1364, 1364, 1364, 1364, 1364, 1364, 1364, 1364,
    1364, 1364, 1364, 1364, 1364, 1364, 1364, 1364, 1364, 1364, 1364, 1364,
    1364, 1364, 1364, 1364, 1364, 1364, 1364, 1364, 1364, 1364, 1364, 1364,
    1364, 1364, 1364, 1364, 1364, 1364, 1364, 1364, 1364, 1364, 1364, 1364,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  0,    341,  341,  341,  341,  0,    0,    341,  341,  341,  341,
    341,  341,  341,  0,    341,  0,    341,  341,  341,  341,  0,    0,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  0,    341,  341,  341,  341,  0,    0,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  0,    341,  341,
    341,  341,  0,    0,    341,  341,  341,  341,  341,  341,  341,  0,
    341,  0,    341,  341,  341,  341,  0,    0,    341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  0,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  0,    341,  341,
    341,  341,  0,    0,    341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  0,
    0,    542,  542,  542,  1046, 1046, 1046, 1046, 1046, 1046, 1046, 1046,
    1046, 1181, 1181, 1181, 1181, 1181, 1181, 1181, 1181, 1181, 1181, 1181,
    1181, 1181, 1181, 1181, 1181, 1181, 1181, 1181, 1181, 0,    0,    0,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  77,   77,   77,   77,   77,   77,   77,   77,
    77,   77,   0,    0,    0,    0,    0,    0,    1365, 1366, 1367, 1368,
    1369, 1370, 1371, 1372, 1373, 1374, 1375, 1376, 1377, 1378, 1379, 1380,
    1381, 1382, 1383, 1384, 1385, 1386, 1387, 1388, 1389, 1390, 1391, 1392,
    1393, 1394, 1395, 1396, 1397, 1398, 1399, 1400, 1401, 1402, 1403, 1404,
    1405, 1406, 1407, 1408, 1409, 1410, 1411, 1412, 1413, 1414, 1415, 1416,
    1417, 1418, 1419, 1420, 1421, 1422, 1423, 1424, 1425, 1426, 1427, 1428,
    1429, 1430, 1431, 1432, 1433, 1434, 1435, 1436, 1437, 1438, 1439, 1440,
    1441, 1442, 1443, 1444, 1445, 1446, 1447, 1448, 1449, 1450, 0,    0,
    1451, 1452, 1453, 1454, 1455, 1456, 0,    0,    1086, 341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  840,  1046, 341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    8,    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  12,   13,   0,    0,    0,    341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  1046,
    1046, 1046, 1457, 1457, 1457, 341,  341,  341,  341,  341,  341,  341,
    341,  0,    0,    0,    0,    0,    0,    0,    341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  0,    341,  341,
    341,  341,  1150, 1150, 1162, 0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  1150, 1150,
    1162, 1046, 1046, 0,    0,    0,    0,    0,    0,    0,    0,    0,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  1150, 1150, 0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  0,    341,  341,
    341,  0,    1150, 1150, 0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  576,  576,  1154, 1150,
    1150, 1150, 1150, 1150, 1150, 1150, 1154, 1154, 1154, 1154, 1154, 1154,
    1154, 1154, 1150, 1154, 1154, 1150, 1150, 1150, 1150, 1150, 1150, 1150,
    1150, 1150, 1162, 1150, 1046, 1046, 1046, 525,  1046, 1046, 1046, 11,
    341,  542,  0,    0,    1171, 1171, 1171, 1171, 1171, 1171, 1171, 1171,
    1171, 1171, 0,    0,    0,    0,    0,    0,    1210, 1210, 1210, 1210,
    1210, 1210, 1210, 1210, 1210, 1210, 0,    0,    0,    0,    0,    0,
    9,    9,    9,    9,    9,    9,    1086, 9,    9,    9,    9,    576,
    576,  576,  1458, 0,    1171, 1171, 1171, 1171, 1171, 1171, 1171, 1171,
    1171, 1171, 0,    0,    0,    0,    0,    0,    341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  525,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  0,    0,    0,    0,    0,    0,    0,    341,  341,  341,  341,
    341,  1150, 1150, 341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  1088, 341,  0,    0,    0,    0,    0,    341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  0,
    1150, 1150, 1150, 1154, 1154, 1154, 1154, 1150, 1150, 1154, 1154, 1154,
    0,    0,    0,    0,    1154, 1154, 1150, 1154, 1154, 1154, 1154, 1154,
    1154, 1087, 542,  555,  0,    0,    0,    0,    77,   0,    0,    0,
    9,    9,    1171, 1171, 1171, 1171, 1171, 1171, 1171, 1171, 1171, 1171,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  0,    0,    341,  341,  341,  341,
    341,  0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  0,    0,    0,    0,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  0,    0,    0,    0,    0,    0,    1171, 1171, 1171, 1171,
    1171, 1171, 1171, 1171, 1171, 1171, 1181, 0,    0,    0,    77,   77,
    77,   77,   77,   77,   77,   77,   77,   77,   77,   77,   77,   77,
    77,   77,   77,   77,   77,   77,   77,   77,   77,   77,   77,   77,
    77,   77,   77,   77,   77,   77,   77,   77,   341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  542,  555,  1154, 1154, 1150,
    0,    0,    1046, 1046, 341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  1154, 1150, 1154,
    1150, 1150, 1150, 1150, 1150, 1150, 1150, 0,    1162, 1271, 1150, 1271,
    1271, 1150, 1150, 1150, 1150, 1150, 1150, 1150, 1150, 1154, 1154, 1154,
    1154, 1154, 1154, 1150, 1150, 542,  542,  542,  542,  542,  542,  542,
    542,  0,    0,    555,  1171, 1171, 1171, 1171, 1171, 1171, 1171, 1171,
    1171, 1171, 0,    0,    0,    0,    0,    0,    1171, 1171, 1171, 1171,
    1171, 1171, 1171, 1171, 1171, 1171, 0,    0,    0,    0,    0,    0,
    1046, 1046, 1046, 1046, 1046, 1046, 1046, 525,  1046, 1046, 1046, 1046,
    1046, 1046, 0,    0,    542,  542,  542,  542,  542,  555,  555,  555,
    555,  555,  555,  542,  542,  555,  841,  555,  555,  0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    1150, 1150, 1150, 1150, 1154, 1459, 1460, 1461, 1462, 1463, 1464, 1465,
    1466, 1467, 1468, 341,  341,  1469, 1470, 341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  1172, 1471, 1150, 1150, 1150, 1150, 1472, 1473,
    1474, 1475, 1476, 1477, 1478, 1479, 1480, 1481, 1482, 341,  341,  341,
    341,  341,  341,  341,  0,    0,    0,    0,    1171, 1171, 1171, 1171,
    1171, 1171, 1171, 1171, 1171, 1171, 1046, 1046, 1046, 1046, 1046, 1046,
    1046, 840,  840,  840,  840,  840,  840,  840,  840,  840,  840,  542,
    555,  542,  542,  542,  542,  542,  542,  542,  840,  840,  840,  840,
    840,  840,  840,  840,  840,  0,    0,    0,    1150, 1150, 1154, 341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  1154, 1150, 1150, 1150, 1150, 1154, 1154,
    1150, 1150, 1482, 1162, 1150, 1150, 341,  341,  1171, 1171, 1171, 1171,
    1171, 1171, 1171, 1171, 1171, 1171, 341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  1172, 1154, 1150, 1150, 1154, 1154, 1154, 1150, 1154, 1150,
    1150, 1150, 1482, 1482, 0,    0,    0,    0,    0,    0,    0,    0,
    1046, 1046, 1046, 1046, 341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  1154, 1154, 1154, 1154, 1154, 1154, 1154, 1154,
    1150, 1150, 1150, 1150, 1150, 1150, 1150, 1150, 1154, 1154, 1150, 1172,
    0,    0,    0,    1046, 1046, 1046, 1046, 1046, 1171, 1171, 1171, 1171,
    1171, 1171, 1171, 1171, 1171, 1171, 0,    0,    0,    341,  341,  341,
    1171, 1171, 1171, 1171, 1171, 1171, 1171, 1171, 1171, 1171, 341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  525,  525,  525,  525,  525,  525,  1046, 1046,
    1483, 1484, 1485, 1486, 1487, 1487, 1488, 1489, 1490, 0,    0,    0,
    0,    0,    0,    0,    1491, 1492, 1493, 1494, 1495, 1496, 1497, 1498,
    1499, 1500, 1501, 1502, 1503, 1504, 1505, 1506, 1507, 1508, 1509, 1510,
    1511, 1512, 1513, 1514, 1515, 1516, 1517, 1518, 1519, 1520, 1521, 1522,
    1523, 1524, 1525, 1526, 1527, 1528, 1529, 1530, 1531, 1532, 1533, 0,
    0,    1534, 1535, 1536, 1046, 1046, 1046, 1046, 1046, 1046, 1046, 1046,
    0,    0,    0,    0,    0,    0,    0,    0,    542,  542,  542,  1046,
    568,  555,  555,  555,  555,  555,  542,  542,  555,  555,  555,  555,
    542,  1154, 568,  568,  568,  568,  568,  568,  568,  341,  341,  341,
    341,  555,  341,  341,  341,  341,  341,  341,  542,  341,  341,  1154,
    542,  542,  341,  0,    0,    0,    0,    0,    215,  215,  215,  215,
    215,  215,  215,  215,  215,  215,  215,  215,  215,  215,  215,  215,
    215,  215,  215,  215,  215,  215,  215,  215,  215,  215,  215,  215,
    215,  215,  215,  215,  215,  215,  215,  215,  215,  215,  215,  215,
    215,  215,  215,  215,  1537, 1538, 1539, 525,  1540, 1541, 1542, 1543,
    1544, 1545, 1546, 1547, 1548, 1549, 1550, 525,  1551, 1552, 1553, 1554,
    1555, 1556, 1557, 1558, 1559, 1560, 1561, 1562, 1563, 1564, 1565, 1566,
    1567, 1568, 525,  1569, 1570, 1571, 1572, 1573, 1574, 1575, 1576, 1577,
    1578, 1579, 1580, 1581, 1582, 1583, 1584, 1585, 1586, 1587, 1588, 1589,
    1590, 1591, 1592, 1593, 1594, 1595, 1596, 215,  215,  215,  215,  215,
    215,  215,  215,  215,  215,  215,  215,  215,  1597, 1598, 215,  215,
    215,  1599, 215,  215,  215,  215,  215,  215,  215,  215,  215,  215,
    215,  215,  215,  215,  215,  215,  1600, 215,  215,  215,  215,  215,
    215,  215,  215,  215,  215,  215,  215,  1601, 1602, 1603, 1604, 1567,
    1605, 1606, 1607, 1608, 1609, 1610, 1611, 1612, 1613, 1614, 1615, 1616,
    1617, 1618, 1619, 1620, 1621, 1622, 1623, 1624, 1625, 1626, 1627, 1628,
    1629, 1630, 1631, 1632, 1633, 1634, 1635, 1636, 542,  542,  555,  542,
    542,  542,  542,  542,  542,  542,  555,  542,  542,  578,  1637, 555,
    557,  542,  542,  542,  542,  542,  542,  542,  542,  542,  542,  542,
    542,  542,  542,  542,  542,  542,  542,  542,  542,  542,  542,  542,
    542,  542,  542,  542,  542,  542,  542,  542,  542,  542,  542,  542,
    542,  542,  554,  1088, 1088, 555,  0,    542,  577,  555,  542,  555,
    1638, 1639, 1640, 1641, 1642, 1643, 1644, 1645, 1646, 1647, 1648, 1649,
    1650, 1651, 1652, 1653, 1654, 1655, 1656, 1657, 1658, 1659, 1660, 1661,
    1662, 1663, 1664, 1665, 1666, 1667, 1668, 1669, 1670, 1671, 1672, 1673,
    1674, 1675, 1676, 1677, 1678, 1679, 1680, 1681, 1682, 1683, 1684, 1685,
    1686, 1687, 1688, 1689, 1690, 1691, 1692, 1693, 1694, 1695, 1696, 1697,
    1698, 1699, 1700, 1701, 1702, 1703, 1704, 1705, 1706, 1707, 1708, 1709,
    1710, 1711, 1712, 1713, 1714, 1715, 1716, 1717, 1718, 1719, 1720, 1721,
    1722, 1723, 1724, 1725, 1726, 1727, 1728, 1729, 1730, 1731, 1732, 1733,
    1734, 1735, 1736, 1737, 1738, 1739, 1740, 1741, 1742, 1743, 1744, 1745,
    1746, 1747, 1748, 1749, 1750, 1751, 1752, 1753, 1754, 1755, 1756, 1757,
    1758, 1759, 1760, 1761, 1762, 1763, 1764, 1765, 1766, 1767, 1768, 1769,
    1770, 1771, 1772, 1773, 1774, 1775, 1776, 1777, 1778, 1779, 1780, 1781,
    1782, 1783, 1784, 1785, 1786, 1787, 1788, 1789, 1790, 1791, 1792, 1793,
    215,  215,  1794, 215,  1795, 1796, 1797, 1798, 1799, 1800, 1801, 1802,
    1803, 1804, 1805, 1806, 1807, 1808, 1809, 1810, 1811, 1812, 1813, 1814,
    1815, 1816, 1817, 1818, 1819, 1820, 1821, 1822, 1823, 1824, 1825, 1826,
    1827, 1828, 1829, 1830, 1831, 1832, 1833, 1834, 1835, 1836, 1837, 1838,
    1839, 1840, 1841, 1842, 1843, 1844, 1845, 1846, 1847, 1848, 1849, 1850,
    1851, 1852, 1853, 1854, 1855, 1856, 1857, 1858, 1859, 1860, 1861, 1862,
    1863, 1864, 1865, 1866, 1867, 1868, 1869, 1870, 1871, 1872, 1873, 1874,
    1875, 1876, 1877, 1878, 1879, 1880, 1881, 1882, 1883, 1884, 1885, 1886,
    1887, 1888, 1889, 1890, 1891, 1892, 1893, 1894, 1895, 1896, 1897, 1898,
    1899, 1900, 1901, 1902, 1903, 1904, 1905, 1906, 1907, 1908, 1909, 1910,
    1911, 1912, 0,    0,    1913, 1914, 1915, 1916, 1917, 1918, 0,    0,
    1919, 1920, 1921, 1922, 1923, 1924, 1925, 1926, 1927, 1928, 1929, 1930,
    1931, 1932, 1933, 1934, 1935, 1936, 1937, 1938, 1939, 1940, 1941, 1942,
    1943, 1944, 1945, 1946, 1947, 1948, 1949, 1950, 1951, 1952, 1953, 1954,
    1955, 1956, 0,    0,    1957, 1958, 1959, 1960, 1961, 1962, 0,    0,
    1963, 1964, 1965, 1966, 1967, 1968, 1969, 1970, 0,    1971, 0,    1972,
    0,    1973, 0,    1974, 1975, 1976, 1977, 1978, 1979, 1980, 1981, 1982,
    1983, 1984, 1985, 1986, 1987, 1988, 1989, 1990, 1991, 1992, 1993, 1994,
    1995, 1996, 1997, 1998, 1999, 2000, 2001, 2002, 2003, 2004, 0,    0,
    2005, 2006, 2007, 2008, 2009, 2010, 2011, 2012, 2013, 2014, 2015, 2016,
    2017, 2018, 2019, 2020, 2021, 2022, 2023, 2024, 2025, 2026, 2027, 2028,
    2029, 2030, 2031, 2032, 2033, 2034, 2035, 2036, 2037, 2038, 2039, 2040,
    2041, 2042, 2043, 2044, 2045, 2046, 2047, 2048, 2049, 2050, 2051, 2052,
    2053, 2054, 2055, 2056, 2057, 0,    2058, 2059, 2060, 2061, 2062, 2063,
    2064, 2065, 2066, 2067, 2068, 2069, 2070, 2071, 2072, 0,    2073, 2074,
    2075, 2076, 2077, 2078, 2079, 2080, 2081, 2082, 2083, 2084, 2085, 2086,
    0,    0,    2087, 2088, 2089, 2090, 2091, 2092, 0,    2093, 2094, 2095,
    2096, 2097, 2098, 2099, 2100, 2101, 2102, 2103, 2104, 2105, 2106, 2107,
    2108, 2109, 2110, 2111, 0,    0,    2112, 2113, 2114, 0,    2115, 2116,
    2117, 2118, 2119, 2120, 2121, 2122, 2123, 0,    2124, 2125, 2126, 2126,
    2126, 2126, 2126, 2127, 2126, 2126, 2126, 1458, 2128, 2129, 2130, 2131,
    1086, 2132, 1086, 1086, 1086, 1086, 9,    2133, 2134, 2135, 2136, 2134,
    2134, 2135, 2136, 2134, 9,    9,    9,    9,    2137, 2138, 2139, 9,
    2140, 2141, 2142, 2143, 2144, 2145, 2146, 76,   10,   10,   10,   2147,
    2148, 9,    2149, 2150, 9,    81,   93,   9,    2151, 9,    2152, 48,
    48,   9,    9,    9,    2153, 12,   13,   2154, 2155, 2156, 9,    9,
    9,    9,    9,    9,    9,    9,    75,   9,    48,   9,    9,    2157,
    9,    9,    9,    9,    9,    9,    9,    2126, 1458, 1458, 1458, 1458,
    1458, 0,    2158, 2159, 2160, 2161, 1458, 1458, 1458, 1458, 1458, 1458,
    2162, 2163, 0,    0,    2164, 2165, 2166, 2167, 2168, 2169, 2170, 2171,
    2172, 2173, 2174, 2175, 2176, 2177, 2178, 2179, 2180, 2181, 2182, 2183,
    2184, 2185, 2186, 2187, 2188, 2189, 2190, 0,    2191, 2192, 2193, 2194,
    2195, 2196, 2197, 2198, 2199, 2200, 2201, 2202, 2203, 0,    0,    0,
    11,   11,   11,   11,   11,   11,   11,   11,   2204, 11,   11,   11,
    11,   11,   11,   11,   11,   11,   11,   11,   11,   11,   11,   11,
    11,   11,   11,   11,   11,   11,   11,   11,   0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    542,  542,  568,  568,  542,  542,  542,  542,  568,  568,  568,  542,
    542,  841,  841,  841,  841,  542,  841,  841,  841,  568,  568,  542,
    555,  542,  568,  568,  555,  555,  555,  555,  542,  0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    2205, 2206, 2207, 2208, 77,   2209, 2210, 2211, 77,   2212, 2213, 2214,
    2214, 2214, 2215, 2216, 2217, 2217, 2218, 2219, 77,   2220, 2221, 77,
    75,   2222, 2223, 2224, 2224, 2224, 77,   77,   2225, 2226, 2227, 77,
    2228, 77,   2229, 77,   2228, 77,   2230, 2231, 2232, 2207, 84,   2233,
    2234, 2235, 2236, 2237, 2238, 2239, 2240, 2241, 2242, 2243, 77,   2244,
    2245, 2246, 2247, 2248, 2249, 75,   75,   75,   75,   2250, 2251, 2233,
    2252, 2253, 77,   75,   77,   77,   2254, 840,  2255, 2256, 2257, 2258,
    2259, 2260, 2261, 2262, 2263, 2264, 2265, 2266, 2267, 2268, 2269, 2270,
    2271, 2272, 2273, 2274, 2275, 2276, 2277, 2278, 2279, 2280, 2281, 2282,
    2283, 2284, 2285, 2286, 2287, 2288, 2289, 2290, 2291, 2292, 2293, 2294,
    2295, 2296, 2297, 2298, 2299, 2300, 2301, 2302, 1457, 1457, 1457, 2303,
    2304, 1457, 1457, 1457, 1457, 2305, 77,   77,   0,    0,    0,    0,
    2306, 75,   2307, 75,   2308, 79,   79,   79,   79,   79,   2309, 2310,
    77,   77,   77,   77,   75,   77,   77,   75,   77,   77,   75,   77,
    77,   79,   79,   77,   77,   77,   2311, 77,   77,   77,   77,   77,
    77,   77,   77,   77,   77,   77,   77,   77,   77,   77,   77,   77,
    77,   77,   77,   77,   77,   77,   77,   77,   77,   77,   77,   77,
    77,   2312, 2313, 2314, 2315, 77,   2316, 77,   2317, 77,   77,   77,
    77,   77,   77,   77,   77,   77,   77,   77,   77,   77,   77,   77,
    77,   77,   77,   77,   77,   77,   77,   77,   77,   77,   77,   77,
    77,   77,   77,   77,   75,   75,   75,   75,   75,   75,   75,   75,
    75,   75,   75,   75,   75,   2318, 2318, 2319, 2320, 75,   75,   75,
    2321, 2322, 2318, 2323, 2324, 2318, 75,   75,   75,   2318, 14,   85,
    75,   2318, 2318, 75,   75,   75,   2318, 2318, 2318, 2318, 75,   2318,
    2318, 2318, 2318, 2325, 2326, 2327, 2328, 75,   75,   75,   75,   2318,
    2329, 2330, 2318, 2331, 2332, 2318, 2318, 2318, 75,   75,   75,   75,
    75,   2318, 75,   2318, 2333, 2318, 2318, 2318, 2318, 2334, 2318, 2335,
    2336, 2337, 2318, 2338, 2339, 2340, 2318, 2318, 2318, 2341, 75,   75,
    75,   75,   2318, 2318, 2318, 2318, 75,   75,   75,   75,   75,   75,
    75,   75,   75,   2318, 2342, 2343, 2344, 75,   2345, 2346, 2318, 2318,
    2318, 2318, 2318, 2318, 75,   2347, 2348, 2349, 2350, 2351, 2352, 2353,
    2354, 2355, 2356, 2357, 2358, 2359, 2360, 2361, 2362, 2363, 2318, 2318,
    2364, 2365, 2366, 2367, 2368, 2369, 2370, 2371, 2372, 2373, 2318, 2318,
    2318, 75,   75,   2318, 2318, 2374, 2375, 75,   75,   75,   75,   75,
    2318, 75,   75,   75,   75,   75,   75,   75,   75,   75,   2376, 2318,
    75,   75,   2318, 2318, 2377, 2378, 2318, 2379, 2380, 2381, 2382, 2383,
    2318, 2318, 2384, 2385, 2386, 2387, 2318, 2318, 2318, 75,   75,   75,
    75,   75,   2318, 2318, 75,   75,   75,   75,   75,   75,   75,   75,
    75,   2318, 2318, 2318, 2318, 2318, 75,   75,   2318, 2318, 75,   75,
    75,   75,   2318, 2318, 2318, 2318, 2318, 2318, 2318, 2318, 2318, 2318,
    2388, 2389, 2390, 2391, 2318, 2318, 2318, 2318, 2318, 2318, 2392, 2393,
    2394, 2395, 75,   75,   2318, 2318, 2318, 2318, 2318, 2318, 2318, 2318,
    2318, 2318, 2318, 2318, 2318, 2318, 2318, 2318, 77,   77,   77,   77,
    77,   77,   77,   77,   12,   13,   12,   13,   77,   77,   77,   77,
    77,   77,   77,   77,   77,   77,   77,   77,   77,   77,   2396, 2396,
    77,   77,   77,   77,   2318, 2318, 77,   77,   77,   77,   77,   77,
    79,   2397, 2398, 77,   77,   77,   77,   77,   77,   77,   77,   77,
    77,   77,   840,  840,  840,  840,  840,  840,  840,  840,  840,  840,
    840,  840,  840,  840,  840,  840,  840,  840,  840,  840,  840,  840,
    840,  840,  840,  840,  840,  840,  840,  840,  840,  840,  840,  840,
    840,  840,  840,  840,  840,  840,  840,  840,  840,  840,  840,  840,
    840,  840,  840,  840,  840,  840,  840,  840,  840,  840,  840,  840,
    840,  840,  840,  840,  840,  840,  840,  840,  840,  840,  840,  77,
    75,   77,   77,   77,   77,   77,   77,   77,   77,   77,   77,   77,
    79,   77,   77,   77,   77,   77,   77,   77,   77,   77,   77,   77,
    77,   840,  77,   77,   77,   77,   77,   75,   75,   75,   75,   75,
    75,   75,   75,   75,   75,   75,   75,   75,   75,   75,   75,   75,
    75,   75,   75,   75,   75,   75,   75,   75,   77,   77,   77,   77,
    77,   77,   77,   77,   77,   77,   77,   77,   77,   77,   77,   77,
    77,   77,   77,   77,   77,   77,   77,   77,   77,   77,   77,   79,
    77,   77,   77,   77,   77,   77,   77,   77,   77,   77,   77,   77,
    75,   75,   75,   75,   75,   75,   77,   77,   77,   77,   77,   77,
    77,   2396, 2396, 2396, 2396, 79,   79,   79,   2396, 79,   79,   2396,
    77,   77,   77,   77,   79,   79,   79,   77,   77,   77,   77,   77,
    77,   77,   77,   77,   77,   77,   77,   77,   77,   77,   77,   77,
    77,   77,   77,   77,   77,   77,   77,   77,   77,   77,   77,   77,
    77,   77,   77,   77,   77,   77,   77,   77,   77,   77,   77,   77,
    77,   77,   77,   0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    77,   77,   77,   77,   77,   77,   77,   77,
    77,   77,   77,   0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    2399, 2400, 2401, 2402, 2403, 2404, 2405, 2406, 2407, 2408, 2409, 2410,
    2411, 2412, 2413, 2414, 2415, 2416, 2417, 2418, 2419, 2420, 2421, 2422,
    2423, 2424, 2425, 2426, 2427, 2428, 2429, 2430, 2431, 2432, 2433, 2434,
    2435, 2436, 2437, 2438, 2439, 2440, 2441, 2442, 2443, 2444, 2445, 2446,
    2447, 2448, 2449, 2450, 2451, 2452, 2453, 2454, 2455, 2456, 2457, 2458,
    2459, 2460, 2461, 2462, 2463, 2464, 2465, 2466, 2467, 2468, 2469, 2470,
    2471, 2472, 2473, 2474, 2475, 2476, 2477, 2478, 2479, 2480, 2481, 2482,
    2483, 2484, 2485, 2486, 2487, 2488, 2489, 2490, 2491, 2492, 2493, 2494,
    2495, 2496, 2497, 2498, 2499, 2500, 2501, 2502, 2503, 2504, 2505, 2506,
    2507, 2508, 2509, 2510, 2511, 2512, 2513, 2514, 2515, 2516, 2517, 2518,
    2519, 2520, 2521, 2522, 2523, 2524, 2525, 2526, 2527, 2528, 2529, 2530,
    2531, 2532, 2533, 2534, 2535, 2536, 2537, 1210, 1210, 1210, 1210, 1210,
    1210, 1210, 1210, 1210, 1210, 1210, 1210, 1210, 1210, 1210, 1210, 1210,
    1210, 1210, 1210, 1210, 77,   77,   77,   77,   77,   77,   77,   77,
    77,   77,   77,   77,   77,   77,   77,   77,   77,   77,   77,   77,
    77,   77,   77,   77,   77,   77,   77,   77,   77,   77,   77,   77,
    77,   77,   77,   77,   77,   77,   77,   77,   77,   77,   77,   77,
    77,   77,   77,   77,   77,   77,   77,   77,   77,   77,   77,   77,
    77,   77,   77,   77,   77,   77,   77,   77,   77,   77,   77,   77,
    77,   77,   77,   77,   77,   77,   77,   77,   77,   77,   77,   77,
    77,   77,   77,   77,   77,   77,   77,   77,   77,   77,   77,   77,
    77,   77,   77,   77,   77,   77,   77,   77,   77,   77,   77,   77,
    77,   77,   77,   77,   77,   77,   77,   77,   77,   77,   77,   77,
    77,   77,   77,   77,   77,   77,   77,   77,   77,   77,   77,   77,
    77,   77,   77,   77,   77,   77,   77,   77,   77,   77,   77,   77,
    77,   77,   77,   77,   77,   77,   77,   77,   77,   77,   77,   77,
    77,   77,   77,   77,   77,   77,   77,   77,   77,   77,   77,   77,
    77,   77,   77,   77,   77,   77,   79,   79,   77,   77,   77,   77,
    77,   77,   77,   77,   77,   77,   79,   75,   77,   77,   77,   77,
    77,   77,   77,   77,   79,   75,   77,   77,   77,   77,   77,   77,
    77,   77,   77,   77,   77,   77,   77,   77,   77,   77,   77,   77,
    77,   77,   77,   77,   77,   77,   77,   77,   77,   77,   77,   77,
    77,   77,   77,   77,   77,   77,   77,   77,   77,   77,   77,   77,
    77,   77,   77,   77,   77,   77,   77,   77,   77,   77,   77,   77,
    75,   75,   75,   2538, 2538, 2539, 2539, 75,   79,   79,   79,   79,
    79,   79,   77,   79,   79,   79,   79,   79,   79,   79,   79,   79,
    79,   79,   79,   77,   2396, 2396, 79,   79,   79,   79,   79,   79,
    79,   79,   79,   79,   79,   79,   79,   79,   79,   79,   79,   79,
    79,   79,   79,   79,   79,   79,   79,   79,   79,   79,   79,   79,
    79,   79,   79,   79,   79,   79,   79,   79,   79,   79,   79,   79,
    79,   79,   79,   79,   79,   79,   79,   79,   2396, 2396, 2396, 2396,
    2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396, 79,   79,   79,   79,
    79,   79,   79,   79,   79,   79,   79,   79,   79,   79,   79,   79,
    79,   79,   79,   79,   79,   79,   79,   79,   79,   79,   79,   2538,
    79,   79,   79,   79,   79,   79,   79,   79,   79,   79,   79,   79,
    79,   79,   79,   2396, 79,   79,   79,   79,   79,   79,   77,   77,
    77,   77,   77,   77,   77,   77,   77,   77,   79,   79,   79,   2396,
    79,   79,   79,   79,   79,   79,   79,   79,   79,   79,   79,   79,
    79,   2396, 79,   79,   79,   79,   79,   79,   79,   79,   2396, 2396,
    2540, 79,   79,   79,   79,   79,   79,   79,   79,   79,   79,   79,
    79,   79,   79,   79,   79,   2396, 2396, 79,   79,   79,   79,   79,
    2396, 2396, 79,   79,   79,   79,   79,   79,   79,   79,   2396, 79,
    79,   79,   79,   79,   2396, 79,   79,   79,   79,   79,   79,   79,
    79,   79,   79,   79,   79,   79,   79,   79,   79,   79,   79,   79,
    79,   79,   2396, 79,   79,   79,   79,   79,   79,   79,   2396, 2396,
    79,   2396, 79,   79,   79,   79,   2396, 79,   79,   2396, 79,   79,
    79,   79,   79,   79,   79,   2396, 77,   77,   79,   79,   2396, 2396,
    79,   79,   79,   79,   79,   79,   79,   77,   79,   77,   79,   77,
    77,   77,   77,   77,   77,   79,   77,   77,   77,   79,   77,   77,
    77,   77,   77,   77,   2396, 77,   77,   77,   77,   77,   77,   77,
    77,   77,   77,   79,   79,   77,   77,   77,   77,   77,   77,   77,
    77,   77,   77,   77,   77,   77,   77,   77,   79,   77,   77,   79,
    77,   77,   77,   77,   2396, 77,   2396, 77,   77,   77,   77,   2396,
    2396, 2396, 77,   2396, 77,   77,   77,   77,   77,   77,   77,   77,
    77,   77,   77,   79,   79,   79,   79,   79,   12,   13,   12,   13,
    12,   13,   12,   13,   12,   13,   12,   13,   12,   13,   1210, 1210,
    1210, 1210, 1210, 1210, 1210, 1210, 1210, 1210, 1210, 1210, 1210, 1210,
    1210, 1210, 1210, 1210, 1210, 1210, 1210, 1210, 1210, 1210, 1210, 1210,
    1210, 1210, 1210, 1210, 77,   2396, 2396, 2396, 77,   77,   77,   77,
    77,   77,   77,   77,   77,   79,   77,   77,   77,   77,   77,   77,
    77,   77,   77,   77,   77,   77,   77,   77,   2396, 77,   77,   77,
    77,   77,   77,   77,   77,   77,   77,   77,   77,   77,   77,   2396,
    2318, 75,   75,   2318, 2318, 12,   13,   75,   2318, 2318, 75,   2318,
    2318, 2318, 75,   75,   75,   75,   75,   2318, 2318, 2318, 2318, 75,
    75,   75,   75,   75,   2318, 2318, 2318, 75,   75,   75,   2318, 2318,
    2318, 2318, 12,   13,   12,   13,   12,   13,   12,   13,   12,   13,
    75,   75,   75,   75,   75,   75,   75,   75,   75,   75,   75,   75,
    75,   75,   75,   75,   840,  840,  840,  840,  840,  840,  840,  840,
    840,  840,  840,  840,  840,  840,  840,  840,  840,  840,  840,  840,
    840,  840,  840,  840,  840,  840,  840,  840,  840,  840,  840,  840,
    840,  840,  840,  840,  840,  840,  840,  840,  840,  840,  840,  840,
    840,  840,  840,  840,  840,  840,  840,  840,  840,  840,  840,  840,
    840,  840,  840,  840,  840,  840,  840,  840,  840,  840,  840,  840,
    840,  840,  840,  840,  840,  840,  840,  840,  840,  840,  840,  840,
    840,  840,  840,  840,  840,  840,  840,  840,  840,  840,  840,  840,
    840,  840,  840,  840,  840,  840,  840,  840,  840,  840,  840,  840,
    840,  840,  840,  840,  840,  840,  840,  840,  840,  840,  840,  840,
    840,  840,  840,  840,  840,  840,  840,  840,  840,  840,  840,  840,
    840,  840,  840,  840,  840,  840,  840,  840,  840,  840,  840,  840,
    840,  840,  840,  840,  840,  840,  840,  840,  840,  840,  840,  840,
    840,  840,  840,  840,  840,  840,  840,  840,  840,  840,  840,  840,
    840,  840,  840,  840,  840,  840,  840,  840,  840,  840,  840,  840,
    840,  840,  840,  840,  840,  840,  840,  840,  840,  840,  840,  840,
    840,  840,  840,  840,  840,  840,  840,  840,  840,  840,  840,  840,
    840,  840,  840,  840,  840,  840,  840,  840,  840,  840,  840,  840,
    840,  840,  840,  840,  840,  840,  840,  840,  840,  840,  840,  840,
    840,  840,  840,  840,  840,  840,  840,  840,  840,  840,  840,  840,
    840,  840,  840,  840,  840,  840,  840,  840,  840,  840,  840,  840,
    840,  840,  840,  840,  840,  840,  840,  840,  75,   75,   75,   75,
    75,   75,   75,   75,   75,   75,   75,   75,   75,   75,   75,   75,
    75,   75,   75,   75,   75,   75,   75,   75,   75,   75,   75,   75,
    75,   75,   75,   75,   75,   75,   75,   75,   75,   75,   75,   75,
    75,   75,   75,   75,   75,   75,   75,   75,   75,   75,   75,   75,
    2538, 2538, 75,   75,   75,   75,   75,   75,   75,   75,   75,   75,
    75,   75,   75,   75,   75,   75,   75,   75,   75,   75,   75,   75,
    75,   75,   75,   75,   75,   75,   75,   75,   75,   75,   75,   75,
    75,   75,   75,   75,   75,   75,   75,   75,   75,   75,   75,   75,
    75,   75,   75,   75,   75,   75,   75,   75,   75,   75,   75,   75,
    75,   75,   75,   75,   75,   75,   75,   75,   75,   75,   75,   75,
    75,   75,   75,   75,   75,   75,   75,   12,   13,   12,   13,   12,
    13,   12,   13,   12,   13,   12,   13,   12,   13,   12,   13,   12,
    13,   12,   13,   12,   13,   75,   75,   2318, 2318, 2318, 2318, 2318,
    2318, 75,   2318, 2318, 2318, 2318, 2318, 2318, 2318, 2318, 2318, 2318,
    2318, 2318, 2318, 2318, 75,   75,   75,   75,   75,   75,   75,   75,
    2318, 75,   75,   75,   75,   75,   75,   75,   2318, 2318, 2318, 2318,
    2318, 2318, 75,   75,   75,   2318, 75,   75,   75,   75,   2318, 2318,
    2318, 2318, 2318, 75,   2318, 2318, 75,   75,   12,   13,   12,   13,
    2318, 75,   75,   75,   75,   2318, 75,   2318, 2318, 2318, 75,   75,
    2318, 2318, 75,   75,   75,   75,   75,   75,   75,   75,   75,   75,
    2318, 2318, 2318, 2318, 2318, 2318, 75,   75,   12,   13,   75,   75,
    75,   75,   75,   75,   75,   75,   75,   75,   75,   75,   2318, 2318,
    2541, 2318, 2318, 2318, 2318, 2318, 2318, 2318, 2318, 2318, 2318, 2318,
    2318, 2318, 2318, 2318, 2318, 75,   2318, 2318, 2318, 2318, 75,   75,
    2318, 75,   2318, 75,   75,   2318, 75,   2318, 2318, 2318, 2318, 75,
    75,   75,   75,   75,   2318, 2318, 75,   75,   75,   75,   75,   75,
    2318, 2318, 2318, 75,   75,   75,   75,   75,   75,   75,   75,   75,
    75,   75,   75,   75,   75,   75,   75,   75,   75,   75,   75,   75,
    75,   75,   75,   2318, 2318, 75,   75,   75,   75,   75,   75,   75,
    75,   75,   75,   75,   2318, 2318, 75,   75,   75,   75,   2318, 2318,
    2318, 2318, 75,   2318, 2318, 75,   75,   2318, 2542, 2543, 2544, 75,
    75,   2318, 2318, 2318, 2318, 2318, 2318, 2318, 2318, 2318, 2318, 2318,
    2318, 2318, 2318, 2318, 2318, 2318, 2318, 2318, 2318, 2318, 2318, 2318,
    2318, 2318, 2318, 2318, 2318, 2318, 2318, 2318, 2318, 2318, 2318, 2318,
    2318, 2318, 2318, 2318, 2318, 2318, 2318, 2318, 75,   75,   2318, 2318,
    2318, 2318, 2318, 2318, 2318, 2318, 75,   2318, 2318, 2318, 2318, 2318,
    2318, 2318, 2318, 2318, 2318, 2318, 2318, 2318, 2318, 2318, 2318, 2318,
    2318, 2318, 2318, 2318, 2318, 2318, 2318, 2318, 2318, 2318, 2318, 2318,
    2318, 2318, 2318, 2318, 2318, 2318, 2318, 2318, 2318, 2318, 2318, 75,
    75,   75,   75,   75,   2545, 2546, 2318, 75,   75,   75,   2318, 2318,
    2318, 2318, 2318, 75,   75,   75,   75,   75,   2318, 2318, 2318, 75,
    75,   75,   75,   2318, 75,   75,   75,   2318, 2318, 2318, 2318, 2318,
    75,   2318, 75,   75,   77,   77,   77,   77,   77,   79,   79,   79,
    77,   77,   77,   77,   77,   77,   77,   77,   77,   77,   77,   77,
    77,   77,   77,   77,   77,   77,   77,   2396, 2396, 77,   77,   77,
    77,   77,   77,   77,   77,   77,   77,   77,   77,   77,   77,   77,
    77,   77,   77,   77,   75,   75,   75,   75,   75,   75,   75,   75,
    75,   75,   75,   75,   75,   75,   75,   75,   75,   75,   75,   75,
    75,   77,   77,   75,   75,   75,   75,   75,   75,   77,   77,   77,
    2396, 77,   77,   77,   77,   2396, 77,   77,   77,   77,   77,   77,
    77,   77,   77,   77,   77,   77,   77,   77,   77,   77,   77,   77,
    77,   77,   77,   77,   77,   77,   77,   77,   77,   77,   77,   77,
    0,    0,    77,   77,   77,   77,   77,   77,   77,   77,   77,   77,
    77,   77,   77,   77,   77,   77,   77,   77,   77,   77,   77,   77,
    77,   77,   77,   77,   77,   77,   77,   77,   77,   77,   0,    77,
    77,   77,   77,   77,   77,   77,   77,   77,   77,   77,   77,   77,
    77,   77,   77,   77,   77,   77,   77,   77,   77,   77,   77,   77,
    77,   77,   77,   77,   77,   77,   77,   77,   77,   77,   77,   77,
    77,   77,   77,   77,   77,   77,   77,   77,   77,   77,   77,   77,
    77,   77,   77,   77,   77,   77,   77,   77,   77,   77,   77,   77,
    77,   77,   77,   77,   77,   77,   77,   77,   77,   77,   77,   77,
    77,   77,   77,   77,   77,   77,   77,   77,   77,   77,   77,   77,
    77,   77,   77,   77,   77,   77,   77,   77,   77,   77,   77,   77,
    77,   77,   77,   77,   77,   77,   2547, 77,   2548, 2549, 2550, 2551,
    2552, 2553, 2554, 2555, 2556, 2557, 2558, 2559, 2560, 2561, 2562, 2563,
    2564, 2565, 2566, 2567, 2568, 2569, 2570, 2571, 2572, 2573, 2574, 2575,
    2576, 2577, 2578, 2579, 2580, 2581, 2582, 2583, 2584, 2585, 2586, 2587,
    2588, 2589, 2590, 2591, 2592, 2593, 2594, 0,    2595, 2596, 2597, 2598,
    2599, 2600, 2601, 2602, 2603, 2604, 2605, 2606, 2607, 2608, 2609, 2610,
    2611, 2612, 2613, 2614, 2615, 2616, 2617, 2618, 2619, 2620, 2621, 2622,
    2623, 2624, 2625, 2626, 2627, 2628, 2629, 2630, 2631, 2632, 2633, 2634,
    2635, 2636, 2637, 2638, 2639, 2640, 2641, 0,    2642, 2643, 2644, 2645,
    2646, 2647, 2648, 2649, 2650, 2651, 2652, 2653, 2654, 2655, 2656, 2657,
    2658, 215,  2659, 2660, 215,  2661, 2662, 215,  215,  215,  215,  215,
    2663, 2664, 2665, 2666, 2667, 2668, 2669, 2670, 2671, 2672, 2673, 2674,
    2675, 2676, 2677, 2678, 2679, 2680, 2681, 2682, 2683, 2684, 2685, 2686,
    2687, 2688, 2689, 2690, 2691, 2692, 2693, 2694, 2695, 2696, 2697, 2698,
    2699, 2700, 2701, 2702, 2703, 2704, 2705, 2706, 2707, 2708, 2709, 2710,
    2711, 2712, 2713, 2714, 2715, 2716, 2717, 2718, 2719, 2720, 2721, 2722,
    2723, 2724, 2725, 2726, 2727, 2728, 2729, 2730, 2731, 2732, 2733, 2734,
    2735, 2736, 2737, 2738, 2739, 2740, 2741, 2742, 2743, 2744, 2745, 2746,
    2747, 2748, 2749, 2750, 2751, 2752, 2753, 2754, 2755, 2756, 2757, 2758,
    2759, 2760, 2761, 2762, 2763, 2764, 2765, 2766, 215,  77,   77,   77,
    77,   77,   77,   2767, 2768, 2769, 2770, 542,  542,  542,  2771, 2772,
    0,    0,    0,    0,    0,    9,    9,    9,    9,    1210, 9,    9,
    2773, 2774, 2775, 2776, 2777, 2778, 2779, 2780, 2781, 2782, 2783, 2784,
    2785, 2786, 2787, 2788, 2789, 2790, 2791, 2792, 2793, 2794, 2795, 2796,
    2797, 2798, 2799, 2800, 2801, 2802, 2803, 2804, 2805, 2806, 2807, 2808,
    2809, 2810, 0,    2811, 0,    0,    0,    0,    0,    2812, 0,    0,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  0,    0,    0,    0,
    0,    0,    0,    2813, 1046, 0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    1162, 341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  0,    0,    0,    0,    0,
    0,    0,    0,    0,    341,  341,  341,  341,  341,  341,  341,  0,
    341,  341,  341,  341,  341,  341,  341,  0,    341,  341,  341,  341,
    341,  341,  341,  0,    341,  341,  341,  341,  341,  341,  341,  0,
    341,  341,  341,  341,  341,  341,  341,  0,    341,  341,  341,  341,
    341,  341,  341,  0,    341,  341,  341,  341,  341,  341,  341,  0,
    341,  341,  341,  341,  341,  341,  341,  0,    542,  542,  542,  542,
    542,  542,  542,  542,  542,  542,  542,  542,  542,  542,  542,  542,
    542,  542,  542,  542,  542,  542,  542,  542,  542,  542,  542,  542,
    542,  542,  542,  542,  9,    9,    81,   93,   81,   93,   9,    9,
    9,    81,   93,   9,    81,   93,   9,    9,    9,    9,    9,    9,
    9,    9,    9,    1086, 9,    9,    1086, 9,    81,   93,   9,    9,
    81,   93,   12,   13,   12,   13,   12,   13,   12,   13,   9,    9,
    9,    9,    9,    524,  9,    9,    9,    9,    9,    9,    9,    9,
    9,    9,    1086, 1086, 9,    9,    9,    9,    1086, 9,    2136, 9,
    9,    9,    9,    9,    9,    9,    9,    9,    9,    9,    9,    9,
    77,   77,   9,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    2814, 2814, 2814, 2814, 2814, 2814, 2814, 2814, 2814, 2814, 2814, 2814,
    2814, 2814, 2814, 2814, 2814, 2814, 2814, 2814, 2814, 2814, 2814, 2814,
    2814, 2814, 0,    2814, 2814, 2814, 2814, 2815, 2814, 2814, 2814, 2814,
    2814, 2814, 2814, 2814, 2814, 2814, 2814, 2814, 2814, 2814, 2814, 2814,
    2814, 2814, 2814, 2814, 2814, 2814, 2814, 2814, 2814, 2814, 2814, 2814,
    2814, 2814, 2814, 2814, 2814, 2814, 2814, 2814, 2814, 2814, 2814, 2814,
    2814, 2814, 2814, 2814, 2814, 2814, 2814, 2814, 2814, 2814, 2814, 2814,
    2814, 2814, 2814, 2814, 2814, 2814, 2814, 2814, 2814, 2814, 2814, 2814,
    2814, 2814, 2814, 2814, 2814, 2814, 2814, 2814, 2814, 2814, 2814, 2814,
    2814, 2814, 2814, 2814, 2814, 2814, 2814, 2816, 0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    2817, 2818, 2819, 2820,
    2821, 2822, 2823, 2824, 2825, 2826, 2827, 2828, 2829, 2830, 2831, 2832,
    2833, 2834, 2835, 2836, 2837, 2838, 2839, 2840, 2841, 2842, 2843, 2844,
    2845, 2846, 2847, 2848, 2849, 2850, 2851, 2852, 2853, 2854, 2855, 2856,
    2857, 2858, 2859, 2860, 2861, 2862, 2863, 2864, 2865, 2866, 2867, 2868,
    2869, 2870, 2871, 2872, 2873, 2874, 2875, 2876, 2877, 2878, 2879, 2880,
    2881, 2882, 2883, 2884, 2885, 2886, 2887, 2888, 2889, 2890, 2891, 2892,
    2893, 2894, 2895, 2896, 2897, 2898, 2899, 2900, 2901, 2902, 2903, 2904,
    2905, 2906, 2907, 2908, 2909, 2910, 2911, 2912, 2913, 2914, 2915, 2916,
    2917, 2918, 2919, 2920, 2921, 2922, 2923, 2924, 2925, 2926, 2927, 2928,
    2929, 2930, 2931, 2932, 2933, 2934, 2935, 2936, 2937, 2938, 2939, 2940,
    2941, 2942, 2943, 2944, 2945, 2946, 2947, 2948, 2949, 2950, 2951, 2952,
    2953, 2954, 2955, 2956, 2957, 2958, 2959, 2960, 2961, 2962, 2963, 2964,
    2965, 2966, 2967, 2968, 2969, 2970, 2971, 2972, 2973, 2974, 2975, 2976,
    2977, 2978, 2979, 2980, 2981, 2982, 2983, 2984, 2985, 2986, 2987, 2988,
    2989, 2990, 2991, 2992, 2993, 2994, 2995, 2996, 2997, 2998, 2999, 3000,
    3001, 3002, 3003, 3004, 3005, 3006, 3007, 3008, 3009, 3010, 3011, 3012,
    3013, 3014, 3015, 3016, 3017, 3018, 3019, 3020, 3021, 3022, 3023, 3024,
    3025, 3026, 3027, 3028, 3029, 3030, 0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    2814, 2814, 2814, 2814,
    2814, 2814, 2814, 2814, 2814, 2814, 2814, 2814, 0,    0,    0,    0,
    3031, 3032, 3032, 3032, 2814, 3033, 3034, 3035, 3036, 3037, 3036, 3037,
    3036, 3037, 3036, 3037, 3036, 3037, 2814, 2814, 3036, 3037, 3036, 3037,
    3036, 3037, 3036, 3037, 3038, 3039, 3040, 3040, 2814, 3035, 3035, 3035,
    3035, 3035, 3035, 3035, 3035, 3035, 3041, 1088, 554,  1087, 3042, 3042,
    3043, 3033, 3033, 3033, 3033, 3033, 3044, 2814, 3045, 3046, 3047, 3033,
    3034, 3048, 2814, 77,   0,    3034, 3034, 3034, 3034, 3034, 3049, 3034,
    3034, 3034, 3034, 3050, 3051, 3052, 3053, 3054, 3055, 3056, 3057, 3058,
    3059, 3060, 3061, 3062, 3063, 3064, 3065, 3066, 3067, 3068, 3069, 3070,
    3071, 3072, 3073, 3034, 3074, 3075, 3076, 3077, 3078, 3079, 3034, 3034,
    3034, 3034, 3034, 3080, 3081, 3082, 3083, 3084, 3085, 3086, 3087, 3088,
    3089, 3090, 3091, 3092, 3093, 3094, 3034, 3034, 3034, 3034, 3034, 3034,
    3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034,
    3034, 3034, 3034, 3034, 3095, 3034, 3034, 0,    0,    3096, 3097, 3098,
    3099, 3100, 3101, 3102, 3038, 3034, 3034, 3034, 3034, 3034, 3103, 3034,
    3034, 3034, 3034, 3104, 3105, 3106, 3107, 3108, 3109, 3110, 3111, 3112,
    3113, 3114, 3115, 3116, 3117, 3118, 3119, 3120, 3121, 3122, 3123, 3124,
    3125, 3126, 3127, 3034, 3128, 3129, 3130, 3131, 3132, 3133, 3034, 3034,
    3034, 3034, 3034, 3134, 3135, 3136, 3137, 3138, 3139, 3140, 3141, 3142,
    3143, 3144, 3145, 3146, 3147, 3148, 3034, 3034, 3034, 3034, 3034, 3034,
    3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3149,
    3150, 3151, 3152, 3034, 3153, 3034, 3034, 3154, 3155, 3156, 3157, 3032,
    3033, 3158, 3159, 3160, 0,    0,    0,    0,    0,    3034, 3034, 3034,
    3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034,
    3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034,
    3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034,
    3034, 3034, 3034, 3034, 0,    3161, 3162, 3163, 3164, 3165, 3166, 3167,
    3168, 3169, 3170, 3171, 3172, 3173, 3174, 3175, 3176, 3177, 3178, 3179,
    3180, 3181, 3182, 3183, 3184, 3185, 3186, 3187, 3188, 3189, 3190, 3191,
    3192, 3193, 3194, 3195, 3196, 3197, 3198, 3199, 3200, 3201, 3202, 3203,
    3204, 3205, 3206, 3207, 3208, 3209, 3210, 3211, 3212, 3213, 3214, 3215,
    3216, 3217, 3218, 3219, 3220, 3221, 3222, 3223, 3224, 3225, 3226, 3227,
    3228, 3229, 3230, 3231, 3232, 3233, 3234, 3235, 3236, 3237, 3238, 3239,
    3240, 3241, 3242, 3243, 3244, 3245, 3246, 3247, 3248, 3249, 3250, 3251,
    3252, 3253, 3254, 0,    3255, 3255, 3256, 3257, 3258, 3259, 3260, 3261,
    3262, 3263, 3264, 3265, 3266, 3267, 3268, 3269, 3034, 3034, 3034, 3034,
    3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034,
    3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034,
    3034, 3034, 3034, 3034, 2814, 2814, 2814, 2814, 2814, 2814, 2814, 2814,
    2814, 2814, 2814, 2814, 2814, 2814, 2814, 2814, 2814, 2814, 2814, 2814,
    2814, 2814, 2814, 2814, 2814, 2814, 2814, 2814, 2814, 2814, 2814, 2814,
    2814, 2814, 2814, 2814, 0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034,
    3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3270, 3271, 3272, 3273,
    3274, 3275, 3276, 3277, 3278, 3279, 3280, 3281, 3282, 3283, 3284, 3285,
    3286, 3287, 3288, 3289, 3290, 3291, 3292, 3293, 3294, 3295, 3296, 3297,
    3298, 3299, 3300, 0,    3301, 3302, 3303, 3304, 3305, 3306, 3307, 3308,
    3309, 3310, 3311, 3312, 3313, 3314, 3315, 3316, 3317, 3318, 3319, 3320,
    3321, 3322, 3323, 3324, 3325, 3326, 3327, 3328, 3329, 3330, 3331, 3332,
    3333, 3334, 3335, 3336, 3337, 3338, 3339, 3340, 1181, 1181, 1181, 1181,
    1181, 1181, 1181, 1181, 3341, 3342, 3343, 3344, 3345, 3346, 3347, 3348,
    3349, 3350, 3351, 3352, 3353, 3354, 3355, 3356, 3357, 3358, 3359, 3360,
    3361, 3362, 3363, 3364, 3365, 3366, 3367, 3368, 3369, 3370, 3371, 3372,
    3373, 3374, 3375, 3376, 3377, 3378, 3379, 3380, 3381, 3382, 3383, 3384,
    3385, 3386, 3387, 3255, 3388, 3389, 3390, 3391, 3392, 3393, 3394, 3395,
    3396, 3397, 3398, 3399, 3400, 3401, 3402, 3403, 3404, 3405, 3406, 3407,
    3408, 3409, 3410, 3411, 3412, 3413, 3414, 3415, 3416, 3417, 3418, 3419,
    3420, 3421, 3422, 3423, 3424, 3425, 3426, 3427, 3428, 3429, 3430, 3431,
    3432, 3433, 3434, 3435, 3436, 3437, 3438, 3439, 3440, 3441, 3442, 3443,
    3444, 3445, 3446, 3447, 3448, 3449, 3450, 3451, 3452, 3453, 3454, 3455,
    3456, 3457, 3458, 3459, 3460, 3461, 3462, 3463, 3464, 3465, 3466, 3467,
    3468, 3469, 3470, 3471, 3472, 3473, 3474, 3475, 3476, 3477, 3478, 3479,
    3480, 3481, 3482, 3483, 3484, 3485, 3486, 3487, 3488, 3489, 3490, 3491,
    3492, 3493, 3494, 3495, 3496, 3497, 3498, 3499, 3500, 3501, 3502, 3503,
    3504, 3505, 3506, 3507, 3508, 3509, 3510, 3511, 3512, 3513, 3514, 3515,
    3516, 3517, 3518, 3519, 3520, 3521, 3522, 3523, 3524, 3525, 3526, 3527,
    3528, 3529, 3530, 3531, 3532, 3533, 3534, 3535, 3536, 3537, 3538, 3539,
    3540, 3541, 3542, 3543, 3544, 3545, 3546, 3547, 3548, 3549, 3550, 3551,
    3552, 3553, 3554, 3555, 3556, 3557, 3558, 3559, 3560, 3561, 3562, 3563,
    3564, 3565, 3566, 3567, 3568, 3569, 3570, 3571, 3572, 3573, 3574, 3575,
    3576, 3577, 3578, 3579, 3580, 3581, 3582, 3583, 3584, 3585, 3586, 3587,
    3588, 3589, 3590, 3591, 3592, 3593, 3594, 3595, 3596, 3597, 3598, 3599,
    3600, 3601, 3602, 3603, 3604, 3605, 3606, 3607, 3608, 3609, 3610, 3611,
    3612, 3613, 3614, 3615, 3616, 3617, 3618, 3619, 3620, 3621, 3622, 3623,
    3624, 3625, 3626, 3627, 3628, 3629, 3630, 3631, 3632, 3633, 3634, 3635,
    3636, 3637, 3638, 3639, 3640, 3641, 3642, 3643, 3644, 3645, 3646, 3647,
    3648, 3649, 3650, 3651, 3652, 3653, 3654, 3655, 3656, 3657, 3658, 3659,
    3660, 3661, 3662, 3663, 3664, 3665, 3666, 3667, 3668, 3669, 3670, 3671,
    3672, 3673, 3674, 3675, 3676, 3677, 3678, 3679, 3680, 3681, 3682, 3683,
    3684, 3685, 3686, 3687, 3688, 3689, 3690, 3691, 3692, 3693, 3694, 3695,
    3696, 3697, 3698, 3699, 3700, 3701, 3702, 3703, 3704, 3705, 3706, 3707,
    3708, 3709, 3710, 3711, 3712, 3713, 3714, 3715, 3716, 3717, 3718, 3719,
    3720, 3721, 3722, 3723, 3724, 3725, 3726, 3727, 3728, 3729, 3730, 3731,
    3732, 3733, 3734, 3735, 3736, 3737, 3738, 3739, 3740, 3741, 3742, 3743,
    3744, 3745, 3746, 3747, 3748, 3749, 3750, 3751, 3752, 3753, 3754, 3755,
    3756, 3757, 3758, 3759, 3760, 3761, 3762, 3763, 3764, 3765, 3766, 3767,
    3768, 3769, 3770, 3771, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034,
    3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034,
    3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034,
    3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034,
    3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034,
    3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034,
    3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034,
    3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034,
    3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034,
    3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034,
    3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034,
    3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034,
    3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034,
    3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034,
    3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034,
    3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034,
    3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034,
    3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034,
    3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034,
    3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034,
    3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034,
    3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034,
    3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034,
    3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034,
    3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034,
    3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034,
    3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034,
    3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034,
    3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034,
    3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034,
    3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034,
    3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034,
    3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034,
    3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034,
    3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034,
    3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034,
    3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034,
    3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 77,   77,   77,   77,
    77,   77,   77,   77,   77,   77,   77,   77,   77,   77,   77,   77,
    77,   77,   77,   77,   77,   77,   77,   77,   77,   77,   77,   77,
    77,   77,   77,   77,   77,   77,   77,   77,   77,   77,   77,   77,
    77,   77,   77,   77,   77,   77,   77,   77,   77,   77,   77,   77,
    77,   77,   77,   77,   77,   77,   77,   77,   77,   77,   77,   77,
    3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034,
    3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034,
    3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034,
    3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034,
    3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034,
    3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034,
    3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034,
    3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034,
    3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034,
    3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034,
    3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034,
    3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034,
    3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034,
    3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034,
    3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034,
    3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034,
    3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034,
    3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034,
    3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034,
    3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034,
    3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034,
    3034, 0,    0,    0,    3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034,
    3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034,
    3034, 3033, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034,
    3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034,
    3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034,
    3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034,
    3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034,
    3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034,
    3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034,
    3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034,
    3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034,
    3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034,
    3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034,
    3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034,
    3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034,
    3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034,
    3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034,
    3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034,
    3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034,
    3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034,
    3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034,
    3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034,
    3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034,
    3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034,
    3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034,
    3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034,
    3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034,
    3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034,
    3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034,
    3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034,
    3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034,
    3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034,
    3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034,
    3034, 3034, 3034, 3034, 3034, 0,    0,    0,    2814, 2814, 2814, 2814,
    2814, 2814, 2814, 2814, 2814, 2814, 2814, 2814, 2814, 2814, 2814, 2814,
    2814, 2814, 2814, 2814, 2814, 2814, 2814, 2814, 2814, 2814, 2814, 2814,
    2814, 2814, 2814, 2814, 2814, 2814, 2814, 2814, 2814, 2814, 2814, 2814,
    2814, 2814, 2814, 2814, 2814, 2814, 2814, 2814, 2814, 2814, 2814, 2814,
    2814, 2814, 2814, 0,    0,    0,    0,    0,    0,    0,    0,    0,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  525,  525,  525,  525,  525,  525,  1046, 1046,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    525,  9,    9,    9,    341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  1171, 1171, 1171, 1171,
    1171, 1171, 1171, 1171, 1171, 1171, 341,  341,  0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    3772, 3773, 3774, 3775, 3776, 3777, 3778, 3779,
    3780, 3781, 3782, 3783, 3784, 3785, 3786, 3787, 3788, 3789, 3790, 3791,
    3792, 3793, 3794, 3795, 3796, 3797, 3798, 3799, 3800, 3801, 3802, 3803,
    3804, 3805, 3806, 3807, 3808, 3809, 3810, 3811, 3812, 3813, 3814, 3815,
    3816, 3817, 341,  542,  841,  841,  841,  9,    542,  542,  542,  542,
    542,  542,  542,  542,  542,  542,  9,    524,  3818, 3819, 3820, 3821,
    3822, 3823, 3824, 3825, 3826, 3827, 3828, 3829, 3830, 3831, 3832, 3833,
    3834, 3835, 3836, 3837, 3838, 3839, 3840, 3841, 3842, 3843, 3844, 3845,
    3846, 3847, 542,  542,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  1457, 1457, 1457, 1457, 1457, 1457, 1457, 1457, 1457, 1457,
    542,  542,  1046, 1046, 1046, 1046, 1046, 1046, 0,    0,    0,    0,
    0,    0,    0,    0,    47,   47,   47,   47,   47,   47,   47,   47,
    47,   47,   47,   47,   47,   47,   47,   47,   47,   47,   47,   47,
    47,   47,   47,   524,  524,  524,  524,  524,  524,  524,  524,  524,
    47,   47,   3848, 3849, 3850, 3851, 3852, 3853, 3854, 3855, 3856, 3857,
    3858, 3859, 3860, 3861, 215,  215,  3862, 3863, 3864, 3865, 3866, 3867,
    3868, 3869, 3870, 3871, 3872, 3873, 3874, 3875, 3876, 3877, 3878, 3879,
    3880, 3881, 3882, 3883, 3884, 3885, 3886, 3887, 3888, 3889, 3890, 3891,
    3892, 3893, 3894, 3895, 3896, 3897, 3898, 3899, 3900, 3901, 3902, 3903,
    3904, 3905, 3906, 3907, 3908, 3909, 3910, 3911, 3912, 3913, 3914, 3915,
    3916, 3917, 3918, 3919, 3920, 3921, 3922, 3923, 3924, 215,  215,  215,
    215,  215,  215,  215,  215,  3925, 3926, 3927, 3928, 3929, 3930, 3931,
    3932, 3933, 3934, 3935, 3936, 3937, 3938, 3939, 524,  3940, 3940, 3941,
    3942, 3943, 215,  341,  3944, 3945, 3946, 3947, 3948, 215,  3949, 3950,
    3951, 3952, 3953, 3954, 3955, 3956, 3957, 3958, 3959, 3960, 3961, 3962,
    3963, 3964, 3965, 3966, 3967, 3968, 3969, 3970, 3971, 3972, 3973, 215,
    3974, 3975, 3976, 3977, 3978, 3979, 3980, 3981, 3982, 3983, 3984, 3985,
    3986, 3987, 3988, 3989, 0,    0,    3990, 3991, 3992, 3993, 3994, 3995,
    3996, 3997, 3998, 0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    3999, 4000, 341,
    4001, 4002, 215,  341,  341,  341,  341,  341,  341,  341,  1150, 341,
    341,  341,  1162, 341,  341,  341,  341,  1150, 341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  1154, 1154, 1150, 1150, 1154,
    77,   77,   77,   77,   1162, 0,    0,    0,    1181, 1181, 1181, 1181,
    1181, 1181, 840,  840,  11,   84,   0,    0,    0,    0,    0,    0,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  9,    9,    9,    9,    0,    0,    0,    0,
    0,    0,    0,    0,    1154, 1154, 341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  1154, 1154, 1154, 1154,
    1154, 1154, 1154, 1154, 1154, 1154, 1154, 1154, 1154, 1154, 1154, 1154,
    1162, 1150, 0,    0,    0,    0,    0,    0,    0,    0,    1046, 1046,
    1171, 1171, 1171, 1171, 1171, 1171, 1171, 1171, 1171, 1171, 0,    0,
    0,    0,    0,    0,    542,  542,  542,  542,  542,  542,  542,  542,
    542,  542,  542,  542,  542,  542,  542,  542,  542,  542,  341,  341,
    341,  341,  341,  341,  1046, 1046, 1046, 341,  1046, 341,  341,  1150,
    1171, 1171, 1171, 1171, 1171, 1171, 1171, 1171, 1171, 1171, 341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  1150, 1150, 1150, 1150, 1150, 555,  555,  555,  1046, 1046,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  1150,
    1150, 1150, 1150, 1150, 1150, 1150, 1150, 1150, 1150, 1150, 1154, 1482,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    1046,
    1360, 1360, 1360, 1360, 1360, 1360, 1360, 1360, 1360, 1360, 1360, 1360,
    1360, 1360, 1360, 1360, 1360, 1360, 1360, 1360, 1360, 1360, 1360, 1360,
    1360, 1360, 1360, 1360, 1360, 0,    0,    0,    1150, 1150, 1150, 1154,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  1172,
    1154, 1154, 1150, 1150, 1150, 1150, 1154, 1154, 1150, 1150, 1154, 1154,
    1482, 1046, 1046, 1046, 1046, 1046, 1046, 1046, 1046, 1046, 1046, 1046,
    1046, 1046, 0,    525,  1171, 1171, 1171, 1171, 1171, 1171, 1171, 1171,
    1171, 1171, 0,    0,    0,    0,    1046, 1046, 341,  341,  341,  341,
    341,  1150, 525,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    1171, 1171, 1171, 1171, 1171, 1171, 1171, 1171, 1171, 1171, 341,  341,
    341,  341,  341,  0,    341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  1150, 1150, 1150,
    1150, 1150, 1150, 1154, 1154, 1150, 1150, 1154, 1154, 1150, 1150, 0,
    0,    0,    0,    0,    0,    0,    0,    0,    341,  341,  341,  1150,
    341,  341,  341,  341,  341,  341,  341,  341,  1150, 1154, 0,    0,
    1171, 1171, 1171, 1171, 1171, 1171, 1171, 1171, 1171, 1171, 0,    0,
    1046, 1046, 1046, 1046, 341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  525,  341,  341,  341,
    341,  341,  341,  840,  840,  840,  341,  1271, 1150, 1271, 341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    542,  341,  542,  542,  555,  341,  341,  542,  542,  341,  341,  341,
    341,  341,  542,  542,  341,  542,  341,  0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    341,  341,  525,  1046, 1046,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  1154,
    1150, 1150, 1154, 1154, 1046, 1046, 341,  525,  525,  1154, 1162, 0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    341,  341,  341,
    341,  341,  341,  0,    0,    341,  341,  341,  341,  341,  341,  0,
    0,    341,  341,  341,  341,  341,  341,  0,    0,    0,    0,    0,
    0,    0,    0,    0,    341,  341,  341,  341,  341,  341,  341,  0,
    341,  341,  341,  341,  341,  341,  341,  0,    215,  215,  215,  215,
    215,  215,  215,  215,  215,  215,  215,  215,  215,  215,  215,  215,
    215,  215,  215,  215,  215,  215,  215,  215,  215,  215,  215,  215,
    215,  215,  215,  215,  215,  215,  215,  4003, 215,  215,  215,  215,
    215,  215,  215,  3940, 4004, 4005, 4006, 4007, 215,  215,  215,  215,
    215,  215,  215,  215,  215,  4008, 47,   47,   0,    0,    0,    0,
    4009, 4010, 4011, 4012, 4013, 4014, 4015, 4016, 4017, 4018, 4019, 4020,
    4021, 4022, 4023, 4024, 4025, 4026, 4027, 4028, 4029, 4030, 4031, 4032,
    4033, 4034, 4035, 4036, 4037, 4038, 4039, 4040, 4041, 4042, 4043, 4044,
    4045, 4046, 4047, 4048, 4049, 4050, 4051, 4052, 4053, 4054, 4055, 4056,
    4057, 4058, 4059, 4060, 4061, 4062, 4063, 4064, 4065, 4066, 4067, 4068,
    4069, 4070, 4071, 4072, 4073, 4074, 4075, 4076, 4077, 4078, 4079, 4080,
    4081, 4082, 4083, 4084, 4085, 4086, 4087, 4088, 341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  1154, 1154, 1150, 1154, 1154,
    1150, 1154, 1154, 1046, 1154, 1162, 0,    0,    1171, 1171, 1171, 1171,
    1171, 1171, 1171, 1171, 1171, 1171, 0,    0,    0,    0,    0,    0,
    4089, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090,
    4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090,
    4090, 4090, 4090, 4090, 4089, 4090, 4090, 4090, 4090, 4090, 4090, 4090,
    4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090,
    4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4089, 4090, 4090, 4090,
    4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090,
    4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090,
    4089, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090,
    4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090,
    4090, 4090, 4090, 4090, 4089, 4090, 4090, 4090, 4090, 4090, 4090, 4090,
    4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090,
    4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4089, 4090, 4090, 4090,
    4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090,
    4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090,
    4089, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090,
    4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090,
    4090, 4090, 4090, 4090, 4089, 4090, 4090, 4090, 4090, 4090, 4090, 4090,
    4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090,
    4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4089, 4090, 4090, 4090,
    4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090,
    4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090,
    4089, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090,
    4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090,
    4090, 4090, 4090, 4090, 4089, 4090, 4090, 4090, 4090, 4090, 4090, 4090,
    4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090,
    4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4089, 4090, 4090, 4090,
    4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090,
    4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090,
    4089, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090,
    4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090,
    4090, 4090, 4090, 4090, 4089, 4090, 4090, 4090, 4090, 4090, 4090, 4090,
    4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090,
    4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4089, 4090, 4090, 4090,
    4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090,
    4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090,
    4089, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090,
    4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090,
    4090, 4090, 4090, 4090, 4089, 4090, 4090, 4090, 4090, 4090, 4090, 4090,
    4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090,
    4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4089, 4090, 4090, 4090,
    4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090,
    4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090,
    4089, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090,
    4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090,
    4090, 4090, 4090, 4090, 4089, 4090, 4090, 4090, 4090, 4090, 4090, 4090,
    4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090,
    4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4089, 4090, 4090, 4090,
    4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090,
    4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090,
    4089, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090,
    4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090,
    4090, 4090, 4090, 4090, 4089, 4090, 4090, 4090, 4090, 4090, 4090, 4090,
    4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090,
    4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4089, 4090, 4090, 4090,
    4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090,
    4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090,
    4089, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090,
    4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090,
    4090, 4090, 4090, 4090, 4089, 4090, 4090, 4090, 4090, 4090, 4090, 4090,
    4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090,
    4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4089, 4090, 4090, 4090,
    4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090,
    4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090,
    4089, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090,
    4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090,
    4090, 4090, 4090, 4090, 4089, 4090, 4090, 4090, 4090, 4090, 4090, 4090,
    4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090,
    4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4089, 4090, 4090, 4090,
    4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090,
    4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090,
    4089, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090,
    4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090,
    4090, 4090, 4090, 4090, 4089, 4090, 4090, 4090, 4090, 4090, 4090, 4090,
    4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090,
    4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4089, 4090, 4090, 4090,
    4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090,
    4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090,
    4089, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090,
    4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090,
    4090, 4090, 4090, 4090, 4089, 4090, 4090, 4090, 4090, 4090, 4090, 4090,
    4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090,
    4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4089, 4090, 4090, 4090,
    4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090,
    4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090,
    4089, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090,
    4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090,
    4090, 4090, 4090, 4090, 4089, 4090, 4090, 4090, 4090, 4090, 4090, 4090,
    4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090,
    4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4089, 4090, 4090, 4090,
    4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090,
    4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090,
    4089, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090,
    4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090,
    4090, 4090, 4090, 4090, 4089, 4090, 4090, 4090, 4090, 4090, 4090, 4090,
    4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090,
    4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4089, 4090, 4090, 4090,
    4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090,
    4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090,
    4089, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090,
    4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090,
    4090, 4090, 4090, 4090, 4089, 4090, 4090, 4090, 4090, 4090, 4090, 4090,
    4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090,
    4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4089, 4090, 4090, 4090,
    4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090,
    4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090,
    4089, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090,
    4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090,
    4090, 4090, 4090, 4090, 4089, 4090, 4090, 4090, 4090, 4090, 4090, 4090,
    4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090,
    4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4089, 4090, 4090, 4090,
    4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090,
    4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090,
    4089, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090,
    4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090,
    4090, 4090, 4090, 4090, 4089, 4090, 4090, 4090, 4090, 4090, 4090, 4090,
    4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090,
    4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4089, 4090, 4090, 4090,
    4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090,
    4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090,
    4089, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090,
    4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090,
    4090, 4090, 4090, 4090, 4089, 4090, 4090, 4090, 4090, 4090, 4090, 4090,
    4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090,
    4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4089, 4090, 4090, 4090,
    4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090,
    4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090,
    4089, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090,
    4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090,
    4090, 4090, 4090, 4090, 4089, 4090, 4090, 4090, 4090, 4090, 4090, 4090,
    4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090,
    4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4089, 4090, 4090, 4090,
    4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090,
    4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090,
    4089, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090,
    4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090,
    4090, 4090, 4090, 4090, 4089, 4090, 4090, 4090, 4090, 4090, 4090, 4090,
    4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090,
    4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4089, 4090, 4090, 4090,
    4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090,
    4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090,
    4089, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090,
    4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090,
    4090, 4090, 4090, 4090, 4089, 4090, 4090, 4090, 4090, 4090, 4090, 4090,
    4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090,
    4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4089, 4090, 4090, 4090,
    4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090,
    4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090,
    4089, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090,
    4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090,
    4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090,
    4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090,
    4090, 4090, 4090, 4090, 4089, 4090, 4090, 4090, 4090, 4090, 4090, 4090,
    4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090,
    4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4089, 4090, 4090, 4090,
    4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090,
    4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090,
    4089, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090,
    4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090,
    4090, 4090, 4090, 4090, 4089, 4090, 4090, 4090, 4090, 4090, 4090, 4090,
    4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090,
    4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4089, 4090, 4090, 4090,
    4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090,
    4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090, 4090,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    1363, 1363, 1363, 1363, 1363, 1363, 1363, 1363, 1363, 1363, 1363, 1363,
    1363, 1363, 1363, 1363, 1363, 1363, 1363, 1363, 1363, 1363, 1363, 0,
    0,    0,    0,    1364, 1364, 1364, 1364, 1364, 1364, 1364, 1364, 1364,
    1364, 1364, 1364, 1364, 1364, 1364, 1364, 1364, 1364, 1364, 1364, 1364,
    1364, 1364, 1364, 1364, 1364, 1364, 1364, 1364, 1364, 1364, 1364, 1364,
    1364, 1364, 1364, 1364, 1364, 1364, 1364, 1364, 1364, 1364, 1364, 1364,
    1364, 1364, 1364, 1364, 0,    0,    0,    0,    4091, 4091, 4091, 4091,
    4091, 4091, 4091, 4091, 4091, 4091, 4091, 4091, 4091, 4091, 4091, 4091,
    4091, 4091, 4091, 4091, 4091, 4091, 4091, 4091, 4091, 4091, 4091, 4091,
    4091, 4091, 4091, 4091, 4091, 4091, 4091, 4091, 4091, 4091, 4091, 4091,
    4091, 4091, 4091, 4091, 4091, 4091, 4091, 4091, 4091, 4091, 4091, 4091,
    4091, 4091, 4091, 4091, 4091, 4091, 4091, 4091, 4091, 4091, 4091, 4091,
    4091, 4091, 4091, 4091, 4091, 4091, 4091, 4091, 4091, 4091, 4091, 4091,
    4091, 4091, 4091, 4091, 4091, 4091, 4091, 4091, 4091, 4091, 4091, 4091,
    4091, 4091, 4091, 4091, 4091, 4091, 4091, 4091, 4091, 4091, 4091, 4091,
    4091, 4091, 4091, 4091, 4091, 4091, 4091, 4091, 4091, 4091, 4091, 4091,
    4091, 4091, 4091, 4091, 4091, 4091, 4091, 4091, 4091, 4091, 4091, 4091,
    4091, 4091, 4091, 4091, 4091, 4091, 4091, 4091, 4091, 4091, 4091, 4091,
    4091, 4091, 4091, 4091, 4091, 4091, 4091, 4091, 4091, 4091, 4091, 4091,
    4091, 4091, 4091, 4091, 4091, 4091, 4091, 4091, 4091, 4091, 4091, 4091,
    4091, 4091, 4091, 4091, 4091, 4091, 4091, 4091, 4091, 4091, 4091, 4091,
    4091, 4091, 4091, 4091, 4091, 4091, 4091, 4091, 4091, 4091, 4091, 4091,
    4091, 4091, 4091, 4091, 4091, 4091, 4091, 4091, 4091, 4091, 4091, 4091,
    4091, 4091, 4091, 4091, 4091, 4091, 4091, 4091, 4091, 4091, 4091, 4091,
    4091, 4091, 4091, 4091, 4091, 4091, 4091, 4091, 4091, 4091, 4091, 4091,
    4091, 4091, 4091, 4091, 4091, 4091, 4091, 4091, 4091, 4091, 4091, 4091,
    4091, 4091, 4091, 4091, 4091, 4091, 4091, 4091, 4091, 4091, 4091, 4091,
    4091, 4091, 4091, 4091, 4091, 4091, 4091, 4091, 4091, 4091, 4091, 4091,
    4092, 4092, 4092, 4092, 4092, 4092, 4092, 4092, 4092, 4092, 4092, 4092,
    4092, 4092, 4092, 4092, 4092, 4092, 4092, 4092, 4092, 4092, 4092, 4092,
    4092, 4092, 4092, 4092, 4092, 4092, 4092, 4092, 4092, 4092, 4092, 4092,
    4092, 4092, 4092, 4092, 4092, 4092, 4092, 4092, 4092, 4092, 4092, 4092,
    4092, 4092, 4092, 4092, 4092, 4092, 4092, 4092, 4092, 4092, 4092, 4092,
    4092, 4092, 4092, 4092, 4092, 4092, 4092, 4092, 4092, 4092, 4092, 4092,
    4092, 4092, 4092, 4092, 4092, 4092, 4092, 4092, 4092, 4092, 4092, 4092,
    4092, 4092, 4092, 4092, 4092, 4092, 4092, 4092, 4092, 4092, 4092, 4092,
    4092, 4092, 4092, 4092, 4092, 4092, 4092, 4092, 4092, 4092, 4092, 4092,
    4092, 4092, 4092, 4092, 4092, 4092, 4092, 4092, 4092, 4092, 4092, 4092,
    4092, 4092, 4092, 4092, 4092, 4092, 4092, 4092, 4092, 4092, 4092, 4092,
    4092, 4092, 4092, 4092, 4092, 4092, 4092, 4092, 4092, 4092, 4092, 4092,
    4092, 4092, 4092, 4092, 4092, 4092, 4092, 4092, 4092, 4092, 4092, 4092,
    4092, 4092, 4092, 4092, 4092, 4092, 4092, 4092, 4092, 4092, 4092, 4092,
    4092, 4092, 4092, 4092, 4092, 4092, 4092, 4092, 4092, 4092, 4092, 4092,
    4092, 4092, 4092, 4092, 4092, 4092, 4092, 4092, 4092, 4092, 4092, 4092,
    4092, 4092, 4092, 4092, 4092, 4092, 4092, 4092, 4092, 4092, 4092, 4092,
    4092, 4092, 4092, 4092, 4092, 4092, 4092, 4092, 4092, 4092, 4092, 4092,
    4092, 4092, 4092, 4092, 4092, 4092, 4092, 4092, 4092, 4092, 4092, 4092,
    4092, 4092, 4092, 4092, 4092, 4092, 4092, 4092, 4092, 4092, 4092, 4092,
    4092, 4092, 4092, 4092, 4092, 4092, 4092, 4092, 4092, 4092, 4092, 4092,
    4092, 4092, 4092, 4092, 4093, 4094, 4095, 4096, 4097, 4098, 4099, 4100,
    4100, 4101, 4102, 4103, 4104, 4105, 4106, 4107, 4108, 4109, 4110, 4111,
    4112, 4113, 4114, 4115, 4116, 4117, 4118, 4119, 4120, 4121, 4122, 4123,
    4124, 4125, 4126, 4127, 4128, 4129, 4130, 4131, 4132, 4133, 4134, 4135,
    4136, 4137, 4138, 4139, 4140, 4141, 4142, 4143, 4144, 4145, 4146, 4147,
    4148, 4149, 4150, 4151, 4152, 4153, 4154, 4155, 4156, 4157, 4158, 4159,
    4160, 4161, 4162, 4163, 4164, 4165, 4166, 4167, 4168, 4169, 4170, 4171,
    4172, 4173, 4174, 4175, 4176, 4177, 4178, 4179, 4180, 4181, 4182, 4183,
    4112, 4184, 4185, 4186, 4187, 4188, 4189, 4190, 4191, 4192, 4193, 4194,
    4195, 4196, 4197, 4198, 4199, 4200, 4201, 4202, 4203, 4204, 4205, 4206,
    4207, 4208, 4209, 4210, 4211, 4212, 4213, 4214, 4215, 4216, 4217, 4218,
    4219, 4220, 4221, 4222, 4223, 4224, 4225, 4226, 4227, 4228, 4229, 4230,
    4231, 4232, 4233, 4234, 4235, 4236, 4237, 4238, 4239, 4240, 4241, 4242,
    4243, 4244, 4245, 4246, 4247, 4248, 4249, 4250, 4251, 4202, 4252, 4253,
    4254, 4255, 4256, 4257, 4258, 4259, 4186, 4260, 4261, 4262, 4263, 4264,
    4265, 4266, 4267, 4268, 4269, 4270, 4271, 4272, 4273, 4274, 4275, 4276,
    4277, 4278, 4279, 4112, 4280, 4281, 4282, 4283, 4284, 4285, 4286, 4287,
    4288, 4289, 4290, 4291, 4292, 4293, 4294, 4295, 4296, 4297, 4298, 4299,
    4300, 4301, 4302, 4303, 4304, 4305, 4306, 4188, 4307, 4308, 4309, 4310,
    4311, 4312, 4313, 4314, 4315, 4316, 4317, 4318, 4319, 4320, 4321, 4322,
    4323, 4324, 4325, 4326, 4327, 4328, 4329, 4330, 4331, 4332, 4333, 4334,
    4335, 4336, 4337, 4338, 4339, 4340, 4341, 4342, 4343, 4344, 4345, 4346,
    4347, 4348, 4349, 4350, 4351, 4352, 4353, 4354, 4355, 4356, 3034, 3034,
    4357, 3034, 4358, 3034, 3034, 4359, 4360, 4361, 4362, 4363, 4364, 4365,
    4366, 4367, 4368, 3034, 4369, 3034, 4370, 3034, 3034, 4371, 4372, 3034,
    3034, 3034, 4373, 4374, 4375, 4376, 4377, 4378, 4379, 4380, 4381, 4382,
    4383, 4384, 4385, 4386, 4387, 4388, 4389, 4390, 4391, 4392, 4393, 4394,
    4395, 4396, 4397, 4398, 4399, 4400, 4401, 4402, 4403, 4404, 4405, 4406,
    4407, 4408, 4409, 4410, 4411, 4412, 4413, 4414, 4415, 4416, 4417, 4241,
    4418, 4419, 4420, 4421, 4422, 4423, 4423, 4424, 4425, 4426, 4427, 4428,
    4429, 4430, 4431, 4371, 4432, 4433, 4434, 4435, 4436, 4437, 0,    0,
    4438, 4439, 4440, 4441, 4442, 4443, 4444, 4445, 4385, 4446, 4447, 4448,
    4357, 4449, 4450, 4451, 4452, 4453, 4454, 4455, 4456, 4457, 4458, 4459,
    4460, 4394, 4461, 4395, 4462, 4463, 4464, 4465, 4466, 4358, 4133, 4467,
    4468, 4469, 4203, 4290, 4470, 4471, 4402, 4472, 4403, 4473, 4474, 4475,
    4360, 4476, 4477, 4478, 4479, 4480, 4361, 4481, 4482, 4483, 4484, 4485,
    4486, 4417, 4487, 4488, 4241, 4489, 4421, 4490, 4491, 4492, 4493, 4494,
    4426, 4495, 4370, 4496, 4427, 4184, 4497, 4428, 4498, 4430, 4499, 4500,
    4501, 4502, 4503, 4432, 4366, 4504, 4433, 4505, 4434, 4506, 4100, 4507,
    4508, 4509, 4510, 4511, 4512, 4513, 4514, 4515, 4516, 4517, 0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    4518, 4519, 4520, 4521, 4522, 4523, 4524, 0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    4525, 4526, 4527, 4528, 4529,
    0,    0,    0,    0,    0,    4530, 4531, 4532, 4533, 4534, 4535, 4536,
    4537, 4538, 4539, 4540, 4541, 4542, 4543, 4544, 4545, 4546, 4547, 4548,
    4549, 4550, 4551, 4552, 4553, 4554, 4555, 0,    4556, 4557, 4558, 4559,
    4560, 0,    4561, 0,    4562, 4563, 0,    4564, 4565, 0,    4566, 4567,
    4568, 4569, 4570, 4571, 4572, 4573, 4574, 4575, 4576, 4577, 4578, 4579,
    4580, 4581, 4582, 4583, 4584, 4585, 4586, 4587, 4588, 4589, 4590, 4591,
    4592, 4593, 4594, 4595, 4596, 4597, 4598, 4599, 4600, 4601, 4602, 4603,
    4604, 4605, 4606, 4607, 4608, 4609, 4610, 4611, 4612, 4613, 4614, 4615,
    4616, 4617, 4618, 4619, 4620, 4621, 4622, 4623, 4624, 4625, 4626, 4627,
    4628, 4629, 4630, 4631, 4632, 4633, 4634, 4635, 4636, 4637, 4638, 4639,
    4640, 4641, 4642, 4643, 4644, 4645, 4646, 4647, 4648, 4649, 4650, 4651,
    4652, 4653, 4654, 4655, 4656, 4657, 4658, 4659, 4660, 4661, 4662, 4663,
    4664, 4665, 4666, 4667, 4668, 4669, 4670, 4671, 4672, 4673, 4674, 4674,
    4674, 4674, 4674, 4674, 4674, 4674, 4674, 4674, 4674, 4674, 4674, 4674,
    4674, 4674, 0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    4675, 4676, 4677, 4678, 4679,
    4680, 4681, 4682, 4683, 4684, 4685, 4686, 4687, 4688, 4689, 4690, 4691,
    4692, 4693, 4694, 4695, 4696, 4697, 4698, 4699, 4700, 4701, 4702, 4703,
    4704, 4705, 4706, 4707, 4708, 4709, 4710, 4711, 4712, 4713, 4714, 4715,
    4716, 4717, 4718, 4719, 4720, 4721, 4722, 4713, 4723, 4724, 4725, 4726,
    4727, 4728, 4729, 4730, 4731, 4732, 4733, 4734, 4735, 4736, 4737, 4738,
    4739, 4740, 4741, 4742, 4743, 4744, 4745, 4746, 4747, 4748, 4749, 4750,
    4751, 4752, 4753, 4754, 4755, 4756, 4757, 4758, 4759, 4760, 4761, 4762,
    4763, 4764, 4765, 4766, 4767, 4768, 4769, 4770, 4771, 4772, 4773, 4774,
    4775, 4776, 4777, 4778, 4779, 4780, 4781, 4782, 4783, 4784, 4785, 4786,
    4787, 4788, 4789, 4790, 4791, 4792, 4793, 4794, 4795, 4796, 4797, 4798,
    4799, 4800, 4801, 4802, 4803, 4804, 4805, 4806, 4807, 4808, 4809, 4810,
    4811, 4812, 4813, 4814, 4815, 4816, 4817, 4818, 4819, 4820, 4821, 4822,
    4714, 4823, 4824, 4825, 4826, 4827, 4828, 4829, 4830, 4831, 4832, 4833,
    4834, 4835, 4836, 4837, 4838, 4839, 4840, 4841, 4842, 4843, 4844, 4845,
    4846, 4847, 4848, 4849, 4850, 4851, 4852, 4853, 4854, 4855, 4856, 4857,
    4858, 4859, 4860, 4861, 4862, 4863, 4864, 4865, 4866, 4867, 4868, 4869,
    4870, 4871, 4872, 4873, 4874, 4875, 4876, 4877, 4878, 4879, 4880, 4881,
    4882, 4883, 4884, 4885, 4886, 4887, 4888, 4889, 4890, 4891, 4892, 4893,
    4894, 4895, 4896, 4897, 4898, 4899, 4900, 4901, 4902, 4903, 4904, 4905,
    4906, 4907, 4908, 4909, 4910, 4911, 4912, 4913, 4914, 4915, 4916, 4917,
    4918, 4919, 4920, 4921, 4922, 4923, 4924, 4925, 4926, 4927, 4928, 4929,
    4930, 4931, 4932, 4933, 4934, 4935, 4936, 4937, 4938, 4939, 4940, 4941,
    4942, 4943, 4944, 4945, 4946, 4947, 4948, 4949, 4950, 4951, 4952, 4953,
    4954, 4955, 4956, 4957, 4958, 4959, 4960, 4961, 4962, 4963, 4964, 4965,
    4966, 4967, 4968, 4969, 4970, 4971, 4972, 4973, 4974, 4975, 4976, 4977,
    4978, 4979, 4980, 4981, 4982, 4983, 4984, 4985, 4986, 4987, 4988, 4989,
    4990, 4991, 4992, 4993, 4994, 4995, 4996, 4997, 4998, 4999, 5000, 5001,
    5002, 5003, 5004, 5005, 5006, 5007, 5008, 5009, 5010, 5011, 5012, 5013,
    5014, 5015, 5016, 5017, 5018, 5019, 5020, 5021, 5022, 5023, 5024, 5025,
    5026, 5027, 5028, 5029, 5030, 5031, 5032, 5033, 5034, 5035, 5036, 2136,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    5037, 5038, 5039, 5040, 5041, 5042, 5043, 5044,
    5045, 5046, 5047, 5048, 5049, 5050, 5051, 5052, 5053, 5054, 5055, 5056,
    5057, 5058, 5059, 5060, 5061, 5062, 5063, 5064, 5065, 5066, 5067, 5068,
    5069, 5070, 5071, 5072, 5073, 5074, 5075, 5076, 5077, 5078, 5079, 5080,
    5081, 5082, 5083, 5084, 5085, 5086, 5087, 5088, 5089, 5090, 5091, 5092,
    5093, 5094, 5095, 5096, 5097, 5098, 5099, 5100, 0,    0,    5101, 5102,
    5103, 5104, 5105, 5106, 5107, 5108, 5109, 5110, 5111, 5112, 5113, 5114,
    5115, 5116, 5117, 5118, 5119, 5120, 5121, 5122, 5123, 5124, 5125, 5126,
    5127, 5128, 5129, 5130, 5131, 5132, 5133, 5134, 5135, 5136, 5137, 5138,
    5139, 5140, 5141, 5142, 5143, 5144, 5145, 5146, 5147, 5148, 5149, 5150,
    5151, 5152, 5153, 5154, 0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    5155, 5156, 5157, 5158,
    5159, 5160, 5161, 5162, 5163, 5164, 5165, 5166, 5167, 77,   0,    0,
    576,  576,  576,  576,  576,  576,  576,  576,  576,  576,  576,  576,
    576,  576,  576,  576,  5168, 5169, 5170, 5171, 5172, 5173, 5174, 5175,
    5176, 5177, 0,    0,    0,    0,    0,    0,    542,  542,  542,  542,
    542,  542,  542,  555,  555,  555,  555,  555,  555,  555,  542,  542,
    5178, 5179, 5180, 5181, 5181, 5182, 5183, 5184, 5185, 5186, 5187, 5188,
    5189, 5190, 5191, 5192, 5193, 5194, 5195, 5196, 5197, 3032, 3032, 5198,
    5199, 5200, 5200, 5200, 5200, 5201, 5201, 5201, 5202, 5203, 5204, 0,
    5205, 5206, 5207, 5208, 5209, 5210, 5211, 5212, 5213, 5214, 5215, 5216,
    5217, 5218, 5219, 5220, 5221, 5222, 5223, 0,    5224, 5225, 5226, 5227,
    0,    0,    0,    0,    5228, 5229, 5230, 1116, 5231, 0,    5232, 5233,
    5234, 5235, 5236, 5237, 5238, 5239, 5240, 5241, 5242, 5243, 5244, 5245,
    5246, 5247, 5248, 5249, 5250, 5251, 5252, 5253, 5254, 5255, 5256, 5257,
    5258, 5259, 5260, 5261, 5262, 5263, 5264, 5265, 5266, 5267, 5268, 5269,
    5270, 5271, 5272, 5273, 5274, 5275, 5276, 5277, 5278, 5279, 5280, 5281,
    5282, 5283, 5284, 5285, 5286, 5287, 5288, 5289, 5290, 5291, 5292, 5293,
    5294, 5295, 5296, 5297, 5298, 5299, 5300, 5301, 5302, 5303, 5304, 5305,
    5306, 5307, 5308, 5309, 5310, 5311, 5312, 5313, 5314, 5315, 5316, 5317,
    5318, 5319, 5320, 5321, 5322, 5323, 5324, 5325, 5326, 5327, 5328, 5329,
    5330, 5331, 5332, 5333, 5334, 5335, 5336, 5337, 5338, 5339, 5340, 5341,
    5342, 5343, 5344, 5345, 5346, 5347, 5348, 5349, 5350, 5351, 5352, 5353,
    5354, 5355, 5356, 5357, 5358, 5359, 5360, 5361, 5362, 5363, 5364, 5365,
    5366, 0,    0,    1458, 0,    5367, 5368, 5369, 5370, 5371, 5372, 5373,
    5374, 5375, 5376, 5377, 5378, 5379, 5380, 5381, 5382, 5383, 5384, 5385,
    5386, 5387, 5388, 5389, 5390, 5391, 5392, 5393, 5394, 5395, 5396, 5397,
    5398, 5399, 5400, 5401, 5402, 5403, 5404, 5405, 5406, 5407, 5408, 5409,
    5410, 5411, 5412, 5413, 5414, 5415, 5416, 5417, 5418, 5419, 5420, 5421,
    5422, 5423, 5424, 5425, 5426, 5427, 5428, 5429, 5430, 5431, 5432, 5433,
    5434, 5435, 5436, 5437, 5438, 5439, 5440, 5441, 5442, 5443, 5444, 5445,
    5446, 5447, 5448, 5449, 5450, 5451, 5452, 5453, 5454, 5455, 5456, 5457,
    5458, 5459, 5460, 5461, 5462, 5463, 5464, 5465, 5466, 5467, 5468, 5469,
    5470, 5471, 5472, 5473, 5474, 5475, 5476, 5477, 5478, 5479, 5480, 5481,
    5482, 5483, 5484, 5485, 5486, 5487, 5488, 5489, 5490, 5491, 5492, 5493,
    5494, 5495, 5496, 5497, 5498, 5499, 5500, 5501, 5502, 5503, 5504, 5505,
    5506, 5507, 5508, 5509, 5510, 5511, 5512, 5513, 5514, 5515, 5516, 5517,
    5518, 5519, 5520, 5521, 5522, 5523, 5524, 5525, 5526, 5527, 5528, 5529,
    5530, 5531, 5532, 5533, 5534, 5535, 5536, 5537, 5538, 5539, 5540, 5541,
    5542, 5543, 5544, 5545, 5546, 5547, 5548, 5549, 5550, 5551, 5552, 5553,
    5554, 5555, 5556, 0,    0,    0,    5557, 5558, 5559, 5560, 5561, 5562,
    0,    0,    5563, 5564, 5565, 5566, 5567, 5568, 0,    0,    5569, 5570,
    5571, 5572, 5573, 5574, 0,    0,    5575, 5576, 5577, 0,    0,    0,
    5578, 5579, 5580, 5581, 5582, 5583, 5584, 0,    5585, 5586, 5587, 5588,
    5589, 5590, 5591, 0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    5592, 5592, 5592, 77,   77,   0,    0,    341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  0,    341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  0,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  0,    341,  341,  0,    341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  0,    0,    341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  0,    0,    0,    0,    0,
    1046, 9,    1046, 0,    0,    0,    0,    1181, 1181, 1181, 1181, 1181,
    1181, 1181, 1181, 1181, 1181, 1181, 1181, 1181, 1181, 1181, 1181, 1181,
    1181, 1181, 1181, 1181, 1181, 1181, 1181, 1181, 1181, 1181, 1181, 1181,
    1181, 1181, 1181, 1181, 1181, 1181, 1181, 1181, 1181, 1181, 1181, 1181,
    1181, 1181, 1181, 1181, 0,    0,    0,    840,  840,  840,  840,  840,
    840,  840,  840,  840,  5593, 5593, 5593, 5593, 5593, 5593, 5593, 5593,
    5593, 5593, 5593, 5593, 5593, 5593, 5593, 5593, 5593, 5593, 5593, 5593,
    5593, 5593, 5593, 5593, 5593, 5593, 5593, 5593, 5593, 5593, 5593, 5593,
    5593, 5593, 5593, 5593, 5593, 5593, 5593, 5593, 5593, 5593, 5593, 5593,
    5593, 5593, 5593, 5593, 5593, 5593, 5593, 5593, 5593, 1210, 1210, 1210,
    1210, 77,   77,   77,   77,   77,   77,   77,   77,   77,   77,   77,
    77,   77,   77,   77,   77,   77,   1210, 1210, 77,   840,  840,  0,
    77,   77,   77,   77,   77,   77,   77,   77,   77,   77,   77,   77,
    77,   0,    0,    0,    77,   0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    840,  840,  840,  840,  840,  840,  840,  840,
    840,  840,  840,  840,  840,  840,  840,  840,  840,  840,  840,  840,
    840,  840,  840,  840,  840,  840,  840,  840,  840,  840,  840,  840,
    840,  840,  840,  840,  840,  840,  840,  840,  840,  840,  840,  840,
    840,  555,  0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  0,    0,    0,    341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    555,  5594, 5594, 5594, 5594, 5594, 5594, 5594, 5594, 5594, 5594, 5594,
    5594, 5594, 5594, 5594, 5594, 5594, 5594, 5594, 5594, 5594, 5594, 5594,
    5594, 5594, 5594, 5594, 0,    0,    0,    0,    341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  1181, 1181, 1181, 1181, 0,    0,    0,    0,
    0,    0,    0,    0,    0,    341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  1457, 341,  341,  341,  341,  341,  341,  341,  341,  1457, 0,
    0,    0,    0,    0,    341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  542,  542,  542,  542,  542,  0,
    0,    0,    0,    0,    341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  0,    1046,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    0,    0,    0,    0,    341,  341,  341,  341,  341,  341,  341,  341,
    1046, 1457, 1457, 1457, 1457, 1457, 0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    5595, 5596, 5597, 5598, 5599, 5600, 5601, 5602, 5603, 5604, 5605, 5606,
    5607, 5608, 5609, 5610, 5611, 5612, 5613, 5614, 5615, 5616, 5617, 5618,
    5619, 5620, 5621, 5622, 5623, 5624, 5625, 5626, 5627, 5628, 5629, 5630,
    5631, 5632, 5633, 5634, 5635, 5636, 5637, 5638, 5639, 5640, 5641, 5642,
    5643, 5644, 5645, 5646, 5647, 5648, 5649, 5650, 5651, 5652, 5653, 5654,
    5655, 5656, 5657, 5658, 5659, 5660, 5661, 5662, 5663, 5664, 5665, 5666,
    5667, 5668, 5669, 5670, 5671, 5672, 5673, 5674, 341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  0,    0,    1171, 1171, 1171, 1171, 1171, 1171, 1171, 1171,
    1171, 1171, 0,    0,    0,    0,    0,    0,    5675, 5676, 5677, 5678,
    5679, 5680, 5681, 5682, 5683, 5684, 5685, 5686, 5687, 5688, 5689, 5690,
    5691, 5692, 5693, 5694, 5695, 5696, 5697, 5698, 5699, 5700, 5701, 5702,
    5703, 5704, 5705, 5706, 5707, 5708, 5709, 5710, 0,    0,    0,    0,
    5711, 5712, 5713, 5714, 5715, 5716, 5717, 5718, 5719, 5720, 5721, 5722,
    5723, 5724, 5725, 5726, 5727, 5728, 5729, 5730, 5731, 5732, 5733, 5734,
    5735, 5736, 5737, 5738, 5739, 5740, 5741, 5742, 5743, 5744, 5745, 5746,
    0,    0,    0,    0,    341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  0,    0,    0,    0,
    0,    0,    0,    0,    341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    1046, 0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  0,    0,    0,    0,    0,    0,    0,    0,    0,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    341,  341,  341,  341,
    341,  341,  341,  341,  0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    1107, 1107, 1107, 1107, 1107, 1107, 0,    0,    1107, 0,    1107, 1107,
    1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107,
    1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107,
    1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107,
    1107, 1107, 1107, 1107, 1107, 1107, 0,    1107, 1107, 0,    0,    0,
    1107, 0,    0,    1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107,
    1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107,
    1107, 1107, 0,    1104, 5747, 5747, 5747, 5747, 5747, 5747, 5747, 5747,
    1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107,
    1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107, 5748,
    5748, 5747, 5747, 5747, 5747, 5747, 5747, 5747, 1107, 1107, 1107, 1107,
    1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107,
    1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107,
    1107, 1107, 1107, 0,    0,    0,    0,    0,    0,    0,    0,    5747,
    5747, 5747, 5747, 5747, 5747, 5747, 5747, 5747, 0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    1107, 1107, 1107, 1107,
    1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107,
    1107, 1107, 1107, 0,    1107, 1107, 0,    0,    0,    0,    0,    5747,
    5747, 5747, 5747, 5747, 1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107,
    1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107,
    1107, 1107, 5747, 5747, 5747, 5747, 5747, 5747, 0,    0,    0,    9,
    1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107,
    1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107,
    1107, 1107, 0,    0,    0,    0,    0,    1104, 0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107,
    1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107,
    1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107,
    1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107,
    1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107, 0,    0,    0,    0,
    5747, 5747, 1107, 1107, 5747, 5747, 5747, 5747, 5747, 5747, 5747, 5747,
    5747, 5747, 5747, 5747, 5747, 5747, 5747, 5747, 0,    0,    5747, 5747,
    5747, 5747, 5747, 5747, 5747, 5747, 5747, 5747, 5747, 5747, 5747, 5747,
    5747, 5747, 5747, 5747, 5747, 5747, 5747, 5747, 5747, 5747, 5747, 5747,
    5747, 5747, 5747, 5747, 5747, 5747, 5747, 5747, 5747, 5747, 5747, 5747,
    5747, 5747, 5747, 5747, 5747, 5747, 5747, 5747, 1107, 1150, 1150, 1150,
    0,    1150, 1150, 0,    0,    0,    0,    0,    1150, 555,  1150, 542,
    1107, 1107, 1107, 1107, 0,    1107, 1107, 1107, 0,    1107, 1107, 1107,
    1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107,
    1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107,
    1107, 1107, 0,    0,    542,  568,  555,  0,    0,    0,    0,    1162,
    5747, 5747, 5747, 5747, 5747, 5747, 5747, 5747, 5747, 0,    0,    0,
    0,    0,    0,    0,    1104, 1104, 1104, 1104, 1104, 1104, 1104, 1104,
    1104, 0,    0,    0,    0,    0,    0,    0,    1107, 1107, 1107, 1107,
    1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107,
    1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107,
    1107, 5747, 5747, 1104, 1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107,
    1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107,
    1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107, 5747, 5747, 5747,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    1107, 1107, 1107, 1107,
    1107, 1107, 1107, 1107, 5748, 1107, 1107, 1107, 1107, 1107, 1107, 1107,
    1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107,
    1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107, 542,  555,  0,
    0,    0,    0,    5747, 5747, 5747, 5747, 5747, 1104, 1104, 1104, 1104,
    1104, 1104, 1104, 0,    0,    0,    0,    0,    0,    0,    0,    0,
    1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107,
    1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107,
    1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107,
    1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107,
    1107, 1107, 1107, 1107, 1107, 1107, 0,    0,    0,    9,    9,    9,
    9,    9,    9,    9,    1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107,
    1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107,
    1107, 1107, 0,    0,    5747, 5747, 5747, 5747, 5747, 5747, 5747, 5747,
    1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107,
    1107, 1107, 1107, 1107, 1107, 1107, 1107, 0,    0,    0,    0,    0,
    5747, 5747, 5747, 5747, 5747, 5747, 5747, 5747, 1107, 1107, 1107, 1107,
    1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107,
    1107, 1107, 0,    0,    0,    0,    0,    0,    0,    1104, 1104, 1104,
    1104, 0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    5747, 5747, 5747, 5747, 5747, 5747, 5747, 0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107,
    1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107,
    1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107,
    1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107,
    1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107,
    1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107,
    1107, 1107, 1107, 1107, 1107, 0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    5749, 5750, 5751, 5752, 5753, 5754, 5755, 5756, 5757, 5758, 5759, 5760,
    5761, 5762, 5763, 5764, 5765, 5766, 5767, 5768, 5769, 5770, 5771, 5772,
    5773, 5774, 5775, 5776, 5777, 5778, 5779, 5780, 5781, 5782, 5783, 5784,
    5785, 5786, 5787, 5788, 5789, 5790, 5791, 5792, 5793, 5794, 5795, 5796,
    5797, 5798, 5799, 0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    5800, 5801, 5802, 5803, 5804, 5805, 5806, 5807,
    5808, 5809, 5810, 5811, 5812, 5813, 5814, 5815, 5816, 5817, 5818, 5819,
    5820, 5821, 5822, 5823, 5824, 5825, 5826, 5827, 5828, 5829, 5830, 5831,
    5832, 5833, 5834, 5835, 5836, 5837, 5838, 5839, 5840, 5841, 5842, 5843,
    5844, 5845, 5846, 5847, 5848, 5849, 5850, 0,    0,    0,    0,    0,
    0,    0,    5747, 5747, 5747, 5747, 5747, 5747, 1116, 1116, 1116, 1116,
    1116, 1116, 1116, 1116, 1116, 1116, 1116, 1116, 1116, 1116, 1116, 1116,
    1116, 1116, 1116, 1116, 1116, 1116, 1116, 1116, 1116, 1116, 1116, 1116,
    1116, 1116, 1116, 1116, 1116, 1116, 1116, 1116, 542,  542,  542,  542,
    0,    0,    0,    0,    0,    0,    0,    0,    1134, 1134, 1134, 1134,
    1134, 1134, 1134, 1134, 1134, 1134, 0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    5851, 5851, 5851, 5851, 5851, 5851, 5851, 5851, 5851, 5851, 5851, 5851,
    5851, 5851, 5851, 5851, 5851, 5851, 5851, 5851, 5851, 5851, 5851, 5851,
    5851, 5851, 5851, 5851, 5851, 5851, 5851, 0,    1107, 1107, 1107, 1107,
    1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107,
    1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107,
    1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107,
    1107, 1107, 0,    542,  542,  1102, 0,    0,    1107, 1107, 0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107,
    1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107,
    1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107, 5747, 5747, 5747,
    5747, 5747, 5747, 5747, 5747, 5747, 5747, 1107, 0,    0,    0,    0,
    0,    0,    0,    0,    1116, 1116, 1116, 1116, 1116, 1116, 1116, 1116,
    1116, 1116, 1116, 1116, 1116, 1116, 1116, 1116, 1116, 1116, 1116, 1116,
    1116, 1116, 555,  555,  542,  542,  542,  555,  542,  555,  555,  555,
    555,  5852, 5852, 5852, 5852, 1111, 1111, 1111, 1111, 1111, 0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107,
    1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107, 5747, 5747, 5747,
    5747, 5747, 5747, 5747, 0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107,
    1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107, 0,
    0,    0,    0,    0,    0,    0,    0,    0,    1154, 1150, 1154, 341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  1150, 1150, 1150, 1150, 1150, 1150, 1150, 1150,
    1150, 1150, 1150, 1150, 1150, 1150, 1162, 1046, 1046, 1046, 1046, 1046,
    1046, 1046, 0,    0,    0,    0,    1210, 1210, 1210, 1210, 1210, 1210,
    1210, 1210, 1210, 1210, 1210, 1210, 1210, 1210, 1210, 1210, 1210, 1210,
    1210, 1210, 1171, 1171, 1171, 1171, 1171, 1171, 1171, 1171, 1171, 1171,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    1162, 1150, 1150, 1154, 341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  5853, 5854, 5855, 5856, 341,  341,  341,
    341,  341,  341,  341,  341,  5857, 341,  341,  341,  341,  341,  5858,
    341,  341,  341,  341,  1154, 1154, 1154, 1150, 1150, 1150, 1150, 1154,
    1154, 1162, 5859, 1046, 1046, 5860, 1046, 1046, 1046, 1046, 0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    5860, 0,    0,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  0,    0,    0,    0,    0,    0,    0,    1171, 1171, 1171, 1171,
    1171, 1171, 1171, 1171, 1171, 1171, 0,    0,    0,    0,    0,    0,
    542,  542,  542,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  5861, 1150, 1150, 1150, 1150, 1154, 1150, 5862, 5863,
    1150, 5864, 5865, 1162, 1162, 0,    1171, 1171, 1171, 1171, 1171, 1171,
    1171, 1171, 1171, 1171, 1046, 1046, 1046, 1046, 341,  1154, 1154, 341,
    0,    0,    0,    0,    0,    0,    0,    0,    341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  1172, 1046, 1046, 341,  0,
    0,    0,    0,    0,    0,    0,    0,    0,    1150, 1150, 1154, 341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  1154,
    1154, 1154, 1150, 1150, 1150, 1150, 1150, 1150, 1150, 1150, 1150, 1154,
    1482, 341,  1227, 1227, 341,  1046, 1046, 1046, 1046, 1150, 1172, 1150,
    1150, 1046, 1154, 1150, 1171, 1171, 1171, 1171, 1171, 1171, 1171, 1171,
    1171, 1171, 341,  1046, 341,  1046, 1046, 1046, 0,    1181, 1181, 1181,
    1181, 1181, 1181, 1181, 1181, 1181, 1181, 1181, 1181, 1181, 1181, 1181,
    1181, 1181, 1181, 1181, 1181, 0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  0,    341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    1154, 1154, 1154, 1150, 1150, 1150, 1154, 1154, 1150, 1482, 1172, 1150,
    1046, 1046, 1046, 1046, 1046, 1046, 1150, 0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    341,  341,  341,  341,  341,  341,  341,  0,    341,  0,    341,  341,
    341,  341,  0,    341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  0,    341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  1046, 0,    0,    0,    0,    0,    0,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  1150,
    1154, 1154, 1154, 1150, 1150, 1150, 1150, 1150, 1150, 1172, 1162, 0,
    0,    0,    0,    0,    1171, 1171, 1171, 1171, 1171, 1171, 1171, 1171,
    1171, 1171, 0,    0,    0,    0,    0,    0,    1150, 1150, 1154, 1154,
    0,    341,  341,  341,  341,  341,  341,  341,  341,  0,    0,    341,
    341,  0,    0,    341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  0,    341,  341,  341,  341,  341,  341,  341,  0,    341,  341,
    0,    341,  341,  341,  341,  341,  0,    1172, 1172, 341,  5866, 1154,
    1150, 1154, 1154, 1154, 1154, 0,    0,    5867, 1154, 0,    0,    5868,
    5869, 1482, 0,    0,    341,  0,    0,    0,    0,    0,    0,    5870,
    0,    0,    0,    0,    0,    341,  341,  341,  341,  341,  1154, 1154,
    0,    0,    542,  542,  542,  542,  542,  542,  542,  0,    0,    0,
    542,  542,  542,  542,  542,  0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  1154, 1154, 1154, 1150, 1150, 1150, 1150,
    1150, 1150, 1150, 1150, 1154, 1154, 1162, 1150, 1150, 1154, 1172, 341,
    341,  341,  341,  1046, 1046, 1046, 1046, 1046, 1171, 1171, 1171, 1171,
    1171, 1171, 1171, 1171, 1171, 1171, 1046, 1046, 0,    1046, 542,  341,
    341,  341,  0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  5871, 1154, 1154, 1150,
    1150, 1150, 1150, 1150, 1150, 5872, 5873, 5874, 5875, 5876, 5877, 1150,
    1150, 1154, 1162, 1172, 341,  341,  1046, 341,  0,    0,    0,    0,
    0,    0,    0,    0,    1171, 1171, 1171, 1171, 1171, 1171, 1171, 1171,
    1171, 1171, 0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  5878,
    1154, 1154, 1150, 1150, 1150, 1150, 0,    0,    5879, 5880, 5881, 5882,
    1150, 1150, 1154, 1162, 1172, 1046, 1046, 1046, 1046, 1046, 1046, 1046,
    1046, 1046, 1046, 1046, 1046, 1046, 1046, 1046, 1046, 1046, 1046, 1046,
    1046, 1046, 1046, 1046, 341,  341,  341,  341,  1150, 1150, 0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  1154, 1154, 1154, 1150,
    1150, 1150, 1150, 1150, 1150, 1150, 1150, 1154, 1154, 1150, 1154, 1162,
    1150, 1046, 1046, 1046, 341,  0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    1171, 1171, 1171, 1171, 1171, 1171, 1171, 1171,
    1171, 1171, 0,    0,    0,    0,    0,    0,    9,    9,    9,    9,
    9,    9,    9,    9,    9,    9,    9,    9,    9,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  1150,
    1154, 1150, 1154, 1154, 1150, 1150, 1150, 1150, 1150, 1150, 1482, 1172,
    341,  0,    0,    0,    0,    0,    0,    0,    1171, 1171, 1171, 1171,
    1171, 1171, 1171, 1171, 1171, 1171, 0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  0,    0,    1150, 1150, 1150, 1154, 1154, 1150, 1150,
    1150, 1150, 1154, 1150, 1150, 1150, 1150, 1162, 0,    0,    0,    0,
    1171, 1171, 1171, 1171, 1171, 1171, 1171, 1171, 1171, 1171, 1181, 1181,
    1046, 1046, 1046, 840,  0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    1154, 1154, 1154, 1150, 1150, 1150, 1150, 1150, 1150, 1150, 1150, 1150,
    1154, 1162, 1172, 1046, 0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    5883, 5884, 5885, 5886,
    5887, 5888, 5889, 5890, 5891, 5892, 5893, 5894, 5895, 5896, 5897, 5898,
    5899, 5900, 5901, 5902, 5903, 5904, 5905, 5906, 5907, 5908, 5909, 5910,
    5911, 5912, 5913, 5914, 5915, 5916, 5917, 5918, 5919, 5920, 5921, 5922,
    5923, 5924, 5925, 5926, 5927, 5928, 5929, 5930, 5931, 5932, 5933, 5934,
    5935, 5936, 5937, 5938, 5939, 5940, 5941, 5942, 5943, 5944, 5945, 5946,
    1171, 1171, 1171, 1171, 1171, 1171, 1171, 1171, 1171, 1171, 1181, 1181,
    1181, 1181, 1181, 1181, 1181, 1181, 1181, 0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    341,  341,  341,  341,  341,
    341,  341,  341,  0,    0,    341,  0,    0,    341,  341,  341,  341,
    341,  341,  341,  341,  0,    341,  341,  0,    341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  5947, 1154, 1154, 1154,
    1154, 5948, 0,    1154, 5949, 0,    0,    1150, 1150, 1482, 1162, 1227,
    1154, 1227, 1154, 1172, 1046, 1046, 1046, 0,    0,    0,    0,    0,
    0,    0,    0,    0,    1171, 1171, 1171, 1171, 1171, 1171, 1171, 1171,
    1171, 1171, 0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    341,  341,  341,  341,  341,  341,  341,  341,  0,    0,    341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  1154, 1154, 1154, 1150, 1150, 1150, 1150, 0,    0,    1150, 1150,
    1154, 1154, 1154, 1154, 1162, 341,  1046, 341,  1154, 0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    341,  1150, 1150, 1150, 1150, 1150, 1150, 5950, 5950, 1150, 1150, 341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  1150, 1162, 1150, 1150, 1150, 1150, 1154, 1227, 1150,
    1150, 1150, 1150, 1046, 1046, 1046, 1046, 1046, 1046, 1046, 1046, 1162,
    0,    0,    0,    0,    0,    0,    0,    0,    341,  1150, 1150, 1150,
    1150, 1150, 1150, 1154, 1154, 1150, 1150, 1150, 341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    1227, 1227, 1227, 1227, 1227, 1227, 1150, 1150, 1150, 1150, 1150, 1150,
    1150, 1150, 1150, 1150, 1150, 1150, 1150, 1154, 1150, 1162, 1046, 1046,
    1046, 341,  1046, 1046, 1046, 1046, 1046, 0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    341,  341,  341,  341,
    341,  341,  341,  341,  341,  0,    341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  1154, 1150, 1150, 1150, 1150,
    1150, 1150, 1150, 0,    1150, 1150, 1150, 1150, 1150, 1150, 1154, 5951,
    341,  1046, 1046, 1046, 1046, 1046, 0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    1171, 1171, 1171, 1171, 1171, 1171, 1171, 1171,
    1171, 1171, 1181, 1181, 1181, 1181, 1181, 1181, 1181, 1181, 1181, 1181,
    1181, 1181, 1181, 1181, 1181, 1181, 1181, 1181, 1181, 0,    0,    0,
    1046, 1046, 341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  0,    0,    1150, 1150,
    1150, 1150, 1150, 1150, 1150, 1150, 1150, 1150, 1150, 1150, 1150, 1150,
    1150, 1150, 1150, 1150, 1150, 1150, 1150, 1150, 0,    1154, 1150, 1150,
    1150, 1150, 1150, 1150, 1150, 1154, 1150, 1150, 1154, 1150, 1150, 0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    341,  341,  341,  341,  341,  341,  341,  0,    341,  341,  0,    341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  1150, 1150, 1150, 1150, 1150, 1150, 0,    0,    0,    1150, 0,
    1150, 1150, 0,    1150, 1150, 1150, 1172, 1150, 1162, 1162, 1227, 1150,
    0,    0,    0,    0,    0,    0,    0,    0,    1171, 1171, 1171, 1171,
    1171, 1171, 1171, 1171, 1171, 1171, 0,    0,    0,    0,    0,    0,
    341,  341,  341,  341,  341,  341,  0,    341,  341,  0,    341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  1154, 1154, 1154, 1154, 1154, 0,
    1150, 1150, 0,    1154, 1154, 1150, 1154, 1162, 341,  0,    0,    0,
    0,    0,    0,    0,    1171, 1171, 1171, 1171, 1171, 1171, 1171, 1171,
    1171, 1171, 0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  1150, 1150, 1154, 1154, 1046,
    1046, 0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    341,  0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    1181, 1181, 1181, 1181,
    1181, 1181, 1181, 1181, 1181, 1181, 1181, 1181, 1181, 1181, 1181, 1181,
    1181, 1181, 1181, 1181, 1181, 77,   77,   77,   77,   77,   77,   77,
    77,   11,   11,   11,   11,   77,   77,   77,   77,   77,   77,   77,
    77,   77,   77,   77,   77,   77,   77,   77,   77,   77,   0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    1046,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    1457, 1457, 1457, 1457, 1457, 1457, 1457, 1457,
    1457, 1457, 1457, 1457, 1457, 1457, 1457, 1457, 1457, 1457, 1457, 1457,
    1457, 1457, 1457, 1457, 1457, 1457, 1457, 1457, 1457, 1457, 1457, 1457,
    1457, 1457, 1457, 1457, 1457, 1457, 1457, 1457, 1457, 1457, 1457, 1457,
    1457, 1457, 1457, 1457, 1457, 1457, 1457, 1457, 1457, 1457, 1457, 1457,
    1457, 1457, 1457, 1457, 1457, 1457, 1457, 1457, 1457, 1457, 1457, 1457,
    1457, 1457, 1457, 1457, 1457, 1457, 1457, 1457, 1457, 1457, 1457, 1457,
    1457, 1457, 1457, 1457, 1457, 1457, 1457, 1457, 1457, 1457, 1457, 1457,
    1457, 1457, 1457, 1457, 1457, 1457, 1457, 1457, 1457, 1457, 1457, 1457,
    1457, 1457, 1457, 1457, 1457, 1457, 1457, 0,    1046, 1046, 1046, 1046,
    1046, 0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  0,
    5952, 5952, 5952, 5952, 5952, 5952, 5952, 5952, 5952, 0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  0,    0,    0,    0,    0,    0,    0,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  0,    1171, 1171, 1171, 1171,
    1171, 1171, 1171, 1171, 1171, 1171, 0,    0,    0,    0,    1046, 1046,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  0,    0,    568,  568,  568,  568,
    568,  1046, 0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    542,  542,  542,  542,  542,  542,  542,  1046, 1046, 1046, 1046, 1046,
    840,  840,  840,  840,  525,  525,  525,  525,  1046, 840,  0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    1171, 1171, 1171, 1171,
    1171, 1171, 1171, 1171, 1171, 1171, 0,    1181, 1181, 1181, 1181, 1181,
    1181, 1181, 0,    341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    0,    0,    0,    0,    0,    341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    5953, 5954, 5955, 5956,
    5957, 5958, 5959, 5960, 5961, 5962, 5963, 5964, 5965, 5966, 5967, 5968,
    5969, 5970, 5971, 5972, 5973, 5974, 5975, 5976, 5977, 5978, 5979, 5980,
    5981, 5982, 5983, 5984, 5985, 5986, 5987, 5988, 5989, 5990, 5991, 5992,
    5993, 5994, 5995, 5996, 5997, 5998, 5999, 6000, 6001, 6002, 6003, 6004,
    6005, 6006, 6007, 6008, 6009, 6010, 6011, 6012, 6013, 6014, 6015, 6016,
    1181, 1181, 1181, 1181, 1181, 1181, 1181, 1181, 1181, 1181, 1181, 1181,
    1181, 1181, 1181, 1181, 1181, 1181, 1181, 1181, 1181, 1181, 1181, 1046,
    1046, 1046, 1046, 0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  0,
    0,    0,    0,    1150, 341,  1154, 1154, 1154, 1154, 1154, 1154, 1154,
    1154, 1154, 1154, 1154, 1154, 1154, 1154, 1154, 1154, 1154, 1154, 1154,
    1154, 1154, 1154, 1154, 1154, 1154, 1154, 1154, 1154, 1154, 1154, 1154,
    1154, 1154, 1154, 1154, 1154, 1154, 1154, 1154, 1154, 1154, 1154, 1154,
    1154, 1154, 1154, 1154, 1154, 1154, 1154, 1154, 1154, 1154, 1154, 1154,
    0,    0,    0,    0,    0,    0,    0,    1150, 1150, 1150, 1150, 525,
    525,  525,  525,  525,  525,  525,  525,  525,  525,  525,  525,  525,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    3033, 3033, 3032, 3033, 1150, 0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    6017, 6017, 0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034,
    3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034,
    3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034,
    3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034,
    3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034,
    3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034,
    3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034,
    3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034,
    3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034,
    3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034,
    3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034,
    3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034,
    3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034,
    3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034,
    3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034,
    3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034,
    3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034,
    3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034,
    3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034,
    3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034,
    3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 0,    0,    0,    0,
    0,    0,    0,    0,    3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034,
    3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034,
    3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034,
    3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034,
    3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034,
    3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034,
    3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034,
    3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034,
    3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034,
    3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034,
    3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034,
    3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034,
    3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034,
    3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034,
    3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034,
    3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034,
    3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034,
    3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034,
    3034, 3034, 0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    3034, 3034, 3034, 3034,
    3034, 3034, 3034, 3034, 3034, 0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034,
    3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034,
    3034, 3034, 3034, 3034, 3034, 3034, 3034, 0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    3034, 3034, 3034, 0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    3034, 3034, 3034, 3034, 0,    0,    0,    0,
    0,    0,    0,    0,    3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034,
    3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034,
    3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034,
    3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034,
    3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034,
    3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034,
    3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034,
    3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034,
    3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034,
    3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034,
    3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034,
    3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034,
    3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034,
    3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034,
    3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034,
    3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034,
    3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034,
    3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034,
    3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034,
    3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034,
    3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034,
    3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034,
    3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034,
    3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034,
    3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034,
    3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034,
    3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034,
    3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034,
    3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034,
    3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034,
    3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034,
    3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034,
    3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034,
    3034, 3034, 3034, 3034, 0,    0,    0,    0,    341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  0,    0,    0,    0,    0,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  0,    0,    0,    341,  341,  341,  341,  341,  341,  341,  341,
    341,  0,    0,    0,    0,    0,    0,    0,    341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  0,    0,    840,  1150, 568,  1046,
    1458, 1458, 1458, 1458, 0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    840,  840,  840,  840,  840,  840,  840,  840,  840,  840,  840,  840,
    840,  840,  840,  840,  840,  840,  840,  840,  840,  840,  840,  840,
    840,  840,  840,  840,  840,  840,  840,  840,  840,  840,  840,  840,
    840,  840,  840,  840,  840,  840,  840,  840,  840,  840,  840,  840,
    840,  840,  840,  840,  840,  840,  840,  840,  840,  840,  840,  840,
    840,  840,  840,  840,  840,  840,  840,  840,  840,  840,  840,  840,
    840,  840,  840,  840,  840,  840,  840,  840,  840,  840,  840,  840,
    840,  840,  840,  840,  840,  840,  840,  840,  840,  840,  840,  840,
    840,  840,  840,  840,  840,  840,  840,  840,  840,  840,  840,  840,
    840,  840,  840,  840,  840,  840,  840,  840,  840,  840,  840,  840,
    840,  840,  840,  840,  840,  840,  840,  840,  840,  840,  840,  840,
    840,  840,  840,  840,  840,  840,  840,  840,  840,  840,  840,  840,
    840,  840,  840,  840,  840,  840,  840,  840,  840,  840,  840,  840,
    840,  840,  840,  840,  840,  840,  840,  840,  840,  840,  840,  840,
    840,  840,  840,  840,  840,  840,  840,  840,  840,  840,  840,  840,
    840,  840,  840,  840,  840,  840,  840,  840,  840,  840,  840,  840,
    840,  840,  840,  840,  840,  840,  840,  840,  840,  840,  840,  840,
    840,  840,  840,  840,  840,  840,  840,  840,  840,  840,  840,  840,
    840,  840,  840,  840,  840,  840,  840,  840,  840,  840,  840,  840,
    840,  840,  840,  840,  840,  840,  840,  840,  840,  840,  840,  840,
    840,  840,  840,  840,  840,  840,  0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    840,  840,  840,  840,  840,  840,  840,  840,
    840,  840,  840,  840,  840,  840,  840,  840,  840,  840,  840,  840,
    840,  840,  840,  840,  840,  840,  840,  840,  840,  840,  840,  840,
    840,  840,  840,  840,  840,  840,  840,  0,    0,    840,  840,  840,
    840,  840,  840,  840,  840,  840,  840,  840,  840,  840,  840,  840,
    840,  840,  840,  840,  840,  840,  840,  840,  840,  840,  840,  840,
    840,  840,  840,  840,  840,  840,  840,  840,  840,  840,  840,  840,
    840,  840,  840,  840,  840,  840,  840,  6018, 6019, 840,  840,  840,
    840,  840,  6020, 6021, 6022, 6023, 6024, 6025, 6026, 6027, 6028, 568,
    568,  568,  840,  840,  840,  6029, 6030, 6031, 6032, 6033, 6034, 1458,
    1458, 1458, 1458, 1458, 1458, 1458, 1458, 555,  555,  555,  555,  555,
    555,  555,  555,  840,  840,  542,  542,  542,  542,  542,  555,  555,
    840,  840,  840,  840,  840,  840,  840,  840,  840,  840,  840,  840,
    840,  840,  840,  840,  840,  840,  840,  840,  840,  840,  840,  840,
    840,  840,  840,  840,  840,  840,  542,  542,  542,  542,  840,  840,
    840,  840,  840,  840,  840,  840,  840,  840,  840,  6035, 6036, 6037,
    6038, 6039, 6040, 6041, 6042, 840,  840,  840,  840,  840,  840,  840,
    840,  840,  840,  840,  840,  840,  840,  840,  840,  840,  840,  840,
    840,  840,  840,  840,  840,  840,  840,  840,  840,  840,  840,  840,
    840,  840,  840,  840,  840,  840,  840,  840,  840,  0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    77,   77,   77,   77,
    77,   77,   77,   77,   77,   77,   77,   77,   77,   77,   77,   77,
    77,   77,   77,   77,   77,   77,   77,   77,   77,   77,   77,   77,
    77,   77,   77,   77,   77,   77,   77,   77,   77,   77,   77,   77,
    77,   77,   77,   77,   77,   77,   77,   77,   77,   77,   77,   77,
    77,   77,   77,   77,   77,   77,   77,   77,   77,   77,   77,   77,
    77,   77,   542,  542,  542,  77,   0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    1181, 1181, 1181, 1181, 1181, 1181, 1181, 1181,
    1181, 1181, 1181, 1181, 1181, 1181, 1181, 1181, 1181, 1181, 1181, 1181,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    77,   77,   77,   77,   77,   77,   77,   77,   77,   77,   77,   77,
    77,   77,   77,   77,   77,   77,   77,   77,   77,   77,   77,   77,
    77,   77,   77,   77,   77,   77,   77,   77,   77,   77,   77,   77,
    77,   77,   77,   77,   77,   77,   77,   77,   77,   77,   77,   77,
    77,   77,   77,   77,   77,   77,   77,   77,   77,   77,   77,   77,
    77,   77,   77,   77,   77,   77,   77,   77,   77,   77,   77,   77,
    77,   77,   77,   77,   77,   77,   77,   77,   77,   77,   77,   77,
    77,   77,   77,   0,    0,    0,    0,    0,    0,    0,    0,    0,
    1181, 1181, 1181, 1181, 1181, 1181, 1181, 1181, 1181, 1181, 1181, 1181,
    1181, 1181, 1181, 1181, 1181, 1181, 1181, 1181, 1181, 1181, 1181, 1181,
    1181, 0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    6043, 2232, 2207, 2250, 2234, 2235, 6044, 2214,
    2217, 6045, 6046, 2218, 2237, 2220, 6047, 2222, 2223, 2224, 6048, 6049,
    6050, 6051, 6052, 6053, 6054, 2228, 6055, 6056, 6057, 2251, 2233, 6058,
    2213, 2215, 2252, 2253, 6059, 2219, 6060, 6061, 2238, 6062, 6063, 6064,
    6065, 6066, 6067, 6068, 6069, 6070, 6071, 6072, 6043, 2232, 2207, 2250,
    2234, 2235, 6044, 2214, 2217, 6045, 6046, 2218, 2237, 2220, 6047, 2222,
    2223, 2224, 6048, 6049, 6050, 6051, 6052, 6053, 6054, 2228, 6055, 6056,
    6057, 2251, 2233, 6058, 2213, 0,    2252, 2253, 6059, 2219, 6060, 6061,
    2238, 6062, 6063, 6064, 6065, 6066, 6067, 6068, 6069, 6070, 6071, 6072,
    6043, 2232, 2207, 2250, 2234, 2235, 6044, 2214, 2217, 6045, 6046, 2218,
    2237, 2220, 6047, 2222, 2223, 2224, 6048, 6049, 6050, 6051, 6052, 6053,
    6054, 2228, 6055, 6056, 6057, 2251, 2233, 6058, 2213, 2215, 2252, 2253,
    6059, 2219, 6060, 6061, 2238, 6062, 6063, 6064, 6065, 6066, 6067, 6068,
    6069, 6070, 6071, 6072, 6043, 0,    2207, 2250, 0,    0,    6044, 0,
    0,    6045, 6046, 0,    0,    2220, 6047, 2222, 2223, 0,    6048, 6049,
    6050, 6051, 6052, 6053, 6054, 2228, 6055, 6056, 6057, 2251, 0,    6058,
    0,    2215, 2252, 2253, 6059, 2219, 6060, 6061, 0,    6062, 6063, 6064,
    6065, 6066, 6067, 6068, 6069, 6070, 6071, 6072, 6043, 2232, 2207, 2250,
    2234, 2235, 6044, 2214, 2217, 6045, 6046, 2218, 2237, 2220, 6047, 2222,
    2223, 2224, 6048, 6049, 6050, 6051, 6052, 6053, 6054, 2228, 6055, 6056,
    6057, 2251, 2233, 6058, 2213, 2215, 2252, 2253, 6059, 2219, 6060, 6061,
    2238, 6062, 6063, 6064, 6065, 6066, 6067, 6068, 6069, 6070, 6071, 6072,
    6043, 2232, 0,    2250, 2234, 2235, 6044, 0,    0,    6045, 6046, 2218,
    2237, 2220, 6047, 2222, 2223, 0,    6048, 6049, 6050, 6051, 6052, 6053,
    6054, 0,    6055, 6056, 6057, 2251, 2233, 6058, 2213, 2215, 2252, 2253,
    6059, 2219, 6060, 6061, 2238, 6062, 6063, 6064, 6065, 6066, 6067, 6068,
    6069, 6070, 6071, 6072, 6043, 2232, 0,    2250, 2234, 2235, 6044, 0,
    2217, 6045, 6046, 2218, 2237, 0,    6047, 0,    0,    0,    6048, 6049,
    6050, 6051, 6052, 6053, 6054, 0,    6055, 6056, 6057, 2251, 2233, 6058,
    2213, 2215, 2252, 2253, 6059, 2219, 6060, 6061, 2238, 6062, 6063, 6064,
    6065, 6066, 6067, 6068, 6069, 6070, 6071, 6072, 6043, 2232, 2207, 2250,
    2234, 2235, 6044, 2214, 2217, 6045, 6046, 2218, 2237, 2220, 6047, 2222,
    2223, 2224, 6048, 6049, 6050, 6051, 6052, 6053, 6054, 2228, 6055, 6056,
    6057, 2251, 2233, 6058, 2213, 2215, 2252, 2253, 6059, 2219, 6060, 6061,
    2238, 6062, 6063, 6064, 6065, 6066, 6067, 6068, 6069, 6070, 6071, 6072,
    6043, 2232, 2207, 2250, 2234, 2235, 6044, 2214, 2217, 6045, 6046, 2218,
    2237, 2220, 6047, 2222, 2223, 2224, 6048, 6049, 6050, 6051, 6052, 6053,
    6054, 2228, 6055, 6056, 6057, 2251, 2233, 6058, 2213, 2215, 2252, 2253,
    6059, 2219, 6060, 6061, 2238, 6062, 6063, 6064, 6065, 6066, 6067, 6068,
    6069, 6070, 6071, 6072, 6043, 2232, 2207, 2250, 2234, 2235, 6044, 2214,
    2217, 6045, 6046, 2218, 2237, 2220, 6047, 2222, 2223, 2224, 6048, 6049,
    6050, 6051, 6052, 6053, 6054, 2228, 6055, 6056, 6057, 2251, 2233, 6058,
    2213, 2215, 2252, 2253, 6059, 2219, 6060, 6061, 2238, 6062, 6063, 6064,
    6065, 6066, 6067, 6068, 6069, 6070, 6071, 6072, 6043, 2232, 2207, 2250,
    2234, 2235, 6044, 2214, 2217, 6045, 6046, 2218, 2237, 2220, 6047, 2222,
    2223, 2224, 6048, 6049, 6050, 6051, 6052, 6053, 6054, 2228, 6055, 6056,
    6057, 2251, 2233, 6058, 2213, 2215, 2252, 2253, 6059, 2219, 6060, 6061,
    2238, 6062, 6063, 6064, 6065, 6066, 6067, 6068, 6069, 6070, 6071, 6072,
    6043, 2232, 2207, 2250, 2234, 2235, 6044, 2214, 2217, 6045, 6046, 2218,
    2237, 2220, 6047, 2222, 2223, 2224, 6048, 6049, 6050, 6051, 6052, 6053,
    6054, 2228, 6055, 6056, 6057, 2251, 2233, 6058, 2213, 2215, 2252, 2253,
    6059, 2219, 6060, 6061, 2238, 6062, 6063, 6064, 6065, 6066, 6067, 6068,
    6069, 6070, 6071, 6072, 6043, 2232, 2207, 2250, 2234, 2235, 6044, 2214,
    2217, 6045, 6046, 2218, 2237, 2220, 6047, 2222, 2223, 2224, 6048, 6049,
    6050, 6051, 6052, 6053, 6054, 2228, 6055, 6056, 6057, 2251, 2233, 6058,
    2213, 2215, 2252, 2253, 6059, 2219, 6060, 6061, 2238, 6062, 6063, 6064,
    6065, 6066, 6067, 6068, 6069, 6070, 6071, 6072, 6073, 6074, 0,    0,
    6075, 6076, 2247, 6077, 6078, 6079, 6080, 6081, 6082, 6083, 6084, 6085,
    6086, 6087, 6088, 2248, 6089, 6090, 6091, 6092, 6093, 6094, 6095, 6096,
    6097, 6098, 6099, 6100, 2246, 6101, 6102, 6103, 6104, 6105, 6106, 6107,
    6108, 6109, 6110, 6111, 6112, 2245, 6113, 6114, 6115, 6116, 6117, 6118,
    6119, 6120, 6121, 6122, 6123, 6124, 6125, 6126, 6127, 6128, 6075, 6076,
    2247, 6077, 6078, 6079, 6080, 6081, 6082, 6083, 6084, 6085, 6086, 6087,
    6088, 2248, 6089, 6090, 6091, 6092, 6093, 6094, 6095, 6096, 6097, 6098,
    6099, 6100, 2246, 6101, 6102, 6103, 6104, 6105, 6106, 6107, 6108, 6109,
    6110, 6111, 6112, 2245, 6113, 6114, 6115, 6116, 6117, 6118, 6119, 6120,
    6121, 6122, 6123, 6124, 6125, 6126, 6127, 6128, 6075, 6076, 2247, 6077,
    6078, 6079, 6080, 6081, 6082, 6083, 6084, 6085, 6086, 6087, 6088, 2248,
    6089, 6090, 6091, 6092, 6093, 6094, 6095, 6096, 6097, 6098, 6099, 6100,
    2246, 6101, 6102, 6103, 6104, 6105, 6106, 6107, 6108, 6109, 6110, 6111,
    6112, 2245, 6113, 6114, 6115, 6116, 6117, 6118, 6119, 6120, 6121, 6122,
    6123, 6124, 6125, 6126, 6127, 6128, 6075, 6076, 2247, 6077, 6078, 6079,
    6080, 6081, 6082, 6083, 6084, 6085, 6086, 6087, 6088, 2248, 6089, 6090,
    6091, 6092, 6093, 6094, 6095, 6096, 6097, 6098, 6099, 6100, 2246, 6101,
    6102, 6103, 6104, 6105, 6106, 6107, 6108, 6109, 6110, 6111, 6112, 2245,
    6113, 6114, 6115, 6116, 6117, 6118, 6119, 6120, 6121, 6122, 6123, 6124,
    6125, 6126, 6127, 6128, 6075, 6076, 2247, 6077, 6078, 6079, 6080, 6081,
    6082, 6083, 6084, 6085, 6086, 6087, 6088, 2248, 6089, 6090, 6091, 6092,
    6093, 6094, 6095, 6096, 6097, 6098, 6099, 6100, 2246, 6101, 6102, 6103,
    6104, 6105, 6106, 6107, 6108, 6109, 6110, 6111, 6112, 2245, 6113, 6114,
    6115, 6116, 6117, 6118, 6119, 6120, 6121, 6122, 6123, 6124, 6125, 6126,
    6127, 6128, 6129, 6130, 0,    0,    6131, 6132, 6133, 6134, 6135, 6136,
    6137, 6138, 6139, 6140, 6131, 6132, 6133, 6134, 6135, 6136, 6137, 6138,
    6139, 6140, 6131, 6132, 6133, 6134, 6135, 6136, 6137, 6138, 6139, 6140,
    6131, 6132, 6133, 6134, 6135, 6136, 6137, 6138, 6139, 6140, 6131, 6132,
    6133, 6134, 6135, 6136, 6137, 6138, 6139, 6140, 1150, 1150, 1150, 1150,
    1150, 1150, 1150, 1150, 1150, 1150, 1150, 1150, 1150, 1150, 1150, 1150,
    1150, 1150, 1150, 1150, 1150, 1150, 1150, 1150, 1150, 1150, 1150, 1150,
    1150, 1150, 1150, 1150, 1150, 1150, 1150, 1150, 1150, 1150, 1150, 1150,
    1150, 1150, 1150, 1150, 1150, 1150, 1150, 1150, 1150, 1150, 1150, 1150,
    1150, 1150, 1150, 840,  840,  840,  840,  1150, 1150, 1150, 1150, 1150,
    1150, 1150, 1150, 1150, 1150, 1150, 1150, 1150, 1150, 1150, 1150, 1150,
    1150, 1150, 1150, 1150, 1150, 1150, 1150, 1150, 1150, 1150, 1150, 1150,
    1150, 1150, 1150, 1150, 1150, 1150, 1150, 1150, 1150, 1150, 1150, 1150,
    1150, 1150, 1150, 1150, 1150, 1150, 1150, 1150, 1150, 840,  840,  840,
    840,  840,  840,  840,  840,  1150, 840,  840,  840,  840,  840,  840,
    840,  840,  840,  840,  840,  840,  840,  840,  1150, 840,  840,  1046,
    1046, 1046, 1046, 1046, 0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    1150, 1150, 1150, 1150, 1150,
    0,    1150, 1150, 1150, 1150, 1150, 1150, 1150, 1150, 1150, 1150, 1150,
    1150, 1150, 1150, 1150, 0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    542,  542,  542,  542,  542,  542,  542,  0,    542,  542,  542,  542,
    542,  542,  542,  542,  542,  542,  542,  542,  542,  542,  542,  542,
    542,  0,    0,    542,  542,  542,  542,  542,  542,  542,  0,    542,
    542,  0,    542,  542,  542,  542,  542,  0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  0,    0,    0,    542,  542,  542,  542,  542,  542,  542,  525,
    525,  525,  525,  525,  525,  525,  0,    0,    1171, 1171, 1171, 1171,
    1171, 1171, 1171, 1171, 1171, 1171, 0,    0,    0,    0,    341,  840,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,  341,
    341,  341,  341,  341,  542,  542,  542,  542,  1171, 1171, 1171, 1171,
    1171, 1171, 1171, 1171, 1171, 1171, 0,    0,    0,    0,    0,    11,
    1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107,
    1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107,
    1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107,
    1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107,
    1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107,
    1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107,
    1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107,
    1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107,
    1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107,
    1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107,
    1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107,
    1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107,
    1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107,
    1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107,
    1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107,
    1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107, 1107,
    1107, 1107, 1107, 1107, 1107, 0,    0,    5747, 5747, 5747, 5747, 5747,
    5747, 5747, 5747, 5747, 555,  555,  555,  555,  555,  555,  555,  0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    6141, 6142, 6143, 6144, 6145, 6146, 6147, 6148,
    6149, 6150, 6151, 6152, 6153, 6154, 6155, 6156, 6157, 6158, 6159, 6160,
    6161, 6162, 6163, 6164, 6165, 6166, 6167, 6168, 6169, 6170, 6171, 6172,
    6173, 6174, 6175, 6176, 6177, 6178, 6179, 6180, 6181, 6182, 6183, 6184,
    6185, 6186, 6187, 6188, 6189, 6190, 6191, 6192, 6193, 6194, 6195, 6196,
    6197, 6198, 6199, 6200, 6201, 6202, 6203, 6204, 6205, 6206, 6207, 6208,
    542,  542,  542,  542,  542,  542,  1172, 1152, 0,    0,    0,    0,
    1151, 1151, 1151, 1151, 1151, 1151, 1151, 1151, 1151, 1151, 0,    0,
    0,    0,    1104, 1104, 0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    5852, 5852, 5852, 5852, 5852, 5852, 5852, 5852, 5852, 5852, 5852,
    5852, 5852, 5852, 5852, 5852, 5852, 5852, 5852, 5852, 5852, 5852, 5852,
    5852, 5852, 5852, 5852, 5852, 5852, 5852, 5852, 5852, 5852, 5852, 5852,
    5852, 5852, 5852, 5852, 5852, 5852, 5852, 5852, 5852, 5852, 5852, 5852,
    5852, 5852, 5852, 5852, 5852, 5852, 5852, 5852, 5852, 5852, 5852, 5852,
    1147, 5852, 5852, 5852, 1110, 5852, 5852, 5852, 5852, 0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    5852, 5852, 5852, 5852, 5852, 5852, 5852, 5852, 5852, 5852, 5852,
    5852, 5852, 5852, 5852, 5852, 5852, 5852, 5852, 5852, 5852, 5852, 5852,
    5852, 5852, 5852, 5852, 5852, 5852, 5852, 5852, 5852, 5852, 5852, 5852,
    5852, 5852, 5852, 5852, 5852, 5852, 5852, 5852, 5852, 5852, 1147, 5852,
    5852, 5852, 5852, 5852, 5852, 5852, 5852, 5852, 5852, 5852, 5852, 5852,
    5852, 5852, 0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    6209, 6210, 6211, 6212, 0,    6213, 6214, 6215,
    6216, 6217, 6218, 6219, 6220, 6221, 6222, 6223, 6224, 6225, 6226, 6227,
    6228, 6229, 6230, 6231, 6232, 6233, 6234, 6235, 6236, 6237, 6238, 6239,
    0,    6210, 6211, 0,    6240, 0,    0,    6215, 0,    6217, 6218, 6219,
    6220, 6221, 6222, 6223, 6224, 6225, 6226, 0,    6228, 6229, 6230, 6231,
    0,    6233, 0,    6235, 0,    0,    0,    0,    0,    0,    6211, 0,
    0,    0,    0,    6215, 0,    6217, 0,    6219, 0,    6221, 6222, 6223,
    0,    6225, 6226, 0,    6228, 0,    0,    6231, 0,    6233, 0,    6235,
    0,    6237, 0,    6239, 0,    6210, 6211, 0,    6240, 0,    0,    6215,
    6216, 6217, 6218, 0,    6220, 6221, 6222, 6223, 6224, 6225, 6226, 0,
    6228, 6229, 6230, 6231, 0,    6233, 6234, 6235, 6236, 0,    6238, 0,
    6209, 6210, 6211, 6212, 6240, 6213, 6214, 6215, 6216, 6217, 0,    6219,
    6220, 6221, 6222, 6223, 6224, 6225, 6226, 6227, 6228, 6229, 6230, 6231,
    6232, 6233, 6234, 6235, 0,    0,    0,    0,    0,    6210, 6211, 6212,
    0,    6213, 6214, 6215, 6216, 6217, 0,    6219, 6220, 6221, 6222, 6223,
    6224, 6225, 6226, 6227, 6228, 6229, 6230, 6231, 6232, 6233, 6234, 6235,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    75,   75,   0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    79,   79,   79,   79,
    2396, 79,   79,   79,   79,   79,   79,   79,   79,   79,   79,   79,
    79,   79,   79,   79,   79,   79,   79,   79,   79,   79,   79,   79,
    79,   79,   79,   79,   79,   79,   79,   79,   79,   79,   79,   79,
    79,   79,   79,   79,   0,    0,    0,    0,    79,   79,   79,   79,
    79,   79,   79,   79,   79,   79,   79,   79,   79,   79,   79,   79,
    79,   79,   79,   79,   79,   79,   79,   79,   79,   79,   79,   79,
    79,   79,   79,   79,   79,   79,   79,   79,   79,   79,   79,   79,
    79,   79,   79,   79,   79,   79,   79,   79,   79,   79,   79,   79,
    79,   79,   79,   79,   79,   79,   79,   79,   79,   79,   79,   79,
    79,   79,   79,   79,   79,   79,   79,   79,   79,   79,   79,   79,
    79,   79,   79,   79,   79,   79,   79,   79,   79,   79,   79,   79,
    79,   79,   79,   79,   79,   79,   79,   79,   79,   79,   79,   79,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    79,   79,   79,   79,   79,   79,   79,   79,   79,   79,   79,   79,
    79,   79,   79,   0,    0,    79,   79,   79,   79,   79,   79,   79,
    79,   79,   79,   79,   79,   79,   79,   79,   0,    79,   79,   79,
    79,   79,   79,   79,   79,   79,   79,   79,   79,   79,   79,   2396,
    0,    79,   79,   79,   79,   79,   79,   79,   79,   79,   79,   79,
    79,   79,   79,   79,   79,   79,   79,   79,   79,   79,   79,   79,
    79,   79,   79,   79,   79,   79,   79,   79,   79,   79,   79,   79,
    79,   79,   0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    6241, 6242, 6243, 6244, 6245, 6246, 6247, 6248, 6249, 6250, 6251, 1210,
    1210, 79,   79,   79,   6252, 6253, 6254, 6255, 6256, 6257, 6258, 6259,
    6260, 6261, 6262, 6263, 6264, 6265, 6266, 6267, 6268, 6269, 6270, 6271,
    6272, 6273, 6274, 6275, 6276, 6277, 6278, 6279, 6280, 6281, 6282, 79,
    6283, 6284, 6285, 6286, 6287, 6288, 6289, 6290, 6291, 6292, 6293, 6294,
    6295, 6296, 6297, 6298, 6299, 6300, 6301, 6302, 6303, 6304, 6305, 6306,
    6307, 6308, 6309, 6310, 6311, 6312, 6313, 6314, 840,  840,  840,  840,
    840,  840,  840,  840,  840,  840,  840,  840,  840,  840,  840,  840,
    840,  840,  840,  840,  840,  840,  840,  840,  840,  840,  6315, 6316,
    6317, 79,   79,   79,   2540, 2540, 840,  840,  840,  840,  840,  840,
    840,  840,  840,  840,  840,  840,  2540, 2540, 840,  840,  840,  840,
    840,  840,  840,  840,  840,  840,  840,  840,  840,  840,  6318, 840,
    6319, 6318, 6318, 6318, 6318, 6318, 6318, 6318, 6318, 6318, 6318, 840,
    840,  840,  840,  840,  840,  840,  840,  840,  840,  840,  840,  840,
    840,  840,  840,  840,  840,  79,   0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    6320, 6320, 6320, 6320, 6320, 6320, 6320, 6320, 6320, 6320,
    6320, 6320, 6320, 6320, 6320, 6320, 6320, 6320, 6320, 6320, 6320, 6320,
    6320, 6320, 6320, 6320, 6321, 6322, 6323, 0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    6324, 6325, 6326, 6327,
    6328, 6329, 6330, 6331, 6332, 6333, 6334, 6335, 6336, 6337, 6338, 6339,
    6340, 6341, 6342, 6343, 6344, 6345, 6346, 6347, 6348, 6349, 6350, 6351,
    6352, 6353, 6354, 6355, 6356, 6357, 6358, 6359, 6360, 6361, 6362, 6363,
    6364, 6365, 6366, 6367, 0,    0,    0,    0,    6368, 6369, 6370, 6371,
    6372, 6373, 6374, 6375, 6376, 0,    0,    0,    0,    0,    0,    0,
    6377, 6378, 0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    2396, 2396, 2396, 2396, 2396, 2396, 0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    2396, 2396, 2396, 2396,
    2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396,
    2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396,
    2396, 2396, 2396, 2396, 2396, 79,   79,   79,   79,   79,   79,   79,
    79,   79,   79,   79,   79,   2396, 2396, 2396, 2396, 2396, 2396, 2396,
    2396, 2396, 79,   2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396,
    2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396,
    2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396,
    2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396,
    2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396,
    2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396,
    2396, 79,   2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396,
    2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396,
    79,   79,   79,   79,   79,   79,   79,   79,   79,   79,   79,   79,
    2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396,
    2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396,
    2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396,
    2396, 2396, 2396, 2396, 2396, 2396, 2396, 79,   79,   79,   79,   2396,
    2396, 2396, 2396, 2396, 79,   79,   79,   79,   79,   79,   79,   79,
    79,   79,   79,   79,   2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396,
    2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396, 79,   79,   79,
    2396, 79,   79,   79,   2396, 2396, 2396, 6379, 6379, 6379, 6379, 6379,
    2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396,
    2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396,
    2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396,
    2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396,
    2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396,
    2396, 2396, 2396, 79,   2396, 79,   2396, 2396, 2396, 2396, 2396, 2396,
    2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396,
    2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396,
    2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396,
    2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396,
    2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396,
    2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396,
    2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396,
    2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396,
    2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396,
    2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396,
    2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396,
    2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396,
    2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396,
    2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396,
    2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396,
    2396, 79,   79,   2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396,
    2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396,
    2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396,
    2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396,
    2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396,
    2396, 2396, 2396, 2396, 2396, 2396, 77,   77,   77,   77,   77,   77,
    77,   77,   79,   79,   79,   79,   79,   2396, 2396, 2396, 2396, 79,
    2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396,
    2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396,
    79,   79,   79,   79,   79,   79,   79,   79,   79,   79,   79,   79,
    79,   79,   79,   79,   79,   79,   2396, 79,   79,   79,   79,   79,
    79,   79,   79,   79,   79,   79,   79,   79,   79,   79,   79,   79,
    79,   79,   79,   79,   79,   79,   79,   79,   79,   2396, 2396, 79,
    79,   79,   79,   79,   79,   79,   79,   79,   79,   79,   79,   79,
    2396, 79,   79,   79,   79,   79,   79,   79,   79,   79,   79,   79,
    79,   79,   79,   79,   79,   79,   79,   79,   79,   79,   79,   79,
    79,   79,   79,   79,   79,   79,   79,   79,   79,   79,   79,   79,
    79,   79,   79,   79,   79,   79,   79,   79,   79,   79,   79,   79,
    79,   79,   79,   79,   79,   79,   79,   79,   79,   79,   79,   79,
    79,   79,   79,   79,   79,   79,   79,   79,   79,   79,   79,   79,
    79,   79,   79,   79,   79,   79,   79,   79,   79,   79,   79,   79,
    79,   79,   79,   2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396,
    2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396,
    2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396,
    2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396,
    2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396,
    2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396,
    2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396,
    2396, 2396, 2396, 2396, 77,   77,   77,   77,   77,   77,   77,   77,
    77,   77,   77,   77,   77,   77,   77,   77,   77,   77,   77,   77,
    77,   77,   77,   77,   77,   77,   77,   77,   77,   77,   77,   77,
    77,   77,   77,   77,   77,   77,   77,   77,   77,   77,   77,   77,
    77,   77,   77,   77,   2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396,
    2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396,
    2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396,
    2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396,
    2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396,
    2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396,
    2396, 2396, 79,   79,   79,   79,   79,   79,   2396, 79,   79,   79,
    2396, 2396, 2396, 79,   79,   2396, 2396, 2396, 0,    0,    0,    0,
    0,    0,    0,    0,    79,   79,   79,   79,   79,   79,   79,   79,
    79,   79,   79,   2396, 2396, 0,    0,    0,    79,   79,   79,   79,
    2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396, 0,    0,    0,
    77,   77,   77,   77,   77,   77,   77,   77,   77,   77,   77,   77,
    77,   77,   77,   77,   77,   77,   77,   77,   77,   77,   77,   77,
    77,   77,   77,   77,   77,   77,   77,   77,   77,   77,   77,   77,
    77,   77,   77,   77,   77,   77,   77,   77,   77,   77,   77,   77,
    77,   77,   77,   77,   77,   77,   77,   77,   77,   77,   77,   77,
    77,   77,   77,   77,   77,   77,   77,   77,   77,   77,   77,   77,
    77,   77,   77,   77,   77,   77,   77,   77,   77,   77,   77,   77,
    77,   77,   77,   77,   77,   77,   77,   77,   77,   77,   77,   77,
    77,   77,   77,   77,   77,   77,   77,   77,   77,   77,   77,   77,
    77,   77,   77,   77,   77,   77,   77,   77,   0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    77,   77,   77,   77,
    77,   77,   77,   77,   77,   77,   77,   77,   77,   77,   77,   77,
    77,   77,   77,   77,   77,   77,   77,   77,   77,   77,   77,   77,
    77,   77,   77,   77,   77,   77,   77,   77,   77,   77,   77,   77,
    77,   77,   77,   77,   77,   77,   77,   77,   77,   77,   77,   77,
    77,   77,   77,   77,   77,   77,   77,   77,   77,   77,   77,   77,
    77,   77,   77,   77,   77,   77,   77,   77,   77,   77,   77,   77,
    77,   77,   77,   77,   77,   77,   77,   77,   77,   79,   79,   79,
    79,   0,    0,    0,    0,    0,    0,    0,    2396, 2396, 2396, 2396,
    2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396, 0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    77,   77,   77,   77,   77,   77,   77,   77,
    77,   77,   77,   77,   0,    0,    0,    0,    77,   77,   77,   77,
    77,   77,   77,   77,   77,   77,   77,   77,   77,   77,   77,   77,
    77,   77,   77,   77,   77,   77,   77,   77,   77,   77,   77,   77,
    77,   77,   77,   77,   77,   77,   77,   77,   77,   77,   77,   77,
    77,   77,   77,   77,   77,   77,   77,   77,   77,   77,   77,   77,
    77,   77,   77,   77,   0,    0,    0,    0,    0,    0,    0,    0,
    77,   77,   77,   77,   77,   77,   77,   77,   77,   77,   0,    0,
    0,    0,    0,    0,    77,   77,   77,   77,   77,   77,   77,   77,
    77,   77,   77,   77,   77,   77,   77,   77,   77,   77,   77,   77,
    77,   77,   77,   77,   77,   77,   77,   77,   77,   77,   77,   77,
    77,   77,   77,   77,   77,   77,   77,   77,   0,    0,    0,    0,
    0,    0,    0,    0,    77,   77,   77,   77,   77,   77,   77,   77,
    77,   77,   77,   77,   77,   77,   77,   77,   77,   77,   77,   77,
    77,   77,   77,   77,   77,   77,   77,   77,   77,   77,   0,    0,
    79,   79,   0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    77,   77,   77,   77,
    77,   77,   77,   77,   77,   77,   77,   77,   2396, 2396, 2396, 2396,
    2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396,
    2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396,
    2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396,
    2396, 2396, 2396, 2396, 2396, 2396, 2396, 77,   2396, 2396, 2396, 2396,
    2396, 2396, 2396, 2396, 2396, 2396, 77,   2396, 2396, 2396, 2396, 2396,
    2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396,
    2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396,
    2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396,
    2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396, 0,    2396, 2396,
    2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396,
    2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396,
    2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396,
    2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396,
    2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396,
    2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396,
    2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396, 0,    2396, 2396, 2396,
    2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396,
    2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396,
    2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396,
    2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396,
    79,   79,   79,   79,   79,   79,   79,   79,   79,   79,   79,   79,
    79,   79,   79,   79,   79,   79,   79,   79,   79,   79,   79,   79,
    79,   79,   79,   79,   79,   79,   79,   79,   79,   79,   79,   79,
    79,   79,   79,   79,   79,   79,   79,   79,   79,   79,   79,   79,
    79,   79,   79,   79,   79,   79,   79,   79,   79,   79,   79,   79,
    79,   79,   79,   79,   79,   79,   79,   79,   79,   79,   79,   79,
    79,   79,   79,   79,   79,   79,   79,   79,   79,   79,   79,   79,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    79,   79,   79,   79,   79,   79,   79,   79,   79,   79,   79,   79,
    79,   79,   0,    0,    2396, 2396, 2396, 2396, 2396, 0,    0,    0,
    2396, 2396, 2396, 0,    0,    0,    0,    0,    2396, 2396, 2396, 2396,
    2396, 2396, 2396, 0,    0,    0,    0,    0,    0,    0,    0,    0,
    2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396,
    2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396, 2396,
    2396, 0,    0,    0,    0,    0,    0,    0,    2396, 2396, 2396, 2396,
    2396, 2396, 2396, 0,    0,    0,    0,    0,    0,    0,    0,    0,
    2396, 2396, 2396, 0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    2396, 2396, 2396, 2396, 2396, 2396, 2396, 0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    77,   77,   77,   77,   77,   77,   77,   77,
    77,   77,   77,   77,   77,   77,   77,   77,   77,   77,   77,   77,
    77,   77,   77,   77,   77,   77,   77,   77,   77,   77,   77,   77,
    77,   77,   77,   77,   77,   77,   77,   77,   77,   77,   77,   77,
    77,   77,   77,   77,   77,   77,   77,   77,   77,   77,   77,   77,
    77,   77,   77,   77,   77,   77,   77,   77,   77,   77,   77,   77,
    77,   77,   77,   77,   77,   77,   77,   77,   77,   77,   77,   77,
    77,   77,   77,   77,   77,   77,   77,   77,   77,   77,   77,   77,
    77,   77,   77,   77,   77,   77,   77,   77,   77,   77,   77,   77,
    77,   77,   77,   77,   77,   77,   77,   77,   77,   77,   77,   77,
    77,   77,   77,   77,   77,   77,   77,   77,   77,   77,   77,   77,
    77,   77,   77,   77,   77,   77,   77,   77,   77,   77,   77,   77,
    77,   77,   77,   77,   77,   77,   77,   0,    77,   77,   77,   77,
    77,   77,   77,   77,   77,   77,   77,   77,   77,   77,   77,   77,
    77,   77,   77,   77,   77,   77,   77,   77,   77,   77,   77,   77,
    77,   77,   77,   77,   77,   77,   77,   77,   77,   77,   77,   77,
    77,   77,   77,   77,   77,   77,   77,   77,   77,   77,   77,   77,
    77,   77,   77,   0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    6131, 6132, 6133, 6134, 6135, 6136, 6137, 6138,
    6139, 6140, 0,    0,    0,    0,    0,    0,    3034, 3034, 3034, 3034,
    3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034,
    3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034,
    3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034,
    3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034,
    3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034,
    3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034,
    3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034,
    3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034,
    3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034,
    3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034,
    3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034,
    3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034,
    3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034,
    3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034,
    3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034,
    3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034,
    3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034,
    3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034,
    3034, 3034, 0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034,
    3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034,
    3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034,
    3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034,
    3034, 3034, 3034, 3034, 3034, 0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034,
    3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034,
    3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034,
    3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034,
    3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034,
    3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034,
    3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034,
    3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034,
    3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034,
    3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034,
    3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034,
    3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034,
    3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034,
    3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034,
    3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034,
    3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034,
    3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034,
    3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034,
    3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 0,    0,
    3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034,
    3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034,
    3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034,
    3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034,
    3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034,
    3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034,
    3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034,
    3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034,
    3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034,
    3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034,
    3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034,
    3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034,
    3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034,
    3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034,
    3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034,
    3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034,
    3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034,
    3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034,
    3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034,
    3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034,
    3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034,
    3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034,
    3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034,
    3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034,
    3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034,
    3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034,
    3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034,
    3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034,
    3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034,
    3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034,
    3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034,
    3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034,
    3034, 3034, 0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034,
    3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034,
    3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034,
    3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034,
    3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034,
    3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034,
    3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034,
    3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034,
    3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034,
    3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034,
    3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034,
    3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034,
    3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034,
    3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034,
    3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034,
    3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034,
    3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034,
    3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034,
    3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034,
    3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034,
    3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034,
    3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034,
    3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034,
    3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034,
    3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034,
    3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    6380, 6381, 6382, 6383, 6384, 4379, 6385, 6386,
    6387, 6388, 4380, 6389, 6390, 6391, 4381, 6392, 6393, 6394, 6395, 6396,
    6397, 6398, 6399, 6400, 6401, 6402, 6403, 4439, 6404, 6405, 6406, 6407,
    6408, 6409, 6410, 6411, 6412, 4444, 4382, 4383, 4445, 6413, 6414, 4190,
    6415, 4384, 6416, 6417, 6418, 6419, 6419, 6419, 6420, 6421, 6422, 6423,
    6424, 6425, 6426, 6427, 6428, 6429, 6430, 6431, 6432, 6433, 6434, 6435,
    6436, 6437, 6437, 4447, 6438, 6439, 6440, 6441, 4386, 6442, 6443, 6444,
    4343, 6445, 6446, 6447, 6448, 6449, 6450, 6451, 6452, 6453, 6454, 6455,
    6456, 6457, 6458, 6459, 6460, 6461, 6462, 6463, 6464, 6465, 6466, 6467,
    6468, 6469, 6470, 6470, 6471, 6472, 6473, 4186, 6474, 6475, 6476, 6477,
    6478, 6479, 6480, 6481, 4391, 6482, 6483, 6484, 6485, 6486, 6487, 6488,
    6489, 6490, 6491, 6492, 6493, 6494, 6495, 6496, 6497, 6498, 6499, 6500,
    6501, 6502, 4132, 6503, 6504, 6505, 6505, 6506, 6507, 6507, 6508, 6509,
    6510, 6511, 6512, 6513, 6514, 6515, 6516, 6517, 6518, 6519, 6520, 4392,
    6521, 6522, 6523, 6524, 4459, 6524, 6525, 4394, 6526, 6527, 6528, 6529,
    4395, 4105, 6530, 6531, 6532, 6533, 6534, 6535, 6536, 6537, 6538, 6539,
    6540, 6541, 6542, 6543, 6544, 6545, 6546, 6547, 6548, 6549, 6550, 6551,
    4396, 6552, 6553, 6554, 6555, 6556, 6557, 4398, 6558, 6559, 6560, 6561,
    6562, 6563, 6564, 6565, 4133, 4467, 6566, 6567, 6568, 6569, 6570, 6571,
    6572, 6573, 4399, 6574, 6575, 6576, 6577, 4510, 6578, 6579, 6580, 6581,
    6582, 6583, 6584, 6585, 6586, 6587, 6588, 6589, 6590, 4203, 6591, 6592,
    6593, 6594, 6595, 6596, 6597, 6598, 6599, 6600, 6601, 4400, 4290, 6602,
    6603, 6604, 6605, 6606, 6607, 6608, 6609, 4471, 6610, 6611, 6612, 6613,
    6614, 6615, 6616, 6617, 4472, 6618, 6619, 6620, 6621, 6622, 6623, 6624,
    6625, 6626, 6627, 6628, 6629, 4474, 6630, 6631, 6632, 6633, 6634, 6635,
    6636, 6637, 6638, 6639, 6640, 6640, 6641, 6642, 4476, 6643, 6644, 6645,
    6646, 6647, 6648, 6649, 4189, 6650, 6651, 6652, 6653, 6654, 6655, 6656,
    4482, 6657, 6658, 6659, 6660, 6661, 6662, 6662, 4483, 4512, 6663, 6664,
    6665, 6666, 6667, 4151, 4485, 6668, 6669, 4411, 6670, 6671, 4365, 6672,
    6673, 4415, 6674, 6675, 6676, 6677, 6677, 6678, 6679, 6680, 6681, 6682,
    6683, 6684, 6685, 6686, 6687, 6688, 6689, 6690, 6691, 6692, 6693, 6694,
    6695, 6696, 6697, 6698, 6699, 6700, 6701, 6702, 6703, 6704, 4421, 6705,
    6706, 6707, 6708, 6709, 6710, 6711, 6712, 6713, 6714, 6715, 6716, 6717,
    6718, 6719, 6720, 6506, 6721, 6722, 6723, 6724, 6725, 6726, 6727, 6728,
    6729, 6730, 6731, 6732, 4207, 6733, 6734, 6735, 6736, 6737, 6738, 4424,
    6739, 6740, 6741, 6742, 6743, 6744, 6745, 6746, 6747, 6748, 6749, 6750,
    6751, 6752, 6753, 6754, 6755, 6756, 6757, 6758, 4146, 6759, 6760, 6761,
    6762, 6763, 6764, 4492, 6765, 6766, 6767, 6768, 6769, 6770, 6771, 6772,
    6773, 6774, 6775, 6776, 6777, 6778, 6779, 6780, 6781, 6782, 6783, 6784,
    4497, 4498, 6785, 6786, 6787, 6788, 6789, 6790, 6791, 6792, 6793, 6794,
    6795, 6796, 6797, 4499, 6798, 6799, 6800, 6801, 6802, 6803, 6804, 6805,
    6806, 6807, 6808, 6809, 6810, 6811, 6812, 6813, 6814, 6815, 6816, 6817,
    6818, 6819, 6820, 6821, 6822, 6823, 6824, 6825, 6826, 6827, 4505, 4505,
    6828, 6829, 6830, 6831, 6832, 6833, 6834, 6835, 6836, 6837, 4506, 6838,
    6839, 6840, 6841, 6842, 6843, 6844, 6845, 6846, 6847, 6848, 6849, 6850,
    6851, 6852, 6853, 6854, 6855, 6856, 0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034,
    3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034,
    3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034,
    3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034,
    3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034,
    3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034, 3034,
    3034, 3034, 3034, 3034, 3034, 3034, 3034, 0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    1458, 0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    2128, 2128, 2128, 2128, 2128, 2128, 2128, 2128,
    2128, 2128, 2128, 2128, 2128, 2128, 2128, 2128, 2128, 2128, 2128, 2128,
    2128, 2128, 2128, 2128, 2128, 2128, 2128, 2128, 2128, 2128, 2128, 2128,
    2128, 2128, 2128, 2128, 2128, 2128, 2128, 2128, 2128, 2128, 2128, 2128,
    2128, 2128, 2128, 2128, 2128, 2128, 2128, 2128, 2128, 2128, 2128, 2128,
    2128, 2128, 2128, 2128, 2128, 2128, 2128, 2128, 2128, 2128, 2128, 2128,
    2128, 2128, 2128, 2128, 2128, 2128, 2128, 2128, 2128, 2128, 2128, 2128,
    2128, 2128, 2128, 2128, 2128, 2128, 2128, 2128, 2128, 2128, 2128, 2128,
    2128, 2128, 2128, 2128, 0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    576,  576,  576,  576,  576,  576,  576,  576,  576,  576,  576,  576,
    576,  576,  576,  576,  576,  576,  576,  576,  576,  576,  576,  576,
    576,  576,  576,  576,  576,  576,  576,  576,  576,  576,  576,  576,
    576,  576,  576,  576,  576,  576,  576,  576,  576,  576,  576,  576,
    576,  576,  576,  576,  576,  576,  576,  576,  576,  576,  576,  576,
    576,  576,  576,  576,  576,  576,  576,  576,  576,  576,  576,  576,
    576,  576,  576,  576,  576,  576,  576,  576,  576,  576,  576,  576,
    576,  576,  576,  576,  576,  576,  576,  576,  576,  576,  576,  576,
    576,  576,  576,  576,  576,  576,  576,  576,  576,  576,  576,  576,
    576,  576,  576,  576,  576,  576,  576,  576,  576,  576,  576,  576,
    576,  576,  576,  576,  576,  576,  576,  576,  576,  576,  576,  576,
    576,  576,  576,  576,  576,  576,  576,  576,  576,  576,  576,  576,
    576,  576,  576,  576,  576,  576,  576,  576,  576,  576,  576,  576,
    576,  576,  576,  576,  576,  576,  576,  576,  576,  576,  576,  576,
    576,  576,  576,  576,  576,  576,  576,  576,  576,  576,  576,  576,
    576,  576,  576,  576,  576,  576,  576,  576,  576,  576,  576,  576,
    576,  576,  576,  576,  576,  576,  576,  576,  576,  576,  576,  576,
    576,  576,  576,  576,  576,  576,  576,  576,  576,  576,  576,  576,
    576,  576,  576,  576,  576,  576,  576,  576,  576,  576,  576,  576,
    576,  576,  576,  576,  576,  576,  576,  576,  576,  576,  576,  576,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    4092, 4092, 4092, 4092, 4092, 4092, 4092, 4092,
    4092, 4092, 4092, 4092, 4092, 4092, 4092, 4092, 4092, 4092, 4092, 4092,
    4092, 4092, 4092, 4092, 4092, 4092, 4092, 4092, 4092, 4092, 4092, 4092,
    4092, 4092, 4092, 4092, 4092, 4092, 4092, 4092, 4092, 4092, 4092, 4092,
    4092, 4092, 4092, 4092, 4092, 4092, 4092, 4092, 4092, 4092, 4092, 4092,
    4092, 4092, 4092, 4092, 4092, 4092, 4092, 4092, 4092, 4092, 4092, 4092,
    4092, 4092, 4092, 4092, 4092, 4092, 4092, 4092, 4092, 4092, 4092, 4092,
    4092, 4092, 4092, 4092, 4092, 4092, 4092, 4092, 4092, 4092, 4092, 4092,
    4092, 4092, 4092, 4092, 4092, 4092, 4092, 4092, 4092, 4092, 4092, 4092,
    4092, 4092, 4092, 4092, 4092, 4092, 4092, 4092, 4092, 4092, 4092, 4092,
    4092, 4092, 4092, 4092, 4092, 4092, 4092, 4092, 4092, 4092, 4092, 4092,
    4092, 4092, 4092, 4092, 4092, 4092, 4092, 4092, 4092, 4092, 4092, 4092,
    4092, 4092, 4092, 4092, 4092, 4092, 4092, 4092, 4092, 4092, 4092, 4092,
    4092, 4092, 4092, 4092, 4092, 4092, 4092, 4092, 4092, 4092, 4092, 4092,
    4092, 4092, 4092, 4092, 4092, 4092, 4092, 4092, 4092, 4092, 4092, 4092,
    4092, 4092, 4092, 4092, 4092, 4092, 4092, 4092, 4092, 4092, 4092, 4092,
    4092, 4092, 4092, 4092, 4092, 4092, 4092, 4092, 4092, 4092, 4092, 4092,
    4092, 4092, 4092, 4092, 4092, 4092, 4092, 4092, 4092, 4092, 4092, 4092,
    4092, 4092, 4092, 4092, 4092, 4092, 4092, 4092, 4092, 4092, 4092, 4092,
    4092, 4092, 4092, 4092, 4092, 4092, 4092, 4092, 4092, 4092, 4092, 4092,
    4092, 4092, 4092, 4092, 4092, 4092, 4092, 4092, 4092, 4092, 4092, 4092,
    4092, 4092, 4092, 4092, 4092, 4092, 0,    0,
};

static const utf8proc_property_t utf8proc_properties[] = {
    {0,
     0,
     0,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_CC,
     0,
     UTF8PROC_BIDI_CLASS_BN,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     true,
     false,
     true,
     0,
     0,
     UTF8PROC_BOUNDCLASS_CONTROL},
    {UTF8PROC_CATEGORY_CC,
     0,
     UTF8PROC_BIDI_CLASS_BN,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     true,
     0,
     0,
     UTF8PROC_BOUNDCLASS_CONTROL},
    {UTF8PROC_CATEGORY_CC,
     0,
     UTF8PROC_BIDI_CLASS_S,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     true,
     0,
     0,
     UTF8PROC_BOUNDCLASS_CONTROL},
    {UTF8PROC_CATEGORY_CC,
     0,
     UTF8PROC_BIDI_CLASS_B,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     true,
     0,
     0,
     UTF8PROC_BOUNDCLASS_LF},
    {UTF8PROC_CATEGORY_CC,
     0,
     UTF8PROC_BIDI_CLASS_WS,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     true,
     0,
     0,
     UTF8PROC_BOUNDCLASS_CONTROL},
    {UTF8PROC_CATEGORY_CC,
     0,
     UTF8PROC_BIDI_CLASS_B,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     true,
     0,
     0,
     UTF8PROC_BOUNDCLASS_CR},
    {UTF8PROC_CATEGORY_CC,
     0,
     UTF8PROC_BIDI_CLASS_B,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     true,
     0,
     0,
     UTF8PROC_BOUNDCLASS_CONTROL},
    {UTF8PROC_CATEGORY_ZS,
     0,
     UTF8PROC_BIDI_CLASS_WS,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_PO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_PO,
     0,
     UTF8PROC_BIDI_CLASS_ET,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SC,
     0,
     UTF8PROC_BIDI_CLASS_ET,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_PS,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     true,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_PE,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     true,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SM,
     0,
     UTF8PROC_BIDI_CLASS_ES,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_PO,
     0,
     UTF8PROC_BIDI_CLASS_CS,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_PD,
     0,
     UTF8PROC_BIDI_CLASS_ES,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_ND,
     0,
     UTF8PROC_BIDI_CLASS_EN,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SM,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     5093,
     true,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SM,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     5084,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SM,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     5096,
     true,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     0,
     UINT16_MAX,
     0,
     UINT16_MAX,
     0,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     1,
     UINT16_MAX,
     1,
     UINT16_MAX,
     2784,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     2,
     UINT16_MAX,
     2,
     UINT16_MAX,
     49,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     3,
     UINT16_MAX,
     3,
     UINT16_MAX,
     704,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     4,
     UINT16_MAX,
     4,
     UINT16_MAX,
     62,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     5,
     UINT16_MAX,
     5,
     UINT16_MAX,
     2872,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     6,
     UINT16_MAX,
     6,
     UINT16_MAX,
     782,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     7,
     UINT16_MAX,
     7,
     UINT16_MAX,
     808,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     8,
     UINT16_MAX,
     8,
     UINT16_MAX,
     111,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     9,
     UINT16_MAX,
     9,
     UINT16_MAX,
     898,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     10,
     UINT16_MAX,
     10,
     UINT16_MAX,
     913,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     11,
     UINT16_MAX,
     11,
     UINT16_MAX,
     999,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     12,
     UINT16_MAX,
     12,
     UINT16_MAX,
     2890,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     13,
     UINT16_MAX,
     13,
     UINT16_MAX,
     160,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     14,
     UINT16_MAX,
     14,
     UINT16_MAX,
     205,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     15,
     UINT16_MAX,
     15,
     UINT16_MAX,
     2982,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     16,
     UINT16_MAX,
     16,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     17,
     UINT16_MAX,
     17,
     UINT16_MAX,
     1087,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     18,
     UINT16_MAX,
     18,
     UINT16_MAX,
     1173,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     19,
     UINT16_MAX,
     19,
     UINT16_MAX,
     1257,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     20,
     UINT16_MAX,
     20,
     UINT16_MAX,
     254,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     21,
     UINT16_MAX,
     21,
     UINT16_MAX,
     3042,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     22,
     UINT16_MAX,
     22,
     UINT16_MAX,
     1337,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     23,
     UINT16_MAX,
     23,
     UINT16_MAX,
     3122,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     24,
     UINT16_MAX,
     24,
     UINT16_MAX,
     303,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     25,
     UINT16_MAX,
     25,
     UINT16_MAX,
     1423,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SK,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_PC,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     1491,
     UINT16_MAX,
     1491,
     352,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     1493,
     UINT16_MAX,
     1493,
     2818,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     2812,
     UINT16_MAX,
     2812,
     401,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     1494,
     UINT16_MAX,
     1494,
     743,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     1495,
     UINT16_MAX,
     1495,
     414,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     2837,
     UINT16_MAX,
     2837,
     2875,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     1497,
     UINT16_MAX,
     1497,
     795,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     1498,
     UINT16_MAX,
     1498,
     853,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     1499,
     UINT16_MAX,
     1499,
     463,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     1500,
     UINT16_MAX,
     1500,
     901,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     1501,
     UINT16_MAX,
     1501,
     956,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     1502,
     UINT16_MAX,
     1502,
     1043,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     1503,
     UINT16_MAX,
     1503,
     2932,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     1504,
     UINT16_MAX,
     1504,
     512,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     1505,
     UINT16_MAX,
     1505,
     557,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     1507,
     UINT16_MAX,
     1507,
     2994,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     2826,
     UINT16_MAX,
     2826,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     1508,
     UINT16_MAX,
     1508,
     1130,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     3305,
     UINT16_MAX,
     3305,
     1215,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     1509,
     UINT16_MAX,
     1509,
     1296,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     1510,
     UINT16_MAX,
     1510,
     606,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     2908,
     UINT16_MAX,
     2908,
     3082,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     1511,
     UINT16_MAX,
     1511,
     1380,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     2925,
     UINT16_MAX,
     2925,
     3131,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     3312,
     UINT16_MAX,
     3312,
     655,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     2834,
     UINT16_MAX,
     2834,
     1466,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SM,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_ZS,
     0,
     UTF8PROC_BIDI_CLASS_CS,
     UTF8PROC_DECOMP_TYPE_NOBREAK,
     26,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SK,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     8219,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     1621,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_EXTENDED_PICTOGRAPHIC},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SUPER,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_PI,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     true,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_CF,
     0,
     UTF8PROC_BIDI_CLASS_BN,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     true,
     true,
     1,
     0,
     UTF8PROC_BOUNDCLASS_CONTROL},
    {UTF8PROC_CATEGORY_SK,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     8221,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_ET,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SM,
     0,
     UTF8PROC_BIDI_CLASS_ET,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_NO,
     0,
     UTF8PROC_BIDI_CLASS_EN,
     UTF8PROC_DECOMP_TYPE_SUPER,
     31,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_NO,
     0,
     UTF8PROC_BIDI_CLASS_EN,
     UTF8PROC_DECOMP_TYPE_SUPER,
     32,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SK,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     8225,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     35,
     35,
     7228,
     UINT16_MAX,
     7228,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SK,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     8228,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_NO,
     0,
     UTF8PROC_BIDI_CLASS_EN,
     UTF8PROC_DECOMP_TYPE_SUPER,
     38,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SUPER,
     14,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_PF,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     true,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_NO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_FRACTION,
     16423,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_NO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_FRACTION,
     16426,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_NO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_FRACTION,
     16429,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     8240,
     50,
     UINT16_MAX,
     50,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     8243,
     53,
     UINT16_MAX,
     53,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     8246,
     56,
     UINT16_MAX,
     56,
     UINT16_MAX,
     3143,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     8249,
     59,
     UINT16_MAX,
     59,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     8252,
     62,
     UINT16_MAX,
     62,
     UINT16_MAX,
     1537,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     8255,
     65,
     UINT16_MAX,
     65,
     UINT16_MAX,
     1579,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     66,
     UINT16_MAX,
     66,
     UINT16_MAX,
     1549,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     8259,
     69,
     UINT16_MAX,
     69,
     UINT16_MAX,
     2852,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     8262,
     72,
     UINT16_MAX,
     72,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     8265,
     75,
     UINT16_MAX,
     75,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     8268,
     78,
     UINT16_MAX,
     78,
     UINT16_MAX,
     3357,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     8271,
     81,
     UINT16_MAX,
     81,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     8274,
     84,
     UINT16_MAX,
     84,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     8277,
     87,
     UINT16_MAX,
     87,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     8280,
     90,
     UINT16_MAX,
     90,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     8283,
     93,
     UINT16_MAX,
     93,
     UINT16_MAX,
     2878,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     94,
     UINT16_MAX,
     94,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     8287,
     97,
     UINT16_MAX,
     97,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     8290,
     100,
     UINT16_MAX,
     100,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     8293,
     103,
     UINT16_MAX,
     103,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     8296,
     106,
     UINT16_MAX,
     106,
     UINT16_MAX,
     3461,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     8299,
     109,
     UINT16_MAX,
     109,
     UINT16_MAX,
     1597,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     8302,
     112,
     UINT16_MAX,
     112,
     UINT16_MAX,
     1591,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     113,
     UINT16_MAX,
     113,
     UINT16_MAX,
     1585,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     8306,
     116,
     UINT16_MAX,
     116,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     8309,
     119,
     UINT16_MAX,
     119,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     8312,
     122,
     UINT16_MAX,
     122,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     8315,
     125,
     UINT16_MAX,
     125,
     UINT16_MAX,
     1509,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     8318,
     128,
     UINT16_MAX,
     128,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     129,
     UINT16_MAX,
     129,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     8322,
     8077,
     UINT16_MAX,
     8077,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     8324,
     UINT16_MAX,
     8078,
     UINT16_MAX,
     8078,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     8326,
     UINT16_MAX,
     8079,
     UINT16_MAX,
     8079,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     8328,
     UINT16_MAX,
     8080,
     UINT16_MAX,
     8080,
     3192,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     8330,
     UINT16_MAX,
     8081,
     UINT16_MAX,
     8081,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     8332,
     UINT16_MAX,
     8082,
     UINT16_MAX,
     8082,
     1540,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     8334,
     UINT16_MAX,
     2836,
     UINT16_MAX,
     2836,
     1582,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     1492,
     UINT16_MAX,
     1492,
     1558,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     8336,
     UINT16_MAX,
     8083,
     UINT16_MAX,
     8083,
     2855,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     8338,
     UINT16_MAX,
     8084,
     UINT16_MAX,
     8084,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     8340,
     UINT16_MAX,
     8085,
     UINT16_MAX,
     8085,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     8342,
     UINT16_MAX,
     8086,
     UINT16_MAX,
     8086,
     3406,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     8344,
     UINT16_MAX,
     8087,
     UINT16_MAX,
     8087,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     8346,
     UINT16_MAX,
     8088,
     UINT16_MAX,
     8088,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     8348,
     UINT16_MAX,
     8089,
     UINT16_MAX,
     8089,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     8350,
     UINT16_MAX,
     8090,
     UINT16_MAX,
     8090,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     8352,
     UINT16_MAX,
     8091,
     UINT16_MAX,
     8091,
     2881,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8092,
     UINT16_MAX,
     8092,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     8354,
     UINT16_MAX,
     8093,
     UINT16_MAX,
     8093,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     8356,
     UINT16_MAX,
     8094,
     UINT16_MAX,
     8094,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     8358,
     UINT16_MAX,
     8095,
     UINT16_MAX,
     8095,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     8360,
     UINT16_MAX,
     8096,
     UINT16_MAX,
     8096,
     3510,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     8362,
     UINT16_MAX,
     8097,
     UINT16_MAX,
     8097,
     1606,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     8364,
     UINT16_MAX,
     8098,
     UINT16_MAX,
     8098,
     1594,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8099,
     UINT16_MAX,
     8099,
     1588,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     8366,
     UINT16_MAX,
     8100,
     UINT16_MAX,
     8100,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     8368,
     UINT16_MAX,
     8101,
     UINT16_MAX,
     8101,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     8370,
     UINT16_MAX,
     8102,
     UINT16_MAX,
     8102,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     8372,
     UINT16_MAX,
     8103,
     UINT16_MAX,
     8103,
     1523,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     8374,
     UINT16_MAX,
     8104,
     UINT16_MAX,
     8104,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8105,
     UINT16_MAX,
     8105,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     8376,
     UINT16_MAX,
     8106,
     UINT16_MAX,
     8106,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     8378,
     188,
     UINT16_MAX,
     188,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     8381,
     UINT16_MAX,
     8107,
     UINT16_MAX,
     8107,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     8383,
     193,
     UINT16_MAX,
     193,
     UINT16_MAX,
     3259,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     8386,
     UINT16_MAX,
     8108,
     UINT16_MAX,
     8108,
     3308,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     8388,
     198,
     UINT16_MAX,
     198,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     8391,
     UINT16_MAX,
     8109,
     UINT16_MAX,
     8109,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     8393,
     203,
     UINT16_MAX,
     203,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     8396,
     UINT16_MAX,
     8110,
     UINT16_MAX,
     8110,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     8398,
     208,
     UINT16_MAX,
     208,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     8401,
     UINT16_MAX,
     8111,
     UINT16_MAX,
     8111,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     8403,
     213,
     UINT16_MAX,
     213,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     8406,
     UINT16_MAX,
     8112,
     UINT16_MAX,
     8112,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     8408,
     218,
     UINT16_MAX,
     218,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     8411,
     UINT16_MAX,
     8113,
     UINT16_MAX,
     8113,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     8413,
     223,
     UINT16_MAX,
     223,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     8416,
     UINT16_MAX,
     8114,
     UINT16_MAX,
     8114,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     226,
     UINT16_MAX,
     226,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8115,
     UINT16_MAX,
     8115,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     8419,
     229,
     UINT16_MAX,
     229,
     UINT16_MAX,
     2858,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     8422,
     UINT16_MAX,
     8116,
     UINT16_MAX,
     8116,
     2862,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     8424,
     234,
     UINT16_MAX,
     234,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     8427,
     UINT16_MAX,
     8117,
     UINT16_MAX,
     8117,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     8429,
     239,
     UINT16_MAX,
     239,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     8432,
     UINT16_MAX,
     8118,
     UINT16_MAX,
     8118,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     8434,
     244,
     UINT16_MAX,
     244,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     8437,
     UINT16_MAX,
     8119,
     UINT16_MAX,
     8119,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     8439,
     249,
     UINT16_MAX,
     249,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     8442,
     UINT16_MAX,
     8120,
     UINT16_MAX,
     8120,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     8444,
     254,
     UINT16_MAX,
     254,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     8447,
     UINT16_MAX,
     8121,
     UINT16_MAX,
     8121,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     8449,
     259,
     UINT16_MAX,
     259,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     8452,
     UINT16_MAX,
     8122,
     UINT16_MAX,
     8122,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     8454,
     264,
     UINT16_MAX,
     264,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     8457,
     UINT16_MAX,
     8123,
     UINT16_MAX,
     8123,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     8459,
     269,
     UINT16_MAX,
     269,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     8462,
     UINT16_MAX,
     8124,
     UINT16_MAX,
     8124,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     8464,
     274,
     UINT16_MAX,
     274,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     8467,
     UINT16_MAX,
     8125,
     UINT16_MAX,
     8125,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     277,
     UINT16_MAX,
     277,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     5167,
     UINT16_MAX,
     5167,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     8470,
     280,
     UINT16_MAX,
     280,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     8473,
     UINT16_MAX,
     8126,
     UINT16_MAX,
     8126,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     8475,
     285,
     UINT16_MAX,
     285,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     8478,
     UINT16_MAX,
     8127,
     UINT16_MAX,
     8127,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     8480,
     290,
     UINT16_MAX,
     290,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     8483,
     UINT16_MAX,
     8128,
     UINT16_MAX,
     8128,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     8485,
     295,
     UINT16_MAX,
     295,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     8488,
     UINT16_MAX,
     8129,
     UINT16_MAX,
     8129,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     8490,
     8492,
     UINT16_MAX,
     8,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     1499,
     UINT16_MAX,
     1499,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     8494,
     304,
     UINT16_MAX,
     304,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     8497,
     UINT16_MAX,
     8130,
     UINT16_MAX,
     8130,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     8499,
     309,
     UINT16_MAX,
     309,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     8502,
     UINT16_MAX,
     8131,
     UINT16_MAX,
     8131,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     8504,
     314,
     UINT16_MAX,
     314,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     8507,
     UINT16_MAX,
     8132,
     UINT16_MAX,
     8132,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     8509,
     319,
     UINT16_MAX,
     319,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     8512,
     UINT16_MAX,
     8133,
     UINT16_MAX,
     8133,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     8514,
     324,
     UINT16_MAX,
     324,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     8517,
     UINT16_MAX,
     8134,
     UINT16_MAX,
     8134,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     8519,
     329,
     UINT16_MAX,
     329,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     8522,
     UINT16_MAX,
     8135,
     UINT16_MAX,
     8135,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     8524,
     334,
     UINT16_MAX,
     334,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     8527,
     UINT16_MAX,
     8136,
     UINT16_MAX,
     8136,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     337,
     UINT16_MAX,
     337,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8137,
     UINT16_MAX,
     8137,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     8530,
     340,
     UINT16_MAX,
     340,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     8533,
     UINT16_MAX,
     8138,
     UINT16_MAX,
     8138,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     8535,
     345,
     UINT16_MAX,
     345,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     8538,
     UINT16_MAX,
     8139,
     UINT16_MAX,
     8139,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     8540,
     350,
     UINT16_MAX,
     350,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     8543,
     UINT16_MAX,
     8140,
     UINT16_MAX,
     8140,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     8545,
     8545,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     355,
     UINT16_MAX,
     355,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8141,
     UINT16_MAX,
     8141,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     8548,
     358,
     UINT16_MAX,
     358,
     UINT16_MAX,
     2974,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     8551,
     UINT16_MAX,
     8142,
     UINT16_MAX,
     8142,
     2978,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     8553,
     363,
     UINT16_MAX,
     363,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     8556,
     UINT16_MAX,
     8143,
     UINT16_MAX,
     8143,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     8558,
     368,
     UINT16_MAX,
     368,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     8561,
     UINT16_MAX,
     8144,
     UINT16_MAX,
     8144,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     371,
     UINT16_MAX,
     371,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8145,
     UINT16_MAX,
     8145,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     8564,
     374,
     UINT16_MAX,
     374,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     8567,
     UINT16_MAX,
     8146,
     UINT16_MAX,
     8146,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     8569,
     379,
     UINT16_MAX,
     379,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     8572,
     UINT16_MAX,
     8147,
     UINT16_MAX,
     8147,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     8574,
     384,
     UINT16_MAX,
     384,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     8577,
     UINT16_MAX,
     8148,
     UINT16_MAX,
     8148,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     8579,
     389,
     UINT16_MAX,
     389,
     UINT16_MAX,
     3012,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     8582,
     UINT16_MAX,
     8149,
     UINT16_MAX,
     8149,
     3015,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     8584,
     394,
     UINT16_MAX,
     394,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     8587,
     UINT16_MAX,
     8150,
     UINT16_MAX,
     8150,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     8589,
     399,
     UINT16_MAX,
     399,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     8592,
     UINT16_MAX,
     8151,
     UINT16_MAX,
     8151,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     8594,
     404,
     UINT16_MAX,
     404,
     UINT16_MAX,
     3018,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     8597,
     UINT16_MAX,
     8152,
     UINT16_MAX,
     8152,
     3021,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     8599,
     409,
     UINT16_MAX,
     409,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     8602,
     UINT16_MAX,
     8153,
     UINT16_MAX,
     8153,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     8604,
     414,
     UINT16_MAX,
     414,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     8607,
     UINT16_MAX,
     8154,
     UINT16_MAX,
     8154,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     417,
     UINT16_MAX,
     417,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8155,
     UINT16_MAX,
     8155,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     8610,
     420,
     UINT16_MAX,
     420,
     UINT16_MAX,
     3030,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     8613,
     UINT16_MAX,
     8156,
     UINT16_MAX,
     8156,
     3033,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     8615,
     425,
     UINT16_MAX,
     425,
     UINT16_MAX,
     3036,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     8618,
     UINT16_MAX,
     8157,
     UINT16_MAX,
     8157,
     3039,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     8620,
     430,
     UINT16_MAX,
     430,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     8623,
     UINT16_MAX,
     8158,
     UINT16_MAX,
     8158,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     8625,
     435,
     UINT16_MAX,
     435,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     8628,
     UINT16_MAX,
     8159,
     UINT16_MAX,
     8159,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     8630,
     440,
     UINT16_MAX,
     440,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     8633,
     UINT16_MAX,
     8160,
     UINT16_MAX,
     8160,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     8635,
     445,
     UINT16_MAX,
     445,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     8638,
     UINT16_MAX,
     8161,
     UINT16_MAX,
     8161,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     8640,
     450,
     UINT16_MAX,
     450,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     8643,
     UINT16_MAX,
     8162,
     UINT16_MAX,
     8162,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     8645,
     455,
     UINT16_MAX,
     455,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     8648,
     UINT16_MAX,
     8163,
     UINT16_MAX,
     8163,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     8650,
     460,
     UINT16_MAX,
     460,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     8653,
     463,
     UINT16_MAX,
     463,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     8656,
     UINT16_MAX,
     8164,
     UINT16_MAX,
     8164,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     8658,
     468,
     UINT16_MAX,
     468,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     8661,
     UINT16_MAX,
     8165,
     UINT16_MAX,
     8165,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     8663,
     473,
     UINT16_MAX,
     473,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     8666,
     UINT16_MAX,
     8166,
     UINT16_MAX,
     8166,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     18,
     18,
     3305,
     UINT16_MAX,
     3305,
     3140,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8167,
     UINT16_MAX,
     8167,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     476,
     UINT16_MAX,
     476,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     477,
     UINT16_MAX,
     477,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8168,
     UINT16_MAX,
     8168,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     478,
     UINT16_MAX,
     478,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8169,
     UINT16_MAX,
     8169,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     479,
     UINT16_MAX,
     479,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     480,
     UINT16_MAX,
     480,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8170,
     UINT16_MAX,
     8170,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     481,
     UINT16_MAX,
     481,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     482,
     UINT16_MAX,
     482,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     483,
     UINT16_MAX,
     483,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8171,
     UINT16_MAX,
     8171,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     484,
     UINT16_MAX,
     484,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     485,
     UINT16_MAX,
     485,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     486,
     UINT16_MAX,
     486,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     487,
     UINT16_MAX,
     487,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8172,
     UINT16_MAX,
     8172,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     488,
     UINT16_MAX,
     488,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     489,
     UINT16_MAX,
     489,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8173,
     UINT16_MAX,
     8173,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     490,
     UINT16_MAX,
     490,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     491,
     UINT16_MAX,
     491,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     492,
     UINT16_MAX,
     492,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8174,
     UINT16_MAX,
     8174,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8175,
     UINT16_MAX,
     8175,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     493,
     UINT16_MAX,
     493,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     494,
     UINT16_MAX,
     494,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8176,
     UINT16_MAX,
     8176,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     495,
     UINT16_MAX,
     495,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     8688,
     498,
     UINT16_MAX,
     498,
     UINT16_MAX,
     3565,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     8691,
     UINT16_MAX,
     8177,
     UINT16_MAX,
     8177,
     3614,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     501,
     UINT16_MAX,
     501,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8178,
     UINT16_MAX,
     8178,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     502,
     UINT16_MAX,
     502,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8179,
     UINT16_MAX,
     8179,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     503,
     UINT16_MAX,
     503,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     504,
     UINT16_MAX,
     504,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8180,
     UINT16_MAX,
     8180,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     505,
     UINT16_MAX,
     505,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     506,
     UINT16_MAX,
     506,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8181,
     UINT16_MAX,
     8181,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     507,
     UINT16_MAX,
     507,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     8700,
     510,
     UINT16_MAX,
     510,
     UINT16_MAX,
     3663,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     8703,
     UINT16_MAX,
     8182,
     UINT16_MAX,
     8182,
     3712,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     513,
     UINT16_MAX,
     513,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     514,
     UINT16_MAX,
     514,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     515,
     UINT16_MAX,
     515,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8183,
     UINT16_MAX,
     8183,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     516,
     UINT16_MAX,
     516,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8184,
     UINT16_MAX,
     8184,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     517,
     UINT16_MAX,
     517,
     UINT16_MAX,
     1573,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     518,
     UINT16_MAX,
     518,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8185,
     UINT16_MAX,
     8185,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     519,
     UINT16_MAX,
     519,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8186,
     UINT16_MAX,
     8186,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8187,
     UINT16_MAX,
     8187,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     8712,
     522,
     UINT16_MAX,
     522,
     8188,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LT,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     8715,
     522,
     8189,
     522,
     8188,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     8717,
     UINT16_MAX,
     8189,
     UINT16_MAX,
     8188,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     8719,
     529,
     UINT16_MAX,
     529,
     8190,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LT,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     8722,
     529,
     8191,
     529,
     8190,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     8724,
     UINT16_MAX,
     8191,
     UINT16_MAX,
     8190,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     8726,
     536,
     UINT16_MAX,
     536,
     8192,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LT,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     8729,
     536,
     8193,
     536,
     8192,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     8731,
     UINT16_MAX,
     8193,
     UINT16_MAX,
     8192,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     8733,
     543,
     UINT16_MAX,
     543,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     8736,
     UINT16_MAX,
     8194,
     UINT16_MAX,
     8194,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     8738,
     548,
     UINT16_MAX,
     548,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     8741,
     UINT16_MAX,
     8195,
     UINT16_MAX,
     8195,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     8743,
     553,
     UINT16_MAX,
     553,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     8746,
     UINT16_MAX,
     8196,
     UINT16_MAX,
     8196,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     8748,
     558,
     UINT16_MAX,
     558,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     8751,
     UINT16_MAX,
     8197,
     UINT16_MAX,
     8197,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     8753,
     563,
     UINT16_MAX,
     563,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     8756,
     UINT16_MAX,
     8198,
     UINT16_MAX,
     8198,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     8758,
     568,
     UINT16_MAX,
     568,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     8761,
     UINT16_MAX,
     8199,
     UINT16_MAX,
     8199,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     8763,
     573,
     UINT16_MAX,
     573,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     8766,
     UINT16_MAX,
     8200,
     UINT16_MAX,
     8200,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     8768,
     578,
     UINT16_MAX,
     578,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     8771,
     UINT16_MAX,
     8201,
     UINT16_MAX,
     8201,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     1496,
     UINT16_MAX,
     1496,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     8773,
     583,
     UINT16_MAX,
     583,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     8776,
     UINT16_MAX,
     8202,
     UINT16_MAX,
     8202,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     8778,
     588,
     UINT16_MAX,
     588,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     8781,
     UINT16_MAX,
     8203,
     UINT16_MAX,
     8203,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     8783,
     593,
     UINT16_MAX,
     593,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     8786,
     UINT16_MAX,
     8204,
     UINT16_MAX,
     8204,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     596,
     UINT16_MAX,
     596,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8205,
     UINT16_MAX,
     8205,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     8789,
     599,
     UINT16_MAX,
     599,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     8792,
     UINT16_MAX,
     8206,
     UINT16_MAX,
     8206,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     8794,
     604,
     UINT16_MAX,
     604,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     8797,
     UINT16_MAX,
     8207,
     UINT16_MAX,
     8207,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     8799,
     609,
     UINT16_MAX,
     609,
     UINT16_MAX,
     1567,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     8802,
     UINT16_MAX,
     8208,
     UINT16_MAX,
     8208,
     1570,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     8804,
     614,
     UINT16_MAX,
     614,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     8807,
     UINT16_MAX,
     8209,
     UINT16_MAX,
     8209,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     8809,
     619,
     UINT16_MAX,
     619,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     8812,
     UINT16_MAX,
     8210,
     UINT16_MAX,
     8210,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     8814,
     8814,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     8816,
     626,
     UINT16_MAX,
     626,
     8211,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LT,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     8819,
     626,
     8212,
     626,
     8211,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     8821,
     UINT16_MAX,
     8212,
     UINT16_MAX,
     8211,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     8823,
     633,
     UINT16_MAX,
     633,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     8826,
     UINT16_MAX,
     8213,
     UINT16_MAX,
     8213,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     636,
     UINT16_MAX,
     636,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     637,
     UINT16_MAX,
     637,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     8830,
     640,
     UINT16_MAX,
     640,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     8833,
     UINT16_MAX,
     8214,
     UINT16_MAX,
     8214,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     8835,
     645,
     UINT16_MAX,
     645,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     8838,
     UINT16_MAX,
     8215,
     UINT16_MAX,
     8215,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     8840,
     650,
     UINT16_MAX,
     650,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     8843,
     UINT16_MAX,
     8216,
     UINT16_MAX,
     8216,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     8845,
     655,
     UINT16_MAX,
     655,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     8848,
     UINT16_MAX,
     8217,
     UINT16_MAX,
     8217,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     8850,
     660,
     UINT16_MAX,
     660,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     8853,
     UINT16_MAX,
     8218,
     UINT16_MAX,
     8218,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     8855,
     665,
     UINT16_MAX,
     665,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     8858,
     UINT16_MAX,
     8219,
     UINT16_MAX,
     8219,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     8860,
     670,
     UINT16_MAX,
     670,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     8863,
     UINT16_MAX,
     8220,
     UINT16_MAX,
     8220,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     8865,
     675,
     UINT16_MAX,
     675,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     8868,
     UINT16_MAX,
     8221,
     UINT16_MAX,
     8221,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     8870,
     680,
     UINT16_MAX,
     680,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     8873,
     UINT16_MAX,
     8222,
     UINT16_MAX,
     8222,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     8875,
     685,
     UINT16_MAX,
     685,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     8878,
     UINT16_MAX,
     8223,
     UINT16_MAX,
     8223,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     8880,
     690,
     UINT16_MAX,
     690,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     8883,
     UINT16_MAX,
     8224,
     UINT16_MAX,
     8224,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     8885,
     695,
     UINT16_MAX,
     695,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     8888,
     UINT16_MAX,
     8225,
     UINT16_MAX,
     8225,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     8890,
     700,
     UINT16_MAX,
     700,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     8893,
     UINT16_MAX,
     8226,
     UINT16_MAX,
     8226,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     8895,
     705,
     UINT16_MAX,
     705,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     8898,
     UINT16_MAX,
     8227,
     UINT16_MAX,
     8227,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     8900,
     710,
     UINT16_MAX,
     710,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     8903,
     UINT16_MAX,
     8228,
     UINT16_MAX,
     8228,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     8905,
     715,
     UINT16_MAX,
     715,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     8908,
     UINT16_MAX,
     8229,
     UINT16_MAX,
     8229,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     8910,
     720,
     UINT16_MAX,
     720,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     8913,
     UINT16_MAX,
     8230,
     UINT16_MAX,
     8230,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     8915,
     725,
     UINT16_MAX,
     725,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     8918,
     UINT16_MAX,
     8231,
     UINT16_MAX,
     8231,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     728,
     UINT16_MAX,
     728,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8232,
     UINT16_MAX,
     8232,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     8921,
     731,
     UINT16_MAX,
     731,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     8924,
     UINT16_MAX,
     8233,
     UINT16_MAX,
     8233,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     734,
     UINT16_MAX,
     734,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     735,
     UINT16_MAX,
     735,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     1506,
     UINT16_MAX,
     1506,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     736,
     UINT16_MAX,
     736,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8234,
     UINT16_MAX,
     8234,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     8929,
     739,
     UINT16_MAX,
     739,
     UINT16_MAX,
     1543,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     8932,
     UINT16_MAX,
     8235,
     UINT16_MAX,
     8235,
     1546,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     8934,
     744,
     UINT16_MAX,
     744,
     UINT16_MAX,
     2866,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     8937,
     UINT16_MAX,
     8236,
     UINT16_MAX,
     8236,
     2869,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     8939,
     749,
     UINT16_MAX,
     749,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     8942,
     UINT16_MAX,
     8237,
     UINT16_MAX,
     8237,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     8944,
     754,
     UINT16_MAX,
     754,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     8947,
     UINT16_MAX,
     8238,
     UINT16_MAX,
     8238,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     8949,
     759,
     UINT16_MAX,
     759,
     UINT16_MAX,
     1615,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     8952,
     UINT16_MAX,
     8239,
     UINT16_MAX,
     8239,
     1618,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     8954,
     764,
     UINT16_MAX,
     764,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     8957,
     UINT16_MAX,
     8240,
     UINT16_MAX,
     8240,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     8959,
     769,
     UINT16_MAX,
     769,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     8962,
     UINT16_MAX,
     8241,
     UINT16_MAX,
     8241,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     772,
     UINT16_MAX,
     772,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     773,
     UINT16_MAX,
     773,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8242,
     UINT16_MAX,
     8242,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     774,
     UINT16_MAX,
     774,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     775,
     UINT16_MAX,
     775,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8243,
     UINT16_MAX,
     8243,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8244,
     UINT16_MAX,
     8244,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     776,
     UINT16_MAX,
     776,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8245,
     UINT16_MAX,
     8245,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     777,
     UINT16_MAX,
     777,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     778,
     UINT16_MAX,
     778,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     779,
     UINT16_MAX,
     779,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     780,
     UINT16_MAX,
     780,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8246,
     UINT16_MAX,
     8246,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     781,
     UINT16_MAX,
     781,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8247,
     UINT16_MAX,
     8247,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     782,
     UINT16_MAX,
     782,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8248,
     UINT16_MAX,
     8248,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     783,
     UINT16_MAX,
     783,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8249,
     UINT16_MAX,
     8249,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     784,
     UINT16_MAX,
     784,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8250,
     UINT16_MAX,
     8250,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8251,
     UINT16_MAX,
     8251,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8252,
     UINT16_MAX,
     8252,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8253,
     UINT16_MAX,
     8253,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8254,
     UINT16_MAX,
     8254,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8255,
     UINT16_MAX,
     8255,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8256,
     UINT16_MAX,
     8256,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8257,
     UINT16_MAX,
     8257,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8258,
     UINT16_MAX,
     8258,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     2821,
     UINT16_MAX,
     2821,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8259,
     UINT16_MAX,
     8259,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8260,
     UINT16_MAX,
     8260,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8261,
     UINT16_MAX,
     8261,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8262,
     UINT16_MAX,
     8262,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8263,
     UINT16_MAX,
     8263,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8264,
     UINT16_MAX,
     8264,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8265,
     UINT16_MAX,
     8265,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8266,
     UINT16_MAX,
     8266,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8267,
     UINT16_MAX,
     8267,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8268,
     UINT16_MAX,
     8268,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8269,
     UINT16_MAX,
     8269,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8270,
     UINT16_MAX,
     8270,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8271,
     UINT16_MAX,
     8271,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8272,
     UINT16_MAX,
     8272,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8273,
     UINT16_MAX,
     8273,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8274,
     UINT16_MAX,
     8274,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8275,
     UINT16_MAX,
     8275,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8276,
     UINT16_MAX,
     8276,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8277,
     UINT16_MAX,
     8277,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8278,
     UINT16_MAX,
     8278,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8279,
     UINT16_MAX,
     8279,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8280,
     UINT16_MAX,
     8280,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8281,
     UINT16_MAX,
     8281,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8282,
     UINT16_MAX,
     8282,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8283,
     UINT16_MAX,
     8283,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8284,
     UINT16_MAX,
     8284,
     1576,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8285,
     UINT16_MAX,
     8285,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8286,
     UINT16_MAX,
     8286,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LM,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SUPER,
     7,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LM,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SUPER,
     785,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LM,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SUPER,
     9,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LM,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SUPER,
     17,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LM,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SUPER,
     786,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LM,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SUPER,
     787,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LM,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SUPER,
     788,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LM,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SUPER,
     22,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LM,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SUPER,
     24,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LM,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LM,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SK,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     8981,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SK,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     8983,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SK,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     8985,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SK,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     8987,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SK,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     8989,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SK,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     8991,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LM,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SUPER,
     489,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LM,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SUPER,
     11,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LM,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SUPER,
     18,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LM,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SUPER,
     23,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LM,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SUPER,
     801,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_MN,
     230,
     UTF8PROC_BIDI_CLASS_NSM,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     32768,
     false,
     false,
     false,
     false,
     0,
     0,
     UTF8PROC_BOUNDCLASS_EXTEND},
    {UTF8PROC_CATEGORY_MN,
     230,
     UTF8PROC_BIDI_CLASS_NSM,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     32769,
     false,
     false,
     false,
     false,
     0,
     0,
     UTF8PROC_BOUNDCLASS_EXTEND},
    {UTF8PROC_CATEGORY_MN,
     230,
     UTF8PROC_BIDI_CLASS_NSM,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     32770,
     false,
     false,
     false,
     false,
     0,
     0,
     UTF8PROC_BOUNDCLASS_EXTEND},
    {UTF8PROC_CATEGORY_MN,
     230,
     UTF8PROC_BIDI_CLASS_NSM,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     32771,
     false,
     false,
     false,
     false,
     0,
     0,
     UTF8PROC_BOUNDCLASS_EXTEND},
    {UTF8PROC_CATEGORY_MN,
     230,
     UTF8PROC_BIDI_CLASS_NSM,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     32775,
     false,
     false,
     false,
     false,
     0,
     0,
     UTF8PROC_BOUNDCLASS_EXTEND},
    {UTF8PROC_CATEGORY_MN,
     230,
     UTF8PROC_BIDI_CLASS_NSM,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     0,
     0,
     UTF8PROC_BOUNDCLASS_EXTEND},
    {UTF8PROC_CATEGORY_MN,
     230,
     UTF8PROC_BIDI_CLASS_NSM,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     32776,
     false,
     false,
     false,
     false,
     0,
     0,
     UTF8PROC_BOUNDCLASS_EXTEND},
    {UTF8PROC_CATEGORY_MN,
     230,
     UTF8PROC_BIDI_CLASS_NSM,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     32778,
     false,
     false,
     false,
     false,
     0,
     0,
     UTF8PROC_BOUNDCLASS_EXTEND},
    {UTF8PROC_CATEGORY_MN,
     230,
     UTF8PROC_BIDI_CLASS_NSM,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     32772,
     false,
     false,
     false,
     false,
     0,
     0,
     UTF8PROC_BOUNDCLASS_EXTEND},
    {UTF8PROC_CATEGORY_MN,
     230,
     UTF8PROC_BIDI_CLASS_NSM,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     32814,
     false,
     false,
     false,
     false,
     0,
     0,
     UTF8PROC_BOUNDCLASS_EXTEND},
    {UTF8PROC_CATEGORY_MN,
     230,
     UTF8PROC_BIDI_CLASS_NSM,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     32773,
     false,
     false,
     false,
     false,
     0,
     0,
     UTF8PROC_BOUNDCLASS_EXTEND},
    {UTF8PROC_CATEGORY_MN,
     230,
     UTF8PROC_BIDI_CLASS_NSM,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     32780,
     false,
     false,
     false,
     false,
     0,
     0,
     UTF8PROC_BOUNDCLASS_EXTEND},
    {UTF8PROC_CATEGORY_MN,
     230,
     UTF8PROC_BIDI_CLASS_NSM,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     32779,
     false,
     false,
     false,
     false,
     0,
     0,
     UTF8PROC_BOUNDCLASS_EXTEND},
    {UTF8PROC_CATEGORY_MN,
     230,
     UTF8PROC_BIDI_CLASS_NSM,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     32782,
     false,
     false,
     false,
     false,
     0,
     0,
     UTF8PROC_BOUNDCLASS_EXTEND},
    {UTF8PROC_CATEGORY_MN,
     230,
     UTF8PROC_BIDI_CLASS_NSM,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     32783,
     false,
     false,
     false,
     false,
     0,
     0,
     UTF8PROC_BOUNDCLASS_EXTEND},
    {UTF8PROC_CATEGORY_MN,
     230,
     UTF8PROC_BIDI_CLASS_NSM,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     32815,
     false,
     false,
     false,
     false,
     0,
     0,
     UTF8PROC_BOUNDCLASS_EXTEND},
    {UTF8PROC_CATEGORY_MN,
     230,
     UTF8PROC_BIDI_CLASS_NSM,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     32816,
     false,
     false,
     false,
     false,
     0,
     0,
     UTF8PROC_BOUNDCLASS_EXTEND},
    {UTF8PROC_CATEGORY_MN,
     232,
     UTF8PROC_BIDI_CLASS_NSM,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     0,
     0,
     UTF8PROC_BOUNDCLASS_EXTEND},
    {UTF8PROC_CATEGORY_MN,
     220,
     UTF8PROC_BIDI_CLASS_NSM,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     0,
     0,
     UTF8PROC_BOUNDCLASS_EXTEND},
    {UTF8PROC_CATEGORY_MN,
     216,
     UTF8PROC_BIDI_CLASS_NSM,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     32781,
     false,
     false,
     false,
     false,
     0,
     0,
     UTF8PROC_BOUNDCLASS_EXTEND},
    {UTF8PROC_CATEGORY_MN,
     202,
     UTF8PROC_BIDI_CLASS_NSM,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     0,
     0,
     UTF8PROC_BOUNDCLASS_EXTEND},
    {UTF8PROC_CATEGORY_MN,
     220,
     UTF8PROC_BIDI_CLASS_NSM,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     32808,
     false,
     false,
     false,
     false,
     0,
     0,
     UTF8PROC_BOUNDCLASS_EXTEND},
    {UTF8PROC_CATEGORY_MN,
     220,
     UTF8PROC_BIDI_CLASS_NSM,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     32813,
     false,
     false,
     false,
     false,
     0,
     0,
     UTF8PROC_BOUNDCLASS_EXTEND},
    {UTF8PROC_CATEGORY_MN,
     220,
     UTF8PROC_BIDI_CLASS_NSM,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     32807,
     false,
     false,
     false,
     false,
     0,
     0,
     UTF8PROC_BOUNDCLASS_EXTEND},
    {UTF8PROC_CATEGORY_MN,
     220,
     UTF8PROC_BIDI_CLASS_NSM,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     32784,
     false,
     false,
     false,
     false,
     0,
     0,
     UTF8PROC_BOUNDCLASS_EXTEND},
    {UTF8PROC_CATEGORY_MN,
     202,
     UTF8PROC_BIDI_CLASS_NSM,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     32774,
     false,
     false,
     false,
     false,
     0,
     0,
     UTF8PROC_BOUNDCLASS_EXTEND},
    {UTF8PROC_CATEGORY_MN,
     202,
     UTF8PROC_BIDI_CLASS_NSM,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     32777,
     false,
     false,
     false,
     false,
     0,
     0,
     UTF8PROC_BOUNDCLASS_EXTEND},
    {UTF8PROC_CATEGORY_MN,
     220,
     UTF8PROC_BIDI_CLASS_NSM,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     32810,
     false,
     false,
     false,
     false,
     0,
     0,
     UTF8PROC_BOUNDCLASS_EXTEND},
    {UTF8PROC_CATEGORY_MN,
     220,
     UTF8PROC_BIDI_CLASS_NSM,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     32812,
     false,
     false,
     false,
     false,
     0,
     0,
     UTF8PROC_BOUNDCLASS_EXTEND},
    {UTF8PROC_CATEGORY_MN,
     220,
     UTF8PROC_BIDI_CLASS_NSM,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     32811,
     false,
     false,
     false,
     false,
     0,
     0,
     UTF8PROC_BOUNDCLASS_EXTEND},
    {UTF8PROC_CATEGORY_MN,
     220,
     UTF8PROC_BIDI_CLASS_NSM,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     32809,
     false,
     false,
     false,
     false,
     0,
     0,
     UTF8PROC_BOUNDCLASS_EXTEND},
    {UTF8PROC_CATEGORY_MN,
     1,
     UTF8PROC_BIDI_CLASS_NSM,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     0,
     0,
     UTF8PROC_BOUNDCLASS_EXTEND},
    {UTF8PROC_CATEGORY_MN,
     1,
     UTF8PROC_BIDI_CLASS_NSM,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     32819,
     false,
     false,
     false,
     false,
     0,
     0,
     UTF8PROC_BOUNDCLASS_EXTEND},
    {UTF8PROC_CATEGORY_MN,
     230,
     UTF8PROC_BIDI_CLASS_NSM,
     0,
     802,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     0,
     0,
     UTF8PROC_BOUNDCLASS_EXTEND},
    {UTF8PROC_CATEGORY_MN,
     230,
     UTF8PROC_BIDI_CLASS_NSM,
     0,
     803,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     0,
     0,
     UTF8PROC_BOUNDCLASS_EXTEND},
    {UTF8PROC_CATEGORY_MN,
     230,
     UTF8PROC_BIDI_CLASS_NSM,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     32817,
     false,
     false,
     false,
     false,
     0,
     0,
     UTF8PROC_BOUNDCLASS_EXTEND},
    {UTF8PROC_CATEGORY_MN,
     230,
     UTF8PROC_BIDI_CLASS_NSM,
     0,
     804,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     0,
     0,
     UTF8PROC_BOUNDCLASS_EXTEND},
    {UTF8PROC_CATEGORY_MN,
     230,
     UTF8PROC_BIDI_CLASS_NSM,
     0,
     8997,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     0,
     0,
     UTF8PROC_BOUNDCLASS_EXTEND},
    {UTF8PROC_CATEGORY_MN,
     240,
     UTF8PROC_BIDI_CLASS_NSM,
     0,
     UINT16_MAX,
     807,
     7225,
     UINT16_MAX,
     7225,
     32818,
     false,
     false,
     false,
     false,
     0,
     0,
     UTF8PROC_BOUNDCLASS_EXTEND},
    {UTF8PROC_CATEGORY_MN,
     0,
     UTF8PROC_BIDI_CLASS_NSM,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     true,
     false,
     0,
     0,
     UTF8PROC_BOUNDCLASS_EXTEND},
    {UTF8PROC_CATEGORY_MN,
     233,
     UTF8PROC_BIDI_CLASS_NSM,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     0,
     0,
     UTF8PROC_BOUNDCLASS_EXTEND},
    {UTF8PROC_CATEGORY_MN,
     234,
     UTF8PROC_BIDI_CLASS_NSM,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     0,
     0,
     UTF8PROC_BOUNDCLASS_EXTEND},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     808,
     UINT16_MAX,
     808,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8287,
     UINT16_MAX,
     8287,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     809,
     UINT16_MAX,
     809,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8288,
     UINT16_MAX,
     8288,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LM,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     0,
     810,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     811,
     UINT16_MAX,
     811,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8289,
     UINT16_MAX,
     8289,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LM,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     9004,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8290,
     UINT16_MAX,
     8290,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8291,
     UINT16_MAX,
     8291,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8292,
     UINT16_MAX,
     8292,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_PO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     0,
     814,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     815,
     UINT16_MAX,
     815,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SK,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     0,
     9008,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     9010,
     820,
     UINT16_MAX,
     820,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_PO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     0,
     821,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     9014,
     824,
     UINT16_MAX,
     824,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     9017,
     827,
     UINT16_MAX,
     827,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     9020,
     830,
     UINT16_MAX,
     830,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     9023,
     833,
     UINT16_MAX,
     833,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     9026,
     836,
     UINT16_MAX,
     836,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     9029,
     839,
     UINT16_MAX,
     839,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     9032,
     17226,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     845,
     UINT16_MAX,
     845,
     UINT16_MAX,
     1673,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     846,
     UINT16_MAX,
     846,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     847,
     UINT16_MAX,
     847,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     848,
     UINT16_MAX,
     848,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     849,
     UINT16_MAX,
     849,
     UINT16_MAX,
     1726,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     850,
     UINT16_MAX,
     850,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     851,
     UINT16_MAX,
     851,
     UINT16_MAX,
     1777,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     852,
     UINT16_MAX,
     852,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     807,
     UINT16_MAX,
     807,
     UINT16_MAX,
     1830,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     853,
     UINT16_MAX,
     853,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     854,
     UINT16_MAX,
     854,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     35,
     UINT16_MAX,
     35,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     855,
     UINT16_MAX,
     855,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     856,
     UINT16_MAX,
     856,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     857,
     UINT16_MAX,
     857,
     UINT16_MAX,
     1881,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     858,
     UINT16_MAX,
     858,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     859,
     UINT16_MAX,
     859,
     UINT16_MAX,
     5027,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     860,
     UINT16_MAX,
     860,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     861,
     UINT16_MAX,
     861,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     862,
     UINT16_MAX,
     862,
     UINT16_MAX,
     1932,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     863,
     UINT16_MAX,
     863,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     864,
     UINT16_MAX,
     864,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     865,
     UINT16_MAX,
     865,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     866,
     UINT16_MAX,
     866,
     UINT16_MAX,
     1983,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     9059,
     869,
     UINT16_MAX,
     869,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     9062,
     872,
     UINT16_MAX,
     872,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     9065,
     UINT16_MAX,
     2603,
     UINT16_MAX,
     2603,
     4904,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     9067,
     UINT16_MAX,
     2635,
     UINT16_MAX,
     2635,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     9069,
     UINT16_MAX,
     2640,
     UINT16_MAX,
     2640,
     4913,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     9071,
     UINT16_MAX,
     2676,
     UINT16_MAX,
     2676,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     9073,
     17267,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     7219,
     UINT16_MAX,
     7219,
     2088,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     7220,
     UINT16_MAX,
     7220,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     2846,
     UINT16_MAX,
     2846,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     7221,
     UINT16_MAX,
     7221,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     7222,
     UINT16_MAX,
     7222,
     2141,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     7223,
     UINT16_MAX,
     7223,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     7224,
     UINT16_MAX,
     7224,
     2192,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     915,
     UINT16_MAX,
     915,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     7225,
     UINT16_MAX,
     7225,
     2245,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     7226,
     UINT16_MAX,
     7226,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     7227,
     UINT16_MAX,
     7227,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     7228,
     UINT16_MAX,
     7228,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     7229,
     UINT16_MAX,
     7229,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     7230,
     UINT16_MAX,
     7230,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     7231,
     UINT16_MAX,
     7231,
     2401,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     2847,
     UINT16_MAX,
     2847,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     7232,
     UINT16_MAX,
     7232,
     5023,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     860,
     917,
     UINT16_MAX,
     917,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     917,
     UINT16_MAX,
     917,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     7234,
     UINT16_MAX,
     7234,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     897,
     UINT16_MAX,
     897,
     2349,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     7235,
     UINT16_MAX,
     7235,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     7236,
     UINT16_MAX,
     7236,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     7237,
     UINT16_MAX,
     7237,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     2835,
     UINT16_MAX,
     2835,
     2452,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     9078,
     UINT16_MAX,
     8293,
     UINT16_MAX,
     8293,
     2036,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     9080,
     UINT16_MAX,
     8294,
     UINT16_MAX,
     8294,
     2297,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     9082,
     UINT16_MAX,
     2745,
     UINT16_MAX,
     2745,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     9084,
     UINT16_MAX,
     2714,
     UINT16_MAX,
     2714,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     9086,
     UINT16_MAX,
     2750,
     UINT16_MAX,
     2750,
     5033,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     896,
     UINT16_MAX,
     896,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     846,
     846,
     7220,
     UINT16_MAX,
     7220,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     852,
     852,
     915,
     UINT16_MAX,
     915,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     897,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     2505,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     9090,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     9092,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     863,
     863,
     7235,
     UINT16_MAX,
     7235,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     858,
     858,
     2847,
     UINT16_MAX,
     2847,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8295,
     UINT16_MAX,
     8295,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     902,
     UINT16_MAX,
     902,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8296,
     UINT16_MAX,
     8296,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     903,
     UINT16_MAX,
     903,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8297,
     UINT16_MAX,
     8297,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     904,
     UINT16_MAX,
     904,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     7246,
     UINT16_MAX,
     7246,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     905,
     UINT16_MAX,
     905,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8298,
     UINT16_MAX,
     8298,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     906,
     UINT16_MAX,
     906,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8299,
     UINT16_MAX,
     8299,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     907,
     UINT16_MAX,
     907,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8300,
     UINT16_MAX,
     8300,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     908,
     UINT16_MAX,
     908,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8301,
     UINT16_MAX,
     8301,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     909,
     UINT16_MAX,
     909,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8302,
     UINT16_MAX,
     8302,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     910,
     UINT16_MAX,
     910,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8303,
     UINT16_MAX,
     8303,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     911,
     UINT16_MAX,
     911,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8304,
     UINT16_MAX,
     8304,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     912,
     UINT16_MAX,
     912,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8305,
     UINT16_MAX,
     8305,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     913,
     UINT16_MAX,
     913,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8306,
     UINT16_MAX,
     8306,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     853,
     853,
     7226,
     UINT16_MAX,
     7226,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     859,
     859,
     7232,
     UINT16_MAX,
     7232,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     914,
     UINT16_MAX,
     8307,
     UINT16_MAX,
     8307,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8308,
     UINT16_MAX,
     8308,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     915,
     852,
     UINT16_MAX,
     852,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     849,
     849,
     7222,
     UINT16_MAX,
     7222,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     916,
     UINT16_MAX,
     916,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8309,
     UINT16_MAX,
     8309,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     917,
     918,
     UINT16_MAX,
     918,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     919,
     UINT16_MAX,
     919,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8310,
     UINT16_MAX,
     8310,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     920,
     UINT16_MAX,
     920,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     921,
     UINT16_MAX,
     921,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     922,
     UINT16_MAX,
     922,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     9115,
     925,
     UINT16_MAX,
     925,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     9118,
     928,
     UINT16_MAX,
     928,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     929,
     UINT16_MAX,
     929,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     9122,
     932,
     UINT16_MAX,
     932,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     933,
     UINT16_MAX,
     933,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     934,
     UINT16_MAX,
     934,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     935,
     UINT16_MAX,
     935,
     UINT16_MAX,
     2525,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     9128,
     938,
     UINT16_MAX,
     938,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     939,
     UINT16_MAX,
     939,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     940,
     UINT16_MAX,
     940,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     941,
     UINT16_MAX,
     941,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     942,
     UINT16_MAX,
     942,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     9135,
     945,
     UINT16_MAX,
     945,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     9138,
     948,
     UINT16_MAX,
     948,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     9141,
     951,
     UINT16_MAX,
     951,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     952,
     UINT16_MAX,
     952,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     953,
     UINT16_MAX,
     953,
     UINT16_MAX,
     2615,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     954,
     UINT16_MAX,
     954,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     955,
     UINT16_MAX,
     955,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     956,
     UINT16_MAX,
     956,
     UINT16_MAX,
     2522,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     957,
     UINT16_MAX,
     957,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     958,
     UINT16_MAX,
     958,
     UINT16_MAX,
     2511,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     959,
     UINT16_MAX,
     959,
     UINT16_MAX,
     2601,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     960,
     UINT16_MAX,
     960,
     UINT16_MAX,
     2635,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     961,
     UINT16_MAX,
     961,
     UINT16_MAX,
     2531,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     9154,
     964,
     UINT16_MAX,
     964,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     965,
     UINT16_MAX,
     965,
     UINT16_MAX,
     2528,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     966,
     UINT16_MAX,
     966,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     967,
     UINT16_MAX,
     967,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     968,
     UINT16_MAX,
     968,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     969,
     UINT16_MAX,
     969,
     UINT16_MAX,
     2641,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     970,
     UINT16_MAX,
     970,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     971,
     UINT16_MAX,
     971,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     972,
     UINT16_MAX,
     972,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     973,
     UINT16_MAX,
     973,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     974,
     UINT16_MAX,
     974,
     UINT16_MAX,
     2542,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     975,
     UINT16_MAX,
     975,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     976,
     UINT16_MAX,
     976,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     977,
     UINT16_MAX,
     977,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     978,
     UINT16_MAX,
     978,
     UINT16_MAX,
     2659,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     979,
     UINT16_MAX,
     979,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     980,
     UINT16_MAX,
     980,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     981,
     UINT16_MAX,
     981,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     982,
     UINT16_MAX,
     982,
     UINT16_MAX,
     2665,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     983,
     UINT16_MAX,
     983,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     984,
     UINT16_MAX,
     984,
     UINT16_MAX,
     2653,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     985,
     UINT16_MAX,
     985,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     986,
     UINT16_MAX,
     986,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8311,
     UINT16_MAX,
     8311,
     2622,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8312,
     UINT16_MAX,
     8312,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8313,
     UINT16_MAX,
     8313,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8314,
     UINT16_MAX,
     8314,
     2575,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8315,
     UINT16_MAX,
     8315,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8316,
     UINT16_MAX,
     8316,
     2564,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8317,
     UINT16_MAX,
     8317,
     2608,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8318,
     UINT16_MAX,
     8318,
     2638,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8319,
     UINT16_MAX,
     8319,
     2553,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     9179,
     UINT16_MAX,
     8320,
     UINT16_MAX,
     8320,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8321,
     UINT16_MAX,
     8321,
     2581,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8322,
     UINT16_MAX,
     8322,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8323,
     UINT16_MAX,
     8323,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8324,
     UINT16_MAX,
     8324,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8325,
     UINT16_MAX,
     8325,
     2644,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8326,
     UINT16_MAX,
     8326,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8327,
     UINT16_MAX,
     8327,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8328,
     UINT16_MAX,
     8328,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8329,
     UINT16_MAX,
     8329,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8330,
     UINT16_MAX,
     8330,
     2584,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8331,
     UINT16_MAX,
     8331,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8332,
     UINT16_MAX,
     8332,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8333,
     UINT16_MAX,
     8333,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8334,
     UINT16_MAX,
     8334,
     2662,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8335,
     UINT16_MAX,
     8335,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8336,
     UINT16_MAX,
     8336,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8337,
     UINT16_MAX,
     8337,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8338,
     UINT16_MAX,
     8338,
     2668,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8339,
     UINT16_MAX,
     8339,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8340,
     UINT16_MAX,
     8340,
     2656,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8341,
     UINT16_MAX,
     8341,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8342,
     UINT16_MAX,
     8342,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     9181,
     UINT16_MAX,
     8343,
     UINT16_MAX,
     8343,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     9183,
     UINT16_MAX,
     8344,
     UINT16_MAX,
     8344,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8345,
     UINT16_MAX,
     8345,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     9185,
     UINT16_MAX,
     8346,
     UINT16_MAX,
     8346,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8347,
     UINT16_MAX,
     8347,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8348,
     UINT16_MAX,
     8348,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8349,
     UINT16_MAX,
     8349,
     2578,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     9187,
     UINT16_MAX,
     8350,
     UINT16_MAX,
     8350,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8351,
     UINT16_MAX,
     8351,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8352,
     UINT16_MAX,
     8352,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8353,
     UINT16_MAX,
     8353,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8354,
     UINT16_MAX,
     8354,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     9189,
     UINT16_MAX,
     8355,
     UINT16_MAX,
     8355,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     9191,
     UINT16_MAX,
     8356,
     UINT16_MAX,
     8356,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     9193,
     UINT16_MAX,
     8357,
     UINT16_MAX,
     8357,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8358,
     UINT16_MAX,
     8358,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     1003,
     UINT16_MAX,
     1003,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8359,
     UINT16_MAX,
     8359,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     1004,
     UINT16_MAX,
     1004,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8360,
     UINT16_MAX,
     8360,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     1005,
     UINT16_MAX,
     1005,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8361,
     UINT16_MAX,
     8361,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     1006,
     UINT16_MAX,
     1006,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8362,
     UINT16_MAX,
     8362,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     1007,
     UINT16_MAX,
     1007,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8363,
     UINT16_MAX,
     8363,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     1008,
     UINT16_MAX,
     1008,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8364,
     UINT16_MAX,
     8364,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     1009,
     UINT16_MAX,
     1009,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8365,
     UINT16_MAX,
     8365,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     1010,
     UINT16_MAX,
     1010,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8366,
     UINT16_MAX,
     8366,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     1011,
     UINT16_MAX,
     1011,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8367,
     UINT16_MAX,
     8367,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     1012,
     UINT16_MAX,
     1012,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8368,
     UINT16_MAX,
     8368,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     1013,
     UINT16_MAX,
     1013,
     UINT16_MAX,
     2595,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8369,
     UINT16_MAX,
     8369,
     2598,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     9206,
     1016,
     UINT16_MAX,
     1016,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     9209,
     UINT16_MAX,
     8370,
     UINT16_MAX,
     8370,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     1019,
     UINT16_MAX,
     1019,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8371,
     UINT16_MAX,
     8371,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     1020,
     UINT16_MAX,
     1020,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8372,
     UINT16_MAX,
     8372,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     1021,
     UINT16_MAX,
     1021,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8373,
     UINT16_MAX,
     8373,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     1022,
     UINT16_MAX,
     1022,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8374,
     UINT16_MAX,
     8374,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     1023,
     UINT16_MAX,
     1023,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8375,
     UINT16_MAX,
     8375,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_ME,
     0,
     UTF8PROC_BIDI_CLASS_NSM,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     0,
     0,
     UTF8PROC_BOUNDCLASS_EXTEND},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     1024,
     UINT16_MAX,
     1024,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8376,
     UINT16_MAX,
     8376,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     1025,
     UINT16_MAX,
     1025,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8377,
     UINT16_MAX,
     8377,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     1026,
     UINT16_MAX,
     1026,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8378,
     UINT16_MAX,
     8378,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     1027,
     UINT16_MAX,
     1027,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8379,
     UINT16_MAX,
     8379,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     1028,
     UINT16_MAX,
     1028,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8380,
     UINT16_MAX,
     8380,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     1029,
     UINT16_MAX,
     1029,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8381,
     UINT16_MAX,
     8381,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     1030,
     UINT16_MAX,
     1030,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8382,
     UINT16_MAX,
     8382,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     1031,
     UINT16_MAX,
     1031,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8383,
     UINT16_MAX,
     8383,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     1032,
     UINT16_MAX,
     1032,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8384,
     UINT16_MAX,
     8384,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     1033,
     UINT16_MAX,
     1033,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8385,
     UINT16_MAX,
     8385,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     1034,
     UINT16_MAX,
     1034,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8386,
     UINT16_MAX,
     8386,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     1035,
     UINT16_MAX,
     1035,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8387,
     UINT16_MAX,
     8387,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     1036,
     UINT16_MAX,
     1036,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8388,
     UINT16_MAX,
     8388,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     1037,
     UINT16_MAX,
     1037,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8389,
     UINT16_MAX,
     8389,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     1038,
     UINT16_MAX,
     1038,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8390,
     UINT16_MAX,
     8390,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     1039,
     UINT16_MAX,
     1039,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8391,
     UINT16_MAX,
     8391,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     1040,
     UINT16_MAX,
     1040,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8392,
     UINT16_MAX,
     8392,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     1041,
     UINT16_MAX,
     1041,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8393,
     UINT16_MAX,
     8393,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     1042,
     UINT16_MAX,
     1042,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8394,
     UINT16_MAX,
     8394,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     1043,
     UINT16_MAX,
     1043,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8395,
     UINT16_MAX,
     8395,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     1044,
     UINT16_MAX,
     1044,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8396,
     UINT16_MAX,
     8396,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     1045,
     UINT16_MAX,
     1045,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8397,
     UINT16_MAX,
     8397,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     1046,
     UINT16_MAX,
     1046,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8398,
     UINT16_MAX,
     8398,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     1047,
     UINT16_MAX,
     1047,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8399,
     UINT16_MAX,
     8399,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     1048,
     UINT16_MAX,
     1048,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8400,
     UINT16_MAX,
     8400,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     1049,
     UINT16_MAX,
     1049,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8401,
     UINT16_MAX,
     8401,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     1050,
     UINT16_MAX,
     1050,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8402,
     UINT16_MAX,
     8402,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     1051,
     UINT16_MAX,
     1051,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     9244,
     1054,
     UINT16_MAX,
     1054,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     9247,
     UINT16_MAX,
     8403,
     UINT16_MAX,
     8403,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     1057,
     UINT16_MAX,
     1057,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8404,
     UINT16_MAX,
     8404,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     1058,
     UINT16_MAX,
     1058,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8405,
     UINT16_MAX,
     8405,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     1059,
     UINT16_MAX,
     1059,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8406,
     UINT16_MAX,
     8406,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     1060,
     UINT16_MAX,
     1060,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8407,
     UINT16_MAX,
     8407,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     1061,
     UINT16_MAX,
     1061,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8408,
     UINT16_MAX,
     8408,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     1062,
     UINT16_MAX,
     1062,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8409,
     UINT16_MAX,
     8409,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8410,
     UINT16_MAX,
     8410,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     9255,
     1065,
     UINT16_MAX,
     1065,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     9258,
     UINT16_MAX,
     8411,
     UINT16_MAX,
     8411,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     9260,
     1070,
     UINT16_MAX,
     1070,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     9263,
     UINT16_MAX,
     8412,
     UINT16_MAX,
     8412,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     1073,
     UINT16_MAX,
     1073,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8413,
     UINT16_MAX,
     8413,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     9266,
     1076,
     UINT16_MAX,
     1076,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     9269,
     UINT16_MAX,
     8414,
     UINT16_MAX,
     8414,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     1079,
     UINT16_MAX,
     1079,
     UINT16_MAX,
     2629,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8415,
     UINT16_MAX,
     8415,
     2632,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     9272,
     1082,
     UINT16_MAX,
     1082,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     9275,
     UINT16_MAX,
     8416,
     UINT16_MAX,
     8416,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     9277,
     1087,
     UINT16_MAX,
     1087,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     9280,
     UINT16_MAX,
     8417,
     UINT16_MAX,
     8417,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     9282,
     1092,
     UINT16_MAX,
     1092,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     9285,
     UINT16_MAX,
     8418,
     UINT16_MAX,
     8418,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     1095,
     UINT16_MAX,
     1095,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8419,
     UINT16_MAX,
     8419,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     9288,
     1098,
     UINT16_MAX,
     1098,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     9291,
     UINT16_MAX,
     8420,
     UINT16_MAX,
     8420,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     9293,
     1103,
     UINT16_MAX,
     1103,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     9296,
     UINT16_MAX,
     8421,
     UINT16_MAX,
     8421,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     9298,
     1108,
     UINT16_MAX,
     1108,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     9301,
     UINT16_MAX,
     8422,
     UINT16_MAX,
     8422,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     1111,
     UINT16_MAX,
     1111,
     UINT16_MAX,
     2647,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8423,
     UINT16_MAX,
     8423,
     2650,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     9304,
     1114,
     UINT16_MAX,
     1114,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     9307,
     UINT16_MAX,
     8424,
     UINT16_MAX,
     8424,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     9309,
     1119,
     UINT16_MAX,
     1119,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     9312,
     UINT16_MAX,
     8425,
     UINT16_MAX,
     8425,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     9314,
     1124,
     UINT16_MAX,
     1124,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     9317,
     UINT16_MAX,
     8426,
     UINT16_MAX,
     8426,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     9319,
     1129,
     UINT16_MAX,
     1129,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     9322,
     UINT16_MAX,
     8427,
     UINT16_MAX,
     8427,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     9324,
     1134,
     UINT16_MAX,
     1134,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     9327,
     UINT16_MAX,
     8428,
     UINT16_MAX,
     8428,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     9329,
     1139,
     UINT16_MAX,
     1139,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     9332,
     UINT16_MAX,
     8429,
     UINT16_MAX,
     8429,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     1142,
     UINT16_MAX,
     1142,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8430,
     UINT16_MAX,
     8430,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     9335,
     1145,
     UINT16_MAX,
     1145,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     9338,
     UINT16_MAX,
     8431,
     UINT16_MAX,
     8431,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     1148,
     UINT16_MAX,
     1148,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8432,
     UINT16_MAX,
     8432,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     1149,
     UINT16_MAX,
     1149,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8433,
     UINT16_MAX,
     8433,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     1150,
     UINT16_MAX,
     1150,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8434,
     UINT16_MAX,
     8434,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     1151,
     UINT16_MAX,
     1151,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8435,
     UINT16_MAX,
     8435,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     1152,
     UINT16_MAX,
     1152,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8436,
     UINT16_MAX,
     8436,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     1153,
     UINT16_MAX,
     1153,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8437,
     UINT16_MAX,
     8437,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     1154,
     UINT16_MAX,
     1154,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8438,
     UINT16_MAX,
     8438,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     1155,
     UINT16_MAX,
     1155,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8439,
     UINT16_MAX,
     8439,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     1156,
     UINT16_MAX,
     1156,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8440,
     UINT16_MAX,
     8440,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     1157,
     UINT16_MAX,
     1157,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8441,
     UINT16_MAX,
     8441,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     1158,
     UINT16_MAX,
     1158,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8442,
     UINT16_MAX,
     8442,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     1159,
     UINT16_MAX,
     1159,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8443,
     UINT16_MAX,
     8443,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     1160,
     UINT16_MAX,
     1160,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8444,
     UINT16_MAX,
     8444,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     1161,
     UINT16_MAX,
     1161,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8445,
     UINT16_MAX,
     8445,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     1162,
     UINT16_MAX,
     1162,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8446,
     UINT16_MAX,
     8446,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     1163,
     UINT16_MAX,
     1163,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8447,
     UINT16_MAX,
     8447,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     1164,
     UINT16_MAX,
     1164,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8448,
     UINT16_MAX,
     8448,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     1165,
     UINT16_MAX,
     1165,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8449,
     UINT16_MAX,
     8449,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     1166,
     UINT16_MAX,
     1166,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8450,
     UINT16_MAX,
     8450,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     1167,
     UINT16_MAX,
     1167,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8451,
     UINT16_MAX,
     8451,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     1168,
     UINT16_MAX,
     1168,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8452,
     UINT16_MAX,
     8452,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     1169,
     UINT16_MAX,
     1169,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8453,
     UINT16_MAX,
     8453,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     1170,
     UINT16_MAX,
     1170,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8454,
     UINT16_MAX,
     8454,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     1171,
     UINT16_MAX,
     1171,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8455,
     UINT16_MAX,
     8455,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     1172,
     UINT16_MAX,
     1172,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8456,
     UINT16_MAX,
     8456,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     1173,
     UINT16_MAX,
     1173,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8457,
     UINT16_MAX,
     8457,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     1174,
     UINT16_MAX,
     1174,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8458,
     UINT16_MAX,
     8458,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     1175,
     UINT16_MAX,
     1175,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     1176,
     UINT16_MAX,
     1176,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     1177,
     UINT16_MAX,
     1177,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     1178,
     UINT16_MAX,
     1178,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     1179,
     UINT16_MAX,
     1179,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     1180,
     UINT16_MAX,
     1180,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     1181,
     UINT16_MAX,
     1181,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     1182,
     UINT16_MAX,
     1182,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     1183,
     UINT16_MAX,
     1183,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     1184,
     UINT16_MAX,
     1184,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     1185,
     UINT16_MAX,
     1185,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     1186,
     UINT16_MAX,
     1186,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     1187,
     UINT16_MAX,
     1187,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     1188,
     UINT16_MAX,
     1188,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     1189,
     UINT16_MAX,
     1189,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     1190,
     UINT16_MAX,
     1190,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     1191,
     UINT16_MAX,
     1191,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     1192,
     UINT16_MAX,
     1192,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     1193,
     UINT16_MAX,
     1193,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     1194,
     UINT16_MAX,
     1194,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     1195,
     UINT16_MAX,
     1195,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     1196,
     UINT16_MAX,
     1196,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     1197,
     UINT16_MAX,
     1197,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     1198,
     UINT16_MAX,
     1198,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     1199,
     UINT16_MAX,
     1199,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     1200,
     UINT16_MAX,
     1200,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     1201,
     UINT16_MAX,
     1201,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     1202,
     UINT16_MAX,
     1202,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     1203,
     UINT16_MAX,
     1203,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     1204,
     UINT16_MAX,
     1204,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     1205,
     UINT16_MAX,
     1205,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     1206,
     UINT16_MAX,
     1206,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     1207,
     UINT16_MAX,
     1207,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     1208,
     UINT16_MAX,
     1208,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     1209,
     UINT16_MAX,
     1209,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     1210,
     UINT16_MAX,
     1210,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     1211,
     UINT16_MAX,
     1211,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     1212,
     UINT16_MAX,
     1212,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_PO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8459,
     UINT16_MAX,
     8459,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8460,
     UINT16_MAX,
     8460,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8461,
     UINT16_MAX,
     8461,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8462,
     UINT16_MAX,
     8462,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8463,
     UINT16_MAX,
     8463,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8464,
     UINT16_MAX,
     8464,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8465,
     UINT16_MAX,
     8465,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8466,
     UINT16_MAX,
     8466,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8467,
     UINT16_MAX,
     8467,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8468,
     UINT16_MAX,
     8468,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8469,
     UINT16_MAX,
     8469,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8470,
     UINT16_MAX,
     8470,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8471,
     UINT16_MAX,
     8471,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8472,
     UINT16_MAX,
     8472,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8473,
     UINT16_MAX,
     8473,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8474,
     UINT16_MAX,
     8474,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8475,
     UINT16_MAX,
     8475,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8476,
     UINT16_MAX,
     8476,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8477,
     UINT16_MAX,
     8477,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8478,
     UINT16_MAX,
     8478,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8479,
     UINT16_MAX,
     8479,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8480,
     UINT16_MAX,
     8480,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8481,
     UINT16_MAX,
     8481,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8482,
     UINT16_MAX,
     8482,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8483,
     UINT16_MAX,
     8483,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8484,
     UINT16_MAX,
     8484,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8485,
     UINT16_MAX,
     8485,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8486,
     UINT16_MAX,
     8486,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8487,
     UINT16_MAX,
     8487,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8488,
     UINT16_MAX,
     8488,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8489,
     UINT16_MAX,
     8489,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8490,
     UINT16_MAX,
     8490,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8491,
     UINT16_MAX,
     8491,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8492,
     UINT16_MAX,
     8492,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8493,
     UINT16_MAX,
     8493,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8494,
     UINT16_MAX,
     8494,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8495,
     UINT16_MAX,
     8495,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8496,
     UINT16_MAX,
     8496,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     9405,
     9405,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_PD,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_MN,
     222,
     UTF8PROC_BIDI_CLASS_NSM,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     0,
     0,
     UTF8PROC_BOUNDCLASS_EXTEND},
    {UTF8PROC_CATEGORY_MN,
     228,
     UTF8PROC_BIDI_CLASS_NSM,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     0,
     0,
     UTF8PROC_BOUNDCLASS_EXTEND},
    {UTF8PROC_CATEGORY_MN,
     10,
     UTF8PROC_BIDI_CLASS_NSM,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     0,
     0,
     UTF8PROC_BOUNDCLASS_EXTEND},
    {UTF8PROC_CATEGORY_MN,
     11,
     UTF8PROC_BIDI_CLASS_NSM,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     0,
     0,
     UTF8PROC_BOUNDCLASS_EXTEND},
    {UTF8PROC_CATEGORY_MN,
     12,
     UTF8PROC_BIDI_CLASS_NSM,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     0,
     0,
     UTF8PROC_BOUNDCLASS_EXTEND},
    {UTF8PROC_CATEGORY_MN,
     13,
     UTF8PROC_BIDI_CLASS_NSM,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     0,
     0,
     UTF8PROC_BOUNDCLASS_EXTEND},
    {UTF8PROC_CATEGORY_MN,
     14,
     UTF8PROC_BIDI_CLASS_NSM,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     0,
     0,
     UTF8PROC_BOUNDCLASS_EXTEND},
    {UTF8PROC_CATEGORY_MN,
     15,
     UTF8PROC_BIDI_CLASS_NSM,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     0,
     0,
     UTF8PROC_BOUNDCLASS_EXTEND},
    {UTF8PROC_CATEGORY_MN,
     16,
     UTF8PROC_BIDI_CLASS_NSM,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     0,
     0,
     UTF8PROC_BOUNDCLASS_EXTEND},
    {UTF8PROC_CATEGORY_MN,
     17,
     UTF8PROC_BIDI_CLASS_NSM,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     0,
     0,
     UTF8PROC_BOUNDCLASS_EXTEND},
    {UTF8PROC_CATEGORY_MN,
     18,
     UTF8PROC_BIDI_CLASS_NSM,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     0,
     0,
     UTF8PROC_BOUNDCLASS_EXTEND},
    {UTF8PROC_CATEGORY_MN,
     19,
     UTF8PROC_BIDI_CLASS_NSM,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     0,
     0,
     UTF8PROC_BOUNDCLASS_EXTEND},
    {UTF8PROC_CATEGORY_MN,
     20,
     UTF8PROC_BIDI_CLASS_NSM,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     0,
     0,
     UTF8PROC_BOUNDCLASS_EXTEND},
    {UTF8PROC_CATEGORY_MN,
     21,
     UTF8PROC_BIDI_CLASS_NSM,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     0,
     0,
     UTF8PROC_BOUNDCLASS_EXTEND},
    {UTF8PROC_CATEGORY_MN,
     22,
     UTF8PROC_BIDI_CLASS_NSM,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     0,
     0,
     UTF8PROC_BOUNDCLASS_EXTEND},
    {UTF8PROC_CATEGORY_PD,
     0,
     UTF8PROC_BIDI_CLASS_R,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_MN,
     23,
     UTF8PROC_BIDI_CLASS_NSM,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     0,
     0,
     UTF8PROC_BOUNDCLASS_EXTEND},
    {UTF8PROC_CATEGORY_PO,
     0,
     UTF8PROC_BIDI_CLASS_R,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_MN,
     24,
     UTF8PROC_BIDI_CLASS_NSM,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     0,
     0,
     UTF8PROC_BOUNDCLASS_EXTEND},
    {UTF8PROC_CATEGORY_MN,
     25,
     UTF8PROC_BIDI_CLASS_NSM,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     0,
     0,
     UTF8PROC_BOUNDCLASS_EXTEND},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_R,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_CF,
     0,
     UTF8PROC_BIDI_CLASS_AN,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     true,
     0,
     0,
     UTF8PROC_BOUNDCLASS_PREPEND},
    {UTF8PROC_CATEGORY_SM,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SC,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_PO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_MN,
     30,
     UTF8PROC_BIDI_CLASS_NSM,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     0,
     0,
     UTF8PROC_BOUNDCLASS_EXTEND},
    {UTF8PROC_CATEGORY_MN,
     31,
     UTF8PROC_BIDI_CLASS_NSM,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     0,
     0,
     UTF8PROC_BOUNDCLASS_EXTEND},
    {UTF8PROC_CATEGORY_MN,
     32,
     UTF8PROC_BIDI_CLASS_NSM,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     0,
     0,
     UTF8PROC_BOUNDCLASS_EXTEND},
    {UTF8PROC_CATEGORY_CF,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     true,
     true,
     0,
     0,
     UTF8PROC_BOUNDCLASS_CONTROL},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     0,
     9407,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     0,
     9409,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     0,
     9411,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     0,
     9413,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     0,
     9415,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     2671,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LM,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     2676,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     2679,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_MN,
     27,
     UTF8PROC_BIDI_CLASS_NSM,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     0,
     0,
     UTF8PROC_BOUNDCLASS_EXTEND},
    {UTF8PROC_CATEGORY_MN,
     28,
     UTF8PROC_BIDI_CLASS_NSM,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     0,
     0,
     UTF8PROC_BOUNDCLASS_EXTEND},
    {UTF8PROC_CATEGORY_MN,
     29,
     UTF8PROC_BIDI_CLASS_NSM,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     0,
     0,
     UTF8PROC_BOUNDCLASS_EXTEND},
    {UTF8PROC_CATEGORY_MN,
     33,
     UTF8PROC_BIDI_CLASS_NSM,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     0,
     0,
     UTF8PROC_BOUNDCLASS_EXTEND},
    {UTF8PROC_CATEGORY_MN,
     34,
     UTF8PROC_BIDI_CLASS_NSM,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     0,
     0,
     UTF8PROC_BOUNDCLASS_EXTEND},
    {UTF8PROC_CATEGORY_MN,
     230,
     UTF8PROC_BIDI_CLASS_NSM,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     32785,
     false,
     false,
     false,
     false,
     0,
     0,
     UTF8PROC_BOUNDCLASS_EXTEND},
    {UTF8PROC_CATEGORY_MN,
     230,
     UTF8PROC_BIDI_CLASS_NSM,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     32786,
     false,
     false,
     false,
     false,
     0,
     0,
     UTF8PROC_BOUNDCLASS_EXTEND},
    {UTF8PROC_CATEGORY_MN,
     220,
     UTF8PROC_BIDI_CLASS_NSM,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     32787,
     false,
     false,
     false,
     false,
     0,
     0,
     UTF8PROC_BOUNDCLASS_EXTEND},
    {UTF8PROC_CATEGORY_ND,
     0,
     UTF8PROC_BIDI_CLASS_AN,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_PO,
     0,
     UTF8PROC_BIDI_CLASS_AN,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_MN,
     35,
     UTF8PROC_BIDI_CLASS_NSM,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     0,
     0,
     UTF8PROC_BOUNDCLASS_EXTEND},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     9417,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     9419,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     9421,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     9423,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     0,
     9425,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     2685,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     0,
     9427,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     2688,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     0,
     9429,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     2682,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_CF,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     true,
     0,
     0,
     UTF8PROC_BOUNDCLASS_PREPEND},
    {UTF8PROC_CATEGORY_MN,
     36,
     UTF8PROC_BIDI_CLASS_NSM,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     0,
     0,
     UTF8PROC_BOUNDCLASS_EXTEND},
    {UTF8PROC_CATEGORY_MN,
     0,
     UTF8PROC_BIDI_CLASS_NSM,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     0,
     0,
     UTF8PROC_BOUNDCLASS_EXTEND},
    {UTF8PROC_CATEGORY_ND,
     0,
     UTF8PROC_BIDI_CLASS_R,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LM,
     0,
     UTF8PROC_BIDI_CLASS_R,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SC,
     0,
     UTF8PROC_BIDI_CLASS_R,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_MC,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     0,
     0,
     UTF8PROC_BOUNDCLASS_SPACINGMARK},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     2691,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     9431,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     2694,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     9433,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     2697,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     9435,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_MN,
     7,
     UTF8PROC_BIDI_CLASS_NSM,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     32788,
     false,
     false,
     false,
     false,
     0,
     0,
     UTF8PROC_BOUNDCLASS_EXTEND},
    {UTF8PROC_CATEGORY_MN,
     9,
     UTF8PROC_BIDI_CLASS_NSM,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     0,
     0,
     UTF8PROC_BOUNDCLASS_EXTEND},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     9437,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     true,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     9439,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     true,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     9441,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     true,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     9443,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     true,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     9445,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     true,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     9447,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     true,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     9449,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     true,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     9451,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     true,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_ND,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_MN,
     7,
     UTF8PROC_BIDI_CLASS_NSM,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     0,
     0,
     UTF8PROC_BOUNDCLASS_EXTEND},
    {UTF8PROC_CATEGORY_MC,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     32789,
     false,
     false,
     false,
     false,
     0,
     0,
     UTF8PROC_BOUNDCLASS_EXTEND},
    {UTF8PROC_CATEGORY_MC,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     2700,
     false,
     false,
     false,
     false,
     0,
     0,
     UTF8PROC_BOUNDCLASS_SPACINGMARK},
    {UTF8PROC_CATEGORY_MC,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     9453,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     0,
     0,
     UTF8PROC_BOUNDCLASS_SPACINGMARK},
    {UTF8PROC_CATEGORY_MC,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     9455,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     0,
     0,
     UTF8PROC_BOUNDCLASS_SPACINGMARK},
    {UTF8PROC_CATEGORY_MC,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     32790,
     false,
     false,
     false,
     false,
     0,
     0,
     UTF8PROC_BOUNDCLASS_EXTEND},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     9457,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     true,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     9459,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     true,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     9461,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     true,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_NO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     9463,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     true,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     9465,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     true,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     9467,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     true,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     9469,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     true,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     9471,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     true,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     9473,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     true,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_MC,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     32792,
     false,
     false,
     false,
     false,
     0,
     0,
     UTF8PROC_BOUNDCLASS_EXTEND},
    {UTF8PROC_CATEGORY_MC,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     2704,
     false,
     false,
     false,
     false,
     0,
     0,
     UTF8PROC_BOUNDCLASS_SPACINGMARK},
    {UTF8PROC_CATEGORY_MC,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     9475,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     0,
     0,
     UTF8PROC_BOUNDCLASS_SPACINGMARK},
    {UTF8PROC_CATEGORY_MC,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     9477,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     0,
     0,
     UTF8PROC_BOUNDCLASS_SPACINGMARK},
    {UTF8PROC_CATEGORY_MC,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     9479,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     0,
     0,
     UTF8PROC_BOUNDCLASS_SPACINGMARK},
    {UTF8PROC_CATEGORY_MN,
     0,
     UTF8PROC_BIDI_CLASS_NSM,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     32791,
     false,
     false,
     false,
     false,
     0,
     0,
     UTF8PROC_BOUNDCLASS_EXTEND},
    {UTF8PROC_CATEGORY_MC,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     32793,
     false,
     false,
     false,
     false,
     0,
     0,
     UTF8PROC_BOUNDCLASS_EXTEND},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     9481,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     true,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     9483,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     true,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     2709,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     9485,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_MC,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     32795,
     false,
     false,
     false,
     false,
     0,
     0,
     UTF8PROC_BOUNDCLASS_EXTEND},
    {UTF8PROC_CATEGORY_MC,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     2712,
     false,
     false,
     false,
     false,
     0,
     0,
     UTF8PROC_BOUNDCLASS_SPACINGMARK},
    {UTF8PROC_CATEGORY_MC,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     2716,
     false,
     false,
     false,
     false,
     0,
     0,
     UTF8PROC_BOUNDCLASS_SPACINGMARK},
    {UTF8PROC_CATEGORY_MC,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     9487,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     0,
     0,
     UTF8PROC_BOUNDCLASS_SPACINGMARK},
    {UTF8PROC_CATEGORY_MC,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     9489,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     0,
     0,
     UTF8PROC_BOUNDCLASS_SPACINGMARK},
    {UTF8PROC_CATEGORY_MC,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     9491,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     0,
     0,
     UTF8PROC_BOUNDCLASS_SPACINGMARK},
    {UTF8PROC_CATEGORY_MC,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     32794,
     false,
     false,
     false,
     false,
     0,
     0,
     UTF8PROC_BOUNDCLASS_EXTEND},
    {UTF8PROC_CATEGORY_MN,
     0,
     UTF8PROC_BIDI_CLASS_NSM,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     2719,
     false,
     false,
     false,
     false,
     0,
     0,
     UTF8PROC_BOUNDCLASS_EXTEND},
    {UTF8PROC_CATEGORY_MN,
     0,
     UTF8PROC_BIDI_CLASS_NSM,
     0,
     9493,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     0,
     0,
     UTF8PROC_BOUNDCLASS_EXTEND},
    {UTF8PROC_CATEGORY_MN,
     84,
     UTF8PROC_BIDI_CLASS_NSM,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     0,
     0,
     UTF8PROC_BOUNDCLASS_EXTEND},
    {UTF8PROC_CATEGORY_MN,
     91,
     UTF8PROC_BIDI_CLASS_NSM,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     32796,
     false,
     false,
     false,
     false,
     0,
     0,
     UTF8PROC_BOUNDCLASS_EXTEND},
    {UTF8PROC_CATEGORY_NO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_MN,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     2722,
     false,
     false,
     false,
     false,
     0,
     0,
     UTF8PROC_BOUNDCLASS_EXTEND},
    {UTF8PROC_CATEGORY_MC,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     9495,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     0,
     0,
     UTF8PROC_BOUNDCLASS_SPACINGMARK},
    {UTF8PROC_CATEGORY_MC,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     32799,
     false,
     false,
     false,
     false,
     0,
     0,
     UTF8PROC_BOUNDCLASS_EXTEND},
    {UTF8PROC_CATEGORY_MN,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     2725,
     false,
     false,
     false,
     false,
     0,
     0,
     UTF8PROC_BOUNDCLASS_EXTEND},
    {UTF8PROC_CATEGORY_MC,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     9497,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     0,
     0,
     UTF8PROC_BOUNDCLASS_SPACINGMARK},
    {UTF8PROC_CATEGORY_MC,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     9499,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     0,
     0,
     UTF8PROC_BOUNDCLASS_SPACINGMARK},
    {UTF8PROC_CATEGORY_MC,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     9501,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     2730,
     false,
     false,
     false,
     false,
     0,
     0,
     UTF8PROC_BOUNDCLASS_SPACINGMARK},
    {UTF8PROC_CATEGORY_MC,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     9503,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     0,
     0,
     UTF8PROC_BOUNDCLASS_SPACINGMARK},
    {UTF8PROC_CATEGORY_MC,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     32797,
     false,
     false,
     false,
     false,
     0,
     0,
     UTF8PROC_BOUNDCLASS_EXTEND},
    {UTF8PROC_CATEGORY_MC,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     32798,
     false,
     false,
     false,
     false,
     0,
     0,
     UTF8PROC_BOUNDCLASS_EXTEND},
    {UTF8PROC_CATEGORY_MC,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     32800,
     false,
     false,
     false,
     false,
     0,
     0,
     UTF8PROC_BOUNDCLASS_EXTEND},
    {UTF8PROC_CATEGORY_MC,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     2733,
     false,
     false,
     false,
     false,
     0,
     0,
     UTF8PROC_BOUNDCLASS_SPACINGMARK},
    {UTF8PROC_CATEGORY_MC,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     2737,
     false,
     false,
     false,
     false,
     0,
     0,
     UTF8PROC_BOUNDCLASS_SPACINGMARK},
    {UTF8PROC_CATEGORY_MC,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     9505,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     0,
     0,
     UTF8PROC_BOUNDCLASS_SPACINGMARK},
    {UTF8PROC_CATEGORY_MC,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     9507,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     0,
     0,
     UTF8PROC_BOUNDCLASS_SPACINGMARK},
    {UTF8PROC_CATEGORY_MC,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     9509,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     0,
     0,
     UTF8PROC_BOUNDCLASS_SPACINGMARK},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_PREPEND},
    {UTF8PROC_CATEGORY_MC,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     32801,
     false,
     false,
     false,
     false,
     0,
     0,
     UTF8PROC_BOUNDCLASS_EXTEND},
    {UTF8PROC_CATEGORY_MN,
     9,
     UTF8PROC_BIDI_CLASS_NSM,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     32802,
     false,
     false,
     false,
     false,
     0,
     0,
     UTF8PROC_BOUNDCLASS_EXTEND},
    {UTF8PROC_CATEGORY_MC,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     32803,
     false,
     false,
     false,
     false,
     0,
     0,
     UTF8PROC_BOUNDCLASS_EXTEND},
    {UTF8PROC_CATEGORY_MC,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     2740,
     false,
     false,
     false,
     false,
     0,
     0,
     UTF8PROC_BOUNDCLASS_SPACINGMARK},
    {UTF8PROC_CATEGORY_MC,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     9511,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     0,
     0,
     UTF8PROC_BOUNDCLASS_SPACINGMARK},
    {UTF8PROC_CATEGORY_MC,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     9513,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     2745,
     false,
     false,
     false,
     false,
     0,
     0,
     UTF8PROC_BOUNDCLASS_SPACINGMARK},
    {UTF8PROC_CATEGORY_MC,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     9515,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     0,
     0,
     UTF8PROC_BOUNDCLASS_SPACINGMARK},
    {UTF8PROC_CATEGORY_MC,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     9517,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     0,
     0,
     UTF8PROC_BOUNDCLASS_SPACINGMARK},
    {UTF8PROC_CATEGORY_MC,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     32804,
     false,
     false,
     false,
     false,
     0,
     0,
     UTF8PROC_BOUNDCLASS_EXTEND},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     9519,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_SPACINGMARK},
    {UTF8PROC_CATEGORY_MN,
     103,
     UTF8PROC_BIDI_CLASS_NSM,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     0,
     0,
     UTF8PROC_BOUNDCLASS_EXTEND},
    {UTF8PROC_CATEGORY_MN,
     107,
     UTF8PROC_BIDI_CLASS_NSM,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     0,
     0,
     UTF8PROC_BOUNDCLASS_EXTEND},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     9521,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_SPACINGMARK},
    {UTF8PROC_CATEGORY_MN,
     118,
     UTF8PROC_BIDI_CLASS_NSM,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     0,
     0,
     UTF8PROC_BOUNDCLASS_EXTEND},
    {UTF8PROC_CATEGORY_MN,
     122,
     UTF8PROC_BIDI_CLASS_NSM,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     0,
     0,
     UTF8PROC_BOUNDCLASS_EXTEND},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     9523,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     9525,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_PO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_NOBREAK,
     1335,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_MN,
     216,
     UTF8PROC_BIDI_CLASS_NSM,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     0,
     0,
     UTF8PROC_BOUNDCLASS_EXTEND},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     9528,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     true,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     9530,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     true,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     9532,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     true,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     9534,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     true,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     9536,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     true,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     9538,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     true,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_MN,
     129,
     UTF8PROC_BIDI_CLASS_NSM,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     0,
     0,
     UTF8PROC_BOUNDCLASS_EXTEND},
    {UTF8PROC_CATEGORY_MN,
     130,
     UTF8PROC_BIDI_CLASS_NSM,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     0,
     0,
     UTF8PROC_BOUNDCLASS_EXTEND},
    {UTF8PROC_CATEGORY_MN,
     0,
     UTF8PROC_BIDI_CLASS_NSM,
     0,
     9540,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     0,
     0,
     UTF8PROC_BOUNDCLASS_EXTEND},
    {UTF8PROC_CATEGORY_MN,
     132,
     UTF8PROC_BIDI_CLASS_NSM,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     0,
     0,
     UTF8PROC_BOUNDCLASS_EXTEND},
    {UTF8PROC_CATEGORY_MN,
     0,
     UTF8PROC_BIDI_CLASS_NSM,
     0,
     9542,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     0,
     0,
     UTF8PROC_BOUNDCLASS_EXTEND},
    {UTF8PROC_CATEGORY_MN,
     0,
     UTF8PROC_BIDI_CLASS_NSM,
     0,
     9544,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     true,
     false,
     false,
     0,
     0,
     UTF8PROC_BOUNDCLASS_EXTEND},
    {UTF8PROC_CATEGORY_MN,
     0,
     UTF8PROC_BIDI_CLASS_NSM,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     9546,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     0,
     0,
     UTF8PROC_BOUNDCLASS_EXTEND},
    {UTF8PROC_CATEGORY_MN,
     0,
     UTF8PROC_BIDI_CLASS_NSM,
     0,
     9548,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     true,
     false,
     false,
     0,
     0,
     UTF8PROC_BOUNDCLASS_EXTEND},
    {UTF8PROC_CATEGORY_MN,
     0,
     UTF8PROC_BIDI_CLASS_NSM,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     9550,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     0,
     0,
     UTF8PROC_BOUNDCLASS_EXTEND},
    {UTF8PROC_CATEGORY_MN,
     0,
     UTF8PROC_BIDI_CLASS_NSM,
     0,
     9552,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     0,
     0,
     UTF8PROC_BOUNDCLASS_EXTEND},
    {UTF8PROC_CATEGORY_MN,
     0,
     UTF8PROC_BIDI_CLASS_NSM,
     0,
     9554,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     true,
     false,
     false,
     0,
     0,
     UTF8PROC_BOUNDCLASS_EXTEND},
    {UTF8PROC_CATEGORY_MN,
     0,
     UTF8PROC_BIDI_CLASS_NSM,
     0,
     9556,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     true,
     false,
     false,
     0,
     0,
     UTF8PROC_BOUNDCLASS_EXTEND},
    {UTF8PROC_CATEGORY_MN,
     0,
     UTF8PROC_BIDI_CLASS_NSM,
     0,
     9558,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     true,
     false,
     false,
     0,
     0,
     UTF8PROC_BOUNDCLASS_EXTEND},
    {UTF8PROC_CATEGORY_MN,
     0,
     UTF8PROC_BIDI_CLASS_NSM,
     0,
     9560,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     true,
     false,
     false,
     0,
     0,
     UTF8PROC_BOUNDCLASS_EXTEND},
    {UTF8PROC_CATEGORY_MN,
     0,
     UTF8PROC_BIDI_CLASS_NSM,
     0,
     9562,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     true,
     false,
     false,
     0,
     0,
     UTF8PROC_BOUNDCLASS_EXTEND},
    {UTF8PROC_CATEGORY_MN,
     0,
     UTF8PROC_BIDI_CLASS_NSM,
     0,
     9564,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     true,
     false,
     false,
     0,
     0,
     UTF8PROC_BOUNDCLASS_EXTEND},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     2748,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     9566,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_MC,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     0,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_MN,
     0,
     UTF8PROC_BIDI_CLASS_NSM,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     32805,
     false,
     false,
     false,
     false,
     0,
     0,
     UTF8PROC_BOUNDCLASS_EXTEND},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     1376,
     UINT16_MAX,
     1376,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     1377,
     UINT16_MAX,
     1377,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     1378,
     UINT16_MAX,
     1378,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     1379,
     UINT16_MAX,
     1379,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     1380,
     UINT16_MAX,
     1380,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     1381,
     UINT16_MAX,
     1381,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     1382,
     UINT16_MAX,
     1382,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     1383,
     UINT16_MAX,
     1383,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     1384,
     UINT16_MAX,
     1384,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     1385,
     UINT16_MAX,
     1385,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     1386,
     UINT16_MAX,
     1386,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     1387,
     UINT16_MAX,
     1387,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     1388,
     UINT16_MAX,
     1388,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     1389,
     UINT16_MAX,
     1389,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     1390,
     UINT16_MAX,
     1390,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     1391,
     UINT16_MAX,
     1391,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     1392,
     UINT16_MAX,
     1392,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     1393,
     UINT16_MAX,
     1393,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     1394,
     UINT16_MAX,
     1394,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     1395,
     UINT16_MAX,
     1395,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     1396,
     UINT16_MAX,
     1396,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     1397,
     UINT16_MAX,
     1397,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     1398,
     UINT16_MAX,
     1398,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     1399,
     UINT16_MAX,
     1399,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     1400,
     UINT16_MAX,
     1400,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     1401,
     UINT16_MAX,
     1401,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     1402,
     UINT16_MAX,
     1402,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     1403,
     UINT16_MAX,
     1403,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     1404,
     UINT16_MAX,
     1404,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     1405,
     UINT16_MAX,
     1405,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     1406,
     UINT16_MAX,
     1406,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     1407,
     UINT16_MAX,
     1407,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     1408,
     UINT16_MAX,
     1408,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     1409,
     UINT16_MAX,
     1409,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     1410,
     UINT16_MAX,
     1410,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     1411,
     UINT16_MAX,
     1411,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     1412,
     UINT16_MAX,
     1412,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     1413,
     UINT16_MAX,
     1413,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     1414,
     UINT16_MAX,
     1414,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     1415,
     UINT16_MAX,
     1415,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8497,
     UINT16_MAX,
     1446,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8498,
     UINT16_MAX,
     1447,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8499,
     UINT16_MAX,
     1448,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8500,
     UINT16_MAX,
     1449,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8501,
     UINT16_MAX,
     1450,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8502,
     UINT16_MAX,
     1451,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8503,
     UINT16_MAX,
     1452,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8504,
     UINT16_MAX,
     1453,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8505,
     UINT16_MAX,
     1454,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8506,
     UINT16_MAX,
     1455,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8507,
     UINT16_MAX,
     1456,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8508,
     UINT16_MAX,
     1457,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8509,
     UINT16_MAX,
     1416,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8510,
     UINT16_MAX,
     1458,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8511,
     UINT16_MAX,
     1459,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8512,
     UINT16_MAX,
     1460,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8513,
     UINT16_MAX,
     1461,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8514,
     UINT16_MAX,
     1462,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8515,
     UINT16_MAX,
     1463,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8516,
     UINT16_MAX,
     1464,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8517,
     UINT16_MAX,
     1465,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8518,
     UINT16_MAX,
     1466,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8519,
     UINT16_MAX,
     1467,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8520,
     UINT16_MAX,
     1468,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8521,
     UINT16_MAX,
     1469,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8522,
     UINT16_MAX,
     1470,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8523,
     UINT16_MAX,
     1471,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8524,
     UINT16_MAX,
     1472,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8525,
     UINT16_MAX,
     1473,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8526,
     UINT16_MAX,
     1474,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8527,
     UINT16_MAX,
     1475,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8528,
     UINT16_MAX,
     1476,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8529,
     UINT16_MAX,
     1477,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8530,
     UINT16_MAX,
     1478,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8531,
     UINT16_MAX,
     1479,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8532,
     UINT16_MAX,
     1480,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8533,
     UINT16_MAX,
     1481,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8534,
     UINT16_MAX,
     1482,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8535,
     UINT16_MAX,
     1483,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8536,
     UINT16_MAX,
     1484,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8537,
     UINT16_MAX,
     1485,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8538,
     UINT16_MAX,
     1486,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8539,
     UINT16_MAX,
     1487,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LM,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SUPER,
     1416,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8540,
     UINT16_MAX,
     1488,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8541,
     UINT16_MAX,
     1489,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8542,
     UINT16_MAX,
     1490,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_L},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     true,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_L},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     true,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_V},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_V},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_T},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     8543,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     8544,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     8545,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     8546,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     8547,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     8548,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     8549,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     8550,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     8551,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     8552,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     8553,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     8554,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     8555,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     8556,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     8557,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     8558,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     8559,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     8560,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     8561,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     8562,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     8563,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     8564,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     8565,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     8566,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     8567,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     8568,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     8569,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     8570,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     8571,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     8572,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     8573,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     8574,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     8575,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     8576,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     8577,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     8578,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     8579,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     8580,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     8581,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     8582,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     8583,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     8584,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     8585,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     8586,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     8587,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     8588,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     8589,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     8590,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     8591,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     8592,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     8593,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     8594,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     8595,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     8596,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     8597,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     8598,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     8599,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     8600,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     8601,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     8602,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     8603,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     8604,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     8605,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     8606,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     8607,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     8608,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     8609,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     8610,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     8611,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     8612,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     8613,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     8614,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     8615,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     8616,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     8617,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     8618,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     8619,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     8620,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     8621,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     8622,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     8623,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     8624,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     8625,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     8626,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     8627,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     8628,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     1417,
     1417,
     UINT16_MAX,
     1417,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     1418,
     1418,
     UINT16_MAX,
     1418,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     1419,
     1419,
     UINT16_MAX,
     1419,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     1420,
     1420,
     UINT16_MAX,
     1420,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     1421,
     1421,
     UINT16_MAX,
     1421,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     1422,
     1422,
     UINT16_MAX,
     1422,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_NL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_CF,
     0,
     UTF8PROC_BIDI_CLASS_BN,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     true,
     true,
     0,
     0,
     UTF8PROC_BOUNDCLASS_CONTROL},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     2751,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     9615,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     2754,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     9617,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     2757,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     9619,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     2760,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     9621,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     2763,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     9623,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     2766,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     9625,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_MC,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     32806,
     false,
     false,
     false,
     false,
     0,
     0,
     UTF8PROC_BOUNDCLASS_EXTEND},
    {UTF8PROC_CATEGORY_MN,
     0,
     UTF8PROC_BIDI_CLASS_NSM,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     2769,
     false,
     false,
     false,
     false,
     0,
     0,
     UTF8PROC_BOUNDCLASS_EXTEND},
    {UTF8PROC_CATEGORY_MC,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     9627,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     0,
     0,
     UTF8PROC_BOUNDCLASS_SPACINGMARK},
    {UTF8PROC_CATEGORY_MN,
     0,
     UTF8PROC_BIDI_CLASS_NSM,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     2772,
     false,
     false,
     false,
     false,
     0,
     0,
     UTF8PROC_BOUNDCLASS_EXTEND},
    {UTF8PROC_CATEGORY_MC,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     9629,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     0,
     0,
     UTF8PROC_BOUNDCLASS_SPACINGMARK},
    {UTF8PROC_CATEGORY_MC,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     2775,
     false,
     false,
     false,
     false,
     0,
     0,
     UTF8PROC_BOUNDCLASS_SPACINGMARK},
    {UTF8PROC_CATEGORY_MC,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     2778,
     false,
     false,
     false,
     false,
     0,
     0,
     UTF8PROC_BOUNDCLASS_SPACINGMARK},
    {UTF8PROC_CATEGORY_MC,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     9631,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     0,
     0,
     UTF8PROC_BOUNDCLASS_SPACINGMARK},
    {UTF8PROC_CATEGORY_MC,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     9633,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     0,
     0,
     UTF8PROC_BOUNDCLASS_SPACINGMARK},
    {UTF8PROC_CATEGORY_MN,
     0,
     UTF8PROC_BIDI_CLASS_NSM,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     2781,
     false,
     false,
     false,
     false,
     0,
     0,
     UTF8PROC_BOUNDCLASS_EXTEND},
    {UTF8PROC_CATEGORY_MC,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     9635,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     0,
     0,
     UTF8PROC_BOUNDCLASS_SPACINGMARK},
    {UTF8PROC_CATEGORY_MC,
     9,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     0,
     0,
     UTF8PROC_BOUNDCLASS_SPACINGMARK},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     955,
     8313,
     UINT16_MAX,
     8313,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     957,
     8315,
     UINT16_MAX,
     8315,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     969,
     8325,
     UINT16_MAX,
     8325,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     972,
     8328,
     UINT16_MAX,
     8328,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     973,
     8329,
     UINT16_MAX,
     8329,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     981,
     8337,
     UINT16_MAX,
     8337,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     1004,
     8360,
     UINT16_MAX,
     8360,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     1445,
     8629,
     UINT16_MAX,
     8629,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     1446,
     UINT16_MAX,
     1446,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     1447,
     UINT16_MAX,
     1447,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     1448,
     UINT16_MAX,
     1448,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     1449,
     UINT16_MAX,
     1449,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     1450,
     UINT16_MAX,
     1450,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     1451,
     UINT16_MAX,
     1451,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     1452,
     UINT16_MAX,
     1452,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     1453,
     UINT16_MAX,
     1453,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     1454,
     UINT16_MAX,
     1454,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     1455,
     UINT16_MAX,
     1455,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     1456,
     UINT16_MAX,
     1456,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     1457,
     UINT16_MAX,
     1457,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     1416,
     UINT16_MAX,
     1416,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     1458,
     UINT16_MAX,
     1458,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     1459,
     UINT16_MAX,
     1459,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     1460,
     UINT16_MAX,
     1460,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     1461,
     UINT16_MAX,
     1461,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     1462,
     UINT16_MAX,
     1462,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     1463,
     UINT16_MAX,
     1463,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     1464,
     UINT16_MAX,
     1464,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     1465,
     UINT16_MAX,
     1465,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     1466,
     UINT16_MAX,
     1466,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     1467,
     UINT16_MAX,
     1467,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     1468,
     UINT16_MAX,
     1468,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     1469,
     UINT16_MAX,
     1469,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     1470,
     UINT16_MAX,
     1470,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     1471,
     UINT16_MAX,
     1471,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     1472,
     UINT16_MAX,
     1472,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     1473,
     UINT16_MAX,
     1473,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     1474,
     UINT16_MAX,
     1474,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     1475,
     UINT16_MAX,
     1475,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     1476,
     UINT16_MAX,
     1476,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     1477,
     UINT16_MAX,
     1477,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     1478,
     UINT16_MAX,
     1478,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     1479,
     UINT16_MAX,
     1479,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     1480,
     UINT16_MAX,
     1480,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     1481,
     UINT16_MAX,
     1481,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     1482,
     UINT16_MAX,
     1482,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     1483,
     UINT16_MAX,
     1483,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     1484,
     UINT16_MAX,
     1484,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     1485,
     UINT16_MAX,
     1485,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     1486,
     UINT16_MAX,
     1486,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     1487,
     UINT16_MAX,
     1487,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     1488,
     UINT16_MAX,
     1488,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     1489,
     UINT16_MAX,
     1489,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     1490,
     UINT16_MAX,
     1490,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LM,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SUPER,
     1491,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LM,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SUPER,
     1492,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LM,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SUPER,
     1493,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LM,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SUPER,
     1494,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LM,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SUPER,
     1495,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LM,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SUPER,
     1496,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LM,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SUPER,
     1497,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LM,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SUPER,
     1498,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LM,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SUPER,
     1499,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LM,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SUPER,
     1500,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LM,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SUPER,
     1501,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LM,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SUPER,
     1502,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LM,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SUPER,
     1503,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LM,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SUPER,
     1504,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LM,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SUPER,
     1505,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LM,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SUPER,
     1506,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LM,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SUPER,
     1507,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LM,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SUPER,
     1508,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LM,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SUPER,
     1509,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LM,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SUPER,
     1510,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LM,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SUPER,
     1511,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LM,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SUPER,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LM,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SUPER,
     1512,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LM,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SUPER,
     1513,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LM,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SUPER,
     1514,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LM,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SUPER,
     1,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LM,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SUPER,
     3,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LM,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SUPER,
     4,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LM,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SUPER,
     485,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LM,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SUPER,
     486,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LM,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SUPER,
     1515,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LM,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SUPER,
     6,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LM,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SUPER,
     10,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LM,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SUPER,
     12,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LM,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SUPER,
     355,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LM,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SUPER,
     14,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LM,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SUPER,
     479,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LM,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SUPER,
     1516,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LM,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SUPER,
     1517,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LM,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SUPER,
     15,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LM,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SUPER,
     19,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LM,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SUPER,
     20,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LM,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SUPER,
     1518,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LM,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SUPER,
     493,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LM,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SUPER,
     21,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LM,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SUPER,
     1519,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LM,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SUPER,
     846,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LM,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SUPER,
     847,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LM,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SUPER,
     848,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LM,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SUPER,
     863,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LM,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SUPER,
     864,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LM,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SUB,
     8,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LM,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SUB,
     17,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LM,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SUB,
     20,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LM,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SUB,
     21,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LM,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SUB,
     846,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LM,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SUB,
     847,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LM,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SUB,
     859,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LM,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SUB,
     863,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LM,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SUB,
     864,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LM,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SUPER,
     968,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8630,
     UINT16_MAX,
     8630,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8631,
     UINT16_MAX,
     8631,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8632,
     UINT16_MAX,
     8632,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LM,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SUPER,
     1520,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LM,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SUPER,
     2,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LM,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SUPER,
     1521,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LM,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SUPER,
     94,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LM,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SUPER,
     5,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LM,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SUPER,
     1522,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LM,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SUPER,
     1523,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LM,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SUPER,
     1524,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LM,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SUPER,
     491,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LM,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SUPER,
     490,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LM,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SUPER,
     1525,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LM,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SUPER,
     1526,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LM,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SUPER,
     1527,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LM,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SUPER,
     1528,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LM,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SUPER,
     1529,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LM,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SUPER,
     1530,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LM,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SUPER,
     1531,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LM,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SUPER,
     1532,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LM,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SUPER,
     494,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LM,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SUPER,
     1533,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LM,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SUPER,
     1534,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LM,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SUPER,
     495,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LM,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SUPER,
     1535,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LM,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SUPER,
     1536,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LM,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SUPER,
     505,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LM,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SUPER,
     1537,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LM,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SUPER,
     778,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LM,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SUPER,
     513,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LM,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SUPER,
     1538,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LM,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SUPER,
     514,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LM,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SUPER,
     779,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LM,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SUPER,
     25,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LM,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SUPER,
     1539,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LM,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SUPER,
     1540,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LM,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SUPER,
     517,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LM,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SUPER,
     852,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_MN,
     214,
     UTF8PROC_BIDI_CLASS_NSM,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     0,
     0,
     UTF8PROC_BOUNDCLASS_EXTEND},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     9733,
     1543,
     UINT16_MAX,
     1543,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     9736,
     UINT16_MAX,
     8633,
     UINT16_MAX,
     8633,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     9738,
     1548,
     UINT16_MAX,
     1548,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     9741,
     UINT16_MAX,
     8634,
     UINT16_MAX,
     8634,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     9743,
     1553,
     UINT16_MAX,
     1553,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     9746,
     UINT16_MAX,
     8635,
     UINT16_MAX,
     8635,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     9748,
     1558,
     UINT16_MAX,
     1558,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     9751,
     UINT16_MAX,
     8636,
     UINT16_MAX,
     8636,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     9753,
     1563,
     UINT16_MAX,
     1563,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     9756,
     UINT16_MAX,
     8637,
     UINT16_MAX,
     8637,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     9758,
     1568,
     UINT16_MAX,
     1568,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     9761,
     UINT16_MAX,
     8638,
     UINT16_MAX,
     8638,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     9763,
     1573,
     UINT16_MAX,
     1573,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     9766,
     UINT16_MAX,
     8639,
     UINT16_MAX,
     8639,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     9768,
     1578,
     UINT16_MAX,
     1578,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     9771,
     UINT16_MAX,
     8640,
     UINT16_MAX,
     8640,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     9773,
     1583,
     UINT16_MAX,
     1583,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     9776,
     UINT16_MAX,
     8641,
     UINT16_MAX,
     8641,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     9778,
     1588,
     UINT16_MAX,
     1588,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     9781,
     UINT16_MAX,
     8642,
     UINT16_MAX,
     8642,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     9783,
     1593,
     UINT16_MAX,
     1593,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     9786,
     UINT16_MAX,
     8643,
     UINT16_MAX,
     8643,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     9788,
     1598,
     UINT16_MAX,
     1598,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     9791,
     UINT16_MAX,
     8644,
     UINT16_MAX,
     8644,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     9793,
     1603,
     UINT16_MAX,
     1603,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     9796,
     UINT16_MAX,
     8645,
     UINT16_MAX,
     8645,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     9798,
     1608,
     UINT16_MAX,
     1608,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     9801,
     UINT16_MAX,
     8646,
     UINT16_MAX,
     8646,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     9803,
     1613,
     UINT16_MAX,
     1613,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     9806,
     UINT16_MAX,
     8647,
     UINT16_MAX,
     8647,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     9808,
     1618,
     UINT16_MAX,
     1618,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     9811,
     UINT16_MAX,
     8648,
     UINT16_MAX,
     8648,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     9813,
     1623,
     UINT16_MAX,
     1623,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     9816,
     UINT16_MAX,
     8649,
     UINT16_MAX,
     8649,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     9818,
     1628,
     UINT16_MAX,
     1628,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     9821,
     UINT16_MAX,
     8650,
     UINT16_MAX,
     8650,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     9823,
     1633,
     UINT16_MAX,
     1633,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     9826,
     UINT16_MAX,
     8651,
     UINT16_MAX,
     8651,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     9828,
     1638,
     UINT16_MAX,
     1638,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     9831,
     UINT16_MAX,
     8652,
     UINT16_MAX,
     8652,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     9833,
     1643,
     UINT16_MAX,
     1643,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     9836,
     UINT16_MAX,
     8653,
     UINT16_MAX,
     8653,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     9838,
     1648,
     UINT16_MAX,
     1648,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     9841,
     UINT16_MAX,
     8654,
     UINT16_MAX,
     8654,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     9843,
     1653,
     UINT16_MAX,
     1653,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     9846,
     UINT16_MAX,
     8655,
     UINT16_MAX,
     8655,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     9848,
     1658,
     UINT16_MAX,
     1658,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     9851,
     UINT16_MAX,
     8656,
     UINT16_MAX,
     8656,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     9853,
     1663,
     UINT16_MAX,
     1663,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     9856,
     UINT16_MAX,
     8657,
     UINT16_MAX,
     8657,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     9858,
     1668,
     UINT16_MAX,
     1668,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     9861,
     UINT16_MAX,
     8658,
     UINT16_MAX,
     8658,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     9863,
     1673,
     UINT16_MAX,
     1673,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     9866,
     UINT16_MAX,
     8659,
     UINT16_MAX,
     8659,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     9868,
     1678,
     UINT16_MAX,
     1678,
     UINT16_MAX,
     2884,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     9871,
     UINT16_MAX,
     8660,
     UINT16_MAX,
     8660,
     2887,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     9873,
     1683,
     UINT16_MAX,
     1683,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     9876,
     UINT16_MAX,
     8661,
     UINT16_MAX,
     8661,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     9878,
     1688,
     UINT16_MAX,
     1688,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     9881,
     UINT16_MAX,
     8662,
     UINT16_MAX,
     8662,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     9883,
     1693,
     UINT16_MAX,
     1693,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     9886,
     UINT16_MAX,
     8663,
     UINT16_MAX,
     8663,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     9888,
     1698,
     UINT16_MAX,
     1698,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     9891,
     UINT16_MAX,
     8664,
     UINT16_MAX,
     8664,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     9893,
     1703,
     UINT16_MAX,
     1703,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     9896,
     UINT16_MAX,
     8665,
     UINT16_MAX,
     8665,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     9898,
     1708,
     UINT16_MAX,
     1708,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     9901,
     UINT16_MAX,
     8666,
     UINT16_MAX,
     8666,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     9903,
     1713,
     UINT16_MAX,
     1713,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     9906,
     UINT16_MAX,
     8667,
     UINT16_MAX,
     8667,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     9908,
     1718,
     UINT16_MAX,
     1718,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     9911,
     UINT16_MAX,
     8668,
     UINT16_MAX,
     8668,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     9913,
     1723,
     UINT16_MAX,
     1723,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     9916,
     UINT16_MAX,
     8669,
     UINT16_MAX,
     8669,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     9918,
     1728,
     UINT16_MAX,
     1728,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     9921,
     UINT16_MAX,
     8670,
     UINT16_MAX,
     8670,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     9923,
     1733,
     UINT16_MAX,
     1733,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     9926,
     UINT16_MAX,
     8671,
     UINT16_MAX,
     8671,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     9928,
     1738,
     UINT16_MAX,
     1738,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     9931,
     UINT16_MAX,
     8672,
     UINT16_MAX,
     8672,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     9933,
     1743,
     UINT16_MAX,
     1743,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     9936,
     UINT16_MAX,
     8673,
     UINT16_MAX,
     8673,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     9938,
     1748,
     UINT16_MAX,
     1748,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     9941,
     UINT16_MAX,
     8674,
     UINT16_MAX,
     8674,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     9943,
     1753,
     UINT16_MAX,
     1753,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     9946,
     UINT16_MAX,
     8675,
     UINT16_MAX,
     8675,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     9948,
     1758,
     UINT16_MAX,
     1758,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     9951,
     UINT16_MAX,
     8676,
     UINT16_MAX,
     8676,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     9953,
     1763,
     UINT16_MAX,
     1763,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     9956,
     UINT16_MAX,
     8677,
     UINT16_MAX,
     8677,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     9958,
     1768,
     UINT16_MAX,
     1768,
     UINT16_MAX,
     3006,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     9961,
     UINT16_MAX,
     8678,
     UINT16_MAX,
     8678,
     3009,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     9963,
     1773,
     UINT16_MAX,
     1773,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     9966,
     UINT16_MAX,
     8679,
     UINT16_MAX,
     8679,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     9968,
     1778,
     UINT16_MAX,
     1778,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     9971,
     UINT16_MAX,
     8680,
     UINT16_MAX,
     8680,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     9973,
     1783,
     UINT16_MAX,
     1783,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     9976,
     UINT16_MAX,
     8681,
     UINT16_MAX,
     8681,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     9978,
     1788,
     UINT16_MAX,
     1788,
     UINT16_MAX,
     3024,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     9981,
     UINT16_MAX,
     8682,
     UINT16_MAX,
     8682,
     3027,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     9983,
     1793,
     UINT16_MAX,
     1793,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     9986,
     UINT16_MAX,
     8683,
     UINT16_MAX,
     8683,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     9988,
     1798,
     UINT16_MAX,
     1798,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     9991,
     UINT16_MAX,
     8684,
     UINT16_MAX,
     8684,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     9993,
     1803,
     UINT16_MAX,
     1803,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     9996,
     UINT16_MAX,
     8685,
     UINT16_MAX,
     8685,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     9998,
     1808,
     UINT16_MAX,
     1808,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10001,
     UINT16_MAX,
     8686,
     UINT16_MAX,
     8686,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10003,
     1813,
     UINT16_MAX,
     1813,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10006,
     UINT16_MAX,
     8687,
     UINT16_MAX,
     8687,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10008,
     1818,
     UINT16_MAX,
     1818,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10011,
     UINT16_MAX,
     8688,
     UINT16_MAX,
     8688,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10013,
     1823,
     UINT16_MAX,
     1823,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10016,
     UINT16_MAX,
     8689,
     UINT16_MAX,
     8689,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10018,
     1828,
     UINT16_MAX,
     1828,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10021,
     UINT16_MAX,
     8690,
     UINT16_MAX,
     8690,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10023,
     1833,
     UINT16_MAX,
     1833,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10026,
     UINT16_MAX,
     8691,
     UINT16_MAX,
     8691,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10028,
     1838,
     UINT16_MAX,
     1838,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10031,
     UINT16_MAX,
     8692,
     UINT16_MAX,
     8692,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10033,
     1843,
     UINT16_MAX,
     1843,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10036,
     UINT16_MAX,
     8693,
     UINT16_MAX,
     8693,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10038,
     1848,
     UINT16_MAX,
     1848,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10041,
     UINT16_MAX,
     8694,
     UINT16_MAX,
     8694,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10043,
     1853,
     UINT16_MAX,
     1853,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10046,
     UINT16_MAX,
     8695,
     UINT16_MAX,
     8695,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10048,
     1858,
     UINT16_MAX,
     1858,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10051,
     UINT16_MAX,
     8696,
     UINT16_MAX,
     8696,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10053,
     1863,
     UINT16_MAX,
     1863,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10056,
     UINT16_MAX,
     8697,
     UINT16_MAX,
     8697,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10058,
     1868,
     UINT16_MAX,
     1868,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10061,
     UINT16_MAX,
     8698,
     UINT16_MAX,
     8698,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10063,
     1873,
     UINT16_MAX,
     1873,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10066,
     UINT16_MAX,
     8699,
     UINT16_MAX,
     8699,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10068,
     1878,
     UINT16_MAX,
     1878,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10071,
     UINT16_MAX,
     8700,
     UINT16_MAX,
     8700,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10073,
     1883,
     UINT16_MAX,
     1883,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10076,
     UINT16_MAX,
     8701,
     UINT16_MAX,
     8701,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10078,
     1888,
     UINT16_MAX,
     1888,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10081,
     UINT16_MAX,
     8702,
     UINT16_MAX,
     8702,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10083,
     1893,
     UINT16_MAX,
     1893,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10086,
     UINT16_MAX,
     8703,
     UINT16_MAX,
     8703,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10088,
     1898,
     UINT16_MAX,
     1898,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10091,
     UINT16_MAX,
     8704,
     UINT16_MAX,
     8704,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10093,
     1903,
     UINT16_MAX,
     1903,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10096,
     UINT16_MAX,
     8705,
     UINT16_MAX,
     8705,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10098,
     1908,
     UINT16_MAX,
     1908,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10101,
     UINT16_MAX,
     8706,
     UINT16_MAX,
     8706,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10103,
     1913,
     UINT16_MAX,
     1913,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10106,
     UINT16_MAX,
     8707,
     UINT16_MAX,
     8707,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10108,
     10108,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10110,
     10110,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10112,
     10112,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10114,
     10114,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     10116,
     10116,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10118,
     1783,
     8681,
     UINT16_MAX,
     8681,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     8322,
     UINT16_MAX,
     8708,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10120,
     1930,
     UINT16_MAX,
     1930,
     UINT16_MAX,
     3241,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10123,
     UINT16_MAX,
     8709,
     UINT16_MAX,
     8709,
     3250,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10125,
     1935,
     UINT16_MAX,
     1935,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10128,
     UINT16_MAX,
     8710,
     UINT16_MAX,
     8710,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10130,
     1940,
     UINT16_MAX,
     1940,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10133,
     UINT16_MAX,
     8711,
     UINT16_MAX,
     8711,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10135,
     1945,
     UINT16_MAX,
     1945,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10138,
     UINT16_MAX,
     8712,
     UINT16_MAX,
     8712,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10140,
     1950,
     UINT16_MAX,
     1950,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10143,
     UINT16_MAX,
     8713,
     UINT16_MAX,
     8713,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10145,
     1955,
     UINT16_MAX,
     1955,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10148,
     UINT16_MAX,
     8714,
     UINT16_MAX,
     8714,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10150,
     1960,
     UINT16_MAX,
     1960,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10153,
     UINT16_MAX,
     8715,
     UINT16_MAX,
     8715,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10155,
     1965,
     UINT16_MAX,
     1965,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10158,
     UINT16_MAX,
     8716,
     UINT16_MAX,
     8716,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10160,
     1970,
     UINT16_MAX,
     1970,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10163,
     UINT16_MAX,
     8717,
     UINT16_MAX,
     8717,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10165,
     1975,
     UINT16_MAX,
     1975,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10168,
     UINT16_MAX,
     8718,
     UINT16_MAX,
     8718,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10170,
     1980,
     UINT16_MAX,
     1980,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10173,
     UINT16_MAX,
     8719,
     UINT16_MAX,
     8719,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10175,
     1985,
     UINT16_MAX,
     1985,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10178,
     UINT16_MAX,
     8720,
     UINT16_MAX,
     8720,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10180,
     1990,
     UINT16_MAX,
     1990,
     UINT16_MAX,
     3455,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10183,
     UINT16_MAX,
     8721,
     UINT16_MAX,
     8721,
     3458,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10185,
     1995,
     UINT16_MAX,
     1995,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10188,
     UINT16_MAX,
     8722,
     UINT16_MAX,
     8722,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10190,
     2000,
     UINT16_MAX,
     2000,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10193,
     UINT16_MAX,
     8723,
     UINT16_MAX,
     8723,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10195,
     2005,
     UINT16_MAX,
     2005,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10198,
     UINT16_MAX,
     8724,
     UINT16_MAX,
     8724,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10200,
     2010,
     UINT16_MAX,
     2010,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10203,
     UINT16_MAX,
     8725,
     UINT16_MAX,
     8725,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10205,
     2015,
     UINT16_MAX,
     2015,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10208,
     UINT16_MAX,
     8726,
     UINT16_MAX,
     8726,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10210,
     2020,
     UINT16_MAX,
     2020,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10213,
     UINT16_MAX,
     8727,
     UINT16_MAX,
     8727,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10215,
     2025,
     UINT16_MAX,
     2025,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10218,
     UINT16_MAX,
     8728,
     UINT16_MAX,
     8728,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10220,
     2030,
     UINT16_MAX,
     2030,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10223,
     UINT16_MAX,
     8729,
     UINT16_MAX,
     8729,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10225,
     2035,
     UINT16_MAX,
     2035,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10228,
     UINT16_MAX,
     8730,
     UINT16_MAX,
     8730,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10230,
     2040,
     UINT16_MAX,
     2040,
     UINT16_MAX,
     3559,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10233,
     UINT16_MAX,
     8731,
     UINT16_MAX,
     8731,
     3562,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10235,
     2045,
     UINT16_MAX,
     2045,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10238,
     UINT16_MAX,
     8732,
     UINT16_MAX,
     8732,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10240,
     2050,
     UINT16_MAX,
     2050,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10243,
     UINT16_MAX,
     8733,
     UINT16_MAX,
     8733,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10245,
     2055,
     UINT16_MAX,
     2055,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10248,
     UINT16_MAX,
     8734,
     UINT16_MAX,
     8734,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10250,
     2060,
     UINT16_MAX,
     2060,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10253,
     UINT16_MAX,
     8735,
     UINT16_MAX,
     8735,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10255,
     2065,
     UINT16_MAX,
     2065,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10258,
     UINT16_MAX,
     8736,
     UINT16_MAX,
     8736,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10260,
     2070,
     UINT16_MAX,
     2070,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10263,
     UINT16_MAX,
     8737,
     UINT16_MAX,
     8737,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10265,
     2075,
     UINT16_MAX,
     2075,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10268,
     UINT16_MAX,
     8738,
     UINT16_MAX,
     8738,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10270,
     2080,
     UINT16_MAX,
     2080,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10273,
     UINT16_MAX,
     8739,
     UINT16_MAX,
     8739,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10275,
     2085,
     UINT16_MAX,
     2085,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10278,
     UINT16_MAX,
     8740,
     UINT16_MAX,
     8740,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10280,
     2090,
     UINT16_MAX,
     2090,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10283,
     UINT16_MAX,
     8741,
     UINT16_MAX,
     8741,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10285,
     2095,
     UINT16_MAX,
     2095,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10288,
     UINT16_MAX,
     8742,
     UINT16_MAX,
     8742,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10290,
     2100,
     UINT16_MAX,
     2100,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10293,
     UINT16_MAX,
     8743,
     UINT16_MAX,
     8743,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10295,
     2105,
     UINT16_MAX,
     2105,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10298,
     UINT16_MAX,
     8744,
     UINT16_MAX,
     8744,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10300,
     2110,
     UINT16_MAX,
     2110,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10303,
     UINT16_MAX,
     8745,
     UINT16_MAX,
     8745,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10305,
     2115,
     UINT16_MAX,
     2115,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10308,
     UINT16_MAX,
     8746,
     UINT16_MAX,
     8746,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10310,
     2120,
     UINT16_MAX,
     2120,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10313,
     UINT16_MAX,
     8747,
     UINT16_MAX,
     8747,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10315,
     2125,
     UINT16_MAX,
     2125,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10318,
     UINT16_MAX,
     8748,
     UINT16_MAX,
     8748,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10320,
     2130,
     UINT16_MAX,
     2130,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10323,
     UINT16_MAX,
     8749,
     UINT16_MAX,
     8749,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10325,
     2135,
     UINT16_MAX,
     2135,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10328,
     UINT16_MAX,
     8750,
     UINT16_MAX,
     8750,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10330,
     2140,
     UINT16_MAX,
     2140,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10333,
     UINT16_MAX,
     8751,
     UINT16_MAX,
     8751,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10335,
     2145,
     UINT16_MAX,
     2145,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10338,
     UINT16_MAX,
     8752,
     UINT16_MAX,
     8752,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10340,
     2150,
     UINT16_MAX,
     2150,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10343,
     UINT16_MAX,
     8753,
     UINT16_MAX,
     8753,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     2153,
     UINT16_MAX,
     2153,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8754,
     UINT16_MAX,
     8754,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     2154,
     UINT16_MAX,
     2154,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8755,
     UINT16_MAX,
     8755,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     2155,
     UINT16_MAX,
     2155,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8756,
     UINT16_MAX,
     8756,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10348,
     UINT16_MAX,
     8757,
     UINT16_MAX,
     8757,
     3761,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10350,
     UINT16_MAX,
     8758,
     UINT16_MAX,
     8758,
     3814,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10352,
     UINT16_MAX,
     8759,
     UINT16_MAX,
     8759,
     4793,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10354,
     UINT16_MAX,
     8760,
     UINT16_MAX,
     8760,
     4796,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10356,
     UINT16_MAX,
     8761,
     UINT16_MAX,
     8761,
     4799,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10358,
     UINT16_MAX,
     8762,
     UINT16_MAX,
     8762,
     4802,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10360,
     UINT16_MAX,
     8763,
     UINT16_MAX,
     8763,
     4805,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10362,
     UINT16_MAX,
     8764,
     UINT16_MAX,
     8764,
     4808,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10364,
     2174,
     UINT16_MAX,
     2174,
     UINT16_MAX,
     3867,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10367,
     2177,
     UINT16_MAX,
     2177,
     UINT16_MAX,
     3920,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10370,
     2180,
     UINT16_MAX,
     2180,
     UINT16_MAX,
     4811,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10373,
     2183,
     UINT16_MAX,
     2183,
     UINT16_MAX,
     4814,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10376,
     2186,
     UINT16_MAX,
     2186,
     UINT16_MAX,
     4817,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10379,
     2189,
     UINT16_MAX,
     2189,
     UINT16_MAX,
     4820,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10382,
     2192,
     UINT16_MAX,
     2192,
     UINT16_MAX,
     4823,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10385,
     2195,
     UINT16_MAX,
     2195,
     UINT16_MAX,
     4826,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10388,
     UINT16_MAX,
     8765,
     UINT16_MAX,
     8765,
     3973,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10390,
     UINT16_MAX,
     8766,
     UINT16_MAX,
     8766,
     3977,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10392,
     UINT16_MAX,
     8767,
     UINT16_MAX,
     8767,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10394,
     UINT16_MAX,
     8768,
     UINT16_MAX,
     8768,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10396,
     UINT16_MAX,
     8769,
     UINT16_MAX,
     8769,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10398,
     UINT16_MAX,
     8770,
     UINT16_MAX,
     8770,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10400,
     2210,
     UINT16_MAX,
     2210,
     UINT16_MAX,
     3981,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10403,
     2213,
     UINT16_MAX,
     2213,
     UINT16_MAX,
     3985,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10406,
     2216,
     UINT16_MAX,
     2216,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10409,
     2219,
     UINT16_MAX,
     2219,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10412,
     2222,
     UINT16_MAX,
     2222,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10415,
     2225,
     UINT16_MAX,
     2225,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10418,
     UINT16_MAX,
     8771,
     UINT16_MAX,
     8771,
     3989,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10420,
     UINT16_MAX,
     8772,
     UINT16_MAX,
     8772,
     4042,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10422,
     UINT16_MAX,
     8773,
     UINT16_MAX,
     8773,
     4829,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10424,
     UINT16_MAX,
     8774,
     UINT16_MAX,
     8774,
     4832,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10426,
     UINT16_MAX,
     8775,
     UINT16_MAX,
     8775,
     4835,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10428,
     UINT16_MAX,
     8776,
     UINT16_MAX,
     8776,
     4838,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10430,
     UINT16_MAX,
     8777,
     UINT16_MAX,
     8777,
     4841,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10432,
     UINT16_MAX,
     8778,
     UINT16_MAX,
     8778,
     4844,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10434,
     2244,
     UINT16_MAX,
     2244,
     UINT16_MAX,
     4095,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10437,
     2247,
     UINT16_MAX,
     2247,
     UINT16_MAX,
     4148,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10440,
     2250,
     UINT16_MAX,
     2250,
     UINT16_MAX,
     4847,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10443,
     2253,
     UINT16_MAX,
     2253,
     UINT16_MAX,
     4850,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10446,
     2256,
     UINT16_MAX,
     2256,
     UINT16_MAX,
     4853,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10449,
     2259,
     UINT16_MAX,
     2259,
     UINT16_MAX,
     4856,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10452,
     2262,
     UINT16_MAX,
     2262,
     UINT16_MAX,
     4859,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10455,
     2265,
     UINT16_MAX,
     2265,
     UINT16_MAX,
     4862,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10458,
     UINT16_MAX,
     8779,
     UINT16_MAX,
     8779,
     4201,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10460,
     UINT16_MAX,
     8780,
     UINT16_MAX,
     8780,
     4253,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10462,
     UINT16_MAX,
     8781,
     UINT16_MAX,
     8781,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10464,
     UINT16_MAX,
     8782,
     UINT16_MAX,
     8782,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10466,
     UINT16_MAX,
     8783,
     UINT16_MAX,
     8783,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10468,
     UINT16_MAX,
     8784,
     UINT16_MAX,
     8784,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10470,
     UINT16_MAX,
     8785,
     UINT16_MAX,
     8785,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10472,
     UINT16_MAX,
     8786,
     UINT16_MAX,
     8786,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10474,
     2284,
     UINT16_MAX,
     2284,
     UINT16_MAX,
     4305,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10477,
     2287,
     UINT16_MAX,
     2287,
     UINT16_MAX,
     4357,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10480,
     2290,
     UINT16_MAX,
     2290,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10483,
     2293,
     UINT16_MAX,
     2293,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10486,
     2296,
     UINT16_MAX,
     2296,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10489,
     2299,
     UINT16_MAX,
     2299,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10492,
     2302,
     UINT16_MAX,
     2302,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10495,
     2305,
     UINT16_MAX,
     2305,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10498,
     UINT16_MAX,
     8787,
     UINT16_MAX,
     8787,
     4409,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10500,
     UINT16_MAX,
     8788,
     UINT16_MAX,
     8788,
     4413,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10502,
     UINT16_MAX,
     8789,
     UINT16_MAX,
     8789,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10504,
     UINT16_MAX,
     8790,
     UINT16_MAX,
     8790,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10506,
     UINT16_MAX,
     8791,
     UINT16_MAX,
     8791,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10508,
     UINT16_MAX,
     8792,
     UINT16_MAX,
     8792,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10510,
     2320,
     UINT16_MAX,
     2320,
     UINT16_MAX,
     4417,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10513,
     2323,
     UINT16_MAX,
     2323,
     UINT16_MAX,
     4421,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10516,
     2326,
     UINT16_MAX,
     2326,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10519,
     2329,
     UINT16_MAX,
     2329,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10522,
     2332,
     UINT16_MAX,
     2332,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10525,
     2335,
     UINT16_MAX,
     2335,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10528,
     10528,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     4425,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10530,
     UINT16_MAX,
     8793,
     UINT16_MAX,
     8793,
     4477,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10532,
     18726,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10537,
     UINT16_MAX,
     8794,
     UINT16_MAX,
     8794,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10539,
     18733,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10544,
     UINT16_MAX,
     8795,
     UINT16_MAX,
     8795,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10546,
     18740,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10551,
     UINT16_MAX,
     8796,
     UINT16_MAX,
     8796,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10553,
     2363,
     UINT16_MAX,
     2363,
     UINT16_MAX,
     4529,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10556,
     2366,
     UINT16_MAX,
     2366,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10559,
     2369,
     UINT16_MAX,
     2369,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10562,
     2372,
     UINT16_MAX,
     2372,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10565,
     UINT16_MAX,
     8797,
     UINT16_MAX,
     8797,
     4581,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10567,
     UINT16_MAX,
     8798,
     UINT16_MAX,
     8798,
     4634,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10569,
     UINT16_MAX,
     8799,
     UINT16_MAX,
     8799,
     4865,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10571,
     UINT16_MAX,
     8800,
     UINT16_MAX,
     8800,
     4868,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10573,
     UINT16_MAX,
     8801,
     UINT16_MAX,
     8801,
     4871,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10575,
     UINT16_MAX,
     8802,
     UINT16_MAX,
     8802,
     4874,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10577,
     UINT16_MAX,
     8803,
     UINT16_MAX,
     8803,
     4877,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10579,
     UINT16_MAX,
     8804,
     UINT16_MAX,
     8804,
     4880,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10581,
     2391,
     UINT16_MAX,
     2391,
     UINT16_MAX,
     4687,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10584,
     2394,
     UINT16_MAX,
     2394,
     UINT16_MAX,
     4740,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10587,
     2397,
     UINT16_MAX,
     2397,
     UINT16_MAX,
     4883,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10590,
     2400,
     UINT16_MAX,
     2400,
     UINT16_MAX,
     4886,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10593,
     2403,
     UINT16_MAX,
     2403,
     UINT16_MAX,
     4889,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10596,
     2406,
     UINT16_MAX,
     2406,
     UINT16_MAX,
     4892,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10599,
     2409,
     UINT16_MAX,
     2409,
     UINT16_MAX,
     4895,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10602,
     2412,
     UINT16_MAX,
     2412,
     UINT16_MAX,
     4898,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10605,
     UINT16_MAX,
     8805,
     UINT16_MAX,
     8805,
     4901,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     820,
     UINT16_MAX,
     8806,
     UINT16_MAX,
     8806,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10607,
     UINT16_MAX,
     8807,
     UINT16_MAX,
     8807,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     824,
     UINT16_MAX,
     8808,
     UINT16_MAX,
     8808,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10609,
     UINT16_MAX,
     8809,
     UINT16_MAX,
     8809,
     4910,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     827,
     UINT16_MAX,
     8810,
     UINT16_MAX,
     8810,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10611,
     UINT16_MAX,
     8811,
     UINT16_MAX,
     8811,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     830,
     UINT16_MAX,
     8812,
     UINT16_MAX,
     8812,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10613,
     UINT16_MAX,
     8813,
     UINT16_MAX,
     8813,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     833,
     UINT16_MAX,
     8814,
     UINT16_MAX,
     8814,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10615,
     UINT16_MAX,
     8815,
     UINT16_MAX,
     8815,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     836,
     UINT16_MAX,
     8816,
     UINT16_MAX,
     8816,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10617,
     UINT16_MAX,
     8817,
     UINT16_MAX,
     8817,
     5030,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     839,
     UINT16_MAX,
     8818,
     UINT16_MAX,
     8818,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10619,
     10621,
     8819,
     UINT16_MAX,
     8819,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10623,
     10625,
     8820,
     UINT16_MAX,
     8820,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10627,
     10629,
     8821,
     UINT16_MAX,
     8821,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10631,
     10633,
     8822,
     UINT16_MAX,
     8822,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10635,
     10637,
     8823,
     UINT16_MAX,
     8823,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10639,
     10641,
     8824,
     UINT16_MAX,
     8824,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10643,
     10645,
     8825,
     UINT16_MAX,
     8825,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10647,
     10649,
     8826,
     UINT16_MAX,
     8826,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LT,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10651,
     10621,
     UINT16_MAX,
     8827,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LT,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10653,
     10625,
     UINT16_MAX,
     8828,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LT,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10655,
     10629,
     UINT16_MAX,
     8829,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LT,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10657,
     10633,
     UINT16_MAX,
     8830,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LT,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10659,
     10637,
     UINT16_MAX,
     8831,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LT,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10661,
     10641,
     UINT16_MAX,
     8832,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LT,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10663,
     10645,
     UINT16_MAX,
     8833,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LT,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10665,
     10649,
     UINT16_MAX,
     8834,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10667,
     10669,
     8835,
     UINT16_MAX,
     8835,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10671,
     10673,
     8836,
     UINT16_MAX,
     8836,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10675,
     10677,
     8837,
     UINT16_MAX,
     8837,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10679,
     10681,
     8838,
     UINT16_MAX,
     8838,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10683,
     10685,
     8839,
     UINT16_MAX,
     8839,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10687,
     10689,
     8840,
     UINT16_MAX,
     8840,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10691,
     10693,
     8841,
     UINT16_MAX,
     8841,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10695,
     10697,
     8842,
     UINT16_MAX,
     8842,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LT,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10699,
     10669,
     UINT16_MAX,
     8843,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LT,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10701,
     10673,
     UINT16_MAX,
     8844,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LT,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10703,
     10677,
     UINT16_MAX,
     8845,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LT,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10705,
     10681,
     UINT16_MAX,
     8846,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LT,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10707,
     10685,
     UINT16_MAX,
     8847,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LT,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10709,
     10689,
     UINT16_MAX,
     8848,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LT,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10711,
     10693,
     UINT16_MAX,
     8849,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LT,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10713,
     10697,
     UINT16_MAX,
     8850,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10715,
     10717,
     8851,
     UINT16_MAX,
     8851,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10719,
     10721,
     8852,
     UINT16_MAX,
     8852,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10723,
     10725,
     8853,
     UINT16_MAX,
     8853,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10727,
     10729,
     8854,
     UINT16_MAX,
     8854,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10731,
     10733,
     8855,
     UINT16_MAX,
     8855,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10735,
     10737,
     8856,
     UINT16_MAX,
     8856,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10739,
     10741,
     8857,
     UINT16_MAX,
     8857,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10743,
     10745,
     8858,
     UINT16_MAX,
     8858,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LT,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10747,
     10717,
     UINT16_MAX,
     8859,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LT,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10749,
     10721,
     UINT16_MAX,
     8860,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LT,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10751,
     10725,
     UINT16_MAX,
     8861,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LT,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10753,
     10729,
     UINT16_MAX,
     8862,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LT,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10755,
     10733,
     UINT16_MAX,
     8863,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LT,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10757,
     10737,
     UINT16_MAX,
     8864,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LT,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10759,
     10741,
     UINT16_MAX,
     8865,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LT,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10761,
     10745,
     UINT16_MAX,
     8866,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10763,
     UINT16_MAX,
     8867,
     UINT16_MAX,
     8867,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10765,
     UINT16_MAX,
     8868,
     UINT16_MAX,
     8868,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10767,
     10769,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10771,
     10773,
     8869,
     UINT16_MAX,
     8869,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10775,
     10777,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10779,
     10779,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     4907,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10781,
     18975,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10786,
     2596,
     UINT16_MAX,
     2596,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10789,
     2599,
     UINT16_MAX,
     2599,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10792,
     2602,
     UINT16_MAX,
     2602,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     2603,
     2604,
     UINT16_MAX,
     2604,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LT,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10797,
     10773,
     UINT16_MAX,
     8870,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SK,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     10799,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     807,
     807,
     7225,
     UINT16_MAX,
     7225,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SK,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     10799,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     4919,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SK,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     10801,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SK,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     0,
     10803,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10805,
     10807,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10809,
     10811,
     8871,
     UINT16_MAX,
     8871,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10813,
     10815,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10817,
     10817,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     4916,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10819,
     19013,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10824,
     2634,
     UINT16_MAX,
     2634,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     2635,
     2636,
     UINT16_MAX,
     2636,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10829,
     2639,
     UINT16_MAX,
     2639,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     2640,
     2641,
     UINT16_MAX,
     2641,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LT,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10834,
     10811,
     UINT16_MAX,
     8872,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SK,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     0,
     10836,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SK,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     0,
     10838,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SK,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     0,
     10840,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10842,
     UINT16_MAX,
     8873,
     UINT16_MAX,
     8873,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10844,
     UINT16_MAX,
     8874,
     UINT16_MAX,
     8874,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10846,
     19040,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     2659,
     17226,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10852,
     10852,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10854,
     19048,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10859,
     2669,
     UINT16_MAX,
     2669,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10862,
     2672,
     UINT16_MAX,
     2672,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10865,
     2675,
     UINT16_MAX,
     2675,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     2676,
     2677,
     UINT16_MAX,
     2677,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SK,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     0,
     10870,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SK,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     0,
     10872,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SK,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     0,
     10874,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10876,
     UINT16_MAX,
     8875,
     UINT16_MAX,
     8875,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10878,
     UINT16_MAX,
     8876,
     UINT16_MAX,
     8876,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10880,
     19074,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     2693,
     17267,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10886,
     10886,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10888,
     UINT16_MAX,
     8877,
     UINT16_MAX,
     8877,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10890,
     10890,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10892,
     19086,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10897,
     2707,
     UINT16_MAX,
     2707,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10900,
     2710,
     UINT16_MAX,
     2710,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10903,
     2713,
     UINT16_MAX,
     2713,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     2714,
     2715,
     UINT16_MAX,
     2715,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10908,
     2718,
     UINT16_MAX,
     2718,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SK,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     0,
     10911,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SK,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     0,
     2721,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SK,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     0,
     2722,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10915,
     10917,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10919,
     10921,
     8878,
     UINT16_MAX,
     8878,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10923,
     10925,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10927,
     10927,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     5036,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10929,
     19123,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10934,
     2744,
     UINT16_MAX,
     2744,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     2745,
     2746,
     UINT16_MAX,
     2746,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10939,
     2749,
     UINT16_MAX,
     2749,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     2750,
     2751,
     UINT16_MAX,
     2751,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LT,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     10944,
     10921,
     UINT16_MAX,
     8879,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SK,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     0,
     2754,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SK,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     10947,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     4971,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_ZS,
     0,
     UTF8PROC_BIDI_CLASS_WS,
     0,
     2757,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_ZS,
     0,
     UTF8PROC_BIDI_CLASS_WS,
     0,
     2758,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_ZS,
     0,
     UTF8PROC_BIDI_CLASS_WS,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     26,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_ZS,
     0,
     UTF8PROC_BIDI_CLASS_WS,
     UTF8PROC_DECOMP_TYPE_NOBREAK,
     26,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_CF,
     0,
     UTF8PROC_BIDI_CLASS_BN,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     true,
     true,
     0,
     0,
     UTF8PROC_BOUNDCLASS_EXTEND},
    {UTF8PROC_CATEGORY_CF,
     0,
     UTF8PROC_BIDI_CLASS_BN,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     true,
     true,
     0,
     0,
     UTF8PROC_BOUNDCLASS_ZWJ},
    {UTF8PROC_CATEGORY_CF,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     true,
     true,
     0,
     0,
     UTF8PROC_BOUNDCLASS_CONTROL},
    {UTF8PROC_CATEGORY_CF,
     0,
     UTF8PROC_BIDI_CLASS_R,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     true,
     true,
     0,
     0,
     UTF8PROC_BOUNDCLASS_CONTROL},
    {UTF8PROC_CATEGORY_PD,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_NOBREAK,
     2759,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_PO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     10952,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_PI,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_PF,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_PS,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_PO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     2762,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_PO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     10955,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_PO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     19149,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_ZL,
     0,
     UTF8PROC_BIDI_CLASS_WS,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     true,
     0,
     0,
     UTF8PROC_BOUNDCLASS_CONTROL},
    {UTF8PROC_CATEGORY_ZP,
     0,
     UTF8PROC_BIDI_CLASS_B,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     true,
     0,
     0,
     UTF8PROC_BOUNDCLASS_CONTROL},
    {UTF8PROC_CATEGORY_CF,
     0,
     UTF8PROC_BIDI_CLASS_LRE,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     true,
     true,
     0,
     0,
     UTF8PROC_BOUNDCLASS_CONTROL},
    {UTF8PROC_CATEGORY_CF,
     0,
     UTF8PROC_BIDI_CLASS_RLE,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     true,
     true,
     0,
     0,
     UTF8PROC_BOUNDCLASS_CONTROL},
    {UTF8PROC_CATEGORY_CF,
     0,
     UTF8PROC_BIDI_CLASS_PDF,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     true,
     true,
     0,
     0,
     UTF8PROC_BOUNDCLASS_CONTROL},
    {UTF8PROC_CATEGORY_CF,
     0,
     UTF8PROC_BIDI_CLASS_LRO,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     true,
     true,
     0,
     0,
     UTF8PROC_BOUNDCLASS_CONTROL},
    {UTF8PROC_CATEGORY_CF,
     0,
     UTF8PROC_BIDI_CLASS_RLO,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     true,
     true,
     0,
     0,
     UTF8PROC_BOUNDCLASS_CONTROL},
    {UTF8PROC_CATEGORY_PO,
     0,
     UTF8PROC_BIDI_CLASS_ET,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     10960,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_PO,
     0,
     UTF8PROC_BIDI_CLASS_ET,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     19154,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_PO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     10965,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_PO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     19159,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_PO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     10970,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_EXTENDED_PICTOGRAPHIC},
    {UTF8PROC_CATEGORY_PO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     10972,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SM,
     0,
     UTF8PROC_BIDI_CLASS_CS,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_PO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     10974,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_PO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     10976,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_PO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     10978,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_EXTENDED_PICTOGRAPHIC},
    {UTF8PROC_CATEGORY_PO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     27364,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_CF,
     0,
     UTF8PROC_BIDI_CLASS_LRI,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     true,
     true,
     0,
     0,
     UTF8PROC_BOUNDCLASS_CONTROL},
    {UTF8PROC_CATEGORY_CF,
     0,
     UTF8PROC_BIDI_CLASS_RLI,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     true,
     true,
     0,
     0,
     UTF8PROC_BOUNDCLASS_CONTROL},
    {UTF8PROC_CATEGORY_CF,
     0,
     UTF8PROC_BIDI_CLASS_FSI,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     true,
     true,
     0,
     0,
     UTF8PROC_BOUNDCLASS_CONTROL},
    {UTF8PROC_CATEGORY_CF,
     0,
     UTF8PROC_BIDI_CLASS_PDI,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     true,
     true,
     0,
     0,
     UTF8PROC_BOUNDCLASS_CONTROL},
    {UTF8PROC_CATEGORY_NO,
     0,
     UTF8PROC_BIDI_CLASS_EN,
     UTF8PROC_DECOMP_TYPE_SUPER,
     2792,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LM,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SUPER,
     8,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_NO,
     0,
     UTF8PROC_BIDI_CLASS_EN,
     UTF8PROC_DECOMP_TYPE_SUPER,
     2793,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_NO,
     0,
     UTF8PROC_BIDI_CLASS_EN,
     UTF8PROC_DECOMP_TYPE_SUPER,
     2794,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_NO,
     0,
     UTF8PROC_BIDI_CLASS_EN,
     UTF8PROC_DECOMP_TYPE_SUPER,
     2795,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_NO,
     0,
     UTF8PROC_BIDI_CLASS_EN,
     UTF8PROC_DECOMP_TYPE_SUPER,
     2796,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_NO,
     0,
     UTF8PROC_BIDI_CLASS_EN,
     UTF8PROC_DECOMP_TYPE_SUPER,
     2797,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_NO,
     0,
     UTF8PROC_BIDI_CLASS_EN,
     UTF8PROC_DECOMP_TYPE_SUPER,
     2798,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SM,
     0,
     UTF8PROC_BIDI_CLASS_ES,
     UTF8PROC_DECOMP_TYPE_SUPER,
     2799,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SM,
     0,
     UTF8PROC_BIDI_CLASS_ES,
     UTF8PROC_DECOMP_TYPE_SUPER,
     2800,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SM,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_SUPER,
     2801,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_PS,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_SUPER,
     2802,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     true,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_PE,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_SUPER,
     2803,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     true,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LM,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SUPER,
     13,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_NO,
     0,
     UTF8PROC_BIDI_CLASS_EN,
     UTF8PROC_DECOMP_TYPE_SUB,
     2792,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_NO,
     0,
     UTF8PROC_BIDI_CLASS_EN,
     UTF8PROC_DECOMP_TYPE_SUB,
     38,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_NO,
     0,
     UTF8PROC_BIDI_CLASS_EN,
     UTF8PROC_DECOMP_TYPE_SUB,
     31,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_NO,
     0,
     UTF8PROC_BIDI_CLASS_EN,
     UTF8PROC_DECOMP_TYPE_SUB,
     32,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_NO,
     0,
     UTF8PROC_BIDI_CLASS_EN,
     UTF8PROC_DECOMP_TYPE_SUB,
     2793,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_NO,
     0,
     UTF8PROC_BIDI_CLASS_EN,
     UTF8PROC_DECOMP_TYPE_SUB,
     2794,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_NO,
     0,
     UTF8PROC_BIDI_CLASS_EN,
     UTF8PROC_DECOMP_TYPE_SUB,
     2795,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_NO,
     0,
     UTF8PROC_BIDI_CLASS_EN,
     UTF8PROC_DECOMP_TYPE_SUB,
     2796,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_NO,
     0,
     UTF8PROC_BIDI_CLASS_EN,
     UTF8PROC_DECOMP_TYPE_SUB,
     2797,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_NO,
     0,
     UTF8PROC_BIDI_CLASS_EN,
     UTF8PROC_DECOMP_TYPE_SUB,
     2798,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SM,
     0,
     UTF8PROC_BIDI_CLASS_ES,
     UTF8PROC_DECOMP_TYPE_SUB,
     2799,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SM,
     0,
     UTF8PROC_BIDI_CLASS_ES,
     UTF8PROC_DECOMP_TYPE_SUB,
     2800,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SM,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_SUB,
     2801,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_PS,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_SUB,
     2802,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     true,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_PE,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_SUB,
     2803,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     true,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LM,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SUB,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LM,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SUB,
     4,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LM,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SUB,
     14,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LM,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SUB,
     23,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LM,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SUB,
     485,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LM,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SUB,
     7,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LM,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SUB,
     10,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LM,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SUB,
     11,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LM,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SUB,
     12,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LM,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SUB,
     13,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LM,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SUB,
     15,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LM,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SUB,
     18,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LM,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SUB,
     19,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SC,
     0,
     UTF8PROC_BIDI_CLASS_ET,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     10996,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     19190,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     19193,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_FONT,
     2812,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     11005,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     19199,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     19202,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     2821,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     11014,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_FONT,
     6,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_FONT,
     1498,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_FONT,
     7,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_FONT,
     277,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_FONT,
     1499,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_FONT,
     1502,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_FONT,
     11,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_FONT,
     1504,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     11016,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_FONT,
     1507,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_FONT,
     2826,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_FONT,
     1508,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_SUPER,
     11019,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     19213,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_SUPER,
     11024,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_EXTENDED_PICTOGRAPHIC},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_FONT,
     2834,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     2835,
     866,
     UINT16_MAX,
     866,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     1501,
     10,
     UINT16_MAX,
     10,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     2836,
     65,
     UINT16_MAX,
     65,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_FONT,
     1493,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_FONT,
     4,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_FONT,
     1495,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_FONT,
     2837,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     2838,
     UINT16_MAX,
     2838,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_FONT,
     1503,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_FONT,
     14,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     2839,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     2840,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     2841,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     2842,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_FONT,
     8,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_EXTENDED_PICTOGRAPHIC},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     19227,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_FONT,
     858,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_FONT,
     847,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_FONT,
     2846,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_FONT,
     2847,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SM,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_FONT,
     2848,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     true,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_FONT,
     1494,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_FONT,
     3,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_FONT,
     8,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_FONT,
     9,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8880,
     UINT16_MAX,
     8880,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_NO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_FRACTION,
     19233,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_NO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_FRACTION,
     19236,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_NO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_FRACTION,
     27431,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_NO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_FRACTION,
     19243,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_NO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_FRACTION,
     19246,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_NO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_FRACTION,
     19249,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_NO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_FRACTION,
     19252,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_NO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_FRACTION,
     19255,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_NO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_FRACTION,
     19258,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_NO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_FRACTION,
     19261,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_NO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_FRACTION,
     19264,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_NO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_FRACTION,
     19267,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_NO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_FRACTION,
     19270,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_NO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_FRACTION,
     19273,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_NO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_FRACTION,
     19276,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_NO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_FRACTION,
     11087,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_NL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     1499,
     2897,
     UINT16_MAX,
     2897,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_NL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     11090,
     2900,
     UINT16_MAX,
     2900,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_NL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     19285,
     2904,
     UINT16_MAX,
     2904,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_NL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     11097,
     2907,
     UINT16_MAX,
     2907,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_NL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     2908,
     2909,
     UINT16_MAX,
     2909,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_NL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     11102,
     2912,
     UINT16_MAX,
     2912,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_NL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     19297,
     2916,
     UINT16_MAX,
     2916,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_NL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     27493,
     2921,
     UINT16_MAX,
     2921,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_NL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     11114,
     2924,
     UINT16_MAX,
     2924,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_NL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     2925,
     2926,
     UINT16_MAX,
     2926,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_NL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     11119,
     2929,
     UINT16_MAX,
     2929,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_NL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     19314,
     2933,
     UINT16_MAX,
     2933,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_NL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     1502,
     2934,
     UINT16_MAX,
     2934,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_NL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     2812,
     2935,
     UINT16_MAX,
     2935,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_NL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     1494,
     2936,
     UINT16_MAX,
     2936,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_NL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     1503,
     2937,
     UINT16_MAX,
     2937,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_NL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     8,
     UINT16_MAX,
     8881,
     UINT16_MAX,
     8881,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_NL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     11130,
     UINT16_MAX,
     8882,
     UINT16_MAX,
     8882,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_NL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     19324,
     UINT16_MAX,
     8883,
     UINT16_MAX,
     8883,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_NL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     11135,
     UINT16_MAX,
     8884,
     UINT16_MAX,
     8884,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_NL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     21,
     UINT16_MAX,
     8885,
     UINT16_MAX,
     8885,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_NL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     11137,
     UINT16_MAX,
     8886,
     UINT16_MAX,
     8886,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_NL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     19331,
     UINT16_MAX,
     8887,
     UINT16_MAX,
     8887,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_NL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     27526,
     UINT16_MAX,
     8888,
     UINT16_MAX,
     8888,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_NL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     11146,
     UINT16_MAX,
     8889,
     UINT16_MAX,
     8889,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_NL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     23,
     UINT16_MAX,
     8890,
     UINT16_MAX,
     8890,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_NL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     11148,
     UINT16_MAX,
     8891,
     UINT16_MAX,
     8891,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_NL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     19342,
     UINT16_MAX,
     8892,
     UINT16_MAX,
     8892,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_NL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     11,
     UINT16_MAX,
     8893,
     UINT16_MAX,
     8893,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_NL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     2,
     UINT16_MAX,
     8894,
     UINT16_MAX,
     8894,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_NL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     3,
     UINT16_MAX,
     8895,
     UINT16_MAX,
     8895,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_NL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     12,
     UINT16_MAX,
     8896,
     UINT16_MAX,
     8896,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     2961,
     UINT16_MAX,
     2961,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8897,
     UINT16_MAX,
     8897,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_NO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_FRACTION,
     19346,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SM,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     5039,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SM,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     5042,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SM,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     5045,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_EXTENDED_PICTOGRAPHIC},
    {UTF8PROC_CATEGORY_SM,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     0,
     11157,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SM,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     0,
     11159,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SM,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     0,
     11161,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     0,
     11163,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SM,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     0,
     11165,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SM,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     0,
     11167,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     5048,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SM,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     5054,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SM,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     5051,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SM,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     true,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SM,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     5057,
     true,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SM,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     0,
     11169,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     true,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SM,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     5060,
     true,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SM,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     0,
     11171,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     true,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SM,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     5063,
     true,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SM,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     0,
     11173,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     true,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SM,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     5066,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SM,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     0,
     11175,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     true,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SM,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     5069,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SM,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     0,
     11177,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     true,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SM,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     11179,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     true,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SM,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     19373,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     true,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SM,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     11184,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     true,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SM,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     19378,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     true,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SM,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     5072,
     true,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SM,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     0,
     11189,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     true,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SM,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     5075,
     true,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SM,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     0,
     11191,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     true,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SM,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     5078,
     true,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SM,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     0,
     11193,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     true,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SM,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     5081,
     true,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SM,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     0,
     11195,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     true,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SM,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     5090,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SM,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     0,
     11197,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     true,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SM,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     5087,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SM,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     0,
     11199,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     true,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SM,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     5099,
     true,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SM,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     5102,
     true,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SM,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     0,
     11201,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SM,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     0,
     11203,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     true,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SM,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     0,
     11205,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     true,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SM,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     0,
     11207,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     true,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SM,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     0,
     11209,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     true,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SM,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     5105,
     true,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SM,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     5108,
     true,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SM,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     0,
     11211,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     true,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SM,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     0,
     11213,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     true,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SM,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     5111,
     true,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SM,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     5114,
     true,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SM,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     0,
     11215,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     true,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SM,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     0,
     11217,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     true,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SM,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     5117,
     true,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SM,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     5120,
     true,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SM,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     5147,
     true,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SM,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     5150,
     true,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SM,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     0,
     11219,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     true,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SM,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     0,
     11221,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     true,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SM,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     5123,
     true,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SM,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     5126,
     true,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SM,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     0,
     11223,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     true,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SM,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     0,
     11225,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     true,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SM,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     5129,
     true,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SM,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     5132,
     true,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SM,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     0,
     11227,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     true,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SM,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     0,
     11229,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     true,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SM,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     5153,
     true,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SM,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     5156,
     true,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SM,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     5135,
     true,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SM,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     5138,
     true,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SM,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     5141,
     true,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SM,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     5144,
     true,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SM,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     0,
     11231,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     true,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SM,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     0,
     11233,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     true,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SM,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     0,
     11235,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     true,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SM,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     0,
     11237,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     true,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SM,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     5159,
     true,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SM,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     5162,
     true,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SM,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     5165,
     true,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SM,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     5168,
     true,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SM,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     0,
     11239,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     true,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SM,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     0,
     11241,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     true,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SM,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     0,
     11243,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     true,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SM,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     0,
     11245,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     true,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SM,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     0,
     11247,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     true,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SM,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     0,
     11249,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     true,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SM,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     0,
     11251,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     true,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SM,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     0,
     11253,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     true,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_EXTENDED_PICTOGRAPHIC},
    {UTF8PROC_CATEGORY_PS,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     0,
     3063,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     true,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_PE,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     0,
     3064,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     true,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_NO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_CIRCLE,
     38,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_NO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_CIRCLE,
     31,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_NO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_CIRCLE,
     32,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_NO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_CIRCLE,
     2793,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_NO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_CIRCLE,
     2794,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_NO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_CIRCLE,
     2795,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_NO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_CIRCLE,
     2796,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_NO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_CIRCLE,
     2797,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_NO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_CIRCLE,
     2798,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_NO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_CIRCLE,
     11257,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_NO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_CIRCLE,
     11259,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_NO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_CIRCLE,
     11261,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_NO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_CIRCLE,
     11263,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_NO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_CIRCLE,
     11265,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_NO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_CIRCLE,
     11267,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_NO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_CIRCLE,
     11269,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_NO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_CIRCLE,
     11271,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_NO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_CIRCLE,
     11273,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_NO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_CIRCLE,
     11275,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_NO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_CIRCLE,
     11277,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_NO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     19471,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_NO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     19474,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_NO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     19477,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_NO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     19480,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_NO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     19483,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_NO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     19486,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_NO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     19489,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_NO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     19492,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_NO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     19495,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_NO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     27690,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_NO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     27694,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_NO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     27698,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_NO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     27702,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_NO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     27706,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_NO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     27710,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_NO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     27714,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_NO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     27718,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_NO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     27722,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_NO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     27726,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_NO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     27730,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_NO,
     0,
     UTF8PROC_BIDI_CLASS_EN,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     11350,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_NO,
     0,
     UTF8PROC_BIDI_CLASS_EN,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     11352,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_NO,
     0,
     UTF8PROC_BIDI_CLASS_EN,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     11354,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_NO,
     0,
     UTF8PROC_BIDI_CLASS_EN,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     11356,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_NO,
     0,
     UTF8PROC_BIDI_CLASS_EN,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     11358,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_NO,
     0,
     UTF8PROC_BIDI_CLASS_EN,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     11360,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_NO,
     0,
     UTF8PROC_BIDI_CLASS_EN,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     11362,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_NO,
     0,
     UTF8PROC_BIDI_CLASS_EN,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     11364,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_NO,
     0,
     UTF8PROC_BIDI_CLASS_EN,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     11366,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_NO,
     0,
     UTF8PROC_BIDI_CLASS_EN,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     19560,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_NO,
     0,
     UTF8PROC_BIDI_CLASS_EN,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     19563,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_NO,
     0,
     UTF8PROC_BIDI_CLASS_EN,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     19566,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_NO,
     0,
     UTF8PROC_BIDI_CLASS_EN,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     19569,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_NO,
     0,
     UTF8PROC_BIDI_CLASS_EN,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     19572,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_NO,
     0,
     UTF8PROC_BIDI_CLASS_EN,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     19575,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_NO,
     0,
     UTF8PROC_BIDI_CLASS_EN,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     19578,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_NO,
     0,
     UTF8PROC_BIDI_CLASS_EN,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     19581,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_NO,
     0,
     UTF8PROC_BIDI_CLASS_EN,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     19584,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_NO,
     0,
     UTF8PROC_BIDI_CLASS_EN,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     19587,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_NO,
     0,
     UTF8PROC_BIDI_CLASS_EN,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     19590,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     19593,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     19596,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     19599,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     19602,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     19605,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     19608,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     19611,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     19614,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     19617,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     19620,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     19623,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     19626,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     19629,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     19632,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     19635,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     19638,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     19641,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     19644,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     19647,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     19650,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     19653,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     19656,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     19659,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     19662,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     19665,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     19668,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_CIRCLE,
     1491,
     3287,
     UINT16_MAX,
     3287,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_CIRCLE,
     1493,
     3288,
     UINT16_MAX,
     3288,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_CIRCLE,
     2812,
     3289,
     UINT16_MAX,
     3289,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_CIRCLE,
     1494,
     3290,
     UINT16_MAX,
     3290,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_CIRCLE,
     1495,
     3291,
     UINT16_MAX,
     3291,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_CIRCLE,
     2837,
     3292,
     UINT16_MAX,
     3292,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_CIRCLE,
     1497,
     3293,
     UINT16_MAX,
     3293,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_CIRCLE,
     1498,
     3294,
     UINT16_MAX,
     3294,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_CIRCLE,
     1499,
     3295,
     UINT16_MAX,
     3295,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_CIRCLE,
     1500,
     3296,
     UINT16_MAX,
     3296,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_CIRCLE,
     1501,
     3297,
     UINT16_MAX,
     3297,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_CIRCLE,
     1502,
     3298,
     UINT16_MAX,
     3298,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_CIRCLE,
     1503,
     3299,
     UINT16_MAX,
     3299,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_EXTENDED_PICTOGRAPHIC},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_CIRCLE,
     1504,
     3300,
     UINT16_MAX,
     3300,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_CIRCLE,
     1505,
     3301,
     UINT16_MAX,
     3301,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_CIRCLE,
     1507,
     3302,
     UINT16_MAX,
     3302,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_CIRCLE,
     2826,
     3303,
     UINT16_MAX,
     3303,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_CIRCLE,
     1508,
     3304,
     UINT16_MAX,
     3304,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_CIRCLE,
     3305,
     3306,
     UINT16_MAX,
     3306,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_CIRCLE,
     1509,
     3307,
     UINT16_MAX,
     3307,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_CIRCLE,
     1510,
     3308,
     UINT16_MAX,
     3308,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_CIRCLE,
     2908,
     3309,
     UINT16_MAX,
     3309,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_CIRCLE,
     1511,
     3310,
     UINT16_MAX,
     3310,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_CIRCLE,
     2925,
     3311,
     UINT16_MAX,
     3311,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_CIRCLE,
     3312,
     3313,
     UINT16_MAX,
     3313,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_CIRCLE,
     2834,
     3314,
     UINT16_MAX,
     3314,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_CIRCLE,
     0,
     UINT16_MAX,
     8898,
     UINT16_MAX,
     8898,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_CIRCLE,
     1,
     UINT16_MAX,
     8899,
     UINT16_MAX,
     8899,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_CIRCLE,
     2,
     UINT16_MAX,
     8900,
     UINT16_MAX,
     8900,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_CIRCLE,
     3,
     UINT16_MAX,
     8901,
     UINT16_MAX,
     8901,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_CIRCLE,
     4,
     UINT16_MAX,
     8902,
     UINT16_MAX,
     8902,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_CIRCLE,
     5,
     UINT16_MAX,
     8903,
     UINT16_MAX,
     8903,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_CIRCLE,
     6,
     UINT16_MAX,
     8904,
     UINT16_MAX,
     8904,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_CIRCLE,
     7,
     UINT16_MAX,
     8905,
     UINT16_MAX,
     8905,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_CIRCLE,
     8,
     UINT16_MAX,
     8906,
     UINT16_MAX,
     8906,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_CIRCLE,
     9,
     UINT16_MAX,
     8907,
     UINT16_MAX,
     8907,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_CIRCLE,
     10,
     UINT16_MAX,
     8908,
     UINT16_MAX,
     8908,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_CIRCLE,
     11,
     UINT16_MAX,
     8909,
     UINT16_MAX,
     8909,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_CIRCLE,
     12,
     UINT16_MAX,
     8910,
     UINT16_MAX,
     8910,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_CIRCLE,
     13,
     UINT16_MAX,
     8911,
     UINT16_MAX,
     8911,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_CIRCLE,
     14,
     UINT16_MAX,
     8912,
     UINT16_MAX,
     8912,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_CIRCLE,
     15,
     UINT16_MAX,
     8913,
     UINT16_MAX,
     8913,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_CIRCLE,
     16,
     UINT16_MAX,
     8914,
     UINT16_MAX,
     8914,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_CIRCLE,
     17,
     UINT16_MAX,
     8915,
     UINT16_MAX,
     8915,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_CIRCLE,
     18,
     UINT16_MAX,
     8916,
     UINT16_MAX,
     8916,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_CIRCLE,
     19,
     UINT16_MAX,
     8917,
     UINT16_MAX,
     8917,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_CIRCLE,
     20,
     UINT16_MAX,
     8918,
     UINT16_MAX,
     8918,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_CIRCLE,
     21,
     UINT16_MAX,
     8919,
     UINT16_MAX,
     8919,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_CIRCLE,
     22,
     UINT16_MAX,
     8920,
     UINT16_MAX,
     8920,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_CIRCLE,
     23,
     UINT16_MAX,
     8921,
     UINT16_MAX,
     8921,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_CIRCLE,
     24,
     UINT16_MAX,
     8922,
     UINT16_MAX,
     8922,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_CIRCLE,
     25,
     UINT16_MAX,
     8923,
     UINT16_MAX,
     8923,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_NO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_CIRCLE,
     2792,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SM,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_EXTENDED_PICTOGRAPHIC},
    {UTF8PROC_CATEGORY_SM,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_EXTENDED_PICTOGRAPHIC},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_EXTENDED_PICTOGRAPHIC},
    {UTF8PROC_CATEGORY_SM,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     27891,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     true,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SM,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     19703,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     true,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SM,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     11514,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SM,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     19708,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SM,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     0,
     11519,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     true,
     true,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SM,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     5171,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     true,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     3329,
     UINT16_MAX,
     3329,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     3330,
     UINT16_MAX,
     3330,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     3331,
     UINT16_MAX,
     3331,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     3332,
     UINT16_MAX,
     3332,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     3333,
     UINT16_MAX,
     3333,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     3334,
     UINT16_MAX,
     3334,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     3335,
     UINT16_MAX,
     3335,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     3336,
     UINT16_MAX,
     3336,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     3337,
     UINT16_MAX,
     3337,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     3338,
     UINT16_MAX,
     3338,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     3339,
     UINT16_MAX,
     3339,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     3340,
     UINT16_MAX,
     3340,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     3341,
     UINT16_MAX,
     3341,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     3342,
     UINT16_MAX,
     3342,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     3343,
     UINT16_MAX,
     3343,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     3344,
     UINT16_MAX,
     3344,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     3345,
     UINT16_MAX,
     3345,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     3346,
     UINT16_MAX,
     3346,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     3347,
     UINT16_MAX,
     3347,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     3348,
     UINT16_MAX,
     3348,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     3349,
     UINT16_MAX,
     3349,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     3350,
     UINT16_MAX,
     3350,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     3351,
     UINT16_MAX,
     3351,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     3352,
     UINT16_MAX,
     3352,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     3353,
     UINT16_MAX,
     3353,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     3354,
     UINT16_MAX,
     3354,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     3355,
     UINT16_MAX,
     3355,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     3356,
     UINT16_MAX,
     3356,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     3357,
     UINT16_MAX,
     3357,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     3358,
     UINT16_MAX,
     3358,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     3359,
     UINT16_MAX,
     3359,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     3360,
     UINT16_MAX,
     3360,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     3361,
     UINT16_MAX,
     3361,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     3362,
     UINT16_MAX,
     3362,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     3363,
     UINT16_MAX,
     3363,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     3364,
     UINT16_MAX,
     3364,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     3365,
     UINT16_MAX,
     3365,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     3366,
     UINT16_MAX,
     3366,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     3367,
     UINT16_MAX,
     3367,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     3368,
     UINT16_MAX,
     3368,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     3369,
     UINT16_MAX,
     3369,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     3370,
     UINT16_MAX,
     3370,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     3371,
     UINT16_MAX,
     3371,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     3372,
     UINT16_MAX,
     3372,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     3373,
     UINT16_MAX,
     3373,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     3374,
     UINT16_MAX,
     3374,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     3375,
     UINT16_MAX,
     3375,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8924,
     UINT16_MAX,
     8924,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8925,
     UINT16_MAX,
     8925,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8926,
     UINT16_MAX,
     8926,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8927,
     UINT16_MAX,
     8927,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8928,
     UINT16_MAX,
     8928,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8929,
     UINT16_MAX,
     8929,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8930,
     UINT16_MAX,
     8930,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8931,
     UINT16_MAX,
     8931,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8932,
     UINT16_MAX,
     8932,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8933,
     UINT16_MAX,
     8933,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8934,
     UINT16_MAX,
     8934,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8935,
     UINT16_MAX,
     8935,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8936,
     UINT16_MAX,
     8936,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8937,
     UINT16_MAX,
     8937,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8938,
     UINT16_MAX,
     8938,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8939,
     UINT16_MAX,
     8939,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8940,
     UINT16_MAX,
     8940,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8941,
     UINT16_MAX,
     8941,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8942,
     UINT16_MAX,
     8942,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8943,
     UINT16_MAX,
     8943,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8944,
     UINT16_MAX,
     8944,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8945,
     UINT16_MAX,
     8945,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8946,
     UINT16_MAX,
     8946,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8947,
     UINT16_MAX,
     8947,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8948,
     UINT16_MAX,
     8948,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8949,
     UINT16_MAX,
     8949,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8950,
     UINT16_MAX,
     8950,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8951,
     UINT16_MAX,
     8951,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8952,
     UINT16_MAX,
     8952,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8953,
     UINT16_MAX,
     8953,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8954,
     UINT16_MAX,
     8954,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8955,
     UINT16_MAX,
     8955,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8956,
     UINT16_MAX,
     8956,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8957,
     UINT16_MAX,
     8957,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8958,
     UINT16_MAX,
     8958,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8959,
     UINT16_MAX,
     8959,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8960,
     UINT16_MAX,
     8960,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8961,
     UINT16_MAX,
     8961,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8962,
     UINT16_MAX,
     8962,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8963,
     UINT16_MAX,
     8963,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8964,
     UINT16_MAX,
     8964,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8965,
     UINT16_MAX,
     8965,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8966,
     UINT16_MAX,
     8966,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8967,
     UINT16_MAX,
     8967,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8968,
     UINT16_MAX,
     8968,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8969,
     UINT16_MAX,
     8969,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8970,
     UINT16_MAX,
     8970,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     3376,
     UINT16_MAX,
     3376,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8971,
     UINT16_MAX,
     8971,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     3377,
     UINT16_MAX,
     3377,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     3378,
     UINT16_MAX,
     3378,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     3379,
     UINT16_MAX,
     3379,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8972,
     UINT16_MAX,
     8972,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8973,
     UINT16_MAX,
     8973,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     3380,
     UINT16_MAX,
     3380,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8974,
     UINT16_MAX,
     8974,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     3381,
     UINT16_MAX,
     3381,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8975,
     UINT16_MAX,
     8975,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     3382,
     UINT16_MAX,
     3382,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8976,
     UINT16_MAX,
     8976,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     1513,
     UINT16_MAX,
     1513,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     1531,
     UINT16_MAX,
     1531,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     1512,
     UINT16_MAX,
     1512,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     1520,
     UINT16_MAX,
     1520,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     3383,
     UINT16_MAX,
     3383,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8977,
     UINT16_MAX,
     8977,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     3384,
     UINT16_MAX,
     3384,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8978,
     UINT16_MAX,
     8978,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LM,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SUB,
     9,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LM,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SUPER,
     2908,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     3385,
     UINT16_MAX,
     3385,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     3386,
     UINT16_MAX,
     3386,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     3387,
     UINT16_MAX,
     3387,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8979,
     UINT16_MAX,
     8979,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     3388,
     UINT16_MAX,
     3388,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8980,
     UINT16_MAX,
     8980,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     3389,
     UINT16_MAX,
     3389,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8981,
     UINT16_MAX,
     8981,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     3390,
     UINT16_MAX,
     3390,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8982,
     UINT16_MAX,
     8982,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     3391,
     UINT16_MAX,
     3391,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8983,
     UINT16_MAX,
     8983,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     3392,
     UINT16_MAX,
     3392,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8984,
     UINT16_MAX,
     8984,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     3393,
     UINT16_MAX,
     3393,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8985,
     UINT16_MAX,
     8985,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     3394,
     UINT16_MAX,
     3394,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8986,
     UINT16_MAX,
     8986,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     3395,
     UINT16_MAX,
     3395,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8987,
     UINT16_MAX,
     8987,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     3396,
     UINT16_MAX,
     3396,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8988,
     UINT16_MAX,
     8988,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     3397,
     UINT16_MAX,
     3397,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8989,
     UINT16_MAX,
     8989,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     3398,
     UINT16_MAX,
     3398,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8990,
     UINT16_MAX,
     8990,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     3399,
     UINT16_MAX,
     3399,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8991,
     UINT16_MAX,
     8991,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     3400,
     UINT16_MAX,
     3400,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8992,
     UINT16_MAX,
     8992,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     3401,
     UINT16_MAX,
     3401,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8993,
     UINT16_MAX,
     8993,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     3402,
     UINT16_MAX,
     3402,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8994,
     UINT16_MAX,
     8994,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     3403,
     UINT16_MAX,
     3403,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8995,
     UINT16_MAX,
     8995,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     3404,
     UINT16_MAX,
     3404,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8996,
     UINT16_MAX,
     8996,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     3405,
     UINT16_MAX,
     3405,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8997,
     UINT16_MAX,
     8997,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     3406,
     UINT16_MAX,
     3406,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8998,
     UINT16_MAX,
     8998,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     3407,
     UINT16_MAX,
     3407,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8999,
     UINT16_MAX,
     8999,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     3408,
     UINT16_MAX,
     3408,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9000,
     UINT16_MAX,
     9000,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     3409,
     UINT16_MAX,
     3409,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9001,
     UINT16_MAX,
     9001,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     3410,
     UINT16_MAX,
     3410,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9002,
     UINT16_MAX,
     9002,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     3411,
     UINT16_MAX,
     3411,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9003,
     UINT16_MAX,
     9003,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     3412,
     UINT16_MAX,
     3412,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9004,
     UINT16_MAX,
     9004,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     3413,
     UINT16_MAX,
     3413,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9005,
     UINT16_MAX,
     9005,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     3414,
     UINT16_MAX,
     3414,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9006,
     UINT16_MAX,
     9006,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     3415,
     UINT16_MAX,
     3415,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9007,
     UINT16_MAX,
     9007,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     3416,
     UINT16_MAX,
     3416,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9008,
     UINT16_MAX,
     9008,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     3417,
     UINT16_MAX,
     3417,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9009,
     UINT16_MAX,
     9009,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     3418,
     UINT16_MAX,
     3418,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9010,
     UINT16_MAX,
     9010,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     3419,
     UINT16_MAX,
     3419,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9011,
     UINT16_MAX,
     9011,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     3420,
     UINT16_MAX,
     3420,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9012,
     UINT16_MAX,
     9012,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     3421,
     UINT16_MAX,
     3421,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9013,
     UINT16_MAX,
     9013,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     3422,
     UINT16_MAX,
     3422,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9014,
     UINT16_MAX,
     9014,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     3423,
     UINT16_MAX,
     3423,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9015,
     UINT16_MAX,
     9015,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     3424,
     UINT16_MAX,
     3424,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9016,
     UINT16_MAX,
     9016,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     3425,
     UINT16_MAX,
     3425,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9017,
     UINT16_MAX,
     9017,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     3426,
     UINT16_MAX,
     3426,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9018,
     UINT16_MAX,
     9018,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     3427,
     UINT16_MAX,
     3427,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9019,
     UINT16_MAX,
     9019,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     3428,
     UINT16_MAX,
     3428,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9020,
     UINT16_MAX,
     9020,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     3429,
     UINT16_MAX,
     3429,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9021,
     UINT16_MAX,
     9021,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     3430,
     UINT16_MAX,
     3430,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9022,
     UINT16_MAX,
     9022,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     3431,
     UINT16_MAX,
     3431,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9023,
     UINT16_MAX,
     9023,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     3432,
     UINT16_MAX,
     3432,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9024,
     UINT16_MAX,
     9024,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     3433,
     UINT16_MAX,
     3433,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9025,
     UINT16_MAX,
     9025,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     3434,
     UINT16_MAX,
     3434,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9026,
     UINT16_MAX,
     9026,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     3435,
     UINT16_MAX,
     3435,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9027,
     UINT16_MAX,
     9027,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     3436,
     UINT16_MAX,
     3436,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9028,
     UINT16_MAX,
     9028,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     3437,
     UINT16_MAX,
     3437,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9029,
     UINT16_MAX,
     9029,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     3438,
     UINT16_MAX,
     3438,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9030,
     UINT16_MAX,
     9030,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     3439,
     UINT16_MAX,
     3439,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9031,
     UINT16_MAX,
     9031,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9032,
     UINT16_MAX,
     9032,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9033,
     UINT16_MAX,
     9033,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9034,
     UINT16_MAX,
     9034,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9035,
     UINT16_MAX,
     9035,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9036,
     UINT16_MAX,
     9036,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9037,
     UINT16_MAX,
     9037,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9038,
     UINT16_MAX,
     9038,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9039,
     UINT16_MAX,
     9039,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9040,
     UINT16_MAX,
     9040,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9041,
     UINT16_MAX,
     9041,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9042,
     UINT16_MAX,
     9042,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9043,
     UINT16_MAX,
     9043,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9044,
     UINT16_MAX,
     9044,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9045,
     UINT16_MAX,
     9045,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9046,
     UINT16_MAX,
     9046,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9047,
     UINT16_MAX,
     9047,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9048,
     UINT16_MAX,
     9048,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9049,
     UINT16_MAX,
     9049,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9050,
     UINT16_MAX,
     9050,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9051,
     UINT16_MAX,
     9051,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9052,
     UINT16_MAX,
     9052,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9053,
     UINT16_MAX,
     9053,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9054,
     UINT16_MAX,
     9054,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9055,
     UINT16_MAX,
     9055,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9056,
     UINT16_MAX,
     9056,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9057,
     UINT16_MAX,
     9057,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9058,
     UINT16_MAX,
     9058,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9059,
     UINT16_MAX,
     9059,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9060,
     UINT16_MAX,
     9060,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9061,
     UINT16_MAX,
     9061,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9062,
     UINT16_MAX,
     9062,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9063,
     UINT16_MAX,
     9063,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9064,
     UINT16_MAX,
     9064,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9065,
     UINT16_MAX,
     9065,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9066,
     UINT16_MAX,
     9066,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9067,
     UINT16_MAX,
     9067,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9068,
     UINT16_MAX,
     9068,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9069,
     UINT16_MAX,
     9069,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9070,
     UINT16_MAX,
     9070,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9071,
     UINT16_MAX,
     9071,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LM,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SUPER,
     3440,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     3441,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     3442,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     3443,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     3444,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     3445,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     3446,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     3447,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     3448,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     3449,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     3450,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     3451,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     3452,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     3453,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     3454,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     3455,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     3456,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     3457,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     3458,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     3459,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     3460,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     3461,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     3462,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     3463,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     3464,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     3465,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     3466,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     3467,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     3468,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     3469,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     3470,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     3471,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     3472,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     3473,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     3474,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     3475,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     3476,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     3477,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     3478,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     3479,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     3480,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     3481,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     3482,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     3483,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     3484,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     3485,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     3486,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     3487,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     3488,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     3489,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     3490,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     3491,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     3492,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     3493,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     3494,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     3495,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     3496,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     3497,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     3498,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     3499,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     3500,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     3501,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     3502,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     3503,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     3504,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     3505,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     3506,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     3507,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     3508,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     3509,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     3510,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     3511,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     3512,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     3513,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     3514,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     3515,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     3516,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     3517,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     3518,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     3519,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     3520,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     3521,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     3522,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     3523,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     3524,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     3525,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     3526,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     3527,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     3528,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     3529,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     3530,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     3531,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     3532,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     3533,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     3534,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     3535,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     3536,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     3537,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     3538,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     3539,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     3540,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     3541,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     3542,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     3543,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     3544,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     3545,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     3546,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     3547,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     3548,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     3549,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     3550,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     3551,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     3552,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     3553,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     3554,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     3555,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     3556,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     3557,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     3558,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     3559,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     3560,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     3561,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     3562,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     3563,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     3564,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     3565,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     3566,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     3567,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     3568,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     3569,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     3570,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     3571,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     3572,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     3573,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     3574,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     3575,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     3576,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     3577,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     3578,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     3579,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     3580,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     3581,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     3582,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     3583,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     3584,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     3585,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     3586,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     3587,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     3588,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     3589,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     3590,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     3591,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     3592,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     3593,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     3594,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     3595,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     3596,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     3597,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     3598,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     3599,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     3600,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     3601,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     3602,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     3603,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     3604,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     3605,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     3606,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     3607,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     3608,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     3609,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     3610,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     3611,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     3612,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     3613,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     3614,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     3615,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     3616,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     3617,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     3618,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     3619,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     3620,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     3621,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     3622,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     3623,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     3624,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     3625,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     3626,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     3627,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     3628,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     3629,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     3630,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     3631,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     3632,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     3633,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     3634,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     3635,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     3636,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     3637,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     3638,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     3639,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     3640,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     3641,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     3642,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     3643,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     3644,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     3645,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     3646,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     3647,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     3648,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     3649,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     3650,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     3651,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     3652,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     3653,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     3654,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     3655,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     3656,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_ZS,
     0,
     UTF8PROC_BIDI_CLASS_WS,
     UTF8PROC_DECOMP_TYPE_WIDE,
     26,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_PO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LM,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_NL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_PS,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     true,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_PE,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     true,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_PD,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_PS,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_PE,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_MN,
     218,
     UTF8PROC_BIDI_CLASS_NSM,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     0,
     0,
     UTF8PROC_BOUNDCLASS_EXTEND},
    {UTF8PROC_CATEGORY_MC,
     224,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_EXTEND},
    {UTF8PROC_CATEGORY_PD,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_EXTENDED_PICTOGRAPHIC},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     3657,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_NL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     3466,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_NL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     3658,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_NL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     3659,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_PO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_EXTENDED_PICTOGRAPHIC},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     5239,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     5174,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     11852,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     5177,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     11854,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     5180,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     11856,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     5183,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     11858,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     5186,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     11860,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     5189,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     11862,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     5192,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     11864,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     5195,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     11866,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     5198,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     11868,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     5201,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     11870,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     5204,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     11872,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     5207,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     11874,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     5210,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     11876,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     5213,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     11878,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     5216,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     11880,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     5219,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     11882,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     11884,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     5223,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     11886,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     11888,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     5227,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     11890,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     11892,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     5231,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     11894,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     11896,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     5235,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     11898,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     11900,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     11902,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_MN,
     8,
     UTF8PROC_BIDI_CLASS_NSM,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     32820,
     false,
     false,
     false,
     false,
     0,
     0,
     UTF8PROC_BOUNDCLASS_EXTEND},
    {UTF8PROC_CATEGORY_MN,
     8,
     UTF8PROC_BIDI_CLASS_NSM,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     32821,
     false,
     false,
     false,
     false,
     0,
     0,
     UTF8PROC_BOUNDCLASS_EXTEND},
    {UTF8PROC_CATEGORY_SK,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     11904,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SK,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     11906,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LM,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     5242,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LM,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     11908,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_VERTICAL,
     11910,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     5310,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     5245,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     11912,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     5248,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     11914,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     5251,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     11916,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     5254,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     11918,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     5257,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     11920,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     5260,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     11922,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     5263,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     11924,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     5266,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     11926,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     5269,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     11928,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     5272,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     11930,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     5275,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     11932,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     5278,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     11934,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     5281,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     11936,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     5284,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     11938,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     5287,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     11940,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     5290,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     11942,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     11944,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     5294,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     11946,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     11948,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     5298,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     11950,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     11952,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     5302,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     11954,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     11956,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     5306,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     11958,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     11960,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     5313,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     5316,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     5319,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     5322,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     11962,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     11964,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     11966,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     11968,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     11970,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LM,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     5325,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LM,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     11972,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_VERTICAL,
     11974,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     3784,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     3785,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     3786,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     3787,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     3788,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     3789,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     3790,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     3791,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     3792,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     3793,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     3794,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     3795,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     3796,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     3797,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     3798,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     3799,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     3800,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     3801,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     3802,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     3803,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     3804,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     3805,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     3806,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     3807,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     3808,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     3809,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     3810,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     3811,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     3812,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     3813,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     3814,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     3815,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     3816,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     3817,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     3818,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     3819,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     3820,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     3821,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     3822,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     3823,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     3824,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     3825,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     3826,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     3827,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     3828,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     3829,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     3830,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     3831,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     3832,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     3833,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     3834,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     3835,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     true,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     3836,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     3837,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     3838,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     3839,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     3840,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     3841,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     3842,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     3843,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     3844,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     3845,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     3846,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     3847,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     3848,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     3849,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     3850,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     3851,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     3852,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     3853,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     3854,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     3855,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     3856,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     3857,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     3858,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     3859,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     3860,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     3861,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     3862,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     3863,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     3864,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     3865,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     3866,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     3867,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     3868,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     3869,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     3870,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     3871,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     3872,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     3873,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     3874,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     3875,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     3876,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     3877,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_NO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SUPER,
     3443,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_NO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SUPER,
     3449,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_NO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SUPER,
     3878,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_NO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SUPER,
     3879,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SUPER,
     3880,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SUPER,
     3881,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SUPER,
     3882,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SUPER,
     3883,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SUPER,
     3447,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SUPER,
     3884,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SUPER,
     3885,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SUPER,
     3886,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SUPER,
     3887,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SUPER,
     3451,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     20272,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     20275,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     20278,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     20281,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     20284,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     20287,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     20290,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     20293,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     20296,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     20299,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     20302,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     20305,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     20308,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     20311,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     28506,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     28510,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     28514,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     28518,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     28522,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     28526,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     28530,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     28534,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     28538,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     28542,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     28546,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     28550,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     28554,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     28558,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     28562,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     53142,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     44957,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_NO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     20387,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_NO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     20390,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_NO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     20393,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_NO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     20396,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_NO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     20399,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_NO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     20402,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_NO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     20405,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_NO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     20408,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_NO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     20411,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_NO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     20414,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     20417,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     20420,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     20423,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     20426,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     20429,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     20432,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     20435,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     20438,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     20441,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     20444,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     20447,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     20450,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     20453,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     20456,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     20459,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     20462,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     20465,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     20468,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     20471,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     20474,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     20477,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     20480,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     20483,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     20486,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     20489,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     20492,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_CIRCLE,
     4111,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_CIRCLE,
     4112,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_CIRCLE,
     3509,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_CIRCLE,
     4113,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_SQUARE,
     20498,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_NO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_CIRCLE,
     12309,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_NO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_CIRCLE,
     12311,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_NO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_CIRCLE,
     12313,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_NO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_CIRCLE,
     12315,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_NO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_CIRCLE,
     12317,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_NO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_CIRCLE,
     12319,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_NO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_CIRCLE,
     12321,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_NO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_CIRCLE,
     12323,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_NO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_CIRCLE,
     12325,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_NO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_CIRCLE,
     12327,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_NO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_CIRCLE,
     12329,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_NO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_CIRCLE,
     12331,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_NO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_CIRCLE,
     12333,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_NO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_CIRCLE,
     12335,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_NO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_CIRCLE,
     12337,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_CIRCLE,
     3784,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_CIRCLE,
     3787,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_CIRCLE,
     3790,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_CIRCLE,
     3792,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_CIRCLE,
     3800,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_CIRCLE,
     3801,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_CIRCLE,
     3804,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_CIRCLE,
     3806,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_CIRCLE,
     3807,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_CIRCLE,
     3809,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_CIRCLE,
     3810,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_CIRCLE,
     3811,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_CIRCLE,
     3812,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_CIRCLE,
     3813,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_CIRCLE,
     12339,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_CIRCLE,
     12341,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_CIRCLE,
     12343,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_CIRCLE,
     12345,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_CIRCLE,
     12347,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_CIRCLE,
     12349,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_CIRCLE,
     12351,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_CIRCLE,
     12353,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_CIRCLE,
     12355,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_CIRCLE,
     12357,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_CIRCLE,
     12359,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_CIRCLE,
     12361,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_CIRCLE,
     12363,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_CIRCLE,
     12365,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_CIRCLE,
     36943,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_CIRCLE,
     28756,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_CIRCLE,
     12376,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_NO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_CIRCLE,
     3443,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_NO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_CIRCLE,
     3449,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_NO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_CIRCLE,
     3878,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_NO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_CIRCLE,
     3879,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_NO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_CIRCLE,
     4186,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_NO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_CIRCLE,
     4187,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_NO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_CIRCLE,
     4188,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_NO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_CIRCLE,
     3454,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_NO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_CIRCLE,
     4189,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_NO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_CIRCLE,
     3466,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_CIRCLE,
     3516,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_CIRCLE,
     3528,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_CIRCLE,
     3527,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_CIRCLE,
     3517,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_CIRCLE,
     3609,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_CIRCLE,
     3474,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_CIRCLE,
     3514,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_CIRCLE,
     4190,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_CIRCLE,
     4191,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_CIRCLE,
     4192,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_CIRCLE,
     4193,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_CIRCLE,
     4194,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_CIRCLE,
     4195,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_CIRCLE,
     4196,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_EXTENDED_PICTOGRAPHIC},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_CIRCLE,
     4197,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_CIRCLE,
     4198,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_EXTENDED_PICTOGRAPHIC},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_CIRCLE,
     4199,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_CIRCLE,
     3480,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_CIRCLE,
     4200,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_CIRCLE,
     4201,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_CIRCLE,
     4202,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_CIRCLE,
     4203,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_CIRCLE,
     4204,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_CIRCLE,
     4205,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_CIRCLE,
     4206,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_CIRCLE,
     4207,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_CIRCLE,
     3880,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_CIRCLE,
     3881,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_CIRCLE,
     3882,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_CIRCLE,
     4208,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_CIRCLE,
     4209,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_CIRCLE,
     4210,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_CIRCLE,
     4211,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_CIRCLE,
     4212,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_CIRCLE,
     4213,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_CIRCLE,
     4214,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_CIRCLE,
     4215,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_CIRCLE,
     4216,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_CIRCLE,
     4217,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_NO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_CIRCLE,
     12410,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_NO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_CIRCLE,
     12412,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_NO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_CIRCLE,
     12414,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_NO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_CIRCLE,
     12416,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_NO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_CIRCLE,
     12418,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_NO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_CIRCLE,
     12420,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_NO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_CIRCLE,
     12422,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_NO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_CIRCLE,
     12424,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_NO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_CIRCLE,
     12426,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_NO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_CIRCLE,
     12428,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_NO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_CIRCLE,
     12430,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_NO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_CIRCLE,
     12432,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_NO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_CIRCLE,
     12434,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_NO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_CIRCLE,
     12436,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_NO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_CIRCLE,
     12438,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     12440,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     12442,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     12444,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     12446,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     12448,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     12450,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     12452,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     12454,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     12456,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     20650,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     20653,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     20656,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_SQUARE,
     12467,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_SQUARE,
     20661,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_SQUARE,
     12472,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_SQUARE,
     20666,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_CIRCLE,
     4285,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_CIRCLE,
     4286,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_CIRCLE,
     4287,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_CIRCLE,
     4288,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_CIRCLE,
     4289,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_CIRCLE,
     4290,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_CIRCLE,
     4291,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_CIRCLE,
     4292,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_CIRCLE,
     4293,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_CIRCLE,
     4294,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_CIRCLE,
     4295,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_CIRCLE,
     4296,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_CIRCLE,
     4297,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_CIRCLE,
     4298,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_CIRCLE,
     4299,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_CIRCLE,
     4300,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_CIRCLE,
     4301,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_CIRCLE,
     4302,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_CIRCLE,
     4303,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_CIRCLE,
     4304,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_CIRCLE,
     4305,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_CIRCLE,
     4306,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_CIRCLE,
     4307,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_CIRCLE,
     4308,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_CIRCLE,
     4309,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_CIRCLE,
     4310,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_CIRCLE,
     4311,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_CIRCLE,
     4312,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_CIRCLE,
     4313,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_CIRCLE,
     4314,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_CIRCLE,
     4315,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_CIRCLE,
     4316,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_CIRCLE,
     4317,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_CIRCLE,
     4318,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_CIRCLE,
     4319,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_CIRCLE,
     4320,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_CIRCLE,
     4321,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_CIRCLE,
     4322,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_CIRCLE,
     4323,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_CIRCLE,
     4324,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_CIRCLE,
     4325,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_CIRCLE,
     4326,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_CIRCLE,
     4327,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_CIRCLE,
     4328,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_CIRCLE,
     4329,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_CIRCLE,
     4330,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_CIRCLE,
     4331,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SQUARE,
     12524,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SQUARE,
     28910,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SQUARE,
     28914,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SQUARE,
     28918,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SQUARE,
     20730,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SQUARE,
     28925,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SQUARE,
     20737,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SQUARE,
     20740,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SQUARE,
     37127,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SQUARE,
     28940,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SQUARE,
     20752,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SQUARE,
     20755,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SQUARE,
     20758,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SQUARE,
     28953,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SQUARE,
     28957,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SQUARE,
     20769,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SQUARE,
     20772,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SQUARE,
     12583,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SQUARE,
     20777,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SQUARE,
     28972,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SQUARE,
     28976,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SQUARE,
     12596,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SQUARE,
     37174,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SQUARE,
     45371,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SQUARE,
     37185,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SQUARE,
     20806,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SQUARE,
     37193,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SQUARE,
     37198,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SQUARE,
     29011,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SQUARE,
     20823,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SQUARE,
     20826,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SQUARE,
     20829,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SQUARE,
     29024,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SQUARE,
     37220,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SQUARE,
     29033,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SQUARE,
     20845,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SQUARE,
     20848,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SQUARE,
     20851,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SQUARE,
     12662,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SQUARE,
     12664,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SQUARE,
     12666,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SQUARE,
     12668,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SQUARE,
     20862,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SQUARE,
     20865,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SQUARE,
     37252,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SQUARE,
     20873,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SQUARE,
     29068,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SQUARE,
     37264,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SQUARE,
     20885,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SQUARE,
     12696,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SQUARE,
     12698,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SQUARE,
     37276,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SQUARE,
     29089,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SQUARE,
     37285,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SQUARE,
     20906,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SQUARE,
     37293,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SQUARE,
     12722,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SQUARE,
     20916,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SQUARE,
     20919,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SQUARE,
     20922,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SQUARE,
     20925,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SQUARE,
     20928,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SQUARE,
     29123,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SQUARE,
     20935,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SQUARE,
     12746,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SQUARE,
     20940,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SQUARE,
     20943,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SQUARE,
     20946,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SQUARE,
     29141,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SQUARE,
     20953,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SQUARE,
     20956,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SQUARE,
     20959,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SQUARE,
     37346,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SQUARE,
     29159,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SQUARE,
     12779,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SQUARE,
     37357,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SQUARE,
     12786,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SQUARE,
     29172,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SQUARE,
     29176,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SQUARE,
     20988,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SQUARE,
     20991,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SQUARE,
     20994,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SQUARE,
     29189,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SQUARE,
     12809,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SQUARE,
     21003,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SQUARE,
     29198,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SQUARE,
     12818,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SQUARE,
     37396,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SQUARE,
     21017,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     12828,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     12830,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     12832,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     12834,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     12836,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     12838,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     12840,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     12842,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     12844,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     12846,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     21040,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     21043,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     21046,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     21049,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     21052,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     21055,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     21058,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     21061,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     21064,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     21067,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     21070,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     21073,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     21076,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     21079,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     21082,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SQUARE,
     21085,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SQUARE,
     12896,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SQUARE,
     12898,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SQUARE,
     21092,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SQUARE,
     12903,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SQUARE,
     12905,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_SQUARE,
     12907,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_SQUARE,
     21101,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_SQUARE,
     21104,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_SQUARE,
     12915,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SQUARE,
     12917,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SQUARE,
     12919,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SQUARE,
     12921,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SQUARE,
     12923,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SQUARE,
     29309,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SQUARE,
     12929,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SQUARE,
     12931,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SQUARE,
     12933,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SQUARE,
     12935,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SQUARE,
     12937,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SQUARE,
     12939,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SQUARE,
     12941,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SQUARE,
     12943,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SQUARE,
     21137,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SQUARE,
     29332,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SQUARE,
     12952,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SQUARE,
     12954,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SQUARE,
     12956,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SQUARE,
     12958,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SQUARE,
     12960,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SQUARE,
     12962,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SQUARE,
     12964,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SQUARE,
     21158,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SQUARE,
     21161,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SQUARE,
     21164,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SQUARE,
     21167,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SQUARE,
     12978,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SQUARE,
     12980,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SQUARE,
     12982,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SQUARE,
     12984,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SQUARE,
     12986,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SQUARE,
     12988,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SQUARE,
     12990,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SQUARE,
     12992,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SQUARE,
     12994,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SQUARE,
     12996,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SQUARE,
     21190,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SQUARE,
     21193,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SQUARE,
     13004,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SQUARE,
     21198,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SQUARE,
     21201,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SQUARE,
     21204,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SQUARE,
     13015,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SQUARE,
     21209,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SQUARE,
     21212,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SQUARE,
     29407,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SQUARE,
     13027,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SQUARE,
     21221,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SQUARE,
     21224,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SQUARE,
     21227,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SQUARE,
     21230,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SQUARE,
     37617,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SQUARE,
     45814,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SQUARE,
     13052,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SQUARE,
     13054,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SQUARE,
     13056,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SQUARE,
     13058,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SQUARE,
     13060,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SQUARE,
     13062,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SQUARE,
     13064,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SQUARE,
     13066,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SQUARE,
     13068,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SQUARE,
     13070,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SQUARE,
     13072,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SQUARE,
     13074,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SQUARE,
     13076,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SQUARE,
     13078,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SQUARE,
     13080,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SQUARE,
     13082,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SQUARE,
     13084,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SQUARE,
     13086,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SQUARE,
     29472,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SQUARE,
     13092,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SQUARE,
     13094,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SQUARE,
     13096,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SQUARE,
     29482,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SQUARE,
     21294,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SQUARE,
     13105,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SQUARE,
     13107,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SQUARE,
     13109,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SQUARE,
     13111,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SQUARE,
     13113,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SQUARE,
     13115,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SQUARE,
     13117,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SQUARE,
     13119,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SQUARE,
     13121,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SQUARE,
     13123,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SQUARE,
     21317,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SQUARE,
     13128,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SQUARE,
     13130,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SQUARE,
     21324,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SQUARE,
     21327,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SQUARE,
     13138,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SQUARE,
     29524,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SQUARE,
     21336,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SQUARE,
     13147,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SQUARE,
     13149,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SQUARE,
     13151,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SQUARE,
     13153,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_SQUARE,
     21347,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_SQUARE,
     21350,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     13161,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     13163,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     13165,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     13167,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     13169,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     13171,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     13173,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     13175,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     13177,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     21371,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     21374,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     21377,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     21380,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     21383,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     21386,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     21389,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     21392,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     21395,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     21398,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     21401,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     21404,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     21407,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     21410,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     21413,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     21416,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     21419,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     21422,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     21425,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     21428,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     21431,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     21434,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_SQUARE,
     21437,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     5056,
     UINT16_MAX,
     5056,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9072,
     UINT16_MAX,
     9072,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     5057,
     UINT16_MAX,
     5057,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9073,
     UINT16_MAX,
     9073,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     5058,
     UINT16_MAX,
     5058,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9074,
     UINT16_MAX,
     9074,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     5059,
     UINT16_MAX,
     5059,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9075,
     UINT16_MAX,
     9075,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     5060,
     UINT16_MAX,
     5060,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9076,
     UINT16_MAX,
     9076,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     1445,
     UINT16_MAX,
     1445,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     8629,
     UINT16_MAX,
     8629,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     5061,
     UINT16_MAX,
     5061,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9077,
     UINT16_MAX,
     9077,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     5062,
     UINT16_MAX,
     5062,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9078,
     UINT16_MAX,
     9078,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     5063,
     UINT16_MAX,
     5063,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9079,
     UINT16_MAX,
     9079,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     5064,
     UINT16_MAX,
     5064,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9080,
     UINT16_MAX,
     9080,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     5065,
     UINT16_MAX,
     5065,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9081,
     UINT16_MAX,
     9081,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     5066,
     UINT16_MAX,
     5066,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9082,
     UINT16_MAX,
     9082,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     5067,
     UINT16_MAX,
     5067,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9083,
     UINT16_MAX,
     9083,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     5068,
     UINT16_MAX,
     5068,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9084,
     UINT16_MAX,
     9084,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     5069,
     UINT16_MAX,
     5069,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9085,
     UINT16_MAX,
     9085,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     5070,
     UINT16_MAX,
     5070,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9086,
     UINT16_MAX,
     9086,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     5071,
     UINT16_MAX,
     5071,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9087,
     UINT16_MAX,
     9087,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     5072,
     UINT16_MAX,
     5072,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9088,
     UINT16_MAX,
     9088,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     5073,
     UINT16_MAX,
     5073,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9089,
     UINT16_MAX,
     9089,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     5074,
     UINT16_MAX,
     5074,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9090,
     UINT16_MAX,
     9090,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     5075,
     UINT16_MAX,
     5075,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9091,
     UINT16_MAX,
     9091,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     5076,
     UINT16_MAX,
     5076,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9092,
     UINT16_MAX,
     9092,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     5077,
     UINT16_MAX,
     5077,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9093,
     UINT16_MAX,
     9093,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     5078,
     UINT16_MAX,
     5078,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9094,
     UINT16_MAX,
     9094,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     5079,
     UINT16_MAX,
     5079,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9095,
     UINT16_MAX,
     9095,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     5080,
     UINT16_MAX,
     5080,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9096,
     UINT16_MAX,
     9096,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     5081,
     UINT16_MAX,
     5081,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9097,
     UINT16_MAX,
     9097,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     5082,
     UINT16_MAX,
     5082,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9098,
     UINT16_MAX,
     9098,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     5083,
     UINT16_MAX,
     5083,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9099,
     UINT16_MAX,
     9099,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     5084,
     UINT16_MAX,
     5084,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9100,
     UINT16_MAX,
     9100,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     5085,
     UINT16_MAX,
     5085,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9101,
     UINT16_MAX,
     9101,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     5086,
     UINT16_MAX,
     5086,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9102,
     UINT16_MAX,
     9102,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     5087,
     UINT16_MAX,
     5087,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9103,
     UINT16_MAX,
     9103,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     5088,
     UINT16_MAX,
     5088,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9104,
     UINT16_MAX,
     9104,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     5089,
     UINT16_MAX,
     5089,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9105,
     UINT16_MAX,
     9105,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     5090,
     UINT16_MAX,
     5090,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9106,
     UINT16_MAX,
     9106,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     5091,
     UINT16_MAX,
     5091,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9107,
     UINT16_MAX,
     9107,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LM,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SUPER,
     981,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LM,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SUPER,
     983,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     5092,
     UINT16_MAX,
     5092,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9108,
     UINT16_MAX,
     9108,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     5093,
     UINT16_MAX,
     5093,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9109,
     UINT16_MAX,
     9109,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     5094,
     UINT16_MAX,
     5094,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9110,
     UINT16_MAX,
     9110,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     5095,
     UINT16_MAX,
     5095,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9111,
     UINT16_MAX,
     9111,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     5096,
     UINT16_MAX,
     5096,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9112,
     UINT16_MAX,
     9112,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     5097,
     UINT16_MAX,
     5097,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9113,
     UINT16_MAX,
     9113,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     5098,
     UINT16_MAX,
     5098,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9114,
     UINT16_MAX,
     9114,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     5099,
     UINT16_MAX,
     5099,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9115,
     UINT16_MAX,
     9115,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     5100,
     UINT16_MAX,
     5100,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9116,
     UINT16_MAX,
     9116,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     5101,
     UINT16_MAX,
     5101,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9117,
     UINT16_MAX,
     9117,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     5102,
     UINT16_MAX,
     5102,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9118,
     UINT16_MAX,
     9118,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     5103,
     UINT16_MAX,
     5103,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9119,
     UINT16_MAX,
     9119,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     5104,
     UINT16_MAX,
     5104,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9120,
     UINT16_MAX,
     9120,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     5105,
     UINT16_MAX,
     5105,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9121,
     UINT16_MAX,
     9121,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     5106,
     UINT16_MAX,
     5106,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9122,
     UINT16_MAX,
     9122,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     5107,
     UINT16_MAX,
     5107,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9123,
     UINT16_MAX,
     9123,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     5108,
     UINT16_MAX,
     5108,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9124,
     UINT16_MAX,
     9124,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     5109,
     UINT16_MAX,
     5109,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9125,
     UINT16_MAX,
     9125,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     5110,
     UINT16_MAX,
     5110,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9126,
     UINT16_MAX,
     9126,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     5111,
     UINT16_MAX,
     5111,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9127,
     UINT16_MAX,
     9127,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     5112,
     UINT16_MAX,
     5112,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9128,
     UINT16_MAX,
     9128,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     5113,
     UINT16_MAX,
     5113,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9129,
     UINT16_MAX,
     9129,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     5114,
     UINT16_MAX,
     5114,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9130,
     UINT16_MAX,
     9130,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     5115,
     UINT16_MAX,
     5115,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9131,
     UINT16_MAX,
     9131,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     5116,
     UINT16_MAX,
     5116,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9132,
     UINT16_MAX,
     9132,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     5117,
     UINT16_MAX,
     5117,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9133,
     UINT16_MAX,
     9133,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     5118,
     UINT16_MAX,
     5118,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9134,
     UINT16_MAX,
     9134,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     5119,
     UINT16_MAX,
     5119,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9135,
     UINT16_MAX,
     9135,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     5120,
     UINT16_MAX,
     5120,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9136,
     UINT16_MAX,
     9136,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     5121,
     UINT16_MAX,
     5121,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9137,
     UINT16_MAX,
     9137,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     5122,
     UINT16_MAX,
     5122,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9138,
     UINT16_MAX,
     9138,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     5123,
     UINT16_MAX,
     5123,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9139,
     UINT16_MAX,
     9139,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     5124,
     UINT16_MAX,
     5124,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9140,
     UINT16_MAX,
     9140,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     5125,
     UINT16_MAX,
     5125,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9141,
     UINT16_MAX,
     9141,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     5126,
     UINT16_MAX,
     5126,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9142,
     UINT16_MAX,
     9142,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     5127,
     UINT16_MAX,
     5127,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9143,
     UINT16_MAX,
     9143,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     5128,
     UINT16_MAX,
     5128,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9144,
     UINT16_MAX,
     9144,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     5129,
     UINT16_MAX,
     5129,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9145,
     UINT16_MAX,
     9145,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LM,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SUPER,
     5129,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     5130,
     UINT16_MAX,
     5130,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9146,
     UINT16_MAX,
     9146,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     5131,
     UINT16_MAX,
     5131,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9147,
     UINT16_MAX,
     9147,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     5132,
     UINT16_MAX,
     5132,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     5133,
     UINT16_MAX,
     5133,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9148,
     UINT16_MAX,
     9148,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     5134,
     UINT16_MAX,
     5134,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9149,
     UINT16_MAX,
     9149,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     5135,
     UINT16_MAX,
     5135,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9150,
     UINT16_MAX,
     9150,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     5136,
     UINT16_MAX,
     5136,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9151,
     UINT16_MAX,
     9151,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     5137,
     UINT16_MAX,
     5137,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9152,
     UINT16_MAX,
     9152,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SK,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     5138,
     UINT16_MAX,
     5138,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9153,
     UINT16_MAX,
     9153,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     1524,
     UINT16_MAX,
     1524,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     5139,
     UINT16_MAX,
     5139,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9154,
     UINT16_MAX,
     9154,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     5140,
     UINT16_MAX,
     5140,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9155,
     UINT16_MAX,
     9155,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9156,
     UINT16_MAX,
     9156,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     5141,
     UINT16_MAX,
     5141,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9157,
     UINT16_MAX,
     9157,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     5142,
     UINT16_MAX,
     5142,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9158,
     UINT16_MAX,
     9158,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     5143,
     UINT16_MAX,
     5143,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9159,
     UINT16_MAX,
     9159,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     5144,
     UINT16_MAX,
     5144,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9160,
     UINT16_MAX,
     9160,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     5145,
     UINT16_MAX,
     5145,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9161,
     UINT16_MAX,
     9161,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     5146,
     UINT16_MAX,
     5146,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9162,
     UINT16_MAX,
     9162,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     5147,
     UINT16_MAX,
     5147,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9163,
     UINT16_MAX,
     9163,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     5148,
     UINT16_MAX,
     5148,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9164,
     UINT16_MAX,
     9164,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     5149,
     UINT16_MAX,
     5149,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9165,
     UINT16_MAX,
     9165,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     5150,
     UINT16_MAX,
     5150,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9166,
     UINT16_MAX,
     9166,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     785,
     UINT16_MAX,
     785,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     1515,
     UINT16_MAX,
     1515,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     1523,
     UINT16_MAX,
     1523,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     5151,
     UINT16_MAX,
     5151,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     1525,
     UINT16_MAX,
     1525,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     5152,
     UINT16_MAX,
     5152,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     5153,
     UINT16_MAX,
     5153,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     1527,
     UINT16_MAX,
     1527,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     5154,
     UINT16_MAX,
     5154,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     5155,
     UINT16_MAX,
     5155,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9167,
     UINT16_MAX,
     9167,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     5156,
     UINT16_MAX,
     5156,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9168,
     UINT16_MAX,
     9168,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     5157,
     UINT16_MAX,
     5157,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9169,
     UINT16_MAX,
     9169,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     5158,
     UINT16_MAX,
     5158,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9170,
     UINT16_MAX,
     9170,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     5159,
     UINT16_MAX,
     5159,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9171,
     UINT16_MAX,
     9171,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     5160,
     UINT16_MAX,
     5160,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9172,
     UINT16_MAX,
     9172,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     5161,
     UINT16_MAX,
     5161,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9173,
     UINT16_MAX,
     9173,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     5162,
     UINT16_MAX,
     5162,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     1536,
     UINT16_MAX,
     1536,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     5163,
     UINT16_MAX,
     5163,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     5164,
     UINT16_MAX,
     5164,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9174,
     UINT16_MAX,
     9174,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     5165,
     UINT16_MAX,
     5165,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9175,
     UINT16_MAX,
     9175,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     5166,
     UINT16_MAX,
     5166,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9176,
     UINT16_MAX,
     9176,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LM,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SUPER,
     5167,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LM,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SUPER,
     371,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9177,
     UINT16_MAX,
     9177,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LM,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SUPER,
     5094,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LM,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SUPER,
     5168,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LM,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SUPER,
     3377,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LM,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SUPER,
     5169,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LM,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SUPER,
     5170,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     5171,
     5171,
     UINT16_MAX,
     5171,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     5172,
     5172,
     UINT16_MAX,
     5172,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     5173,
     5173,
     UINT16_MAX,
     5173,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     5174,
     5174,
     UINT16_MAX,
     5174,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     5175,
     5175,
     UINT16_MAX,
     5175,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     5176,
     5176,
     UINT16_MAX,
     5176,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     5177,
     5177,
     UINT16_MAX,
     5177,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     5178,
     5178,
     UINT16_MAX,
     5178,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     5179,
     5179,
     UINT16_MAX,
     5179,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     5180,
     5180,
     UINT16_MAX,
     5180,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     5181,
     5181,
     UINT16_MAX,
     5181,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     5182,
     5182,
     UINT16_MAX,
     5182,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     5183,
     5183,
     UINT16_MAX,
     5183,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     5184,
     5184,
     UINT16_MAX,
     5184,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     5185,
     5185,
     UINT16_MAX,
     5185,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     5186,
     5186,
     UINT16_MAX,
     5186,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     5187,
     5187,
     UINT16_MAX,
     5187,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     5188,
     5188,
     UINT16_MAX,
     5188,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     5189,
     5189,
     UINT16_MAX,
     5189,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     5190,
     5190,
     UINT16_MAX,
     5190,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     5191,
     5191,
     UINT16_MAX,
     5191,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     5192,
     5192,
     UINT16_MAX,
     5192,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     5193,
     5193,
     UINT16_MAX,
     5193,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     5194,
     5194,
     UINT16_MAX,
     5194,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     5195,
     5195,
     UINT16_MAX,
     5195,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     5196,
     5196,
     UINT16_MAX,
     5196,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     5197,
     5197,
     UINT16_MAX,
     5197,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     5198,
     5198,
     UINT16_MAX,
     5198,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     5199,
     5199,
     UINT16_MAX,
     5199,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     5200,
     5200,
     UINT16_MAX,
     5200,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     5201,
     5201,
     UINT16_MAX,
     5201,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     5202,
     5202,
     UINT16_MAX,
     5202,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     5203,
     5203,
     UINT16_MAX,
     5203,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     5204,
     5204,
     UINT16_MAX,
     5204,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     5205,
     5205,
     UINT16_MAX,
     5205,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     5206,
     5206,
     UINT16_MAX,
     5206,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     5207,
     5207,
     UINT16_MAX,
     5207,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     5208,
     5208,
     UINT16_MAX,
     5208,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     5209,
     5209,
     UINT16_MAX,
     5209,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     5210,
     5210,
     UINT16_MAX,
     5210,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     5211,
     5211,
     UINT16_MAX,
     5211,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     5212,
     5212,
     UINT16_MAX,
     5212,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     5213,
     5213,
     UINT16_MAX,
     5213,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     5214,
     5214,
     UINT16_MAX,
     5214,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     5215,
     5215,
     UINT16_MAX,
     5215,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     5216,
     5216,
     UINT16_MAX,
     5216,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     5217,
     5217,
     UINT16_MAX,
     5217,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     5218,
     5218,
     UINT16_MAX,
     5218,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     5219,
     5219,
     UINT16_MAX,
     5219,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     5220,
     5220,
     UINT16_MAX,
     5220,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     5221,
     5221,
     UINT16_MAX,
     5221,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     5222,
     5222,
     UINT16_MAX,
     5222,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     5223,
     5223,
     UINT16_MAX,
     5223,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     5224,
     5224,
     UINT16_MAX,
     5224,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     5225,
     5225,
     UINT16_MAX,
     5225,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     5226,
     5226,
     UINT16_MAX,
     5226,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     5227,
     5227,
     UINT16_MAX,
     5227,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     5228,
     5228,
     UINT16_MAX,
     5228,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     5229,
     5229,
     UINT16_MAX,
     5229,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     5230,
     5230,
     UINT16_MAX,
     5230,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     5231,
     5231,
     UINT16_MAX,
     5231,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     5232,
     5232,
     UINT16_MAX,
     5232,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     5233,
     5233,
     UINT16_MAX,
     5233,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     5234,
     5234,
     UINT16_MAX,
     5234,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     5235,
     5235,
     UINT16_MAX,
     5235,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     5236,
     5236,
     UINT16_MAX,
     5236,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     5237,
     5237,
     UINT16_MAX,
     5237,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     5238,
     5238,
     UINT16_MAX,
     5238,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     5239,
     5239,
     UINT16_MAX,
     5239,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     5240,
     5240,
     UINT16_MAX,
     5240,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     5241,
     5241,
     UINT16_MAX,
     5241,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     5242,
     5242,
     UINT16_MAX,
     5242,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     5243,
     5243,
     UINT16_MAX,
     5243,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     5244,
     5244,
     UINT16_MAX,
     5244,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     5245,
     5245,
     UINT16_MAX,
     5245,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     5246,
     5246,
     UINT16_MAX,
     5246,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     5247,
     5247,
     UINT16_MAX,
     5247,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     5248,
     5248,
     UINT16_MAX,
     5248,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     5249,
     5249,
     UINT16_MAX,
     5249,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     5250,
     5250,
     UINT16_MAX,
     5250,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_LV},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_LVT},
    {UTF8PROC_CATEGORY_CS,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     0,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_CO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5251,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5252,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     3601,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5253,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5254,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5255,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5256,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     3655,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5257,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     3609,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5258,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5259,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5260,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5261,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5262,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5263,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5264,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5265,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5266,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5267,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5268,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5269,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5270,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5271,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5272,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5273,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5274,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5275,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5276,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5277,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5278,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5279,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5280,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5281,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5282,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5283,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5284,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5285,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5286,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5287,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5288,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5289,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5290,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5291,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5292,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5293,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5294,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5295,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5296,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5297,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5298,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     3567,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5299,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5300,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5301,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5302,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5303,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5304,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5305,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5306,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5307,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5308,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5309,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     3640,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5310,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5311,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5312,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5313,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5314,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5315,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5316,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5317,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5318,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5319,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5320,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5321,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5322,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5323,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5324,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5325,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5326,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5327,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5328,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5329,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5330,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5331,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5332,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5333,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5334,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5335,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5336,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5337,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5338,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5339,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5340,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5341,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5342,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5343,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5344,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5345,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5346,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5347,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5348,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5349,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5350,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5351,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5352,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5353,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5354,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5355,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5356,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     3603,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5357,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5358,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5359,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5360,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5361,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5362,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5363,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5364,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5365,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5366,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5367,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5368,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5369,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5370,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5371,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     3480,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5372,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5373,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5374,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5375,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5376,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5377,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5378,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5379,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     3461,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5380,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5381,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5382,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5383,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5384,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5385,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5386,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5387,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5388,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5389,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5390,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5391,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5392,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5393,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5394,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5395,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5396,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5397,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5398,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5399,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5400,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5401,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5402,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5403,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5404,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5405,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5406,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5407,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5408,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5409,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5410,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5411,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5412,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5413,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5414,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5415,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5416,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5417,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5418,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5419,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5420,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5421,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5422,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5423,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5424,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5425,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5426,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5427,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5428,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5429,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5430,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5431,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5432,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5433,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     3654,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5434,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5435,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5436,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5437,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5438,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5439,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5440,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5441,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5442,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5443,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5444,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5445,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     4187,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5446,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5447,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5448,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5449,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5450,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5451,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5452,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5453,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5454,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5455,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5456,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5457,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5458,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5459,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5460,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5461,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5462,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5463,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5464,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5465,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5466,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5467,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     3608,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5468,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5469,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5470,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5471,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5472,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5473,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5474,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5475,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5476,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5477,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5478,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5479,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5480,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     3559,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5481,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5482,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5483,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5484,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5485,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5486,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5487,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5488,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5489,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5490,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5491,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5492,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5493,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5494,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5495,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5496,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     3586,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5497,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     3589,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5498,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5499,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5500,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5501,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5502,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5503,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5504,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5505,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5506,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5507,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5508,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5509,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5510,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5511,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     3566,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5512,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5513,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5514,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5515,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5516,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5517,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5518,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5519,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5520,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5521,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5522,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5523,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5524,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5525,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5526,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5527,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5528,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5529,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5530,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5531,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5532,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5533,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     3487,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5534,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5535,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5536,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5537,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5538,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5539,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5540,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5541,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5542,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5543,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5544,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5545,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5546,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5547,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5548,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     4192,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5549,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5550,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5551,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5552,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     4196,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5553,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5554,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5555,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5556,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5557,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5558,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5559,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5560,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5561,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5562,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5563,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5564,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5565,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5566,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5567,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5568,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5569,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5570,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5571,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5572,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5573,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5574,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5575,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5576,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5578,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5579,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5580,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5581,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5582,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5583,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5584,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5585,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5586,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5587,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5588,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5589,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5590,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5591,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5592,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5593,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5594,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5595,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5596,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5597,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5598,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5599,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5600,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5601,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5602,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5603,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5604,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5605,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5606,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5607,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5608,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5609,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     3520,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5610,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5611,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5612,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5613,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5614,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5615,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5616,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5617,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5618,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5619,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5620,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5621,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5622,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5623,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5624,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5625,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5626,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5627,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5628,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5629,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5630,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5631,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5632,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5633,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5634,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5635,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5636,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5637,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5638,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5639,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5640,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5641,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5642,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5643,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5644,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5645,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5646,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5647,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5649,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5651,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5653,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5654,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5655,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5656,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5658,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5660,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5662,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     5663,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     13856,
     13856,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     13858,
     13858,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     13860,
     13860,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     22054,
     22054,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     22057,
     22057,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     13868,
     13870,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     13870,
     13870,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     13872,
     13872,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     13874,
     13874,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     13876,
     13876,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     13878,
     13878,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     13880,
     13880,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_R,
     0,
     13882,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     true,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_MN,
     26,
     UTF8PROC_BIDI_CLASS_NSM,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     0,
     0,
     UTF8PROC_BOUNDCLASS_EXTEND},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_R,
     0,
     13884,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     true,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_R,
     UTF8PROC_DECOMP_TYPE_FONT,
     5694,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_R,
     UTF8PROC_DECOMP_TYPE_FONT,
     2839,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_R,
     UTF8PROC_DECOMP_TYPE_FONT,
     2842,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_R,
     UTF8PROC_DECOMP_TYPE_FONT,
     5695,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_R,
     UTF8PROC_DECOMP_TYPE_FONT,
     5696,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_R,
     UTF8PROC_DECOMP_TYPE_FONT,
     5697,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_R,
     UTF8PROC_DECOMP_TYPE_FONT,
     5698,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_R,
     UTF8PROC_DECOMP_TYPE_FONT,
     5699,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_R,
     UTF8PROC_DECOMP_TYPE_FONT,
     5700,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SM,
     0,
     UTF8PROC_BIDI_CLASS_ES,
     UTF8PROC_DECOMP_TYPE_FONT,
     2799,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_R,
     0,
     13893,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     true,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_R,
     0,
     13895,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     true,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_R,
     0,
     13897,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     true,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_R,
     0,
     13899,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     true,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_R,
     0,
     13901,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     true,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_R,
     0,
     13903,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     true,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_R,
     0,
     13905,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     true,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_R,
     0,
     13907,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     true,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_R,
     0,
     13909,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     true,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_R,
     0,
     13911,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     true,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_R,
     0,
     13913,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     true,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_R,
     0,
     13915,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     true,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_R,
     0,
     13917,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     true,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_R,
     0,
     13919,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     true,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_R,
     0,
     13921,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     true,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_R,
     0,
     13923,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     true,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_R,
     0,
     13925,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     true,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_R,
     0,
     13927,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     true,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_R,
     0,
     13929,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     true,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_R,
     0,
     13931,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     true,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_R,
     0,
     13933,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     true,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_R,
     0,
     13935,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     true,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_R,
     0,
     13937,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     true,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_R,
     0,
     13939,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     true,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_R,
     0,
     13941,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     true,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_R,
     0,
     13943,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     true,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_R,
     0,
     13945,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     true,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_R,
     0,
     13947,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     true,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_R,
     0,
     13949,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     true,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_R,
     0,
     13951,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     true,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_R,
     0,
     13953,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     true,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_R,
     0,
     13955,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     true,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_R,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     13957,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_ISOLATED,
     5767,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_FINAL,
     5767,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_ISOLATED,
     5768,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_FINAL,
     5768,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_INITIAL,
     5768,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_MEDIAL,
     5768,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_ISOLATED,
     5769,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_FINAL,
     5769,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_INITIAL,
     5769,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_MEDIAL,
     5769,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_ISOLATED,
     5770,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_FINAL,
     5770,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_INITIAL,
     5770,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_MEDIAL,
     5770,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_ISOLATED,
     5771,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_FINAL,
     5771,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_INITIAL,
     5771,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_MEDIAL,
     5771,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_ISOLATED,
     5772,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_FINAL,
     5772,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_INITIAL,
     5772,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_MEDIAL,
     5772,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_ISOLATED,
     5773,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_FINAL,
     5773,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_INITIAL,
     5773,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_MEDIAL,
     5773,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_ISOLATED,
     5774,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_FINAL,
     5774,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_INITIAL,
     5774,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_MEDIAL,
     5774,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_ISOLATED,
     5775,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_FINAL,
     5775,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_INITIAL,
     5775,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_MEDIAL,
     5775,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_ISOLATED,
     5776,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_FINAL,
     5776,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_INITIAL,
     5776,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_MEDIAL,
     5776,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_ISOLATED,
     5777,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_FINAL,
     5777,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_INITIAL,
     5777,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_MEDIAL,
     5777,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_ISOLATED,
     5778,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_FINAL,
     5778,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_INITIAL,
     5778,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_MEDIAL,
     5778,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_ISOLATED,
     5779,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_FINAL,
     5779,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_INITIAL,
     5779,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_MEDIAL,
     5779,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_ISOLATED,
     5780,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_FINAL,
     5780,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_ISOLATED,
     5781,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_FINAL,
     5781,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_ISOLATED,
     5782,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_FINAL,
     5782,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_ISOLATED,
     5783,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_FINAL,
     5783,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_ISOLATED,
     5784,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_FINAL,
     5784,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_ISOLATED,
     5785,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_FINAL,
     5785,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_ISOLATED,
     5786,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_FINAL,
     5786,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_INITIAL,
     5786,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_MEDIAL,
     5786,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_ISOLATED,
     5787,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_FINAL,
     5787,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_INITIAL,
     5787,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_MEDIAL,
     5787,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_ISOLATED,
     5788,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_FINAL,
     5788,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_INITIAL,
     5788,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_MEDIAL,
     5788,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_ISOLATED,
     5789,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_FINAL,
     5789,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_INITIAL,
     5789,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_MEDIAL,
     5789,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_ISOLATED,
     5790,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_FINAL,
     5790,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_ISOLATED,
     5791,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_FINAL,
     5791,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_INITIAL,
     5791,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_MEDIAL,
     5791,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_ISOLATED,
     5792,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_FINAL,
     5792,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_ISOLATED,
     5793,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_FINAL,
     5793,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_INITIAL,
     5793,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_MEDIAL,
     5793,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_ISOLATED,
     5794,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_FINAL,
     5794,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_INITIAL,
     5794,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_MEDIAL,
     5794,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_ISOLATED,
     5795,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_FINAL,
     5795,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_ISOLATED,
     5796,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_FINAL,
     5796,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SK,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_ISOLATED,
     5797,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_FINAL,
     5797,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_INITIAL,
     5797,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_MEDIAL,
     5797,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_ISOLATED,
     5798,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_FINAL,
     5798,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_ISOLATED,
     5799,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_FINAL,
     5799,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_ISOLATED,
     5800,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_FINAL,
     5800,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_ISOLATED,
     5801,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_ISOLATED,
     5802,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_FINAL,
     5802,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_ISOLATED,
     5803,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_FINAL,
     5803,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_ISOLATED,
     5804,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_FINAL,
     5804,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_ISOLATED,
     5805,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_FINAL,
     5805,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_INITIAL,
     5805,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_MEDIAL,
     5805,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_INITIAL,
     5806,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_MEDIAL,
     5806,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_ISOLATED,
     13999,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_FINAL,
     13999,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_ISOLATED,
     14001,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_FINAL,
     14001,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_ISOLATED,
     14003,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_FINAL,
     14003,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_ISOLATED,
     14005,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_FINAL,
     14005,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_ISOLATED,
     14007,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_FINAL,
     14007,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_ISOLATED,
     14009,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_FINAL,
     14009,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_ISOLATED,
     14011,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_FINAL,
     14011,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_INITIAL,
     14011,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_ISOLATED,
     14013,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_FINAL,
     14013,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_INITIAL,
     14013,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_ISOLATED,
     5823,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_FINAL,
     5823,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_INITIAL,
     5823,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_MEDIAL,
     5823,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_ISOLATED,
     14016,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_ISOLATED,
     14018,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_ISOLATED,
     14020,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_ISOLATED,
     14022,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_ISOLATED,
     14024,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_ISOLATED,
     14026,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_ISOLATED,
     14028,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_ISOLATED,
     14030,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_ISOLATED,
     14032,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_ISOLATED,
     14034,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_ISOLATED,
     14036,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_ISOLATED,
     14038,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_ISOLATED,
     14040,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_ISOLATED,
     14042,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_ISOLATED,
     14044,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_ISOLATED,
     14046,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_ISOLATED,
     14048,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_ISOLATED,
     14050,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_ISOLATED,
     14052,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_ISOLATED,
     14054,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_ISOLATED,
     14056,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_ISOLATED,
     14058,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_ISOLATED,
     14060,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_ISOLATED,
     14062,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_ISOLATED,
     14064,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_ISOLATED,
     14066,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_ISOLATED,
     14068,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_ISOLATED,
     14070,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_ISOLATED,
     14072,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_ISOLATED,
     14074,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_ISOLATED,
     14076,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_ISOLATED,
     14078,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_ISOLATED,
     14080,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_ISOLATED,
     14082,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_ISOLATED,
     14084,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_ISOLATED,
     14086,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_ISOLATED,
     14088,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_ISOLATED,
     14090,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_ISOLATED,
     14092,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_ISOLATED,
     14094,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_ISOLATED,
     14096,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_ISOLATED,
     14098,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_ISOLATED,
     14100,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_ISOLATED,
     14102,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_ISOLATED,
     14104,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_ISOLATED,
     14106,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_ISOLATED,
     14108,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_ISOLATED,
     14110,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_ISOLATED,
     14112,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_ISOLATED,
     14114,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_ISOLATED,
     14116,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_ISOLATED,
     14118,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_ISOLATED,
     14120,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_ISOLATED,
     14122,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_ISOLATED,
     14124,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_ISOLATED,
     14126,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_ISOLATED,
     14128,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_ISOLATED,
     14130,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_ISOLATED,
     14132,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_ISOLATED,
     14134,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_ISOLATED,
     14136,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_ISOLATED,
     14138,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_ISOLATED,
     14140,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_ISOLATED,
     14142,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_ISOLATED,
     14144,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_ISOLATED,
     14146,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_ISOLATED,
     14148,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_ISOLATED,
     14150,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_ISOLATED,
     14152,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_ISOLATED,
     14154,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_ISOLATED,
     14156,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_ISOLATED,
     14158,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_ISOLATED,
     14160,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_ISOLATED,
     14162,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_ISOLATED,
     14164,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_ISOLATED,
     14166,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_ISOLATED,
     14168,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_ISOLATED,
     14170,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_ISOLATED,
     14172,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_ISOLATED,
     14174,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_ISOLATED,
     14176,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_ISOLATED,
     14178,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_ISOLATED,
     14180,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_ISOLATED,
     14182,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_ISOLATED,
     14184,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_ISOLATED,
     14186,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_ISOLATED,
     14188,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_ISOLATED,
     14190,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_ISOLATED,
     14192,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_ISOLATED,
     14194,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_ISOLATED,
     14196,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_ISOLATED,
     14198,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_ISOLATED,
     14200,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_ISOLATED,
     22394,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_ISOLATED,
     22397,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_ISOLATED,
     22400,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_ISOLATED,
     22403,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_ISOLATED,
     22406,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_ISOLATED,
     22409,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_FINAL,
     14220,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_FINAL,
     14222,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_FINAL,
     14020,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_FINAL,
     14224,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_FINAL,
     14022,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_FINAL,
     14226,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_FINAL,
     14228,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_FINAL,
     14030,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_FINAL,
     14230,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_FINAL,
     14032,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_FINAL,
     14034,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_FINAL,
     14232,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_FINAL,
     14234,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_FINAL,
     14042,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_FINAL,
     14236,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_FINAL,
     14044,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_FINAL,
     14046,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_FINAL,
     14238,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_FINAL,
     14240,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_FINAL,
     14050,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_FINAL,
     14242,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_FINAL,
     14052,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_FINAL,
     14054,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_FINAL,
     14112,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_FINAL,
     14114,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_FINAL,
     14120,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_FINAL,
     14122,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_FINAL,
     14124,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_FINAL,
     14132,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_FINAL,
     14134,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_FINAL,
     14136,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_FINAL,
     14138,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_FINAL,
     14146,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_FINAL,
     14148,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_FINAL,
     14150,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_FINAL,
     14244,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_FINAL,
     14158,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_FINAL,
     14246,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_FINAL,
     14248,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_FINAL,
     14170,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_FINAL,
     14250,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_FINAL,
     14172,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_FINAL,
     14174,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_FINAL,
     14200,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_FINAL,
     14252,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_FINAL,
     14254,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_FINAL,
     14190,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_FINAL,
     14256,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_FINAL,
     14192,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_FINAL,
     14194,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_INITIAL,
     14016,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_INITIAL,
     14018,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_INITIAL,
     14258,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_INITIAL,
     14020,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_INITIAL,
     14260,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_INITIAL,
     14024,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_INITIAL,
     14026,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_INITIAL,
     14028,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_INITIAL,
     14030,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_INITIAL,
     14262,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_INITIAL,
     14036,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_INITIAL,
     14038,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_INITIAL,
     14040,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_INITIAL,
     14042,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_INITIAL,
     14264,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_INITIAL,
     14050,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_INITIAL,
     14056,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_INITIAL,
     14058,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_INITIAL,
     14060,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_INITIAL,
     14062,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_INITIAL,
     14064,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_INITIAL,
     14068,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_INITIAL,
     14070,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_INITIAL,
     14072,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_INITIAL,
     14074,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_INITIAL,
     14076,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_INITIAL,
     14078,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_INITIAL,
     14266,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_INITIAL,
     14080,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_INITIAL,
     14082,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_INITIAL,
     14084,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_INITIAL,
     14086,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_INITIAL,
     14088,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_INITIAL,
     14090,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_INITIAL,
     14094,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_INITIAL,
     14096,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_INITIAL,
     14098,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_INITIAL,
     14100,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_INITIAL,
     14102,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_INITIAL,
     14104,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_INITIAL,
     14106,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_INITIAL,
     14108,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_INITIAL,
     14110,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_INITIAL,
     14116,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_INITIAL,
     14118,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_INITIAL,
     14126,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_INITIAL,
     14128,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_INITIAL,
     14130,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_INITIAL,
     14132,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_INITIAL,
     14134,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_INITIAL,
     14140,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_INITIAL,
     14142,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_INITIAL,
     14144,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_INITIAL,
     14146,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_INITIAL,
     14268,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_INITIAL,
     14152,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_INITIAL,
     14154,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_INITIAL,
     14156,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_INITIAL,
     14158,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_INITIAL,
     14164,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_INITIAL,
     14166,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_INITIAL,
     14168,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_INITIAL,
     14170,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_INITIAL,
     14270,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_INITIAL,
     14176,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_INITIAL,
     14178,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_INITIAL,
     14272,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_INITIAL,
     14184,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_INITIAL,
     14186,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_INITIAL,
     14188,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_INITIAL,
     14190,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_INITIAL,
     14274,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_MEDIAL,
     14020,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_MEDIAL,
     14260,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_MEDIAL,
     14030,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_MEDIAL,
     14262,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_MEDIAL,
     14042,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_MEDIAL,
     14264,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_MEDIAL,
     14050,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_MEDIAL,
     14276,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_MEDIAL,
     14076,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_MEDIAL,
     14278,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_MEDIAL,
     14280,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_MEDIAL,
     14282,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_MEDIAL,
     14132,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_MEDIAL,
     14134,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_MEDIAL,
     14146,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_MEDIAL,
     14170,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_MEDIAL,
     14270,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_MEDIAL,
     14190,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_MEDIAL,
     14274,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_MEDIAL,
     22476,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_MEDIAL,
     22479,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_MEDIAL,
     22482,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_ISOLATED,
     14293,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_ISOLATED,
     14295,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_ISOLATED,
     14297,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_ISOLATED,
     14299,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_ISOLATED,
     14301,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_ISOLATED,
     14303,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_ISOLATED,
     14305,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_ISOLATED,
     14307,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_ISOLATED,
     14309,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_ISOLATED,
     14311,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_ISOLATED,
     14313,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_ISOLATED,
     14315,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_ISOLATED,
     14317,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_ISOLATED,
     14319,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_ISOLATED,
     14321,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_ISOLATED,
     14323,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_ISOLATED,
     14325,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_ISOLATED,
     14327,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_ISOLATED,
     14329,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_ISOLATED,
     14331,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_ISOLATED,
     14333,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_ISOLATED,
     14335,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_ISOLATED,
     14337,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_ISOLATED,
     14280,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_ISOLATED,
     14339,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_ISOLATED,
     14341,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_ISOLATED,
     14343,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_ISOLATED,
     14345,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_FINAL,
     14293,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_FINAL,
     14295,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_FINAL,
     14297,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_FINAL,
     14299,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_FINAL,
     14301,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_FINAL,
     14303,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_FINAL,
     14305,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_FINAL,
     14307,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_FINAL,
     14309,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_FINAL,
     14311,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_FINAL,
     14313,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_FINAL,
     14315,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_FINAL,
     14317,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_FINAL,
     14319,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_FINAL,
     14321,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_FINAL,
     14323,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_FINAL,
     14325,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_FINAL,
     14327,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_FINAL,
     14329,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_FINAL,
     14331,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_FINAL,
     14333,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_FINAL,
     14335,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_FINAL,
     14337,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_FINAL,
     14280,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_FINAL,
     14339,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_FINAL,
     14341,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_FINAL,
     14343,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_FINAL,
     14345,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_INITIAL,
     14333,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_INITIAL,
     14335,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_INITIAL,
     14337,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_INITIAL,
     14280,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_INITIAL,
     14278,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_INITIAL,
     14282,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_INITIAL,
     14092,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_MEDIAL,
     14070,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_MEDIAL,
     14072,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_MEDIAL,
     14074,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_MEDIAL,
     14333,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_MEDIAL,
     14335,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_MEDIAL,
     14337,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_MEDIAL,
     14092,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_MEDIAL,
     14094,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_FINAL,
     14347,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_ISOLATED,
     14347,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_PE,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_INITIAL,
     22541,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_FINAL,
     22544,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_INITIAL,
     22544,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_INITIAL,
     22547,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_INITIAL,
     22550,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_INITIAL,
     22553,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_INITIAL,
     22556,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_INITIAL,
     22559,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_FINAL,
     22562,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_INITIAL,
     22562,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_FINAL,
     22565,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_FINAL,
     22568,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_INITIAL,
     22571,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_INITIAL,
     22574,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_FINAL,
     22577,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_FINAL,
     22580,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_INITIAL,
     22580,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_INITIAL,
     22583,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_FINAL,
     22586,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_INITIAL,
     22586,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_FINAL,
     22589,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_INITIAL,
     22589,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_FINAL,
     22592,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_FINAL,
     22595,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_INITIAL,
     22595,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_FINAL,
     22598,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_FINAL,
     22601,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_INITIAL,
     22601,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_FINAL,
     22604,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_INITIAL,
     22604,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_FINAL,
     22607,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_FINAL,
     22610,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_INITIAL,
     22610,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_FINAL,
     22613,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_INITIAL,
     22613,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_INITIAL,
     22616,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_FINAL,
     22619,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_FINAL,
     22622,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_FINAL,
     22625,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_INITIAL,
     22625,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_FINAL,
     22628,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_FINAL,
     22631,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_FINAL,
     22634,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_FINAL,
     22637,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_FINAL,
     22640,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_INITIAL,
     22640,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_FINAL,
     22643,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_FINAL,
     22646,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_FINAL,
     22649,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_FINAL,
     22652,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_FINAL,
     22655,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_INITIAL,
     22658,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_FINAL,
     22658,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_FINAL,
     22661,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_INITIAL,
     22661,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_FINAL,
     22664,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_INITIAL,
     22664,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_INITIAL,
     22667,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_INITIAL,
     22670,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_FINAL,
     22673,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_INITIAL,
     22676,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_INITIAL,
     22679,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_INITIAL,
     22682,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_INITIAL,
     22685,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_INITIAL,
     22688,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_INITIAL,
     22691,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_INITIAL,
     22694,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_INITIAL,
     22697,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_FINAL,
     22700,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_FINAL,
     22703,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_INITIAL,
     22703,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_FINAL,
     22706,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_FINAL,
     22709,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_FINAL,
     22712,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_FINAL,
     22715,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_INITIAL,
     22715,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_FINAL,
     22718,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_FINAL,
     22721,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_FINAL,
     22724,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_FINAL,
     22727,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_FINAL,
     22730,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_FINAL,
     22733,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_FINAL,
     22736,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_FINAL,
     22739,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_FINAL,
     22742,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_FINAL,
     22745,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_FINAL,
     22748,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_FINAL,
     22751,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_FINAL,
     22754,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_FINAL,
     22757,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_FINAL,
     22760,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_FINAL,
     22763,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_FINAL,
     22766,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_FINAL,
     22769,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_FINAL,
     22772,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_FINAL,
     22775,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_FINAL,
     22778,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_FINAL,
     22781,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_INITIAL,
     22643,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_INITIAL,
     22649,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_FINAL,
     22784,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_FINAL,
     22787,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_INITIAL,
     22790,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_FINAL,
     22793,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_INITIAL,
     22796,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_FINAL,
     22799,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_FINAL,
     22796,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_FINAL,
     22790,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_FINAL,
     22802,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_FINAL,
     22805,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_FINAL,
     22808,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_FINAL,
     22811,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_FINAL,
     22814,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_INITIAL,
     22799,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_INITIAL,
     22622,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_INITIAL,
     22592,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_FINAL,
     22817,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_FINAL,
     22820,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_ISOLATED,
     22823,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_ISOLATED,
     22826,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_ISOLATED,
     31021,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_ISOLATED,
     31025,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_ISOLATED,
     31029,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_ISOLATED,
     31033,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_ISOLATED,
     31037,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_ISOLATED,
     31041,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_ISOLATED,
     31045,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_ISOLATED,
     22857,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_ISOLATED,
     63820,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_ISOLATED,
     63839,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SC,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_ISOLATED,
     31080,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_PO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_VERTICAL,
     6508,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_PO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_VERTICAL,
     6509,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_PO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_VERTICAL,
     6510,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_PO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_VERTICAL,
     6511,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_PO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_VERTICAL,
     814,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_PO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_VERTICAL,
     6512,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_PO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_VERTICAL,
     6513,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_PS,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_VERTICAL,
     6514,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_PE,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_VERTICAL,
     6515,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_PO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_VERTICAL,
     6516,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_PO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_VERTICAL,
     6517,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_PD,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_VERTICAL,
     6518,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_PD,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_VERTICAL,
     6519,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_PC,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_VERTICAL,
     6520,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_PS,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_VERTICAL,
     2802,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_PE,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_VERTICAL,
     2803,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_PS,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_VERTICAL,
     6521,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_PE,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_VERTICAL,
     6522,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_PS,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_VERTICAL,
     6523,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_PE,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_VERTICAL,
     6524,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_PS,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_VERTICAL,
     6525,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_PE,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_VERTICAL,
     6526,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_PS,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_VERTICAL,
     6527,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_PE,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_VERTICAL,
     6528,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_PS,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_VERTICAL,
     3063,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_PE,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_VERTICAL,
     3064,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_PS,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_VERTICAL,
     6529,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_PE,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_VERTICAL,
     6530,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_PS,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_VERTICAL,
     6531,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_PE,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_VERTICAL,
     6532,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_PS,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_VERTICAL,
     6533,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_PE,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_VERTICAL,
     6534,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_PO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     6535,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_PC,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     6520,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_PO,
     0,
     UTF8PROC_BIDI_CLASS_CS,
     UTF8PROC_DECOMP_TYPE_SMALL,
     6508,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_PO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_SMALL,
     6509,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_PO,
     0,
     UTF8PROC_BIDI_CLASS_CS,
     UTF8PROC_DECOMP_TYPE_SMALL,
     2762,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_PO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_SMALL,
     814,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_PO,
     0,
     UTF8PROC_BIDI_CLASS_CS,
     UTF8PROC_DECOMP_TYPE_SMALL,
     6511,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_PO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_SMALL,
     6513,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_PO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_SMALL,
     6512,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_PD,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_SMALL,
     6518,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_PS,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_SMALL,
     2802,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     true,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_PE,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_SMALL,
     2803,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     true,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_PS,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_SMALL,
     6521,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     true,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_PE,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_SMALL,
     6522,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     true,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_PS,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_SMALL,
     6523,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     true,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_PE,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_SMALL,
     6524,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     true,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_PO,
     0,
     UTF8PROC_BIDI_CLASS_ET,
     UTF8PROC_DECOMP_TYPE_SMALL,
     6536,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_PO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_SMALL,
     6537,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_PO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_SMALL,
     6538,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SM,
     0,
     UTF8PROC_BIDI_CLASS_ES,
     UTF8PROC_DECOMP_TYPE_SMALL,
     2799,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_PD,
     0,
     UTF8PROC_BIDI_CLASS_ES,
     UTF8PROC_DECOMP_TYPE_SMALL,
     6539,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SM,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_SMALL,
     6540,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     true,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SM,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_SMALL,
     6541,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     true,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SM,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_SMALL,
     2801,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_PO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_SMALL,
     6542,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SC,
     0,
     UTF8PROC_BIDI_CLASS_ET,
     UTF8PROC_DECOMP_TYPE_SMALL,
     6543,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_PO,
     0,
     UTF8PROC_BIDI_CLASS_ET,
     UTF8PROC_DECOMP_TYPE_SMALL,
     6544,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_PO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_SMALL,
     6545,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_ISOLATED,
     14738,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_MEDIAL,
     14740,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_ISOLATED,
     14742,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_ISOLATED,
     14744,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_ISOLATED,
     14746,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_MEDIAL,
     14748,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_ISOLATED,
     14750,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_MEDIAL,
     14752,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_ISOLATED,
     14754,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_MEDIAL,
     14756,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_ISOLATED,
     14758,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_MEDIAL,
     14760,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_ISOLATED,
     14762,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_MEDIAL,
     14764,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_ISOLATED,
     6574,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_ISOLATED,
     6575,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_FINAL,
     6575,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_ISOLATED,
     6576,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_FINAL,
     6576,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_ISOLATED,
     6577,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_FINAL,
     6577,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_ISOLATED,
     6578,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_FINAL,
     6578,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_ISOLATED,
     6579,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_FINAL,
     6579,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_INITIAL,
     6579,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_MEDIAL,
     6579,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_ISOLATED,
     6580,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_FINAL,
     6580,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_ISOLATED,
     6581,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_FINAL,
     6581,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_INITIAL,
     6581,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_MEDIAL,
     6581,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_ISOLATED,
     6582,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_FINAL,
     6582,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_ISOLATED,
     6583,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_FINAL,
     6583,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_INITIAL,
     6583,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_MEDIAL,
     6583,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_ISOLATED,
     6584,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_FINAL,
     6584,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_INITIAL,
     6584,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_MEDIAL,
     6584,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_ISOLATED,
     6585,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_FINAL,
     6585,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_INITIAL,
     6585,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_MEDIAL,
     6585,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_ISOLATED,
     6586,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_FINAL,
     6586,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_INITIAL,
     6586,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_MEDIAL,
     6586,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_ISOLATED,
     6587,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_FINAL,
     6587,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_INITIAL,
     6587,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_MEDIAL,
     6587,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_ISOLATED,
     6588,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_FINAL,
     6588,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_ISOLATED,
     6589,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_FINAL,
     6589,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_ISOLATED,
     6590,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_FINAL,
     6590,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_ISOLATED,
     6591,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_FINAL,
     6591,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_ISOLATED,
     6592,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_FINAL,
     6592,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_INITIAL,
     6592,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_MEDIAL,
     6592,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_ISOLATED,
     6593,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_FINAL,
     6593,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_INITIAL,
     6593,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_MEDIAL,
     6593,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_ISOLATED,
     6594,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_FINAL,
     6594,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_INITIAL,
     6594,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_MEDIAL,
     6594,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_ISOLATED,
     6595,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_FINAL,
     6595,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_INITIAL,
     6595,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_MEDIAL,
     6595,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_ISOLATED,
     6596,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_FINAL,
     6596,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_INITIAL,
     6596,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_MEDIAL,
     6596,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_ISOLATED,
     6597,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_FINAL,
     6597,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_INITIAL,
     6597,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_MEDIAL,
     6597,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_ISOLATED,
     6598,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_FINAL,
     6598,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_INITIAL,
     6598,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_MEDIAL,
     6598,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_ISOLATED,
     6599,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_FINAL,
     6599,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_INITIAL,
     6599,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_MEDIAL,
     6599,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_ISOLATED,
     6600,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_FINAL,
     6600,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_INITIAL,
     6600,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_MEDIAL,
     6600,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_ISOLATED,
     6601,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_FINAL,
     6601,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_INITIAL,
     6601,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_MEDIAL,
     6601,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_ISOLATED,
     6602,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_FINAL,
     6602,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_INITIAL,
     6602,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_MEDIAL,
     6602,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_ISOLATED,
     6603,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_FINAL,
     6603,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_INITIAL,
     6603,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_MEDIAL,
     6603,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_ISOLATED,
     6604,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_FINAL,
     6604,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_INITIAL,
     6604,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_MEDIAL,
     6604,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_ISOLATED,
     6605,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_FINAL,
     6605,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_INITIAL,
     6605,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_MEDIAL,
     6605,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_ISOLATED,
     6606,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_FINAL,
     6606,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_INITIAL,
     6606,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_MEDIAL,
     6606,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_ISOLATED,
     6607,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_FINAL,
     6607,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_ISOLATED,
     5806,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_FINAL,
     5806,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_ISOLATED,
     6608,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_FINAL,
     6608,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_INITIAL,
     6608,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_MEDIAL,
     6608,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_ISOLATED,
     14801,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_FINAL,
     14801,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_ISOLATED,
     14803,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_FINAL,
     14803,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_ISOLATED,
     14805,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_FINAL,
     14805,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_ISOLATED,
     14807,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_FINAL,
     14807,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_PO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_WIDE,
     6512,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_PO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_WIDE,
     6617,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_PO,
     0,
     UTF8PROC_BIDI_CLASS_ET,
     UTF8PROC_DECOMP_TYPE_WIDE,
     6536,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SC,
     0,
     UTF8PROC_BIDI_CLASS_ET,
     UTF8PROC_DECOMP_TYPE_WIDE,
     6543,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_PO,
     0,
     UTF8PROC_BIDI_CLASS_ET,
     UTF8PROC_DECOMP_TYPE_WIDE,
     6544,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_PO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_WIDE,
     6537,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_PO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_WIDE,
     6618,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_PS,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_WIDE,
     2802,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     true,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_PE,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_WIDE,
     2803,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     true,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_PO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_WIDE,
     6538,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SM,
     0,
     UTF8PROC_BIDI_CLASS_ES,
     UTF8PROC_DECOMP_TYPE_WIDE,
     2799,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_PO,
     0,
     UTF8PROC_BIDI_CLASS_CS,
     UTF8PROC_DECOMP_TYPE_WIDE,
     6508,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_PD,
     0,
     UTF8PROC_BIDI_CLASS_ES,
     UTF8PROC_DECOMP_TYPE_WIDE,
     6539,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_PO,
     0,
     UTF8PROC_BIDI_CLASS_CS,
     UTF8PROC_DECOMP_TYPE_WIDE,
     2762,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_PO,
     0,
     UTF8PROC_BIDI_CLASS_CS,
     UTF8PROC_DECOMP_TYPE_WIDE,
     6619,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_ND,
     0,
     UTF8PROC_BIDI_CLASS_EN,
     UTF8PROC_DECOMP_TYPE_WIDE,
     2792,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_ND,
     0,
     UTF8PROC_BIDI_CLASS_EN,
     UTF8PROC_DECOMP_TYPE_WIDE,
     38,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_ND,
     0,
     UTF8PROC_BIDI_CLASS_EN,
     UTF8PROC_DECOMP_TYPE_WIDE,
     31,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_ND,
     0,
     UTF8PROC_BIDI_CLASS_EN,
     UTF8PROC_DECOMP_TYPE_WIDE,
     32,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_ND,
     0,
     UTF8PROC_BIDI_CLASS_EN,
     UTF8PROC_DECOMP_TYPE_WIDE,
     2793,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_ND,
     0,
     UTF8PROC_BIDI_CLASS_EN,
     UTF8PROC_DECOMP_TYPE_WIDE,
     2794,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_ND,
     0,
     UTF8PROC_BIDI_CLASS_EN,
     UTF8PROC_DECOMP_TYPE_WIDE,
     2795,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_ND,
     0,
     UTF8PROC_BIDI_CLASS_EN,
     UTF8PROC_DECOMP_TYPE_WIDE,
     2796,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_ND,
     0,
     UTF8PROC_BIDI_CLASS_EN,
     UTF8PROC_DECOMP_TYPE_WIDE,
     2797,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_ND,
     0,
     UTF8PROC_BIDI_CLASS_EN,
     UTF8PROC_DECOMP_TYPE_WIDE,
     2798,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_PO,
     0,
     UTF8PROC_BIDI_CLASS_CS,
     UTF8PROC_DECOMP_TYPE_WIDE,
     6511,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_PO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_WIDE,
     814,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SM,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_WIDE,
     6540,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     true,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SM,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_WIDE,
     2801,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SM,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_WIDE,
     6541,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     true,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_PO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_WIDE,
     6513,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_PO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_WIDE,
     6545,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_WIDE,
     1491,
     6620,
     UINT16_MAX,
     6620,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_WIDE,
     1493,
     6621,
     UINT16_MAX,
     6621,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_WIDE,
     2812,
     6622,
     UINT16_MAX,
     6622,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_WIDE,
     1494,
     6623,
     UINT16_MAX,
     6623,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_WIDE,
     1495,
     6624,
     UINT16_MAX,
     6624,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_WIDE,
     2837,
     6625,
     UINT16_MAX,
     6625,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_WIDE,
     1497,
     6626,
     UINT16_MAX,
     6626,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_WIDE,
     1498,
     6627,
     UINT16_MAX,
     6627,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_WIDE,
     1499,
     6628,
     UINT16_MAX,
     6628,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_WIDE,
     1500,
     6629,
     UINT16_MAX,
     6629,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_WIDE,
     1501,
     6630,
     UINT16_MAX,
     6630,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_WIDE,
     1502,
     6631,
     UINT16_MAX,
     6631,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_WIDE,
     1503,
     6632,
     UINT16_MAX,
     6632,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_WIDE,
     1504,
     6633,
     UINT16_MAX,
     6633,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_WIDE,
     1505,
     6634,
     UINT16_MAX,
     6634,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_WIDE,
     1507,
     6635,
     UINT16_MAX,
     6635,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_WIDE,
     2826,
     6636,
     UINT16_MAX,
     6636,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_WIDE,
     1508,
     6637,
     UINT16_MAX,
     6637,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_WIDE,
     3305,
     6638,
     UINT16_MAX,
     6638,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_WIDE,
     1509,
     6639,
     UINT16_MAX,
     6639,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_WIDE,
     1510,
     6640,
     UINT16_MAX,
     6640,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_WIDE,
     2908,
     6641,
     UINT16_MAX,
     6641,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_WIDE,
     1511,
     6642,
     UINT16_MAX,
     6642,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_WIDE,
     2925,
     6643,
     UINT16_MAX,
     6643,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_WIDE,
     3312,
     6644,
     UINT16_MAX,
     6644,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_WIDE,
     2834,
     6645,
     UINT16_MAX,
     6645,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_PS,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_WIDE,
     6533,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     true,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_PO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_WIDE,
     6542,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_PE,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_WIDE,
     6534,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     true,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SK,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_WIDE,
     6646,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_PC,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_WIDE,
     6520,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SK,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_WIDE,
     2722,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_WIDE,
     0,
     UINT16_MAX,
     9178,
     UINT16_MAX,
     9178,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_WIDE,
     1,
     UINT16_MAX,
     9179,
     UINT16_MAX,
     9179,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_WIDE,
     2,
     UINT16_MAX,
     9180,
     UINT16_MAX,
     9180,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_WIDE,
     3,
     UINT16_MAX,
     9181,
     UINT16_MAX,
     9181,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_WIDE,
     4,
     UINT16_MAX,
     9182,
     UINT16_MAX,
     9182,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_WIDE,
     5,
     UINT16_MAX,
     9183,
     UINT16_MAX,
     9183,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_WIDE,
     6,
     UINT16_MAX,
     9184,
     UINT16_MAX,
     9184,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_WIDE,
     7,
     UINT16_MAX,
     9185,
     UINT16_MAX,
     9185,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_WIDE,
     8,
     UINT16_MAX,
     9186,
     UINT16_MAX,
     9186,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_WIDE,
     9,
     UINT16_MAX,
     9187,
     UINT16_MAX,
     9187,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_WIDE,
     10,
     UINT16_MAX,
     9188,
     UINT16_MAX,
     9188,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_WIDE,
     11,
     UINT16_MAX,
     9189,
     UINT16_MAX,
     9189,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_WIDE,
     12,
     UINT16_MAX,
     9190,
     UINT16_MAX,
     9190,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_WIDE,
     13,
     UINT16_MAX,
     9191,
     UINT16_MAX,
     9191,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_WIDE,
     14,
     UINT16_MAX,
     9192,
     UINT16_MAX,
     9192,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_WIDE,
     15,
     UINT16_MAX,
     9193,
     UINT16_MAX,
     9193,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_WIDE,
     16,
     UINT16_MAX,
     9194,
     UINT16_MAX,
     9194,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_WIDE,
     17,
     UINT16_MAX,
     9195,
     UINT16_MAX,
     9195,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_WIDE,
     18,
     UINT16_MAX,
     9196,
     UINT16_MAX,
     9196,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_WIDE,
     19,
     UINT16_MAX,
     9197,
     UINT16_MAX,
     9197,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_WIDE,
     20,
     UINT16_MAX,
     9198,
     UINT16_MAX,
     9198,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_WIDE,
     21,
     UINT16_MAX,
     9199,
     UINT16_MAX,
     9199,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_WIDE,
     22,
     UINT16_MAX,
     9200,
     UINT16_MAX,
     9200,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_WIDE,
     23,
     UINT16_MAX,
     9201,
     UINT16_MAX,
     9201,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_WIDE,
     24,
     UINT16_MAX,
     9202,
     UINT16_MAX,
     9202,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_WIDE,
     25,
     UINT16_MAX,
     9203,
     UINT16_MAX,
     9203,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_PS,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_WIDE,
     6521,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     true,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SM,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_WIDE,
     6647,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_PE,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_WIDE,
     6522,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     true,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SM,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_WIDE,
     6648,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_PS,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_WIDE,
     6649,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     true,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_PE,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_WIDE,
     6650,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     true,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_PO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_NARROW,
     6510,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_PS,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_NARROW,
     6529,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     true,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_PE,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_NARROW,
     6530,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     true,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_PO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_NARROW,
     6509,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_PO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_NARROW,
     6651,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_NARROW,
     4331,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_NARROW,
     6652,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_NARROW,
     6653,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_NARROW,
     6654,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_NARROW,
     6655,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_NARROW,
     6656,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_NARROW,
     6657,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_NARROW,
     6658,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_NARROW,
     6659,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_NARROW,
     6660,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LM,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_NARROW,
     6661,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_NARROW,
     4285,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_NARROW,
     4286,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_NARROW,
     4287,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_NARROW,
     4288,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_NARROW,
     4289,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_NARROW,
     4290,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_NARROW,
     4291,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_NARROW,
     4292,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_NARROW,
     4293,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_NARROW,
     4294,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_NARROW,
     4295,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_NARROW,
     4296,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_NARROW,
     4297,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_NARROW,
     4298,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_NARROW,
     4299,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_NARROW,
     4300,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_NARROW,
     4301,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_NARROW,
     4302,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_NARROW,
     4303,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_NARROW,
     4304,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_NARROW,
     4305,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_NARROW,
     4306,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_NARROW,
     4307,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_NARROW,
     4308,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_NARROW,
     4309,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_NARROW,
     4310,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_NARROW,
     4311,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_NARROW,
     4312,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_NARROW,
     4313,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_NARROW,
     4314,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_NARROW,
     4315,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_NARROW,
     4316,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_NARROW,
     4317,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_NARROW,
     4318,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_NARROW,
     4319,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_NARROW,
     4320,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_NARROW,
     4321,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_NARROW,
     4322,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_NARROW,
     4323,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_NARROW,
     4324,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_NARROW,
     4325,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_NARROW,
     4326,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_NARROW,
     4327,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_NARROW,
     4328,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_NARROW,
     6662,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LM,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_NARROW,
     6663,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_EXTEND},
    {UTF8PROC_CATEGORY_LM,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_NARROW,
     6664,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_EXTEND},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_NARROW,
     6665,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     true,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_NARROW,
     6666,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_NARROW,
     6667,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_NARROW,
     6668,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_NARROW,
     6669,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_NARROW,
     6670,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_NARROW,
     6671,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_NARROW,
     6672,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_NARROW,
     6673,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_NARROW,
     6674,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_NARROW,
     6675,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_NARROW,
     6676,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_NARROW,
     6677,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_NARROW,
     6678,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_NARROW,
     6679,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_NARROW,
     6680,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_NARROW,
     6681,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_NARROW,
     6682,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_NARROW,
     6683,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_NARROW,
     6684,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_NARROW,
     6685,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_NARROW,
     6686,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_NARROW,
     6687,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_NARROW,
     6688,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_NARROW,
     6689,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_NARROW,
     6690,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_NARROW,
     6691,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_NARROW,
     6692,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_NARROW,
     6693,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_NARROW,
     6694,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_NARROW,
     6695,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_NARROW,
     6696,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_NARROW,
     6697,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_NARROW,
     6698,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_NARROW,
     6699,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_NARROW,
     6700,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_NARROW,
     6701,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_NARROW,
     6702,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_NARROW,
     6703,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_NARROW,
     6704,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_NARROW,
     6705,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_NARROW,
     6706,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_NARROW,
     6707,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_NARROW,
     6708,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_NARROW,
     6709,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_NARROW,
     6710,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_NARROW,
     6711,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_NARROW,
     6712,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_NARROW,
     6713,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_NARROW,
     6714,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_NARROW,
     6715,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_NARROW,
     6716,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SC,
     0,
     UTF8PROC_BIDI_CLASS_ET,
     UTF8PROC_DECOMP_TYPE_WIDE,
     6717,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SC,
     0,
     UTF8PROC_BIDI_CLASS_ET,
     UTF8PROC_DECOMP_TYPE_WIDE,
     6718,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SM,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_WIDE,
     6719,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SK,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_WIDE,
     6720,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_WIDE,
     6721,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SC,
     0,
     UTF8PROC_BIDI_CLASS_ET,
     UTF8PROC_DECOMP_TYPE_WIDE,
     6722,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SC,
     0,
     UTF8PROC_BIDI_CLASS_ET,
     UTF8PROC_DECOMP_TYPE_WIDE,
     6723,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_NARROW,
     6724,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SM,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_NARROW,
     6725,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SM,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_NARROW,
     6726,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SM,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_NARROW,
     6727,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SM,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_NARROW,
     6728,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_NARROW,
     6729,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_NARROW,
     6730,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_CF,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     true,
     0,
     0,
     UTF8PROC_BOUNDCLASS_CONTROL},
    {UTF8PROC_CATEGORY_NL,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_NO,
     0,
     UTF8PROC_BIDI_CLASS_EN,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     6731,
     UINT16_MAX,
     6731,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     6733,
     UINT16_MAX,
     6733,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     6735,
     UINT16_MAX,
     6735,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     6737,
     UINT16_MAX,
     6737,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     6739,
     UINT16_MAX,
     6739,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     6741,
     UINT16_MAX,
     6741,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     6743,
     UINT16_MAX,
     6743,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     6745,
     UINT16_MAX,
     6745,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     6747,
     UINT16_MAX,
     6747,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     6749,
     UINT16_MAX,
     6749,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     6751,
     UINT16_MAX,
     6751,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     6753,
     UINT16_MAX,
     6753,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     6755,
     UINT16_MAX,
     6755,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     6757,
     UINT16_MAX,
     6757,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     6759,
     UINT16_MAX,
     6759,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     6761,
     UINT16_MAX,
     6761,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     6763,
     UINT16_MAX,
     6763,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     6765,
     UINT16_MAX,
     6765,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     6767,
     UINT16_MAX,
     6767,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     6769,
     UINT16_MAX,
     6769,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     6771,
     UINT16_MAX,
     6771,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     6773,
     UINT16_MAX,
     6773,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     6775,
     UINT16_MAX,
     6775,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     6777,
     UINT16_MAX,
     6777,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     6779,
     UINT16_MAX,
     6779,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     6781,
     UINT16_MAX,
     6781,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     6783,
     UINT16_MAX,
     6783,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     6785,
     UINT16_MAX,
     6785,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     6787,
     UINT16_MAX,
     6787,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     6789,
     UINT16_MAX,
     6789,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     6791,
     UINT16_MAX,
     6791,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     6793,
     UINT16_MAX,
     6793,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     6795,
     UINT16_MAX,
     6795,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     6797,
     UINT16_MAX,
     6797,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     6799,
     UINT16_MAX,
     6799,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     6801,
     UINT16_MAX,
     6801,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     6803,
     UINT16_MAX,
     6803,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     6805,
     UINT16_MAX,
     6805,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     6807,
     UINT16_MAX,
     6807,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     6809,
     UINT16_MAX,
     6809,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9204,
     UINT16_MAX,
     9204,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9206,
     UINT16_MAX,
     9206,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9208,
     UINT16_MAX,
     9208,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9210,
     UINT16_MAX,
     9210,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9212,
     UINT16_MAX,
     9212,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9214,
     UINT16_MAX,
     9214,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9216,
     UINT16_MAX,
     9216,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9218,
     UINT16_MAX,
     9218,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9220,
     UINT16_MAX,
     9220,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9222,
     UINT16_MAX,
     9222,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9224,
     UINT16_MAX,
     9224,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9226,
     UINT16_MAX,
     9226,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9228,
     UINT16_MAX,
     9228,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9230,
     UINT16_MAX,
     9230,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9232,
     UINT16_MAX,
     9232,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9234,
     UINT16_MAX,
     9234,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9236,
     UINT16_MAX,
     9236,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9238,
     UINT16_MAX,
     9238,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9240,
     UINT16_MAX,
     9240,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9242,
     UINT16_MAX,
     9242,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9244,
     UINT16_MAX,
     9244,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9246,
     UINT16_MAX,
     9246,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9248,
     UINT16_MAX,
     9248,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9250,
     UINT16_MAX,
     9250,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9252,
     UINT16_MAX,
     9252,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9254,
     UINT16_MAX,
     9254,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9256,
     UINT16_MAX,
     9256,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9258,
     UINT16_MAX,
     9258,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9260,
     UINT16_MAX,
     9260,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9262,
     UINT16_MAX,
     9262,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9264,
     UINT16_MAX,
     9264,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9266,
     UINT16_MAX,
     9266,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9268,
     UINT16_MAX,
     9268,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9270,
     UINT16_MAX,
     9270,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9272,
     UINT16_MAX,
     9272,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9274,
     UINT16_MAX,
     9274,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9276,
     UINT16_MAX,
     9276,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9278,
     UINT16_MAX,
     9278,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9280,
     UINT16_MAX,
     9280,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9282,
     UINT16_MAX,
     9282,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     6811,
     UINT16_MAX,
     6811,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     6813,
     UINT16_MAX,
     6813,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     6815,
     UINT16_MAX,
     6815,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     6817,
     UINT16_MAX,
     6817,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     6819,
     UINT16_MAX,
     6819,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     6821,
     UINT16_MAX,
     6821,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     6823,
     UINT16_MAX,
     6823,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     6825,
     UINT16_MAX,
     6825,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     6827,
     UINT16_MAX,
     6827,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     6829,
     UINT16_MAX,
     6829,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     6831,
     UINT16_MAX,
     6831,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     6833,
     UINT16_MAX,
     6833,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     6835,
     UINT16_MAX,
     6835,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     6837,
     UINT16_MAX,
     6837,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     6839,
     UINT16_MAX,
     6839,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     6841,
     UINT16_MAX,
     6841,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     6843,
     UINT16_MAX,
     6843,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     6845,
     UINT16_MAX,
     6845,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     6847,
     UINT16_MAX,
     6847,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     6849,
     UINT16_MAX,
     6849,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     6851,
     UINT16_MAX,
     6851,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     6853,
     UINT16_MAX,
     6853,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     6855,
     UINT16_MAX,
     6855,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     6857,
     UINT16_MAX,
     6857,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     6859,
     UINT16_MAX,
     6859,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     6861,
     UINT16_MAX,
     6861,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     6863,
     UINT16_MAX,
     6863,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     6865,
     UINT16_MAX,
     6865,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     6867,
     UINT16_MAX,
     6867,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     6869,
     UINT16_MAX,
     6869,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     6871,
     UINT16_MAX,
     6871,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     6873,
     UINT16_MAX,
     6873,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     6875,
     UINT16_MAX,
     6875,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     6877,
     UINT16_MAX,
     6877,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     6879,
     UINT16_MAX,
     6879,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     6881,
     UINT16_MAX,
     6881,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9284,
     UINT16_MAX,
     9284,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9286,
     UINT16_MAX,
     9286,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9288,
     UINT16_MAX,
     9288,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9290,
     UINT16_MAX,
     9290,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9292,
     UINT16_MAX,
     9292,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9294,
     UINT16_MAX,
     9294,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9296,
     UINT16_MAX,
     9296,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9298,
     UINT16_MAX,
     9298,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9300,
     UINT16_MAX,
     9300,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9302,
     UINT16_MAX,
     9302,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9304,
     UINT16_MAX,
     9304,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9306,
     UINT16_MAX,
     9306,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9308,
     UINT16_MAX,
     9308,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9310,
     UINT16_MAX,
     9310,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9312,
     UINT16_MAX,
     9312,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9314,
     UINT16_MAX,
     9314,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9316,
     UINT16_MAX,
     9316,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9318,
     UINT16_MAX,
     9318,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9320,
     UINT16_MAX,
     9320,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9322,
     UINT16_MAX,
     9322,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9324,
     UINT16_MAX,
     9324,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9326,
     UINT16_MAX,
     9326,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9328,
     UINT16_MAX,
     9328,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9330,
     UINT16_MAX,
     9330,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9332,
     UINT16_MAX,
     9332,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9334,
     UINT16_MAX,
     9334,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9336,
     UINT16_MAX,
     9336,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9338,
     UINT16_MAX,
     9338,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9340,
     UINT16_MAX,
     9340,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9342,
     UINT16_MAX,
     9342,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9344,
     UINT16_MAX,
     9344,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9346,
     UINT16_MAX,
     9346,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9348,
     UINT16_MAX,
     9348,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9350,
     UINT16_MAX,
     9350,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9352,
     UINT16_MAX,
     9352,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9354,
     UINT16_MAX,
     9354,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_NO,
     0,
     UTF8PROC_BIDI_CLASS_R,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_R,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_R,
     0,
     UINT16_MAX,
     6883,
     UINT16_MAX,
     6883,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_R,
     0,
     UINT16_MAX,
     6885,
     UINT16_MAX,
     6885,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_R,
     0,
     UINT16_MAX,
     6887,
     UINT16_MAX,
     6887,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_R,
     0,
     UINT16_MAX,
     6889,
     UINT16_MAX,
     6889,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_R,
     0,
     UINT16_MAX,
     6891,
     UINT16_MAX,
     6891,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_R,
     0,
     UINT16_MAX,
     6893,
     UINT16_MAX,
     6893,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_R,
     0,
     UINT16_MAX,
     6895,
     UINT16_MAX,
     6895,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_R,
     0,
     UINT16_MAX,
     6897,
     UINT16_MAX,
     6897,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_R,
     0,
     UINT16_MAX,
     6899,
     UINT16_MAX,
     6899,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_R,
     0,
     UINT16_MAX,
     6901,
     UINT16_MAX,
     6901,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_R,
     0,
     UINT16_MAX,
     6903,
     UINT16_MAX,
     6903,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_R,
     0,
     UINT16_MAX,
     6905,
     UINT16_MAX,
     6905,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_R,
     0,
     UINT16_MAX,
     6907,
     UINT16_MAX,
     6907,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_R,
     0,
     UINT16_MAX,
     6909,
     UINT16_MAX,
     6909,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_R,
     0,
     UINT16_MAX,
     6911,
     UINT16_MAX,
     6911,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_R,
     0,
     UINT16_MAX,
     6913,
     UINT16_MAX,
     6913,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_R,
     0,
     UINT16_MAX,
     6915,
     UINT16_MAX,
     6915,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_R,
     0,
     UINT16_MAX,
     6917,
     UINT16_MAX,
     6917,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_R,
     0,
     UINT16_MAX,
     6919,
     UINT16_MAX,
     6919,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_R,
     0,
     UINT16_MAX,
     6921,
     UINT16_MAX,
     6921,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_R,
     0,
     UINT16_MAX,
     6923,
     UINT16_MAX,
     6923,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_R,
     0,
     UINT16_MAX,
     6925,
     UINT16_MAX,
     6925,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_R,
     0,
     UINT16_MAX,
     6927,
     UINT16_MAX,
     6927,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_R,
     0,
     UINT16_MAX,
     6929,
     UINT16_MAX,
     6929,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_R,
     0,
     UINT16_MAX,
     6931,
     UINT16_MAX,
     6931,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_R,
     0,
     UINT16_MAX,
     6933,
     UINT16_MAX,
     6933,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_R,
     0,
     UINT16_MAX,
     6935,
     UINT16_MAX,
     6935,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_R,
     0,
     UINT16_MAX,
     6937,
     UINT16_MAX,
     6937,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_R,
     0,
     UINT16_MAX,
     6939,
     UINT16_MAX,
     6939,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_R,
     0,
     UINT16_MAX,
     6941,
     UINT16_MAX,
     6941,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_R,
     0,
     UINT16_MAX,
     6943,
     UINT16_MAX,
     6943,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_R,
     0,
     UINT16_MAX,
     6945,
     UINT16_MAX,
     6945,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_R,
     0,
     UINT16_MAX,
     6947,
     UINT16_MAX,
     6947,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_R,
     0,
     UINT16_MAX,
     6949,
     UINT16_MAX,
     6949,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_R,
     0,
     UINT16_MAX,
     6951,
     UINT16_MAX,
     6951,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_R,
     0,
     UINT16_MAX,
     6953,
     UINT16_MAX,
     6953,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_R,
     0,
     UINT16_MAX,
     6955,
     UINT16_MAX,
     6955,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_R,
     0,
     UINT16_MAX,
     6957,
     UINT16_MAX,
     6957,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_R,
     0,
     UINT16_MAX,
     6959,
     UINT16_MAX,
     6959,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_R,
     0,
     UINT16_MAX,
     6961,
     UINT16_MAX,
     6961,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_R,
     0,
     UINT16_MAX,
     6963,
     UINT16_MAX,
     6963,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_R,
     0,
     UINT16_MAX,
     6965,
     UINT16_MAX,
     6965,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_R,
     0,
     UINT16_MAX,
     6967,
     UINT16_MAX,
     6967,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_R,
     0,
     UINT16_MAX,
     6969,
     UINT16_MAX,
     6969,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_R,
     0,
     UINT16_MAX,
     6971,
     UINT16_MAX,
     6971,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_R,
     0,
     UINT16_MAX,
     6973,
     UINT16_MAX,
     6973,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_R,
     0,
     UINT16_MAX,
     6975,
     UINT16_MAX,
     6975,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_R,
     0,
     UINT16_MAX,
     6977,
     UINT16_MAX,
     6977,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_R,
     0,
     UINT16_MAX,
     6979,
     UINT16_MAX,
     6979,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_R,
     0,
     UINT16_MAX,
     6981,
     UINT16_MAX,
     6981,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_R,
     0,
     UINT16_MAX,
     6983,
     UINT16_MAX,
     6983,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_R,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9356,
     UINT16_MAX,
     9356,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_R,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9358,
     UINT16_MAX,
     9358,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_R,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9360,
     UINT16_MAX,
     9360,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_R,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9362,
     UINT16_MAX,
     9362,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_R,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9364,
     UINT16_MAX,
     9364,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_R,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9366,
     UINT16_MAX,
     9366,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_R,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9368,
     UINT16_MAX,
     9368,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_R,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9370,
     UINT16_MAX,
     9370,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_R,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9372,
     UINT16_MAX,
     9372,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_R,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9374,
     UINT16_MAX,
     9374,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_R,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9376,
     UINT16_MAX,
     9376,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_R,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9378,
     UINT16_MAX,
     9378,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_R,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9380,
     UINT16_MAX,
     9380,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_R,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9382,
     UINT16_MAX,
     9382,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_R,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9384,
     UINT16_MAX,
     9384,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_R,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9386,
     UINT16_MAX,
     9386,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_R,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9388,
     UINT16_MAX,
     9388,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_R,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9390,
     UINT16_MAX,
     9390,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_R,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9392,
     UINT16_MAX,
     9392,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_R,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9394,
     UINT16_MAX,
     9394,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_R,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9396,
     UINT16_MAX,
     9396,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_R,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9398,
     UINT16_MAX,
     9398,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_R,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9400,
     UINT16_MAX,
     9400,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_R,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9402,
     UINT16_MAX,
     9402,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_R,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9404,
     UINT16_MAX,
     9404,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_R,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9406,
     UINT16_MAX,
     9406,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_R,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9408,
     UINT16_MAX,
     9408,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_R,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9410,
     UINT16_MAX,
     9410,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_R,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9412,
     UINT16_MAX,
     9412,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_R,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9414,
     UINT16_MAX,
     9414,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_R,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9416,
     UINT16_MAX,
     9416,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_R,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9418,
     UINT16_MAX,
     9418,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_R,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9420,
     UINT16_MAX,
     9420,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_R,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9422,
     UINT16_MAX,
     9422,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_R,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9424,
     UINT16_MAX,
     9424,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_R,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9426,
     UINT16_MAX,
     9426,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_R,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9428,
     UINT16_MAX,
     9428,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_R,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9430,
     UINT16_MAX,
     9430,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_R,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9432,
     UINT16_MAX,
     9432,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_R,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9434,
     UINT16_MAX,
     9434,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_R,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9436,
     UINT16_MAX,
     9436,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_R,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9438,
     UINT16_MAX,
     9438,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_R,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9440,
     UINT16_MAX,
     9440,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_R,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9442,
     UINT16_MAX,
     9442,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_R,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9444,
     UINT16_MAX,
     9444,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_R,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9446,
     UINT16_MAX,
     9446,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_R,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9448,
     UINT16_MAX,
     9448,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_R,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9450,
     UINT16_MAX,
     9450,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_R,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9452,
     UINT16_MAX,
     9452,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_R,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9454,
     UINT16_MAX,
     9454,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_R,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9456,
     UINT16_MAX,
     9456,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_NO,
     0,
     UTF8PROC_BIDI_CLASS_AN,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_NO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     5328,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     15177,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     5332,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     15181,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     5336,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     15185,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_MN,
     7,
     UTF8PROC_BIDI_CLASS_NSM,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     49206,
     false,
     false,
     false,
     false,
     0,
     0,
     UTF8PROC_BOUNDCLASS_EXTEND},
    {UTF8PROC_CATEGORY_CF,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     true,
     0,
     0,
     UTF8PROC_BOUNDCLASS_PREPEND},
    {UTF8PROC_CATEGORY_MN,
     0,
     UTF8PROC_BIDI_CLASS_NSM,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     49208,
     false,
     false,
     false,
     false,
     0,
     0,
     UTF8PROC_BOUNDCLASS_EXTEND},
    {UTF8PROC_CATEGORY_MN,
     0,
     UTF8PROC_BIDI_CLASS_NSM,
     0,
     15189,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     0,
     0,
     UTF8PROC_BOUNDCLASS_EXTEND},
    {UTF8PROC_CATEGORY_MN,
     0,
     UTF8PROC_BIDI_CLASS_NSM,
     0,
     15193,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     0,
     0,
     UTF8PROC_BOUNDCLASS_EXTEND},
    {UTF8PROC_CATEGORY_MN,
     0,
     UTF8PROC_BIDI_CLASS_NSM,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     5340,
     false,
     false,
     false,
     false,
     0,
     0,
     UTF8PROC_BOUNDCLASS_EXTEND},
    {UTF8PROC_CATEGORY_MN,
     0,
     UTF8PROC_BIDI_CLASS_NSM,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     5344,
     false,
     false,
     false,
     false,
     0,
     0,
     UTF8PROC_BOUNDCLASS_EXTEND},
    {UTF8PROC_CATEGORY_MC,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     49210,
     false,
     false,
     false,
     false,
     0,
     0,
     UTF8PROC_BOUNDCLASS_EXTEND},
    {UTF8PROC_CATEGORY_MC,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     5348,
     false,
     false,
     false,
     false,
     0,
     0,
     UTF8PROC_BOUNDCLASS_SPACINGMARK},
    {UTF8PROC_CATEGORY_MC,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     15197,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     0,
     0,
     UTF8PROC_BOUNDCLASS_SPACINGMARK},
    {UTF8PROC_CATEGORY_MC,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     15201,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     0,
     0,
     UTF8PROC_BOUNDCLASS_SPACINGMARK},
    {UTF8PROC_CATEGORY_MC,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     49212,
     false,
     false,
     false,
     false,
     0,
     0,
     UTF8PROC_BOUNDCLASS_EXTEND},
    {UTF8PROC_CATEGORY_MC,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     49216,
     false,
     false,
     false,
     false,
     0,
     0,
     UTF8PROC_BOUNDCLASS_EXTEND},
    {UTF8PROC_CATEGORY_MC,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     5354,
     false,
     false,
     false,
     false,
     0,
     0,
     UTF8PROC_BOUNDCLASS_SPACINGMARK},
    {UTF8PROC_CATEGORY_MN,
     0,
     UTF8PROC_BIDI_CLASS_NSM,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     49214,
     false,
     false,
     false,
     false,
     0,
     0,
     UTF8PROC_BOUNDCLASS_EXTEND},
    {UTF8PROC_CATEGORY_MC,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     15205,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     0,
     0,
     UTF8PROC_BOUNDCLASS_SPACINGMARK},
    {UTF8PROC_CATEGORY_MC,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     15209,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     0,
     0,
     UTF8PROC_BOUNDCLASS_SPACINGMARK},
    {UTF8PROC_CATEGORY_MC,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     49218,
     false,
     false,
     false,
     false,
     0,
     0,
     UTF8PROC_BOUNDCLASS_EXTEND},
    {UTF8PROC_CATEGORY_MC,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     15213,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     0,
     0,
     UTF8PROC_BOUNDCLASS_SPACINGMARK},
    {UTF8PROC_CATEGORY_MC,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     49220,
     false,
     false,
     false,
     false,
     0,
     0,
     UTF8PROC_BOUNDCLASS_EXTEND},
    {UTF8PROC_CATEGORY_MC,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     5362,
     false,
     false,
     false,
     false,
     0,
     0,
     UTF8PROC_BOUNDCLASS_SPACINGMARK},
    {UTF8PROC_CATEGORY_MC,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     5366,
     false,
     false,
     false,
     false,
     0,
     0,
     UTF8PROC_BOUNDCLASS_SPACINGMARK},
    {UTF8PROC_CATEGORY_MC,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     15217,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     0,
     0,
     UTF8PROC_BOUNDCLASS_SPACINGMARK},
    {UTF8PROC_CATEGORY_MC,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     15221,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     0,
     0,
     UTF8PROC_BOUNDCLASS_SPACINGMARK},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     7033,
     UINT16_MAX,
     7033,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     7035,
     UINT16_MAX,
     7035,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     7037,
     UINT16_MAX,
     7037,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     7039,
     UINT16_MAX,
     7039,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     7041,
     UINT16_MAX,
     7041,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     7043,
     UINT16_MAX,
     7043,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     7045,
     UINT16_MAX,
     7045,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     7047,
     UINT16_MAX,
     7047,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     7049,
     UINT16_MAX,
     7049,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     7051,
     UINT16_MAX,
     7051,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     7053,
     UINT16_MAX,
     7053,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     7055,
     UINT16_MAX,
     7055,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     7057,
     UINT16_MAX,
     7057,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     7059,
     UINT16_MAX,
     7059,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     7061,
     UINT16_MAX,
     7061,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     7063,
     UINT16_MAX,
     7063,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     7065,
     UINT16_MAX,
     7065,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     7067,
     UINT16_MAX,
     7067,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     7069,
     UINT16_MAX,
     7069,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     7071,
     UINT16_MAX,
     7071,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     7073,
     UINT16_MAX,
     7073,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     7075,
     UINT16_MAX,
     7075,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     7077,
     UINT16_MAX,
     7077,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     7079,
     UINT16_MAX,
     7079,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     7081,
     UINT16_MAX,
     7081,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     7083,
     UINT16_MAX,
     7083,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     7085,
     UINT16_MAX,
     7085,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     7087,
     UINT16_MAX,
     7087,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     7089,
     UINT16_MAX,
     7089,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     7091,
     UINT16_MAX,
     7091,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     7093,
     UINT16_MAX,
     7093,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     7095,
     UINT16_MAX,
     7095,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9458,
     UINT16_MAX,
     9458,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9460,
     UINT16_MAX,
     9460,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9462,
     UINT16_MAX,
     9462,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9464,
     UINT16_MAX,
     9464,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9466,
     UINT16_MAX,
     9466,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9468,
     UINT16_MAX,
     9468,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9470,
     UINT16_MAX,
     9470,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9472,
     UINT16_MAX,
     9472,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9474,
     UINT16_MAX,
     9474,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9476,
     UINT16_MAX,
     9476,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9478,
     UINT16_MAX,
     9478,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9480,
     UINT16_MAX,
     9480,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9482,
     UINT16_MAX,
     9482,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9484,
     UINT16_MAX,
     9484,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9486,
     UINT16_MAX,
     9486,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9488,
     UINT16_MAX,
     9488,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9490,
     UINT16_MAX,
     9490,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9492,
     UINT16_MAX,
     9492,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9494,
     UINT16_MAX,
     9494,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9496,
     UINT16_MAX,
     9496,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9498,
     UINT16_MAX,
     9498,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9500,
     UINT16_MAX,
     9500,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9502,
     UINT16_MAX,
     9502,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9504,
     UINT16_MAX,
     9504,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9506,
     UINT16_MAX,
     9506,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9508,
     UINT16_MAX,
     9508,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9510,
     UINT16_MAX,
     9510,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9512,
     UINT16_MAX,
     9512,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9514,
     UINT16_MAX,
     9514,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9516,
     UINT16_MAX,
     9516,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9518,
     UINT16_MAX,
     9518,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9520,
     UINT16_MAX,
     9520,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_MC,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     49222,
     false,
     false,
     false,
     false,
     0,
     0,
     UTF8PROC_BOUNDCLASS_EXTEND},
    {UTF8PROC_CATEGORY_MC,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     5370,
     false,
     false,
     false,
     false,
     0,
     0,
     UTF8PROC_BOUNDCLASS_SPACINGMARK},
    {UTF8PROC_CATEGORY_MC,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     15289,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     0,
     0,
     UTF8PROC_BOUNDCLASS_SPACINGMARK},
    {UTF8PROC_CATEGORY_MN,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     0,
     0,
     UTF8PROC_BOUNDCLASS_EXTEND},
    {UTF8PROC_CATEGORY_MN,
     9,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     0,
     0,
     UTF8PROC_BOUNDCLASS_EXTEND},
    {UTF8PROC_CATEGORY_CF,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     true,
     0,
     0,
     UTF8PROC_BOUNDCLASS_CONTROL},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     7101,
     UINT16_MAX,
     7101,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     7103,
     UINT16_MAX,
     7103,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     7105,
     UINT16_MAX,
     7105,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     7107,
     UINT16_MAX,
     7107,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     7109,
     UINT16_MAX,
     7109,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     7111,
     UINT16_MAX,
     7111,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     7113,
     UINT16_MAX,
     7113,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     7115,
     UINT16_MAX,
     7115,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     7117,
     UINT16_MAX,
     7117,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     7119,
     UINT16_MAX,
     7119,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     7121,
     UINT16_MAX,
     7121,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     7123,
     UINT16_MAX,
     7123,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     7125,
     UINT16_MAX,
     7125,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     7127,
     UINT16_MAX,
     7127,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     7129,
     UINT16_MAX,
     7129,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     7131,
     UINT16_MAX,
     7131,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     7133,
     UINT16_MAX,
     7133,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     7135,
     UINT16_MAX,
     7135,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     7137,
     UINT16_MAX,
     7137,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     7139,
     UINT16_MAX,
     7139,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     7141,
     UINT16_MAX,
     7141,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     7143,
     UINT16_MAX,
     7143,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     7145,
     UINT16_MAX,
     7145,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     7147,
     UINT16_MAX,
     7147,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     7149,
     UINT16_MAX,
     7149,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     7151,
     UINT16_MAX,
     7151,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     7153,
     UINT16_MAX,
     7153,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     7155,
     UINT16_MAX,
     7155,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     7157,
     UINT16_MAX,
     7157,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     7159,
     UINT16_MAX,
     7159,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     7161,
     UINT16_MAX,
     7161,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     7163,
     UINT16_MAX,
     7163,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9522,
     UINT16_MAX,
     9522,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9524,
     UINT16_MAX,
     9524,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9526,
     UINT16_MAX,
     9526,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9528,
     UINT16_MAX,
     9528,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9530,
     UINT16_MAX,
     9530,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9532,
     UINT16_MAX,
     9532,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9534,
     UINT16_MAX,
     9534,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9536,
     UINT16_MAX,
     9536,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9538,
     UINT16_MAX,
     9538,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9540,
     UINT16_MAX,
     9540,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9542,
     UINT16_MAX,
     9542,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9544,
     UINT16_MAX,
     9544,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9546,
     UINT16_MAX,
     9546,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9548,
     UINT16_MAX,
     9548,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9550,
     UINT16_MAX,
     9550,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9552,
     UINT16_MAX,
     9552,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9554,
     UINT16_MAX,
     9554,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9556,
     UINT16_MAX,
     9556,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9558,
     UINT16_MAX,
     9558,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9560,
     UINT16_MAX,
     9560,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9562,
     UINT16_MAX,
     9562,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9564,
     UINT16_MAX,
     9564,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9566,
     UINT16_MAX,
     9566,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9568,
     UINT16_MAX,
     9568,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9570,
     UINT16_MAX,
     9570,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9572,
     UINT16_MAX,
     9572,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9574,
     UINT16_MAX,
     9574,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9576,
     UINT16_MAX,
     9576,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9578,
     UINT16_MAX,
     9578,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9580,
     UINT16_MAX,
     9580,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9582,
     UINT16_MAX,
     9582,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9584,
     UINT16_MAX,
     9584,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_MC,
     6,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_SPACINGMARK},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     5374,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     5378,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     15357,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     true,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     15361,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     5382,
     false,
     true,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     15365,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     true,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     15369,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     true,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     15373,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     true,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     15377,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     true,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     15381,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     true,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_MC,
     216,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     49224,
     false,
     false,
     false,
     false,
     0,
     0,
     UTF8PROC_BOUNDCLASS_EXTEND},
    {UTF8PROC_CATEGORY_MC,
     216,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     0,
     0,
     UTF8PROC_BOUNDCLASS_SPACINGMARK},
    {UTF8PROC_CATEGORY_MC,
     226,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     0,
     0,
     UTF8PROC_BOUNDCLASS_SPACINGMARK},
    {UTF8PROC_CATEGORY_MC,
     216,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     49226,
     false,
     false,
     false,
     false,
     0,
     0,
     UTF8PROC_BOUNDCLASS_EXTEND},
    {UTF8PROC_CATEGORY_MC,
     216,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     49228,
     false,
     false,
     false,
     false,
     0,
     0,
     UTF8PROC_BOUNDCLASS_EXTEND},
    {UTF8PROC_CATEGORY_MC,
     216,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     49230,
     false,
     false,
     false,
     false,
     0,
     0,
     UTF8PROC_BOUNDCLASS_EXTEND},
    {UTF8PROC_CATEGORY_MC,
     216,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     49232,
     false,
     false,
     false,
     false,
     0,
     0,
     UTF8PROC_BOUNDCLASS_EXTEND},
    {UTF8PROC_CATEGORY_MC,
     216,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     49234,
     false,
     false,
     false,
     false,
     0,
     0,
     UTF8PROC_BOUNDCLASS_EXTEND},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     5394,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     5398,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     15385,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     5402,
     false,
     true,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     15389,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     5408,
     false,
     true,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     15393,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     true,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     15397,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     true,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     15401,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     true,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     15405,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     true,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_FONT,
     1491,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_FONT,
     1497,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_FONT,
     1500,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_FONT,
     1501,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_FONT,
     1505,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_FONT,
     3305,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_FONT,
     1509,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_FONT,
     1510,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_FONT,
     2908,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_FONT,
     1511,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_FONT,
     2925,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_FONT,
     3312,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_FONT,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_FONT,
     1,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_FONT,
     2,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_FONT,
     5,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_FONT,
     10,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_FONT,
     12,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_FONT,
     13,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_FONT,
     15,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_FONT,
     16,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_FONT,
     17,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_FONT,
     18,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_FONT,
     19,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_FONT,
     20,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_FONT,
     21,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_FONT,
     22,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_FONT,
     23,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_FONT,
     24,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_FONT,
     25,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_FONT,
     7217,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_FONT,
     7218,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_FONT,
     7219,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_FONT,
     7220,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_FONT,
     7221,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_FONT,
     7222,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_FONT,
     7223,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_FONT,
     7224,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_FONT,
     915,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_FONT,
     7225,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_FONT,
     7226,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_FONT,
     7227,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_FONT,
     7228,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_FONT,
     7229,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_FONT,
     7230,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_FONT,
     7231,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_FONT,
     7232,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_FONT,
     7233,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_FONT,
     917,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_FONT,
     7234,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_FONT,
     897,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_FONT,
     7235,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_FONT,
     7236,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_FONT,
     7237,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_FONT,
     2835,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SM,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_FONT,
     7238,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_FONT,
     845,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_FONT,
     846,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_FONT,
     848,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_FONT,
     849,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_FONT,
     850,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_FONT,
     851,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_FONT,
     852,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_FONT,
     807,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_FONT,
     853,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_FONT,
     854,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_FONT,
     35,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_FONT,
     855,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_FONT,
     856,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_FONT,
     857,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_FONT,
     859,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_FONT,
     914,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_FONT,
     860,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_FONT,
     861,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_FONT,
     862,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_FONT,
     863,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_FONT,
     864,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_FONT,
     865,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_FONT,
     866,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SM,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_FONT,
     7239,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     true,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_FONT,
     7240,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_FONT,
     7241,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_FONT,
     7242,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_FONT,
     7243,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_FONT,
     7244,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_FONT,
     7245,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_FONT,
     7246,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_FONT,
     904,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_ND,
     0,
     UTF8PROC_BIDI_CLASS_EN,
     UTF8PROC_DECOMP_TYPE_FONT,
     2792,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_ND,
     0,
     UTF8PROC_BIDI_CLASS_EN,
     UTF8PROC_DECOMP_TYPE_FONT,
     38,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_ND,
     0,
     UTF8PROC_BIDI_CLASS_EN,
     UTF8PROC_DECOMP_TYPE_FONT,
     31,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_ND,
     0,
     UTF8PROC_BIDI_CLASS_EN,
     UTF8PROC_DECOMP_TYPE_FONT,
     32,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_ND,
     0,
     UTF8PROC_BIDI_CLASS_EN,
     UTF8PROC_DECOMP_TYPE_FONT,
     2793,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_ND,
     0,
     UTF8PROC_BIDI_CLASS_EN,
     UTF8PROC_DECOMP_TYPE_FONT,
     2794,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_ND,
     0,
     UTF8PROC_BIDI_CLASS_EN,
     UTF8PROC_DECOMP_TYPE_FONT,
     2795,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_ND,
     0,
     UTF8PROC_BIDI_CLASS_EN,
     UTF8PROC_DECOMP_TYPE_FONT,
     2796,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_ND,
     0,
     UTF8PROC_BIDI_CLASS_EN,
     UTF8PROC_DECOMP_TYPE_FONT,
     2797,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_ND,
     0,
     UTF8PROC_BIDI_CLASS_EN,
     UTF8PROC_DECOMP_TYPE_FONT,
     2798,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_R,
     0,
     UINT16_MAX,
     7247,
     UINT16_MAX,
     7247,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_R,
     0,
     UINT16_MAX,
     7249,
     UINT16_MAX,
     7249,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_R,
     0,
     UINT16_MAX,
     7251,
     UINT16_MAX,
     7251,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_R,
     0,
     UINT16_MAX,
     7253,
     UINT16_MAX,
     7253,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_R,
     0,
     UINT16_MAX,
     7255,
     UINT16_MAX,
     7255,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_R,
     0,
     UINT16_MAX,
     7257,
     UINT16_MAX,
     7257,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_R,
     0,
     UINT16_MAX,
     7259,
     UINT16_MAX,
     7259,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_R,
     0,
     UINT16_MAX,
     7261,
     UINT16_MAX,
     7261,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_R,
     0,
     UINT16_MAX,
     7263,
     UINT16_MAX,
     7263,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_R,
     0,
     UINT16_MAX,
     7265,
     UINT16_MAX,
     7265,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_R,
     0,
     UINT16_MAX,
     7267,
     UINT16_MAX,
     7267,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_R,
     0,
     UINT16_MAX,
     7269,
     UINT16_MAX,
     7269,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_R,
     0,
     UINT16_MAX,
     7271,
     UINT16_MAX,
     7271,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_R,
     0,
     UINT16_MAX,
     7273,
     UINT16_MAX,
     7273,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_R,
     0,
     UINT16_MAX,
     7275,
     UINT16_MAX,
     7275,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_R,
     0,
     UINT16_MAX,
     7277,
     UINT16_MAX,
     7277,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_R,
     0,
     UINT16_MAX,
     7279,
     UINT16_MAX,
     7279,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_R,
     0,
     UINT16_MAX,
     7281,
     UINT16_MAX,
     7281,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_R,
     0,
     UINT16_MAX,
     7283,
     UINT16_MAX,
     7283,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_R,
     0,
     UINT16_MAX,
     7285,
     UINT16_MAX,
     7285,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_R,
     0,
     UINT16_MAX,
     7287,
     UINT16_MAX,
     7287,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_R,
     0,
     UINT16_MAX,
     7289,
     UINT16_MAX,
     7289,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_R,
     0,
     UINT16_MAX,
     7291,
     UINT16_MAX,
     7291,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_R,
     0,
     UINT16_MAX,
     7293,
     UINT16_MAX,
     7293,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_R,
     0,
     UINT16_MAX,
     7295,
     UINT16_MAX,
     7295,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_R,
     0,
     UINT16_MAX,
     7297,
     UINT16_MAX,
     7297,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_R,
     0,
     UINT16_MAX,
     7299,
     UINT16_MAX,
     7299,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_R,
     0,
     UINT16_MAX,
     7301,
     UINT16_MAX,
     7301,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_R,
     0,
     UINT16_MAX,
     7303,
     UINT16_MAX,
     7303,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_R,
     0,
     UINT16_MAX,
     7305,
     UINT16_MAX,
     7305,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_R,
     0,
     UINT16_MAX,
     7307,
     UINT16_MAX,
     7307,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_R,
     0,
     UINT16_MAX,
     7309,
     UINT16_MAX,
     7309,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_R,
     0,
     UINT16_MAX,
     7311,
     UINT16_MAX,
     7311,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LU,
     0,
     UTF8PROC_BIDI_CLASS_R,
     0,
     UINT16_MAX,
     7313,
     UINT16_MAX,
     7313,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_R,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9586,
     UINT16_MAX,
     9586,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_R,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9588,
     UINT16_MAX,
     9588,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_R,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9590,
     UINT16_MAX,
     9590,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_R,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9592,
     UINT16_MAX,
     9592,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_R,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9594,
     UINT16_MAX,
     9594,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_R,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9596,
     UINT16_MAX,
     9596,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_R,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9598,
     UINT16_MAX,
     9598,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_R,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9600,
     UINT16_MAX,
     9600,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_R,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9602,
     UINT16_MAX,
     9602,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_R,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9604,
     UINT16_MAX,
     9604,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_R,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9606,
     UINT16_MAX,
     9606,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_R,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9608,
     UINT16_MAX,
     9608,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_R,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9610,
     UINT16_MAX,
     9610,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_R,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9612,
     UINT16_MAX,
     9612,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_R,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9614,
     UINT16_MAX,
     9614,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_R,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9616,
     UINT16_MAX,
     9616,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_R,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9618,
     UINT16_MAX,
     9618,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_R,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9620,
     UINT16_MAX,
     9620,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_R,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9622,
     UINT16_MAX,
     9622,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_R,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9624,
     UINT16_MAX,
     9624,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_R,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9626,
     UINT16_MAX,
     9626,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_R,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9628,
     UINT16_MAX,
     9628,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_R,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9630,
     UINT16_MAX,
     9630,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_R,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9632,
     UINT16_MAX,
     9632,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_R,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9634,
     UINT16_MAX,
     9634,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_R,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9636,
     UINT16_MAX,
     9636,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_R,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9638,
     UINT16_MAX,
     9638,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_R,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9640,
     UINT16_MAX,
     9640,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_R,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9642,
     UINT16_MAX,
     9642,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_R,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9644,
     UINT16_MAX,
     9644,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_R,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9646,
     UINT16_MAX,
     9646,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_R,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9648,
     UINT16_MAX,
     9648,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_R,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9650,
     UINT16_MAX,
     9650,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LL,
     0,
     UTF8PROC_BIDI_CLASS_R,
     0,
     UINT16_MAX,
     UINT16_MAX,
     9652,
     UINT16_MAX,
     9652,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_FONT,
     6580,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_FONT,
     6581,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_FONT,
     6585,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_FONT,
     6588,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_FONT,
     6607,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_FONT,
     6591,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_FONT,
     6586,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_FONT,
     6596,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_FONT,
     6608,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_FONT,
     6602,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_FONT,
     6603,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_FONT,
     6604,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_FONT,
     6605,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_FONT,
     6592,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_FONT,
     6598,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_FONT,
     6600,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_FONT,
     6594,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_FONT,
     6601,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_FONT,
     6590,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_FONT,
     6593,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_FONT,
     6583,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_FONT,
     6584,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_FONT,
     6587,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_FONT,
     6589,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_FONT,
     6595,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_FONT,
     6597,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_FONT,
     6599,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_FONT,
     7315,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_FONT,
     5790,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_FONT,
     7316,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_FONT,
     7317,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_AL,
     UTF8PROC_DECOMP_TYPE_FONT,
     6606,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_NO,
     0,
     UTF8PROC_BIDI_CLASS_EN,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     15510,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_NO,
     0,
     UTF8PROC_BIDI_CLASS_EN,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     15512,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_NO,
     0,
     UTF8PROC_BIDI_CLASS_EN,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     15514,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_NO,
     0,
     UTF8PROC_BIDI_CLASS_EN,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     15516,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_NO,
     0,
     UTF8PROC_BIDI_CLASS_EN,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     15518,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_NO,
     0,
     UTF8PROC_BIDI_CLASS_EN,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     15520,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_NO,
     0,
     UTF8PROC_BIDI_CLASS_EN,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     15522,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_NO,
     0,
     UTF8PROC_BIDI_CLASS_EN,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     15524,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_NO,
     0,
     UTF8PROC_BIDI_CLASS_EN,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     15526,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_NO,
     0,
     UTF8PROC_BIDI_CLASS_EN,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     15528,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_NO,
     0,
     UTF8PROC_BIDI_CLASS_EN,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     15530,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     23724,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     23727,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     23730,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     23733,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     23736,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     23739,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     23742,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     23745,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     23748,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     23751,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     23754,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     23757,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     23760,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     23763,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     23766,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     23769,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     23772,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     23775,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     23778,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     23781,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     23784,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     23787,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     23790,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     23793,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     23796,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     23799,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     23802,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_CIRCLE,
     2812,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_CIRCLE,
     1508,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_CIRCLE,
     15613,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_CIRCLE,
     15615,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SQUARE,
     1491,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SQUARE,
     1493,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SQUARE,
     2812,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SQUARE,
     1494,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SQUARE,
     1495,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SQUARE,
     2837,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SQUARE,
     1497,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SQUARE,
     1498,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SQUARE,
     1499,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SQUARE,
     1500,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SQUARE,
     1501,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SQUARE,
     1502,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SQUARE,
     1503,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SQUARE,
     1504,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SQUARE,
     1505,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SQUARE,
     1507,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SQUARE,
     2826,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SQUARE,
     1508,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SQUARE,
     3305,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SQUARE,
     1509,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SQUARE,
     1510,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SQUARE,
     2908,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SQUARE,
     1511,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SQUARE,
     2925,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SQUARE,
     3312,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SQUARE,
     2834,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SQUARE,
     15617,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SQUARE,
     13070,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SQUARE,
     15619,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SQUARE,
     15621,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SQUARE,
     23815,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SQUARE,
     15626,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_SUPER,
     15628,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_SUPER,
     15630,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     UTF8PROC_DECOMP_TYPE_SUPER,
     15632,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_EXTENDED_PICTOGRAPHIC},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_EXTENDED_PICTOGRAPHIC},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SQUARE,
     15634,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     1,
     0,
     UTF8PROC_BOUNDCLASS_REGIONAL_INDICATOR},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SQUARE,
     15636,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SQUARE,
     15638,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_EXTENDED_PICTOGRAPHIC},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SQUARE,
     4295,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_EXTENDED_PICTOGRAPHIC},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SQUARE,
     3506,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SQUARE,
     7448,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SQUARE,
     7449,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SQUARE,
     7450,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SQUARE,
     3449,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SQUARE,
     7451,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SQUARE,
     7452,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SQUARE,
     3886,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SQUARE,
     7453,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SQUARE,
     7454,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SQUARE,
     7455,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_EXTENDED_PICTOGRAPHIC},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SQUARE,
     5429,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SQUARE,
     7456,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SQUARE,
     7457,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SQUARE,
     7458,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SQUARE,
     7459,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SQUARE,
     7460,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SQUARE,
     7461,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SQUARE,
     3542,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SQUARE,
     7462,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SQUARE,
     7463,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SQUARE,
     7464,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SQUARE,
     7465,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SQUARE,
     7466,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SQUARE,
     7467,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SQUARE,
     3443,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SQUARE,
     3878,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SQUARE,
     7468,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SQUARE,
     4208,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SQUARE,
     3881,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SQUARE,
     4209,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SQUARE,
     7469,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_EXTENDED_PICTOGRAPHIC},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SQUARE,
     3598,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SQUARE,
     7470,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SQUARE,
     7471,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_EXTENDED_PICTOGRAPHIC},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SQUARE,
     7472,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_EXTENDED_PICTOGRAPHIC},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SQUARE,
     7473,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_EXTENDED_PICTOGRAPHIC},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SQUARE,
     7474,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_EXTENDED_PICTOGRAPHIC},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SQUARE,
     4191,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_EXTENDED_PICTOGRAPHIC},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SQUARE,
     3516,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_EXTENDED_PICTOGRAPHIC},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SQUARE,
     7475,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_EXTENDED_PICTOGRAPHIC},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SQUARE,
     7476,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_EXTENDED_PICTOGRAPHIC},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SQUARE,
     7477,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_EXTENDED_PICTOGRAPHIC},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_SQUARE,
     7478,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     23863,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     23866,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     23869,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     23872,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     23875,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     23878,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     23881,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     23884,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_COMPAT,
     23887,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_CIRCLE,
     7506,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_EXTENDED_PICTOGRAPHIC},
    {UTF8PROC_CATEGORY_SO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     UTF8PROC_DECOMP_TYPE_CIRCLE,
     7507,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_EXTENDED_PICTOGRAPHIC},
    {UTF8PROC_CATEGORY_SK,
     0,
     UTF8PROC_BIDI_CLASS_ON,
     0,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_EXTEND},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7508,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7509,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7510,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7511,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7513,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7514,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7515,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7516,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7517,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7518,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7519,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7520,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7522,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7523,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7524,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7525,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7527,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7528,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7458,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7529,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7531,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7532,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7533,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7534,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7535,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     3459,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7537,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7538,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7539,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7540,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7476,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7541,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7542,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7543,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7544,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7545,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7546,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7547,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7548,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7549,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7550,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7552,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7553,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7554,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7555,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7557,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7558,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7559,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7560,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7561,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7562,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7563,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7564,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7565,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7566,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7567,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7568,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7569,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7570,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7571,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7572,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7573,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7574,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7575,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7576,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7577,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7578,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7579,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7580,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7581,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7582,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7583,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7584,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7585,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7587,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7588,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7589,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7451,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7590,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7591,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7592,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7594,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7596,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7597,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7598,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7599,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7600,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7601,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7602,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7603,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7604,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7605,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7607,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7608,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7609,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7610,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7612,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7613,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7614,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     3485,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7615,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7616,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7617,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7618,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7619,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7621,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7622,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7624,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7625,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7626,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7627,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7628,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7629,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7630,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7631,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7632,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7633,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7634,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7635,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7637,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7638,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7639,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7640,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7641,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     3497,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7643,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7645,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7646,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7647,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7648,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7650,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7652,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7653,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7654,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7655,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7656,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7657,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7658,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7659,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7660,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7661,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7662,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7664,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7665,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7666,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7667,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7668,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7669,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7670,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7671,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7672,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7673,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7674,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7675,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7676,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7677,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7678,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7680,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7681,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7682,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7683,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7684,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7685,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7687,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7688,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7689,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7690,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7691,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7692,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7693,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7694,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7695,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7696,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7697,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7699,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7700,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7701,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7702,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7703,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7704,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7705,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7706,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7707,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7708,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7709,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7710,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7711,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7712,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7713,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7714,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7716,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7717,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7718,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7719,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7720,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7722,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7723,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7724,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7725,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7726,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7727,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7728,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7729,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7731,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7732,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7733,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7734,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7736,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7737,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7738,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7739,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7740,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7741,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7743,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7745,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7747,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7748,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7750,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7751,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7752,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7753,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7754,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7755,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7756,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7757,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7758,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7760,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7761,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7762,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7763,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7764,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7765,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7767,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7768,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7769,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7771,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7773,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7774,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7775,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7776,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7777,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7778,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7779,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7780,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7781,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7783,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7784,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7786,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7787,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7789,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7790,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7791,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7793,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7794,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7795,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7797,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7799,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7800,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7801,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7802,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7803,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7804,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7805,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7806,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7807,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7808,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7809,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7810,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7812,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7813,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7815,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7817,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7818,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7820,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7822,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7824,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7825,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7826,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7828,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7830,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7832,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7834,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7835,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7836,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7837,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7838,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7839,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7841,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7842,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7843,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7845,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7847,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7849,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7850,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7851,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7852,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7853,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7855,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7857,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7858,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7859,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7861,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7862,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7863,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7864,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7866,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7867,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7868,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7869,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7870,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7871,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7873,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7874,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7875,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7876,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7877,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7878,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7879,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7881,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7883,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7884,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7886,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7887,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7889,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7890,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7891,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7893,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7895,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7896,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7898,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7899,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7901,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7902,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7903,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7904,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7905,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7906,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7907,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7909,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7911,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7913,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7915,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7916,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7917,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7918,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7919,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7920,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7921,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7922,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7923,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7924,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7925,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7926,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7928,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7929,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7930,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7931,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7932,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7933,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7934,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7935,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7936,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7937,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7938,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7940,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7942,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7944,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7945,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7946,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7947,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7948,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7950,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7951,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7953,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7954,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7955,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7957,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7959,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7960,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7961,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7962,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7963,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7964,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7965,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7966,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7967,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7968,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7969,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7970,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7971,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7972,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7973,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7974,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     3587,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7975,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7977,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7978,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7979,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7980,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7981,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7982,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7984,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7986,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7987,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7988,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     3594,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7989,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7991,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7992,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7993,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7994,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7995,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7997,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     7999,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     8000,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     8001,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     8002,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     8004,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     8005,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     8007,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     8009,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     8010,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     8011,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     8012,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     8014,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     8015,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     8016,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     8017,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     8018,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     8019,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     8020,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     8021,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     8023,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     8024,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     8025,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     8026,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     8028,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     8029,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     8030,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     8031,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     8032,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     8034,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     8036,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     8037,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     8038,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     8039,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     8041,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     8042,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     8044,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     8045,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     8047,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     8048,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     8049,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     8050,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     8051,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     8052,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     8053,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     8054,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     8056,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     8057,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     8058,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     8059,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     8060,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     8061,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     8063,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     8064,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     8066,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     8068,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     3642,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     8070,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     3646,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     8071,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     8072,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     8073,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     8074,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     3651,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
    {UTF8PROC_CATEGORY_LO,
     0,
     UTF8PROC_BIDI_CLASS_L,
     0,
     8075,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     UINT16_MAX,
     false,
     false,
     false,
     false,
     2,
     0,
     UTF8PROC_BOUNDCLASS_OTHER},
};

static const utf8proc_uint16_t utf8proc_combinations[] = {
    0,     46,    192,   193,   194,   195,   196,   197,   0,     256,   258,
    260,   550,   461,   0,     0,     512,   514,   0,     0,     0,     0,
    0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
    0,     0,     0,     0,     0,     0,     0,     0,     7680,  7840,  0,
    0,     0,     0,     0,     7842,  1,     11,    262,   264,   0,     0,
    0,     199,   0,     0,     0,     266,   268,   0,     46,    200,   201,
    202,   7868,  203,   0,     552,   274,   276,   280,   278,   282,   0,
    0,     516,   518,   0,     0,     0,     0,     0,     0,     0,     0,
    0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
    0,     0,     0,     0,     0,     7864,  0,     7704,  7706,  0,     0,
    7866,  0,     46,    204,   205,   206,   296,   207,   0,     0,     298,
    300,   302,   304,   463,   0,     0,     520,   522,   0,     0,     0,
    0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
    0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     7882,
    0,     0,     7724,  0,     0,     7880,  0,     42,    504,   323,   0,
    209,   0,     0,     325,   0,     0,     0,     7748,  327,   0,     0,
    0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
    0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
    0,     0,     0,     0,     7750,  7752,  7754,  0,     46,    210,   211,
    212,   213,   214,   0,     0,     332,   334,   490,   558,   465,   336,
    416,   524,   526,   0,     0,     0,     0,     0,     0,     0,     0,
    0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
    0,     0,     0,     0,     0,     7884,  0,     0,     0,     0,     0,
    7886,  0,     46,    217,   218,   219,   360,   220,   366,   0,     362,
    364,   370,   0,     467,   368,   431,   532,   534,   0,     0,     0,
    0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
    0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     7908,
    0,     7798,  7796,  0,     7794,  7910,  0,     46,    7922,  221,   374,
    7928,  376,   0,     0,     562,   0,     0,     7822,  0,     0,     0,
    0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
    0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
    0,     0,     0,     0,     7924,  0,     0,     0,     0,     0,     7926,
    0,     46,    224,   225,   226,   227,   228,   229,   0,     257,   259,
    261,   551,   462,   0,     0,     513,   515,   0,     0,     0,     0,
    0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
    0,     0,     0,     0,     0,     0,     0,     0,     7681,  7841,  0,
    0,     0,     0,     0,     7843,  1,     11,    263,   265,   0,     0,
    0,     231,   0,     0,     0,     267,   269,   0,     46,    232,   233,
    234,   7869,  235,   0,     553,   275,   277,   281,   279,   283,   0,
    0,     517,   519,   0,     0,     0,     0,     0,     0,     0,     0,
    0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
    0,     0,     0,     0,     0,     7865,  0,     7705,  7707,  0,     0,
    7867,  0,     46,    236,   237,   238,   297,   239,   0,     0,     299,
    301,   303,   0,     464,   0,     0,     521,   523,   0,     0,     0,
    0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
    0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     7883,
    0,     0,     7725,  0,     0,     7881,  0,     42,    505,   324,   0,
    241,   0,     0,     326,   0,     0,     0,     7749,  328,   0,     0,
    0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
    0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
    0,     0,     0,     0,     7751,  7753,  7755,  0,     46,    242,   243,
    244,   245,   246,   0,     0,     333,   335,   491,   559,   466,   337,
    417,   525,   527,   0,     0,     0,     0,     0,     0,     0,     0,
    0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
    0,     0,     0,     0,     0,     7885,  0,     0,     0,     0,     0,
    7887,  0,     46,    249,   250,   251,   361,   252,   367,   0,     363,
    365,   371,   0,     468,   369,   432,   533,   535,   0,     0,     0,
    0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
    0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     7909,
    0,     7799,  7797,  0,     7795,  7911,  0,     46,    7923,  253,   375,
    7929,  255,   7833,  0,     563,   0,     0,     7823,  0,     0,     0,
    0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
    0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
    0,     0,     0,     0,     7925,  0,     0,     0,     0,     0,     7927,
    6,     42,    7696,  0,     0,     0,     7690,  270,   0,     0,     0,
    0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
    0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
    0,     0,     0,     7692,  7694,  7698,  6,     42,    7697,  0,     0,
    0,     7691,  271,   0,     0,     0,     0,     0,     0,     0,     0,
    0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
    0,     0,     0,     0,     0,     0,     0,     0,     0,     7693,  7695,
    7699,  1,     11,    500,   284,   0,     0,     0,     290,   7712,  286,
    0,     288,   486,   1,     11,    501,   285,   0,     0,     0,     291,
    7713,  287,   0,     289,   487,   2,     44,    292,   0,     7718,  0,
    7720,  0,     0,     0,     7714,  542,   0,     0,     0,     0,     0,
    0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
    0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
    0,     7716,  0,     0,     0,     7722,  2,     44,    293,   0,     7719,
    0,     7721,  0,     0,     0,     7715,  543,   0,     0,     0,     0,
    0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
    0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
    0,     0,     7717,  7830,  0,     0,     7723,  2,     2,     308,   2,
    11,    309,   0,     0,     0,     0,     0,     0,     0,     0,     496,
    1,     41,    7728,  0,     0,     0,     0,     310,   0,     0,     0,
    0,     488,   0,     0,     0,     0,     0,     0,     0,     0,     0,
    0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
    0,     0,     0,     0,     0,     0,     0,     0,     7730,  7732,  1,
    41,    7729,  0,     0,     0,     0,     311,   0,     0,     0,     0,
    489,   0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
    0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
    0,     0,     0,     0,     0,     0,     0,     7731,  7733,  1,     42,
    313,   0,     0,     0,     0,     315,   0,     0,     0,     0,     317,
    0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
    0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
    0,     0,     0,     0,     0,     0,     7734,  7738,  7740,  1,     42,
    314,   0,     0,     0,     0,     316,   0,     0,     0,     0,     318,
    0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
    0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
    0,     0,     0,     0,     0,     0,     7735,  7739,  7741,  1,     41,
    340,   0,     0,     0,     0,     342,   0,     0,     0,     7768,  344,
    0,     0,     528,   530,   0,     0,     0,     0,     0,     0,     0,
    0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
    0,     0,     0,     0,     0,     0,     7770,  7774,  1,     41,    341,
    0,     0,     0,     0,     343,   0,     0,     0,     7769,  345,   0,
    0,     529,   531,   0,     0,     0,     0,     0,     0,     0,     0,
    0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
    0,     0,     0,     0,     0,     7771,  7775,  1,     40,    346,   348,
    0,     0,     0,     350,   0,     0,     0,     7776,  352,   0,     0,
    0,     0,     536,   0,     0,     0,     0,     0,     0,     0,     0,
    0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
    0,     0,     0,     0,     7778,  1,     40,    347,   349,   0,     0,
    0,     351,   0,     0,     0,     7777,  353,   0,     0,     0,     0,
    537,   0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
    0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
    0,     0,     7779,  6,     42,    354,   0,     0,     0,     7786,  356,
    0,     0,     0,     0,     538,   0,     0,     0,     0,     0,     0,
    0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
    0,     0,     0,     0,     0,     0,     7788,  7790,  7792,  4,     42,
    7831,  0,     355,   0,     0,     0,     7787,  357,   0,     0,     0,
    0,     539,   0,     0,     0,     0,     0,     0,     0,     0,     0,
    0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
    0,     0,     0,     7789,  7791,  7793,  0,     40,    7808,  7810,  372,
    0,     7812,  0,     0,     0,     0,     0,     7814,  0,     0,     0,
    0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
    0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
    0,     0,     0,     0,     7816,  0,     40,    7809,  7811,  373,   0,
    7813,  7832,  0,     0,     0,     0,     7815,  0,     0,     0,     0,
    0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
    0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
    0,     0,     0,     7817,  1,     41,    377,   7824,  0,     0,     0,
    0,     0,     0,     0,     379,   381,   0,     0,     0,     0,     0,
    0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
    0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
    0,     7826,  7828,  1,     41,    378,   7825,  0,     0,     0,     0,
    0,     0,     0,     380,   382,   0,     0,     0,     0,     0,     0,
    0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
    0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
    7827,  7829,  0,     11,    475,   471,   0,     0,     0,     0,     0,
    469,   0,     0,     0,     473,   0,     11,    476,   472,   0,     0,
    0,     0,     0,     470,   0,     0,     0,     474,   7,     7,     478,
    7,     7,     479,   7,     7,     480,   7,     7,     481,   1,     7,
    508,   0,     0,     0,     0,     0,     482,   1,     7,     509,   0,
    0,     0,     0,     0,     483,   7,     7,     492,   7,     7,     493,
    11,    11,    494,   11,    11,    495,   1,     1,     506,   1,     1,
    507,   1,     1,     510,   1,     1,     511,   7,     7,     554,   7,
    7,     555,   1,     7,     7756,  0,     0,     7758,  0,     0,     556,
    1,     7,     7757,  0,     0,     7759,  0,     0,     557,   7,     7,
    560,   7,     7,     561,   0,     49,    8173,  901,   0,     0,     0,
    0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
    0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
    0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
    0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
    8129,  0,     50,    8122,  902,   0,     0,     0,     0,     0,     8121,
    8120,  0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
    0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
    0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
    0,     0,     0,     0,     0,     0,     7944,  7945,  0,     8124,  0,
    48,    8136,  904,   0,     0,     0,     0,     0,     0,     0,     0,
    0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
    0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
    0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
    0,     0,     0,     0,     7960,  7961,  0,     50,    8138,  905,   0,
    0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
    0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
    0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
    0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
    7976,  7977,  0,     8140,  0,     48,    8154,  906,   0,     0,     938,
    0,     0,     8153,  8152,  0,     0,     0,     0,     0,     0,     0,
    0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
    0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
    0,     0,     0,     0,     0,     0,     0,     0,     0,     7992,  7993,
    0,     48,    8184,  908,   0,     0,     0,     0,     0,     0,     0,
    0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
    0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
    0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
    0,     0,     0,     0,     0,     8008,  8009,  0,     48,    8170,  910,
    0,     0,     939,   0,     0,     8169,  8168,  0,     0,     0,     0,
    0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
    0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
    0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
    0,     0,     8025,  0,     50,    8186,  911,   0,     0,     0,     0,
    0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
    0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
    0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
    0,     0,     0,     0,     0,     0,     0,     0,     8040,  8041,  0,
    8188,  0,     49,    8146,  912,   0,     0,     0,     0,     0,     0,
    0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
    0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
    0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
    0,     0,     0,     0,     0,     0,     0,     0,     8151,  0,     50,
    8048,  940,   0,     0,     0,     0,     0,     8113,  8112,  0,     0,
    0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
    0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
    0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
    0,     0,     0,     7936,  7937,  8118,  8115,  0,     48,    8050,  941,
    0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
    0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
    0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
    0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
    0,     7952,  7953,  0,     50,    8052,  942,   0,     0,     0,     0,
    0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
    0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
    0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
    0,     0,     0,     0,     0,     0,     0,     0,     7968,  7969,  8134,
    8131,  0,     49,    8054,  943,   0,     0,     970,   0,     0,     8145,
    8144,  0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
    0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
    0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
    0,     0,     0,     0,     0,     0,     7984,  7985,  8150,  0,     49,
    8162,  944,   0,     0,     0,     0,     0,     0,     0,     0,     0,
    0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
    0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
    0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
    0,     0,     0,     0,     0,     8167,  0,     49,    8058,  973,   0,
    0,     971,   0,     0,     8161,  8160,  0,     0,     0,     0,     0,
    0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
    0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
    0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
    8016,  8017,  8166,  0,     48,    8056,  972,   0,     0,     0,     0,
    0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
    0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
    0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
    0,     0,     0,     0,     0,     0,     0,     0,     8000,  8001,  0,
    50,    8060,  974,   0,     0,     0,     0,     0,     0,     0,     0,
    0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
    0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
    0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
    0,     0,     0,     0,     8032,  8033,  8182,  8179,  1,     4,     979,
    0,     0,     980,   0,     8,     1024,  0,     0,     0,     1025,  0,
    0,     0,     1238,  1,     1,     1027,  4,     4,     1031,  1,     1,
    1036,  0,     8,     1037,  0,     0,     0,     1252,  0,     0,     1250,
    1049,  4,     12,    1264,  0,     0,     1262,  1038,  0,     0,     0,
    1266,  0,     8,     1117,  0,     0,     0,     1253,  0,     0,     1251,
    1081,  0,     8,     1104,  0,     0,     0,     1105,  0,     0,     0,
    1239,  1,     1,     1107,  4,     4,     1111,  1,     1,     1116,  4,
    12,    1265,  0,     0,     1263,  1118,  0,     0,     0,     1267,  14,
    14,    1142,  14,    14,    1143,  4,     8,     1244,  0,     0,     0,
    1217,  4,     8,     1245,  0,     0,     0,     1218,  4,     8,     1234,
    0,     0,     0,     1232,  4,     8,     1235,  0,     0,     0,     1233,
    4,     4,     1242,  4,     4,     1243,  4,     4,     1246,  4,     4,
    1247,  4,     4,     1254,  4,     4,     1255,  4,     4,     1258,  4,
    4,     1259,  4,     4,     1260,  4,     4,     1261,  4,     4,     1268,
    4,     4,     1269,  4,     4,     1272,  4,     4,     1273,  17,    19,
    1570,  1571,  1573,  18,    18,    1572,  18,    18,    1574,  18,    18,
    1728,  18,    18,    1730,  18,    18,    1747,  20,    20,    2345,  20,
    20,    2353,  20,    20,    2356,  21,    22,    2507,  2508,  23,    25,
    2888,  2891,  2892,  26,    26,    2964,  26,    27,    3020,  3018,  27,
    27,    3019,  28,    28,    3144,  29,    29,    3264,  29,    31,    3271,
    3272,  3274,  29,    29,    3275,  32,    33,    3402,  3404,  32,    32,
    3403,  34,    36,    3546,  3548,  3550,  34,    34,    3549,  37,    37,
    4134,  38,    38,    6918,  38,    38,    6920,  38,    38,    6922,  38,
    38,    6924,  38,    38,    6926,  38,    38,    6930,  38,    38,    6971,
    38,    38,    6973,  38,    38,    6976,  38,    38,    6977,  38,    38,
    6979,  10,    41,    7682,  0,     0,     0,     0,     0,     0,     0,
    0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
    0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
    7684,  7686,  10,    41,    7683,  0,     0,     0,     0,     0,     0,
    0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
    0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
    0,     7685,  7687,  1,     1,     7688,  1,     1,     7689,  0,     1,
    7700,  7702,  0,     1,     7701,  7703,  8,     8,     7708,  8,     8,
    7709,  10,    10,    7710,  10,    10,    7711,  1,     1,     7726,  1,
    1,     7727,  7,     7,     7736,  7,     7,     7737,  1,     40,    7742,
    0,     0,     0,     0,     0,     0,     0,     0,     7744,  0,     0,
    0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
    0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
    0,     0,     0,     0,     0,     7746,  1,     40,    7743,  0,     0,
    0,     0,     0,     0,     0,     0,     7745,  0,     0,     0,     0,
    0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
    0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
    0,     0,     0,     7747,  0,     1,     7760,  7762,  0,     1,     7761,
    7763,  1,     10,    7764,  0,     0,     0,     0,     0,     0,     0,
    0,     7766,  1,     10,    7765,  0,     0,     0,     0,     0,     0,
    0,     0,     7767,  7,     7,     7772,  7,     7,     7773,  10,    10,
    7780,  10,    10,    7781,  10,    10,    7782,  10,    10,    7783,  10,
    10,    7784,  10,    10,    7785,  1,     1,     7800,  1,     1,     7801,
    4,     4,     7802,  4,     4,     7803,  3,     40,    7804,  0,     0,
    0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
    0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
    0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
    0,     7806,  3,     40,    7805,  0,     0,     0,     0,     0,     0,
    0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
    0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
    0,     0,     0,     0,     0,     0,     0,     0,     7807,  4,     10,
    7820,  0,     0,     0,     0,     0,     7818,  4,     10,    7821,  0,
    0,     0,     0,     0,     7819,  10,    10,    7835,  0,     46,    7846,
    7844,  0,     7850,  0,     0,     0,     0,     0,     0,     0,     0,
    0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
    0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
    0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
    0,     7848,  0,     46,    7847,  7845,  0,     7851,  0,     0,     0,
    0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
    0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
    0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
    0,     0,     0,     0,     0,     0,     7849,  2,     8,     7852,  0,
    0,     0,     0,     0,     7862,  2,     8,     7853,  0,     0,     0,
    0,     0,     7863,  0,     46,    7856,  7854,  0,     7860,  0,     0,
    0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
    0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
    0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
    0,     0,     0,     0,     0,     0,     0,     7858,  0,     46,    7857,
    7855,  0,     7861,  0,     0,     0,     0,     0,     0,     0,     0,
    0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
    0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
    0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
    0,     7859,  0,     46,    7872,  7870,  0,     7876,  0,     0,     0,
    0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
    0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
    0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
    0,     0,     0,     0,     0,     0,     7874,  0,     46,    7873,  7871,
    0,     7877,  0,     0,     0,     0,     0,     0,     0,     0,     0,
    0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
    0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
    0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
    7875,  2,     2,     7878,  2,     2,     7879,  0,     46,    7890,  7888,
    0,     7894,  0,     0,     0,     0,     0,     0,     0,     0,     0,
    0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
    0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
    0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
    7892,  0,     46,    7891,  7889,  0,     7895,  0,     0,     0,     0,
    0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
    0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
    0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
    0,     0,     0,     0,     0,     7893,  2,     2,     7896,  2,     2,
    7897,  0,     46,    7900,  7898,  0,     7904,  0,     0,     0,     0,
    0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
    0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
    0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     7906,
    0,     0,     0,     0,     0,     7902,  0,     46,    7901,  7899,  0,
    7905,  0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
    0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
    0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
    0,     0,     0,     0,     7907,  0,     0,     0,     0,     0,     7903,
    0,     46,    7914,  7912,  0,     7918,  0,     0,     0,     0,     0,
    0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
    0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
    0,     0,     0,     0,     0,     0,     0,     0,     0,     7920,  0,
    0,     0,     0,     0,     7916,  0,     46,    7915,  7913,  0,     7919,
    0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
    0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
    0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
    0,     0,     0,     7921,  0,     0,     0,     0,     0,     7917,  0,
    50,    7938,  7940,  0,     0,     0,     0,     0,     0,     0,     0,
    0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
    0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
    0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
    0,     0,     0,     0,     0,     0,     7942,  8064,  0,     50,    7939,
    7941,  0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
    0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
    0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
    0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
    0,     0,     0,     0,     7943,  8065,  0,     50,    7946,  7948,  0,
    0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
    0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
    0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
    0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
    0,     0,     7950,  8072,  0,     50,    7947,  7949,  0,     0,     0,
    0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
    0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
    0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
    0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
    7951,  8073,  0,     1,     7954,  7956,  0,     1,     7955,  7957,  0,
    1,     7962,  7964,  0,     1,     7963,  7965,  0,     50,    7970,  7972,
    0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
    0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
    0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
    0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
    0,     0,     0,     7974,  8080,  0,     50,    7971,  7973,  0,     0,
    0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
    0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
    0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
    0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
    0,     7975,  8081,  0,     50,    7978,  7980,  0,     0,     0,     0,
    0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
    0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
    0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
    0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     7982,
    8088,  0,     50,    7979,  7981,  0,     0,     0,     0,     0,     0,
    0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
    0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
    0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
    0,     0,     0,     0,     0,     0,     0,     0,     7983,  8089,  0,
    49,    7986,  7988,  0,     0,     0,     0,     0,     0,     0,     0,
    0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
    0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
    0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
    0,     0,     0,     0,     0,     0,     7990,  0,     49,    7987,  7989,
    0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
    0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
    0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
    0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
    0,     0,     0,     7991,  0,     49,    7994,  7996,  0,     0,     0,
    0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
    0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
    0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
    0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
    7998,  0,     49,    7995,  7997,  0,     0,     0,     0,     0,     0,
    0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
    0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
    0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
    0,     0,     0,     0,     0,     0,     0,     0,     7999,  0,     1,
    8002,  8004,  0,     1,     8003,  8005,  0,     1,     8010,  8012,  0,
    1,     8011,  8013,  0,     49,    8018,  8020,  0,     0,     0,     0,
    0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
    0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
    0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
    0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     8022,
    0,     49,    8019,  8021,  0,     0,     0,     0,     0,     0,     0,
    0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
    0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
    0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
    0,     0,     0,     0,     0,     0,     0,     8023,  0,     49,    8027,
    8029,  0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
    0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
    0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
    0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
    0,     0,     0,     0,     8031,  0,     50,    8034,  8036,  0,     0,
    0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
    0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
    0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
    0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
    0,     8038,  8096,  0,     50,    8035,  8037,  0,     0,     0,     0,
    0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
    0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
    0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
    0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     8039,
    8097,  0,     50,    8042,  8044,  0,     0,     0,     0,     0,     0,
    0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
    0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
    0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
    0,     0,     0,     0,     0,     0,     0,     0,     8046,  8104,  0,
    50,    8043,  8045,  0,     0,     0,     0,     0,     0,     0,     0,
    0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
    0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
    0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
    0,     0,     0,     0,     0,     0,     8047,  8105,  50,    50,    8066,
    50,    50,    8067,  50,    50,    8068,  50,    50,    8069,  50,    50,
    8070,  50,    50,    8071,  50,    50,    8074,  50,    50,    8075,  50,
    50,    8076,  50,    50,    8077,  50,    50,    8078,  50,    50,    8079,
    50,    50,    8082,  50,    50,    8083,  50,    50,    8084,  50,    50,
    8085,  50,    50,    8086,  50,    50,    8087,  50,    50,    8090,  50,
    50,    8091,  50,    50,    8092,  50,    50,    8093,  50,    50,    8094,
    50,    50,    8095,  50,    50,    8098,  50,    50,    8099,  50,    50,
    8100,  50,    50,    8101,  50,    50,    8102,  50,    50,    8103,  50,
    50,    8106,  50,    50,    8107,  50,    50,    8108,  50,    50,    8109,
    50,    50,    8110,  50,    50,    8111,  50,    50,    8114,  50,    50,
    8116,  50,    50,    8119,  50,    50,    8130,  50,    50,    8132,  50,
    50,    8135,  0,     49,    8141,  8142,  0,     0,     0,     0,     0,
    0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
    0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
    0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
    0,     0,     0,     0,     0,     0,     0,     0,     0,     8143,  0,
    49,    8157,  8158,  0,     0,     0,     0,     0,     0,     0,     0,
    0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
    0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
    0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
    0,     0,     0,     0,     0,     0,     8159,  47,    48,    8164,  8165,
    48,    48,    8172,  50,    50,    8178,  50,    50,    8180,  50,    50,
    8183,  51,    51,    8602,  51,    51,    8603,  51,    51,    8622,  51,
    51,    8653,  51,    51,    8654,  51,    51,    8655,  51,    51,    8708,
    51,    51,    8713,  51,    51,    8716,  51,    51,    8740,  51,    51,
    8742,  51,    51,    8769,  51,    51,    8772,  51,    51,    8775,  51,
    51,    8777,  51,    51,    8800,  51,    51,    8802,  51,    51,    8813,
    51,    51,    8814,  51,    51,    8815,  51,    51,    8816,  51,    51,
    8817,  51,    51,    8820,  51,    51,    8821,  51,    51,    8824,  51,
    51,    8825,  51,    51,    8832,  51,    51,    8833,  51,    51,    8836,
    51,    51,    8837,  51,    51,    8840,  51,    51,    8841,  51,    51,
    8876,  51,    51,    8877,  51,    51,    8878,  51,    51,    8879,  51,
    51,    8928,  51,    51,    8929,  51,    51,    8930,  51,    51,    8931,
    51,    51,    8938,  51,    51,    8939,  51,    51,    8940,  51,    51,
    8941,  51,    51,    10972, 52,    52,    12364, 52,    52,    12366, 52,
    52,    12368, 52,    52,    12370, 52,    52,    12372, 52,    52,    12374,
    52,    52,    12376, 52,    52,    12378, 52,    52,    12380, 52,    52,
    12382, 52,    52,    12384, 52,    52,    12386, 52,    52,    12389, 52,
    52,    12391, 52,    52,    12393, 52,    53,    12400, 12401, 52,    53,
    12403, 12404, 52,    53,    12406, 12407, 52,    53,    12409, 12410, 52,
    53,    12412, 12413, 52,    52,    12436, 52,    52,    12446, 52,    52,
    12460, 52,    52,    12462, 52,    52,    12464, 52,    52,    12466, 52,
    52,    12468, 52,    52,    12470, 52,    52,    12472, 52,    52,    12474,
    52,    52,    12476, 52,    52,    12478, 52,    52,    12480, 52,    52,
    12482, 52,    52,    12485, 52,    52,    12487, 52,    52,    12489, 52,
    53,    12496, 12497, 52,    53,    12499, 12500, 52,    53,    12502, 12503,
    52,    53,    12505, 12506, 52,    53,    12508, 12509, 52,    52,    12532,
    52,    52,    12535, 52,    52,    12536, 52,    52,    12537, 52,    52,
    12538, 52,    52,    12542, 54,    55,    1,     4250,  54,    55,    1,
    4252,  54,    55,    1,     4267,  56,    57,    1,     4398,  56,    57,
    1,     4399,  58,    61,    1,     4939,  1,     4940,  62,    67,    1,
    5307,  1,     5308,  1,     5310,  68,    69,    1,     5562,  68,    69,
    1,     5563,  70,    71,    1,     6456,  72,    73,    1,     53598, 72,
    73,    1,     53599, 74,    83,    1,     53600, 1,     53601, 1,     53602,
    1,     53603, 1,     53604, 72,    73,    1,     53691, 72,    73,    1,
    53692, 74,    77,    1,     53693, 1,     53695, 74,    77,    1,     53694,
    1,     53696,
};
